/*
 * arch/arm/common/profile.c
 * 
 * profiler
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * 
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>

#include <asm/mach/arch.h>
#include <asm/arch/system.h>
#include <asm/mach/time.h>
#include <asm/arch/time.h>
#if defined(CONFIG_ARCH_CXD4105BASED)	
#include <asm/io.h>
#include <asm/arch/regs-gpio.h>
#endif

#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/kallsyms.h>

typedef struct {
	unsigned long long time;
	const char *caller;
	const char *callee;
	const void *addr; //callee address
} profile_record_t;

#define RECORDS CONFIG_DEBUG_PROFILE_RECORDS
static profile_record_t profile_records[RECORDS];
static char profilie_strings[RECORDS][16];
static unsigned record_index;
#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
static unsigned is_index_wrapped;
#endif
static DECLARE_MUTEX(record_index_sem);

void profile_soft_init(void)
{
	record_index = 0;
#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
	is_index_wrapped = 0;
#endif
}

void profile_printk_init(void)
{
	extern void profile_timer_init(void);
	
	profile_timer_init();
	profile_soft_init();
}

void profile_record(const char *caller,const char *callee)
{
	if(down_interruptible(&record_index_sem))
		return;

#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
	if(record_index >= RECORDS)
	{
		record_index -= RECORDS;
		is_index_wrapped = 1;
	}
#else
	if(record_index >= RECORDS)
		goto out;
#endif
	
	profile_records[record_index].time = sched_clock();
	profile_records[record_index].caller = caller;
	profile_records[record_index].callee = callee;
	profile_records[record_index++].addr = NULL;

 out:
	up(&record_index_sem);
}

EXPORT_SYMBOL(profile_record);

void profile_record_p(const char *caller,const void *addr)
{
	if(down_interruptible(&record_index_sem))
		return;
	
#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
	if(record_index >= RECORDS)
	{
		record_index -= RECORDS;
		is_index_wrapped = 1;
	}
#else
	if(record_index >= RECORDS)
		goto out;
#endif
	
	profile_records[record_index].time = sched_clock();
	profile_records[record_index].caller = caller;
	profile_records[record_index].callee = NULL;
	profile_records[record_index++].addr = addr;
	
 out:
	up(&record_index_sem);
}

EXPORT_SYMBOL(profile_record_p);

static int sched_clock_proc_write(struct file *file, const char *buffer,
				 unsigned long count, void *data)
{
	char *buf;

	if(down_interruptible(&record_index_sem))
		return -EINTR;
	
#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
	if(record_index >= RECORDS)
	{
		record_index -= RECORDS;
		is_index_wrapped = 1;
	}
#else
	if(record_index >= RECORDS)
		goto out;
#endif
	
	buf = profilie_strings[record_index];

	strncpy(buf, buffer, count>15?15:count);
	buf[count>15?15:count] = 0;
	
	
	profile_records[record_index].time = sched_clock();
	profile_records[record_index].caller = "user";
	profile_records[record_index].callee = buf;
	profile_records[record_index++].addr = NULL;

 out:
	up(&record_index_sem);
	return count;
}

static int sched_clock_proc_read(char *page, char **start, off_t off, int count,
				 int *eof, void *data)
{
	unsigned tlen;
	unsigned long long t;
	unsigned long nanosec_rem;
	
	t = sched_clock();
	nanosec_rem = do_div(t, 1000000000);
	tlen = sprintf(page,
		       "[ %5lu.%06lu ] -profile shell\n",
		       (unsigned long)t,
		       nanosec_rem/1000);
	*eof = 1;
	return tlen;
}

static int __init cxd2230ga_proc_shcd_clock(void) 
{
	struct proc_dir_entry *ent;
	ent = create_proc_entry("sched_clock", S_IFREG|S_IRUGO|S_IWUSR, NULL);
	if(!ent) printk(KERN_DEBUG "create sched_clock proc entry failed\n");
	ent->read_proc = sched_clock_proc_read;
	ent->write_proc = sched_clock_proc_write;
	return 0;
}

late_initcall(cxd2230ga_proc_shcd_clock);

static int record_to_string(profile_record_t *record, char *buf, int len)
{
	unsigned long nanosec_rem;
	unsigned long long t;
	t = record->time;
	
	if(record->addr){
		char *modname;
		unsigned long offset,size;
		char namebuf[KSYM_NAME_LEN+1];
		record->callee = kallsyms_lookup((unsigned long)record->addr, &size, &offset, &modname, namebuf);
	}

	nanosec_rem = do_div(t, 1000000000);
	return snprintf(buf,len,
			"[ %5lu.%06lu ] -profile %s %s\n",
			(unsigned long)t,
			nanosec_rem/1000,
			record->caller,
			record->callee);
}

static int profile_record_proc_write(struct file *file, const char *buffer,
				 unsigned long count, void *data)
{
	if(down_interruptible(&record_index_sem))
		return -EINTR;
	
	profile_soft_init();

	up(&record_index_sem);
	return count;
}

static int profile_record_proc_read(char *page, char **start, off_t off, int count,
				 int *eof, void *data)
{
	unsigned long tlen=0;
	profile_record_t *record;
	
	if(down_interruptible(&record_index_sem))
		return -EINTR;
	
#ifdef CONFIG_DEBUG_PROFILE_RINGBUFF
	/*
     * case1: discard mode
	 * case2: ring buffer mode -- if (is_index_wrapped == 1)
	 *
	 *   off= 0               ri            RECORDS 
	 * case1: | 1.from boot   |   unused    |
	 * case2: | 2.wrapped new | 1.used      |
	 *                        |<-adj (cont.)|
	 *        |-(cont.)  adj->|
	 *
	 */
	if (!is_index_wrapped)
	{
		if(off >= record_index){
			*eof = 1;
			goto out;
		}
		record = &profile_records[off];
	}
	else
	{
		unsigned adjust = off + record_index;
		if (adjust >= RECORDS)
			adjust -= RECORDS;
		if(off >= RECORDS)
		{
			*eof = 1;
			goto out;
		}
		record = &profile_records[adjust];
	}
#else
	if(off >= record_index){
		*eof = 1;
		goto out;
	}
	record = &profile_records[off];
#endif
	tlen = record_to_string(record, page, PAGE_SIZE);
	*start = (char *)1;
 out:
	up(&record_index_sem);
	return tlen>count?0:tlen;
}

static int __init debug_proc_profile(void) 
{
	struct proc_dir_entry *ent;
	ent = create_proc_entry("profile_record", S_IFREG|S_IRUGO|S_IWUSR, NULL);
	if(!ent) printk(KERN_DEBUG "create profile_record proc entry failed\n");
	ent->read_proc = profile_record_proc_read;
	ent->write_proc = profile_record_proc_write;
	return 0;
}

void profile_dump_record(void (*print)(char *))
{
	char buf[128];
	int i;
	profile_record(__FUNCTION__, "dump");
	for(i=0; i<record_index; i++){
		record_to_string(&profile_records[i], buf, 128);
		print(buf);
	}
}

late_initcall(debug_proc_profile);

#if defined(CONFIG_DEBUG_PROFILE_HEAP)
static unsigned is_heap_index_wrapped=0;
#ifndef CONFIG_PROFILE_HEAP_RECORDS_NUM
#define CONFIG_PROFILE_HEAP_RECORDS_NUM  7000
#endif
static DECLARE_MUTEX(heap_record_index_sem);
typedef struct {
	unsigned long long time; /* sched_clock */
	unsigned long type;      /* measurement point
				    0:try_to_freepages,
				    1: __vm_enough_memory
				 */
	unsigned int nr_free_pages;
	unsigned long nr_cache_pages;
	unsigned long req_pages;
	int pid; // current pid
} profile_heap_record_t;

static profile_heap_record_t profile_heap_records[CONFIG_PROFILE_HEAP_RECORDS_NUM];
static profile_heap_record_t profile_heap_used_max_record={0,};
static profile_heap_record_t profile_heap_needed_max_record={0,};
static unsigned long used_max = 0;
static unsigned long needed_max = 0;
static unsigned long heap_record_index=0;
extern unsigned long totalram_pages;
#include <linux/pagemap.h>
extern unsigned int nr_free_pages(void);
void profile_heap_record(unsigned long type, unsigned long req_pages)
{
	unsigned long used=totalram_pages;

	if(down_interruptible(&heap_record_index_sem))
		return;


	if(heap_record_index >= CONFIG_PROFILE_HEAP_RECORDS_NUM)
	{
		heap_record_index -= CONFIG_PROFILE_HEAP_RECORDS_NUM;
		is_heap_index_wrapped = 1;
	}

	profile_heap_records[heap_record_index].time = sched_clock();
	profile_heap_records[heap_record_index].type = type;
	profile_heap_records[heap_record_index].nr_free_pages = nr_free_pages();
	used -= profile_heap_records[heap_record_index].nr_free_pages;
	profile_heap_records[heap_record_index].nr_cache_pages = get_page_cache_size();
	used -= profile_heap_records[heap_record_index].nr_cache_pages;
	profile_heap_records[heap_record_index].pid = current->pid;
	profile_heap_records[heap_record_index].req_pages = req_pages;

	if ( used_max <= used ){
		used_max = used;
		profile_heap_used_max_record = profile_heap_records[heap_record_index];
	}
	used += req_pages;
	if ( needed_max <= used ){
		needed_max = used;
		profile_heap_needed_max_record = profile_heap_records[heap_record_index];
	}
	heap_record_index++;

	up(&heap_record_index_sem);
	return ;
}
EXPORT_SYMBOL(profile_heap_record);

extern unsigned long nr_try_to_free_pages;
extern unsigned long nr__alloc_pages_oka;
static int heap_record_to_string(profile_heap_record_t *record, char *buf, int len)
{
	unsigned long nanosec_rem;
	unsigned long long t;
	unsigned long used;

	t = record->time;
	nanosec_rem = do_div(t, 1000000000);

	used = totalram_pages - record->nr_free_pages - record->nr_cache_pages;

	return snprintf(buf,len,
			/* time        used   cache  free  used+req  req    pid     type */
			"%5lu.%06lu,%lu," "%lu"  ",%u" ",%lu"   ",%lu" ",%d"  ",%lu\n",
			(unsigned long)t,
			nanosec_rem/1000,
			used  << (PAGE_SHIFT - 10),
			record->nr_cache_pages << (PAGE_SHIFT - 10),
			record->nr_free_pages << (PAGE_SHIFT - 10),
			(record->req_pages + used) << (PAGE_SHIFT - 10),
			record->req_pages << (PAGE_SHIFT - 10),
			record->pid,
			record->type) ;
}

static unsigned int is_first_line = 1;
#define PROFILE_HEAP_HEADINGS "\ntime,used,cache,free,needed,req,pid,type\n"
#define HEADINGS_LEN sizeof(PROFILE_HEAP_HEADINGS)
#define MAX_NEEDED_STR "Max needed record\n"
#define MAX_NEEDED_STRLEN sizeof(MAX_NEEDED_STR)
#define MAX_USED_STR "\nMax used record\n"
#define MAX_USED_STRLEN sizeof(MAX_USED_STR)
static int profile_heap_proc_read(char *page, char **start, off_t off, int count,
				  int *eof, void *data)
{
	unsigned long tlen=0;
	profile_heap_record_t *record;
	
	if(down_interruptible(&heap_record_index_sem))
		return -EINTR;

	if (!is_heap_index_wrapped)
	{
		if(off >= heap_record_index){
			*eof = 1;
			is_first_line = 1;
			goto out;
		}
		record = &profile_heap_records[off];
	}
	else
	{
		unsigned adjust = off + record_index;
		if (adjust >= CONFIG_PROFILE_HEAP_RECORDS_NUM)
			adjust -= CONFIG_PROFILE_HEAP_RECORDS_NUM;
		if(off >= CONFIG_PROFILE_HEAP_RECORDS_NUM)
		{
			*eof = 1;
			is_first_line = 1;
			goto out;
		}
		record = &profile_heap_records[adjust];
	}
	if ( is_first_line ){
		strncpy(page,MAX_NEEDED_STR,MAX_NEEDED_STRLEN);
		tlen += MAX_NEEDED_STRLEN;
		tlen += heap_record_to_string(&profile_heap_needed_max_record,page+tlen, PAGE_SIZE-tlen);
		strncpy(page+tlen,MAX_USED_STR,MAX_USED_STRLEN);
		tlen += MAX_USED_STRLEN;
		tlen += heap_record_to_string(&profile_heap_used_max_record,page+tlen, PAGE_SIZE-tlen);
		strncpy(page+tlen, PROFILE_HEAP_HEADINGS, HEADINGS_LEN );
		tlen += HEADINGS_LEN;
		is_first_line = 0;
	}

	tlen += heap_record_to_string(record, page+tlen, PAGE_SIZE-tlen);
	*start = (char *)1;
 out:
	up(&heap_record_index_sem);
	return tlen>count?0:tlen;
}


static int __init debug_heap_proc_profile(void) 
{
	struct proc_dir_entry *ent;
	ent = create_proc_entry("profile_heap", S_IFREG|S_IRUGO|S_IWUSR, NULL);
	if(!ent) printk(KERN_DEBUG "create profile_heap proc entry failed\n");
	ent->read_proc = profile_heap_proc_read;
	return 0;
}
late_initcall(debug_heap_proc_profile);
#endif
