/*
 * drivers/serial/amba2-dma.c
 *
 * CXD4105 UART DMA driver
 *
 * Copyright 2005,2006 Sony Corporation
 */
/*
 *  linux/drivers/char/amba.c
 *
 *  Driver for AMBA serial ports
 *
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 *  Copyright 1999 ARM Limited
 *  Copyright (C) 2000 Deep Blue Solutions Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: amba.c,v 1.41 2002/07/28 10:03:27 rmk Exp $
 *
 * This is a generic driver for ARM AMBA-type serial ports.  They
 * have a lot of 16550-like features, but are not register compatible.
 * Note that although they do have CTS, DCD and DSR inputs, they do
 * not have an RI input, nor do they have DTR or RTS outputs.  If
 * required, these have to be supplied via some other means (eg, GPIO)
 * and hooked into this driver.
 */
#include <linux/config.h>

#if defined(CONFIG_SERIAL_AMBA2_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_core.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/device.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>

#include <asm/page.h>
#include <asm/arch/time.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#ifdef CONFIG_ARCH_CXD4105
#include <asm/arch/regs-dmac.h>
#endif /* CONFIG_ARCH_CXD4105 */
#include <asm/hardware/amba_serial.h>
#include <asm/hardware/serial_amba2.h>
#include <asm/arch/uart.h>
#include "amba2.h"

#ifdef CONFIG_SERIAL_AMBA2_TXDMA
#define DMAC_CTRL_VAL_32   0x76099000 /* dst=8bit/16 burst, src=32bit/4 burst */
#define DMAC_CTRL_VAL_8    0x7601b000 /* dst=8bit/16 burst, src=8bit/16 burst */
#define DMAC_CFG_VAL       0x0100c801 /* burst,Int,start */

#endif /* CONFIG_SERIAL_AMBA2_TXDMA */

#ifdef CONFIG_SERIAL_AMBA2_RXDMA
#define N_DMABUF  7

#define DMAC_CTRL_VAL_32RX 0x7940b000 /* dst=32bit/4 burst, src=8bit/16 burst */
#define DMAC_CFG_VAL_RX    0x0100d001 /* burst,Int,start */

/* RX timer */
/* Timer definitions */
#ifdef CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER
# define RXDMA_TIMER_PERIOD  1 /* jiffies */
#else /* CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER */

#ifdef CONFIG_ARCH_CXD4105
# define TIMER_VA  /* TBD */
# define TIMER_IRQ /* TBD */
#endif

#define TIMER_FREQ 1280000 /* HZ */
#define POLL_FREQ     1000 /* HZ */

#endif /* CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER */

#endif /* CONFIG_SERIAL_AMBA2_RXDMA */


#ifdef CONFIG_SERIAL_AMBA2_TXDMA

/*
 *        UART(TX) DMA section
 */

/* Build DMA list and start DMA */
static void dma_tx(struct uart_amba_port *uap)
{
	unsigned long flags;
	struct circ_buf *xmit = &uap->port.info->xmit;
	int ch;
	unsigned int len, n, ctrl, *ll;
	dma_addr_t phys, llphys;

	/*--- Disable interrupt ---*/
	local_irq_save(flags);
	if (uap->tx_dma_state != UART_DMA_IDLE){
		local_irq_restore(flags);
		return;
	}
	uap->tx_dma_state = UART_DMA_BUSY;
	local_irq_restore(flags);
	/*-------------------------*/

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&uap->port);

	len = CIRC_CNT_TO_END(xmit->head, xmit->tail, UART_XMIT_SIZE);
	if (len == 0){
		uap->tx_dma_state = UART_DMA_IDLE;
		return;
	}

	ch = uap->info->tx_dma_ch;
	ll = uap->tx_dma_ll;
	llphys = uap->tx_dma_llphys;
	uap->tx_dma_len = len;

	phys = uap->tx_xmit_phys + xmit->tail;
	writel(phys, DMACCSRCADDR_(ch));
	writel(uap->port.mapbase, DMACCDESTADDR_(ch));
	writel(0, DMACCLLI_(ch));
	if ((phys & 0x3) || (len < 4)){ /* not 32bit aligned  or  fragment */
		n = 4 - (phys & 0x3);
		if (n > len)
			n = len;
		ctrl = DMAC_CTRL_VAL_8 | n;
	} else { /* 32bit aligned */
		n = len >> 2;
		ctrl = DMAC_CTRL_VAL_32 | n;
		n <<= 2;
	}
	phys += n;
	len -= n;
	if (len == 0) /* last ? */
		ctrl |= DMAC_TC_ENABLE;
	writel(ctrl, DMACCCONTROL_(ch));

	if (len > 0){
		writel(llphys, DMACCLLI_(ch));
		n = len >> 2;
		if (n > 0){
			ll[0] = phys;
			ll[1] = uap->port.mapbase;
			ll[2] = 0;
			ll[3] = DMAC_CTRL_VAL_32|n;
			n <<= 2;
			phys += n;
			len -= n;
			if (len == 0){ /* last ? */
				ll[3] |= DMAC_TC_ENABLE;
			} else { /* next */
				ll[2] = llphys + 4 * sizeof (int);
			}
			ll += 4;
			llphys += 4 * sizeof (int);
		}
		if (len > 0){ /* fragment */
			ll[0] = phys;
			ll[1] = uap->port.mapbase;
			ll[2] = 0;
			ll[3] = DMAC_TC_ENABLE|DMAC_CTRL_VAL_8|len;
		}
	}
        writel(DMAC_CFG_VAL|uap->info->tx_dma_dev, DMACCCONFIGURATION_(ch)); /* start */
}

static irqreturn_t on_txdmac_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_amba_port *uap = dev_id;
	struct circ_buf *xmit = &uap->port.info->xmit;
	unsigned int err;

	err = readl(DMACINTERRORSTATUS);
	if (err != 0){
		printk(KERN_ERR "UART:on_txdma_intr: err=0x%x\n", err);
		writel(1 << DMA_IRQ_TO_CH(irq), DMACINTERRCLR);
	}
	writel(1 << DMA_IRQ_TO_CH(irq), DMACINTTCCLR); /* INT clear */

	xmit->tail = (xmit->tail + uap->tx_dma_len) & (UART_XMIT_SIZE - 1);
	uap->port.icount.tx += uap->tx_dma_len;
	uap->tx_dma_len = 0;
	uap->tx_dma_state = UART_DMA_IDLE;

	dma_tx(uap);

	return IRQ_HANDLED;
}

void amba_dmauart_start_tx(struct uart_amba_port *uap)
{
	unsigned long flags;
	unsigned int dmacr;

	dma_tx(uap);

	local_irq_save(flags);
	dmacr = UART_GET_DMACR(&uap->port);
	dmacr |= UART011_TXDMAE;
	UART_PUT_DMACR(&uap->port, dmacr);
	local_irq_restore(flags);
}

void amba_dmauart_stop_tx(struct uart_amba_port *uap)
{
	if (uap->info->tx_use_dma){
		unsigned long flags;
		unsigned int dmacr;

		local_irq_save(flags);
		dmacr = UART_GET_DMACR(&uap->port);
		dmacr &= ~UART011_TXDMAE;
		UART_PUT_DMACR(&uap->port, dmacr);
		local_irq_restore(flags);
	}
}

void amba_dmauart_continue_txdma(struct uart_amba_port *uap)
{
	unsigned int conf;

	if (uap->info->tx_use_dma){
		if (uap->tx_dma_state == UART_DMA_HALT){
			conf = readl(DMACCCONFIGURATION_(uap->info->tx_dma_ch));
			if (conf & DMACCCONFIG_ENABLE){
				/* restart */
				conf &= ~DMACCCONFIG_HALT;
				writel(conf, DMACCCONFIGURATION_(uap->info->tx_dma_ch));
			}
		}
	}
}

static void dmac_halt(struct uart_amba_port *uap, int ch)
{
	unsigned int conf;
	unsigned long flags;

	local_irq_save(flags);
	if (uap->tx_dma_state == UART_DMA_BUSY){
		/* halt DMA */
		conf = readl(DMACCCONFIGURATION_(ch));
		if (conf & DMACCCONFIG_ENABLE){
			conf |= DMACCCONFIG_HALT;
			writel(conf, DMACCCONFIGURATION_(ch));
			conf = readl(DMACCCONFIGURATION_(ch));
			if (conf & DMACCCONFIG_ENABLE)
				uap->tx_dma_state = UART_DMA_HALT;
		}
	}
	local_irq_restore(flags);
}

void amba_dmauart_send_xchar(struct uart_port *port, char ch)
{
	struct uart_amba_port *uap = (struct uart_amba_port *)port;
	unsigned long flags;

	port->x_char = ch;
	if (ch){
		spin_lock_irqsave(&port->lock, flags);

		if (uap->info->tx_use_dma){
#ifdef DEBUG
			printk(KERN_ERR "UART(DEBUG):send_xchar\n");
#endif
			dmac_halt(uap, uap->info->tx_dma_ch);
		}
		ambauart_start_tx(uap);

		spin_unlock_irqrestore(&port->lock, flags);
	}
}

static void tx_dma_stop(struct uart_amba_port *uap)
{
	unsigned long flags;

	local_irq_save(flags);
	dmac_halt(uap, uap->info->tx_dma_ch);
	amba_dmauart_stop_tx(uap);
	writel(0, DMACCCONFIGURATION_(uap->info->tx_dma_ch));
	writel(1 << uap->info->tx_dma_ch, DMACINTTCCLR);
	writel(1 << uap->info->tx_dma_ch, DMACINTERRCLR);
	local_irq_restore(flags);
}
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */


#ifdef CONFIG_SERIAL_AMBA2_RXDMA

/*
 *        UART(RX) DMA section
 */

static void uart_rxdma_enable(struct uart_amba_port *uap)
{
	unsigned long flags;
	unsigned int dmacr;

	local_irq_save(flags);
	dmacr = UART_GET_DMACR(&uap->port);
	dmacr |= UART011_RXDMAE|UART011_DMAONERR;
	UART_PUT_DMACR(&uap->port, dmacr);
	local_irq_restore(flags);
}

static void uart_rxdma_disable(struct uart_amba_port *uap)
{
	unsigned long flags;
	unsigned int dmacr;

	local_irq_save(flags);
	dmacr = UART_GET_DMACR(&uap->port);
	dmacr &= ~(UART011_RXDMAE|UART011_DMAONERR);
	UART_PUT_DMACR(&uap->port, dmacr);
	local_irq_restore(flags);
}

static void rxdmac_start(struct uart_amba_port *uap)
{
	unsigned long flags;
	int ch;

	/* Start DMAC */
	local_irq_save(flags);
	ch = uap->info->rx_dma_ch;
	writel(uap->port.mapbase, DMACCSRCADDR_(ch));
	writel(DMABUF_PUT_PHYS(&uap->rx_dmabuf), DMACCDESTADDR_(ch));
	writel(0, DMACCLLI_(ch));
	writel(DMAC_TC_ENABLE|DMAC_CTRL_VAL_32RX|TTY_FLIPBUF_SIZE, DMACCCONTROL_(ch));
        writel(DMAC_CFG_VAL_RX|uap->info->rx_dma_dev, DMACCCONFIGURATION_(ch)); /* start */

	uart_rxdma_enable(uap);
	uap->rx_dma_state = UART_DMA_BUSY;
	local_irq_restore(flags);
}

#define MAX_DMAC_DELAY 5 /* us */
#define MAX_HALT_WAIT  100
static int rxdmac_stop(struct uart_amba_port *uap)
{
	unsigned long flags;
	unsigned int tmp;
	int n;

	local_irq_save(flags);
	if (!(readl(DMACCCONFIGURATION_(uap->info->rx_dma_ch)) & DMACCCONFIG_ENABLE)){
		local_irq_restore(flags);
		return -1;
	}
	uap->rx_dma_state = UART_DMA_STOP_WAIT;
	/* Stop DMAC */
	tmp = readl(DMACCCONFIGURATION_(uap->info->rx_dma_ch)) & 0x0100fbdf;
	tmp |= DMACCCONFIG_HALT;
	writel(tmp, DMACCCONFIGURATION_(uap->info->rx_dma_ch)); /* HALT */
	local_irq_restore(flags);

#ifdef CONFIG_ARCH_CXD4105
	tmp = readl(DMACCCONFIGURATION_(uap->info->rx_dma_ch)); /* dummy read to wait 5 HCLK */
#else
	udelay(MAX_DMAC_DELAY);
#endif
	n = MAX_HALT_WAIT;
	while (readl(DMACCCONFIGURATION_(uap->info->rx_dma_ch)) & DMACCCONFIG_ACTIVE){
		if (--n == 0)
			break;
	}
	if (n == 0){
		printk(KERN_ERR "UART:rxdmac_stop(): HALT wait timeout\n");
	}
	writel(0, DMACCCONFIGURATION_(uap->info->rx_dma_ch));  /* DMA disable */
	uap->rx_dma_state = UART_DMA_IDLE;

	uart_rxdma_disable(uap);
	return 0;
}

/* SA_INTERRUPT */
static irqreturn_t on_rxdmac_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_amba_port *uap = dev_id;
	unsigned int ch, stat, cnt, rsr;

	ch = uap->info->rx_dma_ch;

	stat = readl(DMACINTERRORSTATUS) & (1 << ch);
	if (stat){
		printk(KERN_ERR "UART: DMA ERROR\n");
		writel(stat, DMACINTERRCLR); /* clear error intr */
	}
	stat = readl(DMACINTTCSTATUS) & (1 << ch);
	if (stat == 0)
		return IRQ_HANDLED;
	writel(stat, DMACINTTCCLR); /* clear intr */

	if (uap->rx_dma_state != UART_DMA_BUSY)
		return IRQ_HANDLED;

	rsr = UART_GET_RSR(&uap->port);
	UART_PUT_RSR(&uap->port, 0); /* write any data to clear */
	uap->rx_dma_state = UART_DMA_IDLE;
	uart_rxdma_disable(uap);

	cnt = TTY_FLIPBUF_SIZE - (readl(DMACCCONTROL_(ch)) & 0xfff);
	DMABUF_PUT_SIZE(&uap->rx_dmabuf) = cnt;
	DMABUF_PUT_RSR(&uap->rx_dmabuf) = rsr;
	uap->port.icount.rx += cnt;
	RINGBUF_PUT(&uap->rx_dmabuf);
	if (RINGBUF_CNT(&uap->rx_dmabuf) < N_DMABUF)
		rxdmac_start(uap);
#ifdef DEBUG
	else
		printk(KERN_ERR "UART(DEBUG):on_rxdmac_intr: rx_dmabuf full\n");
#endif

	schedule_delayed_work(&uap->work, 1);
	return IRQ_HANDLED;
}

/*
 * Call from the beginning of UART interrupt handler
 */
unsigned int amba_dmauart_rxintr_enter(struct uart_amba_port *uap, int *handled)
{
	unsigned int cnt, status;

	cnt = 0;
	if (uap->info->rx_use_dma){
		status = UART_GET_INT_STATUS(&uap->port);
		if (status & (AMBAUART_ERRORS|UART011_RTIS)){
			rxdmac_stop(uap);
			cnt = TTY_FLIPBUF_SIZE - (readl(DMACCCONTROL_(uap->info->rx_dma_ch)) & 0xfff);
			DMABUF_PUT_SIZE(&uap->rx_dmabuf) = cnt;
			DMABUF_PUT_RSR(&uap->rx_dmabuf) = 0;
			uap->port.icount.rx += cnt;
			if (status & AMBAUART_ERRORS){
				DMABUF_PUT_RSR(&uap->rx_dmabuf) = UART_GET_RSR(&uap->port);
				UART_PUT_ICR(&uap->port, AMBAUART_ERRORS);
				*handled = 1;
			}
		}
	}
	return cnt;
}

/*
 * Call from the mid of UART interrupt handler
 */
unsigned int amba_dmauart_rx_chars(struct uart_amba_port *uap)
{
	unsigned int size, max_count, status, rsr;
	unsigned char *t, ch;

	size = DMABUF_PUT_SIZE(&uap->rx_dmabuf);
	max_count = sizeof (uap->rx_dmabuf.bufs[0].buf) - size;
	t = DMABUF_PUT_BUF(&uap->rx_dmabuf) + size;
	rsr = DMABUF_PUT_RSR(&uap->rx_dmabuf);

	status = UART_GET_FR(&uap->port);
	while (UART_RX_DATA(status) && max_count--){
		ch = UART_GET_CHAR(&uap->port);
		uap->port.icount.rx++;

		rsr |= UART_GET_RSR(&uap->port);
		*t++ = ch;
		size++;
		status = UART_GET_FR(&uap->port);
	}
	DMABUF_PUT_SIZE(&uap->rx_dmabuf) = size;
	DMABUF_PUT_RSR(&uap->rx_dmabuf) = rsr;
	return size;
}

/*
 * Call from the end of UART interrupt handler
 */
void amba_dmauart_rxintr_leave(struct uart_amba_port *uap, unsigned int cnt)
{
	unsigned long flags;

	if (uap->info->rx_use_dma){
		if (cnt > 0){
			RINGBUF_PUT(&uap->rx_dmabuf);
		}
		local_irq_save(flags);
		if (RINGBUF_CNT(&uap->rx_dmabuf) > 0)
			schedule_delayed_work(&uap->work, 1);
		if (RINGBUF_CNT(&uap->rx_dmabuf) < N_DMABUF
		    && uap->rx_dma_state == UART_DMA_IDLE)
			rxdmac_start(uap);
#ifdef DEBUG
		else
			printk(KERN_ERR "UART(DEBUG): rxintr_leave: rx_dmabuf full\n");
#endif
		local_irq_restore(flags);
	}
}

/* WORK */
static void do_flip(void *private_)
{
	struct uart_amba_port *uap = (struct uart_amba_port *)private_;
	struct tty_struct *tty = uap->port.info->tty;
	struct dmabuf *s;
	int size, i, n;
	unsigned int getp, flag;
	unsigned long flags;

	if (tty == NULL)
		return;

	if (RINGBUF_CNT(&uap->rx_dmabuf) == 0){
#ifdef DEBUG
		printk(KERN_ERR "UART(DEBUG):do_flip: no data\n");
#endif
		return;
	}
	while (RINGBUF_CNT(&uap->rx_dmabuf) > 0  &&  tty->flip.count == 0){
		/* try to merge */
		size = 0;
		getp = uap->rx_dmabuf.get;
		n = RINGBUF_CNT(&uap->rx_dmabuf);
		for (i = 0; i < n; i++){
			s = &uap->rx_dmabuf.bufs[getp];
			if (size + s->size > TTY_FLIPBUF_SIZE)
				break;
#ifdef DEBUG
			if (i > 0)
				printk(KERN_ERR "UART(DEBUG): merge %d bytes\n", s->size);
#endif
			memcpy(tty->flip.char_buf_ptr, s->buf, s->size);
			memset(tty->flip.flag_buf_ptr, TTY_NORMAL, s->size);
			/* process RSR */
			flag = TTY_NORMAL;
			if (s->rsr & UART01x_RSR_BE) {
				s->rsr &= ~(UART01x_RSR_FE | UART01x_RSR_PE);
				uap->port.icount.brk++;
				uart_handle_break(&uap->port);
			} else if (s->rsr & UART01x_RSR_PE)
				uap->port.icount.parity++;
			else if (s->rsr & UART01x_RSR_FE)
				uap->port.icount.frame++;
			if (s->rsr & UART01x_RSR_OE)
				uap->port.icount.overrun++;

			s->rsr &= uap->port.read_status_mask;
			if (s->rsr & UART01x_RSR_BE)
				flag = TTY_BREAK;
			else if (s->rsr & UART01x_RSR_PE)
				flag = TTY_PARITY;
			else if (s->rsr & UART01x_RSR_FE)
				flag = TTY_FRAME;
			tty->flip.flag_buf_ptr[s->size - 1] = flag;

			tty->flip.char_buf_ptr += s->size;
			tty->flip.flag_buf_ptr += s->size;
			size += s->size;

			if ((s->rsr & UART01x_RSR_OE) &&  size < TTY_FLIPBUF_SIZE){
				*tty->flip.char_buf_ptr++ = 0;
				*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
				size++;
			}

			getp = getp + 1;
			if (getp >= uap->rx_dmabuf.n_buf)
				getp = 0;
		}
		tty->flip.count = size;
		tty_flip_buffer_push(tty);

		/*--- Disable interrupt ---*/
		local_irq_save(flags);
		uap->rx_dmabuf.get = getp;
		uap->rx_dmabuf.cnt -= i;
		if (uap->rx_dma_state == UART_DMA_IDLE)
			rxdmac_start(uap);
		local_irq_restore(flags);
		/*-------------------------*/
	}
	if (RINGBUF_CNT(&uap->rx_dmabuf) > 0  &&  tty->flip.count != 0){
		schedule_delayed_work(&uap->work, 1);
	}
}

/*----- RX gap detect timer -----*/
static void rxdma_timer_proc(struct uart_amba_port *uap)
{
	unsigned long flags;
	unsigned int status, cnt;

	local_irq_save(flags);
	if (uap->rx_dma_state == UART_DMA_IDLE)
		goto out;
	status = UART_GET_INT_STATUS(&uap->port);
	if (status & (AMBAUART_ERRORS|UART011_RTIS)) /* RX intr pending ? */
		goto out;
	status = UART_GET_FR(&uap->port);
	if (UART_RX_DATA(status))  /* FIFO is not empty ? */
		goto out;
	cnt = readl(DMACCCONTROL_(uap->info->rx_dma_ch)) & 0xfff;
	if (cnt == TTY_FLIPBUF_SIZE  ||  cnt == 0)
		goto out;
	rxdmac_stop(uap);

	status = readl(DMACINTTCSTATUS) & (1 << uap->info->rx_dma_ch);

	if (status) /* DMA intr pending ? */
		goto out;

	cnt = TTY_FLIPBUF_SIZE - (readl(DMACCCONTROL_(uap->info->rx_dma_ch)) & 0xfff);
	DMABUF_PUT_SIZE(&uap->rx_dmabuf) = cnt;
	uap->port.icount.rx += cnt;
	RINGBUF_PUT(&uap->rx_dmabuf);
	if (RINGBUF_CNT(&uap->rx_dmabuf) < N_DMABUF)
		rxdmac_start(uap);

	local_irq_restore(flags);
	schedule_delayed_work(&uap->work, 1);
#ifdef DEBUG
	printk(KERN_ERR "UART(DEBUG) port%d:flip by timer\n", uap->port.line);
#endif
	return;

out:
	local_irq_restore(flags);
}

#ifdef CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER
static void on_rxdma_timer(unsigned long arg)
{
	struct uart_amba_port *uap = (struct uart_amba_port *)arg;

	rxdma_timer_proc(uap);
	mod_timer(&uap->rx_dma_timer, jiffies + RXDMA_TIMER_PERIOD);
}
#else /* CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER */
static DEFINE_SPINLOCK(timer_lock);
static unsigned int timer_flag = 0;

static irqreturn_t timer_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	volatile TimerStruct_t *timer = dev_id;
	int i;
	unsigned int mask;
	struct uart_amba_port *uap;
	extern struct uart_amba_port amba2_ports[];

	spin_lock(&timer_lock);

	for (i = 0, mask = 0x1; i < UART_NR; i++, mask <<= 1){
		if (timer_flag & mask){
			uap = &amba2_ports[i];
			rxdma_timer_proc(uap);
		}
	}
	timer->TimerClear = 1;

	spin_unlock(&timer_lock);

	return IRQ_HANDLED;
}

static void timer_start(struct uart_amba_port *uap)
{
	unsigned long flags;

	volatile TimerStruct_t *timer = (volatile TimerStruct_t *)TIMER_VA;

	spin_lock_irqsave(&timer_lock, flags);
	if (timer_flag == 0){
#ifdef DEBUG
		printk(KERN_ERR "UART(DEBUG):TIMER start\n");
#endif
		timer->TimerControl = TM1CKDIV1 | TM1CSPERIODIC | TM1INT | TM1ST;
	}
	timer_flag |= 1 << uap->port.line;
	spin_unlock_irqrestore(&timer_lock, flags);
}

static void timer_stop(struct uart_amba_port *uap)
{
	unsigned long flags;
	volatile TimerStruct_t *timer = (volatile TimerStruct_t *)TIMER_VA;

	spin_lock_irqsave(&timer_lock, flags);
	timer_flag &= ~(1 << uap->port.line);
	if (timer_flag == 0){
		timer->TimerControl = TM1CKDIV1 | TM1CSPERIODIC | TM1INT;
#ifdef DEBUG
		printk(KERN_ERR "UART(DEBUG):TIMER stopped\n");
#endif
	}
	spin_unlock_irqrestore(&timer_lock, flags);
}
#endif /* CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER */

static void rx_dma_stop(struct uart_amba_port *uap)
{
	unsigned long flags;

	local_irq_save(flags);
	uart_rxdma_disable(uap);
	writel(0, DMACCCONFIGURATION_(uap->info->rx_dma_ch));

	writel(1 << uap->info->rx_dma_ch, DMACINTTCCLR);
	writel(1 << uap->info->rx_dma_ch, DMACINTERRCLR);

	local_irq_restore(flags);
}
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */


void amba_dmauart_shutdown(struct uart_amba_port *uap)
{
#ifdef CONFIG_SERIAL_AMBA2_TXDMA
	if (uap->info->tx_use_dma){
		tx_dma_stop(uap);
		free_irq(DMA_CH_TO_IRQ(uap->info->tx_dma_ch), uap);
	}
#endif
#ifdef CONFIG_SERIAL_AMBA2_RXDMA
	if (uap->info->rx_use_dma){
#ifdef CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER
		del_timer(&uap->rx_dma_timer);
#else
		timer_stop(uap);
#endif
		rx_dma_stop(uap);
		cancel_delayed_work(&uap->work);
		free_irq(DMA_CH_TO_IRQ(uap->info->rx_dma_ch), uap);
	}
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
}

void amba_dmauart_startup(struct uart_amba_port *uap)
{
#ifdef CONFIG_SERIAL_AMBA2_DMA
	int retval;
#endif /* CONFIG_SERIAL_AMBA2_DMA */
#ifdef CONFIG_SERIAL_AMBA2_TXDMA
	if (uap->info->tx_use_dma){
		uap->tx_dma_state = UART_DMA_IDLE;
		retval = request_irq(DMA_CH_TO_IRQ(uap->info->tx_dma_ch), on_txdmac_intr, 0,
				     "uart-txdmac", uap);
		if (retval){
			printk(KERN_ERR "uart port%d: can not grab TX-DMAC IRQ."
			       "Fallback to non TX-DMA mode\n", uap->port.line);
			uap->info->tx_use_dma = 0;
		}
	}
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */
#ifdef CONFIG_SERIAL_AMBA2_RXDMA
	if (uap->info->rx_use_dma){
		uap->rx_dma_state = UART_DMA_IDLE;
		retval = request_irq(DMA_CH_TO_IRQ(uap->info->rx_dma_ch), on_rxdmac_intr, SA_INTERRUPT,
				     "uart-rxdmac", uap);
		if (retval){
			printk(KERN_ERR "uart port%d: can not grab RX-DMAC IRQ."
			       "Fallback to non RX-DMA mode\n", uap->port.line);
			uap->info->rx_use_dma = 0;
		}

		if (uap->info->rx_use_dma){
			unsigned int tmp;
			unsigned long flags;

			/* mask single-DMA request */
			local_irq_save(flags);

			tmp = readl(DMACSREQMASK);
			tmp &= ~(1 << (uap->info->rx_dma_dev >> DMACCCONFIG_SRCDEV_SHIFT));
			writel(tmp, DMACSREQMASK);

			local_irq_restore(flags);

			if (uap->info->rx_timeout > 0  &&  uap->info->rx_timeout <= 65536){
				UART_PUT_RTO(&uap->port, UART_RTO_CHAR(uap->info->rx_timeout));
			}
			rxdmac_start(uap);
#ifdef CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER
			init_timer(&uap->rx_dma_timer);
			uap->rx_dma_timer.data = (unsigned long)uap;
			uap->rx_dma_timer.function = on_rxdma_timer;
			mod_timer(&uap->rx_dma_timer, jiffies + RXDMA_TIMER_PERIOD);
#else
			timer_start(uap);
#endif
		}
	}
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
}

#ifdef CONFIG_SERIAL_AMBA2_TXDMA
void *amba_dmauart_alloc_xmitbuf(struct uart_port *port)
{
	struct uart_amba_port *uap = (struct uart_amba_port *)port;
	void *buf;
	dma_addr_t phys;

	buf = NULL;
	if (uap->info->tx_use_dma){
		buf = dma_alloc_coherent(NULL, PAGE_SIZE, &phys, GFP_DMA);
		if (buf == NULL){
			printk(KERN_ERR "uart port%d: can not alloc XMIT buffer."
			       "Fallback to non DMA mode\n", uap->port.line);
			uap->info->tx_use_dma = 0;
		}
		uap->tx_xmitbuf = buf;
		uap->tx_xmit_phys = phys;
#ifdef DEBUG
		printk(KERN_ERR "UART(DEBUG): xmit.buf = 0x%x (phys=0x%x)\n", buf, phys);
#endif
	}
	return buf;
}

int amba_dmauart_free_xmitbuf(struct uart_port *port)
{
	struct uart_amba_port *uap = (struct uart_amba_port *)port;

	if (uap->info->tx_use_dma  &&  uap->tx_xmitbuf){
		dma_free_coherent(NULL, PAGE_SIZE, uap->tx_xmitbuf, uap->tx_xmit_phys);
		return 1;
	}
	return 0;
}
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */

void amba_dmauart_port_init(struct uart_amba_port *uap)
{
	unsigned char *buf;
	unsigned int phys;

	if (!uap->info->tx_use_dma  &&  !uap->info->rx_use_dma){
		uap->dmabuf = NULL;
		return;
	}
	buf = dma_alloc_coherent(NULL, PAGE_SIZE, &phys, GFP_DMA);
	if (buf == NULL){
		printk(KERN_ERR "uart port%d: can not alloc DMA buffer."
		       "Fallback to non DMA mode\n", uap->port.line);
		uap->info->tx_use_dma = 0;
		uap->info->rx_use_dma = 0;
	}
	uap->dmabuf = buf;
	uap->dma_phys = phys;
#ifdef CONFIG_SERIAL_AMBA2_TXDMA
	if (uap->info->tx_use_dma){
		uap->tx_dma_state = UART_DMA_IDLE;
		uap->tx_dma_ll = (unsigned int *)buf;
		uap->tx_dma_llphys = phys;
		buf += sizeof (unsigned int) * 4 * 2; /* 2 linked list, 4 elements each */
		phys += sizeof (unsigned int) * 4 * 2;
		/* at this point, buf and phys must be 4 byte aligned. */
	}
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */
#ifdef CONFIG_SERIAL_AMBA2_RXDMA
	if (uap->info->rx_use_dma){
		int i;
		uap->rx_dma_state = UART_DMA_IDLE;
		INIT_WORK(&uap->work, do_flip, uap);
		/* init RX DMA buffer */
		uap->rx_dmabuf.bufs = (struct dmabuf *)buf;
		for (i = 0; i < N_DMABUF; i++){
			unsigned int off;
			off = uap->rx_dmabuf.bufs[i].buf - buf;
			uap->rx_dmabuf.bufs[i].phys = phys + off;
			uap->rx_dmabuf.bufs[i].rsr = 0;
			uap->rx_dmabuf.bufs[i].size = 0;
		}
		uap->rx_dmabuf.n_buf = N_DMABUF;
		uap->rx_dmabuf.put = 0;
		uap->rx_dmabuf.get = 0;
		uap->rx_dmabuf.cnt = 0;
	}
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
}

void amba_dmauart_port_remove(struct uart_amba_port *uap)
{
	if (uap->dmabuf){
		dma_free_coherent(NULL, PAGE_SIZE, uap->dmabuf, uap->dma_phys);
		uap->dmabuf = NULL;
	}
}

int __init amba_dmauart_init(void)
{
	int ret = 0;
#ifdef CONFIG_SERIAL_AMBA2_RXDMA
#ifndef CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER
	volatile TimerStruct_t *timer = (volatile TimerStruct_t *)TIMER_VA;

	ret = request_irq(TIMER_IRQ, timer_intr, 0, "uart-timer", (void *)timer);
	if (ret)
		return ret;

	timer->TimerControl = 0;
	timer->TimerClear = 0x11; /* clear counter and irq */
	timer->TimerControl = TM1CKDIV1 | TM1CSPERIODIC | TM1INT;
	timer->TimerCompare = TIMER_FREQ/POLL_FREQ;
#endif /* !CONFIG_SERIAL_AMBA2_RXDMA_OSTIMER */
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */

	return ret;
}
