/*
 * drivers/serial/amba2.h
 *
 * CXD4105 UART and DMA ring buffer access macros
 *
 * Copyright 2005,2006 Sony Corporation
 *
 * This code is based on drivers/serial/amba-pl010.c
 */
/*
 *  linux/drivers/char/amba.c
 *
 *  Driver for AMBA serial ports
 *
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 *  Copyright 1999 ARM Limited
 *  Copyright (C) 2000 Deep Blue Solutions Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: amba.c,v 1.41 2002/07/28 10:03:27 rmk Exp $
 *
 * This is a generic driver for ARM AMBA-type serial ports.  They
 * have a lot of 16550-like features, but are not register compatible.
 * Note that although they do have CTS, DCD and DSR inputs, they do
 * not have an RI input, nor do they have DTR or RTS outputs.  If
 * required, these have to be supplied via some other means (eg, GPIO)
 * and hooked into this driver.
 */
#ifndef __AMBA2_UART_H__
#define __AMBA2_UART_H__

#undef DEBUG

/*
 * Access macros for the AMBA UARTs
 */
#define UART_GET_INT_STATUS(p)	readl((p)->membase + UART011_MIS)
#define UART_PUT_ICR(p, c)	writel((c), (p)->membase + UART011_ICR)
#define UART_GET_FR(p)		readl((p)->membase + UART01x_FR)
/* UARTDR bit 8-11 is FE,PE,BE,OE */
#define UART_GET_CHAR(p)	(readl((p)->membase + UART01x_DR) & 0xff)
#define UART_PUT_CHAR(p, c)	writel((c), (p)->membase + UART01x_DR)
#define UART_GET_RSR(p)		readl((p)->membase + UART01x_RSR)
#define UART_PUT_RSR(p, c)	writel((c), (p)->membase + UART01x_RSR)
#define UART_GET_CR(p)		readl((p)->membase + UART011_CR)
#define UART_PUT_CR(p,c)	writel((c), (p)->membase + UART011_CR)
#define UART_GET_IBRD(p)	readl((p)->membase + UART011_IBRD)
#define UART_PUT_IBRD(p,c)	writel((c), (p)->membase + UART011_IBRD)
#define UART_GET_FBRD(p)	readl((p)->membase + UART011_FBRD)
#define UART_PUT_FBRD(p,c)	writel((c), (p)->membase + UART011_FBRD)
#define UART_GET_LCRH(p)	readl((p)->membase + UART011_LCRH)
#define UART_PUT_LCRH(p,c)	writel((c), (p)->membase + UART011_LCRH)
#define UART_GET_IMSC(p)	readl((p)->membase + UART011_IMSC)
#define UART_PUT_IMSC(p,c)	writel((c), (p)->membase + UART011_IMSC)
#define UART_GET_IFLS(p)	readl((p)->membase + UART011_IFLS)
#define UART_PUT_IFLS(p,c)	writel((c), (p)->membase + UART011_IFLS)
#define UART_GET_DMACR(p)	readl((p)->membase + UART011_DMACR)
#define UART_PUT_DMACR(p,c)	writel((c), (p)->membase + UART011_DMACR)
#define UART_GET_RTO(p)		readl((p)->membase + AMBAUART_RTO)
#define UART_PUT_RTO(p,c)	writel((c), (p)->membase + AMBAUART_RTO)

#define UART_RX_DATA(s)		(((s) & UART01x_FR_RXFE) == 0)
#define UART_TX_READY(s)	(((s) & UART01x_FR_TXFF) == 0)

#define AMBAUART_ERRORS (UART011_OEIM|UART011_BEIM|UART011_PEIM|UART011_FEIM)

#ifdef CONFIG_SERIAL_AMBA2_RXDMA
/* RX-DMA buffer */
struct dmabuf {
	unsigned char buf[TTY_FLIPBUF_SIZE];
	unsigned int phys;
	int size;
	unsigned int rsr;
};

struct dmaringbuf {
	struct dmabuf *bufs;
	unsigned int put;
	unsigned int get;
	unsigned int n_buf;
	int cnt;
};
#define DMABUF_GET(p)		((p)->bufs + (p)->get)
#define DMABUF_PUT_SIZE(p)	((p)->bufs[(p)->put].size)
#define DMABUF_PUT_RSR(p)	((p)->bufs[(p)->put].rsr)
#define DMABUF_PUT_BUF(p)	((p)->bufs[(p)->put].buf)
#define DMABUF_PUT_PHYS(p)	((p)->bufs[(p)->put].phys)
#define RINGBUF_PUT(p)	{ 	(p)->put = (p)->put + 1;    \
				if ((p)->put >= (p)->n_buf) \
					(p)->put = 0;	    \
				(p)->cnt++; }
#define RINGBUF_GET(p)	{	(p)->get = (p)->get + 1;    \
				if ((p)->get >= (p)->n_buf) \
					(p)->get = 0;	    \
				(p)->cnt--; }
#define RINGBUF_CNT(p)		((p)->cnt)
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
/*
 * We wrap our port structure around the generic uart_port.
 */
struct uart_amba_port {
	struct uart_port	port;
	unsigned int		im;	/* interrupt mask */
	unsigned int		old_status;
	int			is_console;
	int			txrx_enable;

	struct uart_platform_data *info;
	/* DMA stuff */
#ifdef CONFIG_SERIAL_AMBA2_TXDMA
	int tx_dma_state;
	unsigned int tx_dma_len;
	unsigned int *tx_dma_ll; /* linked list for chain DMA */
	dma_addr_t tx_dma_llphys;
	char *tx_xmitbuf;
	dma_addr_t tx_xmit_phys;
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */
#ifdef CONFIG_SERIAL_AMBA2_RXDMA
	int rx_dma_state;
	struct dmaringbuf rx_dmabuf;
	struct timer_list rx_dma_timer;
	struct work_struct work;
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
#ifdef CONFIG_SERIAL_AMBA2_DMA
	char *dmabuf;
	unsigned int dma_phys;
#endif /* CONFIG_SERIAL_AMBA2_DMA */
};
/* DMA state */
#define UART_DMA_IDLE 0
#define UART_DMA_BUSY 1
#define UART_DMA_HALT 2
#define UART_DMA_STOP_WAIT 3

extern void ambauart_start_tx(struct uart_amba_port *);
#ifdef CONFIG_SERIAL_AMBA2_DMA
extern int amba_dmauart_init(void);
extern void amba_dmauart_port_init(struct uart_amba_port *);
extern void amba_dmauart_port_remove(struct uart_amba_port *);
extern void amba_dmauart_startup(struct uart_amba_port *);
extern void amba_dmauart_shutdown(struct uart_amba_port *);

#ifdef CONFIG_SERIAL_AMBA2_TXDMA
extern void amba_dmauart_send_xchar(struct uart_port *, char);
extern void *amba_dmauart_alloc_xmitbuf(struct uart_port *);
extern int amba_dmauart_free_xmitbuf(struct uart_port *);
extern void amba_dmauart_start_tx(struct uart_amba_port *);
extern void amba_dmauart_stop_tx(struct uart_amba_port *);
extern void amba_dmauart_continue_txdma(struct uart_amba_port *);
#endif /* CONFIG_SERIAL_AMBA2_TXDMA */

#ifdef CONFIG_SERIAL_AMBA2_RXDMA
extern unsigned int amba_dmauart_rxintr_enter(struct uart_amba_port *, int *);
extern void amba_dmauart_rxintr_leave(struct uart_amba_port *, unsigned int);
extern unsigned int amba_dmauart_rx_chars(struct uart_amba_port *);
#endif /* CONFIG_SERIAL_AMBA2_RXDMA */
#endif /* CONFIG_SERIAL_AMBA2_DMA */

#endif
