/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_bgadget.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This code is based on musb_bgadget.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/smp.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#ifdef MUSB_V26
#include <linux/moduleparam.h>
#include <linux/usb.h>
#include <linux/stat.h>
#include <hub.h>
#else
struct usb_tt {
        struct usb_device       *hub;   /* upstream highspeed hub */
        int                     multi;  /* true means one TT per port */
};
#endif

#ifdef MUSB_DMA
#include "usb_mcd_dma.h"
#define USB_REQ_DMA_MODE1 1
#define USB_REQ_DMA_MODE0 2
#define USB_REQ_CPU       3
#define USB_INIT_REQ_PHASE1       1
#define USB_INIT_REQ_PHASE2       2

#define DUMMY_USB_IOCTL_TESTMODE_NORMAL		/* ƥȥ⡼TESTMODE_NORMALϥߡ */
						/* TESTMODE_NORMALꤵƤʤˤ⤷ʤ */
#define MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER	/* 4byteʲΥ硼ȥѥåȤCPUž */
#define MUSB_DMA_FEWBYTE  4
#define USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
//#define MUSB_LOCK_DEBUG
#endif
#include <linux/init.h> /* mvl21 */
#include <linux/usb_ch9.h>
#include "usb_mcd_defs.h"
#include "usb_mcd_fifo.h"
#include "usb_mcd_gadgetdefs.h"
#include <linux/usb/mcd/usb_mcd_ioctl.h>

#ifndef MUSB_V26
#ifndef MUSB_LINUX_MV21
#include "hcd.h"
#endif
#endif


/**************************************************************************
Gadget Functions
**************************************************************************/

/* endpoint operations */
STATIC int MGC_GadgetEnableEnd(struct usb_ep *ep,
	const struct usb_endpoint_descriptor *desc);
STATIC int MGC_GadgetDisableEnd(struct usb_ep *ep);

STATIC struct usb_request* MGC_GadgetAllocRequest(struct usb_ep *ep,
	int gfp_flags);
STATIC void MGC_GadgetFreeRequest(struct usb_ep *ep, struct usb_request *req);
STATIC void* MGC_GadgetAllocBuffer(struct usb_ep *ep, unsigned bytes,
				   dma_addr_t *dma, int gfp_flags);
STATIC void MGC_GadgetFreeBuffer(struct usb_ep *ep, void *buf, dma_addr_t dma,
				 unsigned bytes);
STATIC int MGC_GadgetQueue(struct usb_ep *ep, struct usb_request *req,
	int gfp_flags);
STATIC int MGC_Call_RestartRequest(struct usb_ep *ep, struct usb_request *pRequest,
    int gfp_flags);
STATIC int MGC_GadgetDequeue(struct usb_ep *ep, struct usb_request *req);
int MGC_GadgetSetHalt(struct usb_ep *ep, int value);
STATIC int MGC_GadgetFifoStatus(struct usb_ep *ep);
STATIC void MGC_GadgetFifoFlush(struct usb_ep *ep);
STATIC void MGC_GadgetFifoFlush_FlushOnly(struct usb_ep *ep);

/* general operations */
STATIC int MGC_GadgetGetFrame(struct usb_gadget *gadget);
STATIC int MGC_GadgetWakeup(struct usb_gadget *gadget);
STATIC int MGC_GadgetSetSelfPowered(struct usb_gadget *gadget, 
				    int is_selfpowered);

#if defined(MUSB_OTG) && defined(GADGET_FULL)
STATIC int MGC_GadgetVbusSession(struct usb_gadget *gadget, int is_active);
STATIC int MGC_GadgetVbusDraw(struct usb_gadget *gadget, unsigned mA);
STATIC int MGC_GadgetPullup(struct usb_gadget *gadget, int is_on);
#endif

STATIC int MGC_GadgetIoctl(struct usb_gadget *gadget,
			   unsigned code, unsigned long param);

STATIC void MGC_GadgetDmaSetDefault(void);

static struct usb_ep_ops MGC_GadgetEndpointOperations =
{
	MGC_GadgetEnableEnd,
	MGC_GadgetDisableEnd,
	MGC_GadgetAllocRequest,
	MGC_GadgetFreeRequest,
	MGC_GadgetAllocBuffer,
	MGC_GadgetFreeBuffer,
	MGC_GadgetQueue,
	MGC_GadgetDequeue,
	MGC_GadgetSetHalt,
	MGC_GadgetFifoStatus,
	MGC_GadgetFifoFlush
};

static struct usb_gadget_ops MGC_GadgetOperations =
{
	MGC_GadgetGetFrame,
	MGC_GadgetWakeup,
	MGC_GadgetSetSelfPowered,
#if defined(MUSB_OTG) && defined(GADGET_FULL)
	MGC_GadgetVbusSession,
	MGC_GadgetVbusDraw,
	MGC_GadgetPullup,
#endif
	MGC_GadgetIoctl
};


#ifdef MUSB_DMA
void MGC_TxDmaCallback(MGC_DmaChannel* pChannel,
                                  uint8_t bEnd, 
                                  uint8_t bTransmit,
                                  uint8_t bProtocol,
				  void* pData);

void MGC_RxDmaCallback(MGC_DmaChannel* pChannel,
                                  uint8_t bEnd, 
                                  uint8_t bTransmit,
                                  uint8_t bProtocol,
				  void* pData);
#endif /* MUSB_DMA */

/**************************************************************************
Gadget End points
**************************************************************************/

MGC_GadgetLocalEnd MGC_aGadgetLocalEnd[MUSB_C_NUM_EPS]; /* counters and queues */

static struct {
	spinlock_t lock;
	struct list_head req_list;	
} SchedulerQueue;



static char aGadgetDefaultEndNames[MUSB_C_NUM_EPS][6]; /* the singleton */
static int iGadgetEndNamesLen=-1;
static char* aGadgetEndNames[MUSB_C_NUM_EPS];

#ifdef MUSB_GADGET_PARAMS
module_param_array(
	aGadgetEndNames, 
	charp, 
	iGadgetEndNamesLen, 
	S_IRUGO);
#endif


struct MGC_UsbRequest {
    struct usb_request request;
#if MUSB_DEBUG > 0
    uint32_t dwPadFront;
#endif
    uint8_t bEnd; /* will be used by the completition routine */
    struct list_head completition_list; /* will be used by the completition routine */
#if MUSB_DEBUG > 0
    uint8_t bQueued;
    uint32_t dwPadBack;
#endif
};

/**************************************************************************
Diagnostic
**************************************************************************/

inline int is_corrupted_request(void *req) {
    int rc=0;
    const struct MGC_UsbRequest* pRequest=(struct MGC_UsbRequest*)req;

    if ( !pRequest ) {
    	ERR("pRequest is NULL\n");	
	rc=1;
    }
    
#if MUSB_DEBUG >0 
    if ( pRequest->dwPadFront!=MGC_PAD_FRONT ) {
    	ERR("pRequest->dwPadFront corrupted in pRequest=%p\n", pRequest);	
	rc=1;
    } 
    if ( pRequest->dwPadBack != MGC_PAD_BACK ) {
    	ERR("pRequest->dwPadBack corrupted in pRequest=%p\n", pRequest);
	rc=1;
    }
#endif

     return rc; /* not corrupted */
}

#if MUSB_DEBUG > 0
#define is_queued_for_completition( _req ) (((struct MGC_UsbRequest*)(_req))->bQueued)
#else
#define is_queued_for_completition( _req ) 0
#endif

#if MUSB_DEBUG>0

#ifndef MUSB_LINUX_MV21
char *decode_request(struct usb_ctrlrequest *req)  {
    static char buf [8];
    sprintf(buf, "%s%s",( USB_TYPE_STANDARD==(req->bRequestType&USB_TYPE_MASK )?"S":"N" ),
    	is_tx_request(req) ? "TX" 
		: is_rx_request(req) ? "RX" 
		: is_zerodata_request(req) ? "Z" 
		: "-" );
    return buf;
}
#endif

inline char *decode_usb_ctrlrequest(const struct usb_ctrlrequest *pControlRequest) {
    static char buf[256];

    sprintf(buf, "pControlRequest=%p, bRequest=%02x, bmRequestType=%02x [%sstandard], wValue=%04x, wIndex=%04x, wLength=%04x\n", 
	pControlRequest, pControlRequest->bRequest, pControlRequest->bRequestType, 
	( USB_TYPE_STANDARD==(pControlRequest->bRequestType&USB_TYPE_MASK )) ? "":"no",
	pControlRequest->wValue, pControlRequest->wIndex, pControlRequest->wLength); 

    return buf;
}

#endif

static void
stop_dma(
    MGC_LinuxCd* pThis,
    uint8_t bEnd)
{
#ifdef MUSB_DMA
    uint16_t wCsrVal;
    MGC_DmaChannel* pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;

    if(pDmaChannel) { 
        pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
        pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
        pThis->pDmaController->pfDmaReleaseChannel(pDmaChannel);
        pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = NULL;

        MGC_SelectEnd(pBase, bEnd); 
        wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
        wCsrVal &= ~(MGC_M_TXCSR_MODE);
        wCsrVal &= ~(MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAMODE);
        wCsrVal &= ~(MGC_M_TXCSR_TXPKTRDY);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);

        wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
        wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);
        wCsrVal &= ~(MGC_M_RXCSR_RXPKTRDY);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
    }
#endif
}

inline void
set_enabledma_bulkout(
    MGC_LinuxCd* pThis,
    uint8_t bEnd)
{
    uint16_t wCsrVal;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[bEnd]);

    if (pEnd->bTrafficType == USB_ENDPOINT_XFER_BULK
        && !pEnd->bIsTx
        && pEnd->bDma_ena == USB_MCD_DMA_SETON) {
        /* BulkOutEP ShortPacketꥯdequeueDMA mode1 */
        MGC_SelectEnd(pBase, bEnd); 
        wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
        wCsrVal |= MGC_M_RXCSR_DMAENAB;
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
    }
}

static int
complete_request_immediate(
    struct usb_ep* ep,
    struct usb_request* pRequest,
    int status)
{
    unsigned long flags;
    struct usb_request* req;
    MGC_GadgetLocalEnd* pEnd =(MGC_GadgetLocalEnd*)ep;
    const uint8_t bEnd = EP_NUMBER(ep);
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "<==\n" );
    DBG(3, "dequeuing from nEnd=0x%x, pEnd=%p, pRequest=%p\n", \
        EP_NUMBER(pEnd), pEnd, pRequest); 

    EP_SPIN_LOCK_IRQSAVE(pEnd, flags);

    req = MGC_CurrentRequest(pEnd);

    if (NULL == req ){
        EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
	return -EINVAL;
    }

    if (!bEnd) {
        req->status = status;
        MGC_CompleteEp0Request();
        EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
        MGC_GadgetFifoFlush_FlushOnly(ep);
    }
    else if (req == pRequest) {
        list_del(&req->list);
        if(-EINPROGRESS != req->status){ 
        /* list_del() SchedulerQueue. */
            list_del(&((struct MGC_UsbRequest*)req)->completition_list ); 
        }
        req->status = status;
        stop_dma(pThis, bEnd);
        set_enabledma_bulkout(pThis, bEnd);
        MGC_GadgetFifoFlush_FlushOnly(ep);
        EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
        if (req->complete) {
            req->complete(&pEnd->end_point, req);
        }
    }
    else {
        list_for_each_entry(req, &pEnd->req_list, list) {
            if (req == pRequest) {
                list_del(&req->list);
                if(-EINPROGRESS != req->status){ 
                    /* list_del() SchedulerQueue. */
                    list_del(&((struct MGC_UsbRequest*)req)->completition_list ); 
                }
                req->status = status;
                EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);
                if (req->complete) {
                    req->complete(&pEnd->end_point, req);
                }
                goto OUT_ELSE_DEQUE;
            }
        }
        EP_SPIN_UNLOCK_IRQRESTORE(pEnd, flags);

        return -EINVAL;
    }
  OUT_ELSE_DEQUE:


    return 0;
}

/**************************************************************************
Tasklet
**************************************************************************/

static void gadget_complete_requests(unsigned long data) {
#if MUSB_DEBUG>0
    int success=0, fail=0;
#endif
    struct list_head local_list;
    struct MGC_UsbRequest* req=NULL;
    struct usb_request *pRequest;
    unsigned long flags;

    INIT_LIST_HEAD( &local_list );

     spin_lock_irqsave(&SchedulerQueue.lock, flags);	/* Modyfied for request queue restart */
     while ( !list_empty(&SchedulerQueue.req_list) ) {
	req=list_entry(SchedulerQueue.req_list.next, 
	    struct MGC_UsbRequest, completition_list);
	list_move_tail( &req->completition_list, &local_list);
     }

    /* complete the request first */
    while ( !list_empty(&local_list) ) {
	req=list_entry(local_list.next, 
	    struct MGC_UsbRequest, completition_list);
        pRequest = (struct usb_request *)req;
        list_del( &pRequest->list );
        pRequest = MGC_CurrentRequest(&MGC_aGadgetLocalEnd[req->bEnd]);
        if ( pRequest ) {
             MGC_Call_RestartRequest((struct usb_ep *)&MGC_aGadgetLocalEnd[req->bEnd], pRequest, GFP_KERNEL);
        }
        list_del( &req->completition_list );
#if MUSB_DEBUG>0
	req->bQueued=0;
#endif

	DBG(4, "completing pRequest=%p on bEnd=%d\n", req, req->bEnd);

        if ( !is_corrupted_request( req ) ) {	    
	    if ( req->request.complete ) {	    			    
		req->request.complete(&MGC_aGadgetLocalEnd[req->bEnd].end_point, 
			(struct usb_request*)req);
	    }
#if MUSB_DEBUG>0
	    success++;
	} else {
	    fail++;
#endif
	    DBG(4, "pRequest=%p on bEnd=%d is corrupted!!\n", req, 
		req->bEnd);

	}
    }
    spin_unlock_irqrestore(&SchedulerQueue.lock, flags);
    PRLOCK();
}

DECLARE_TASKLET(MGC_GadgetCompleteRequests, gadget_complete_requests, 0);


/* 
 * @pre spinlock_is_locked( &MGC_aGadgetLocalEnd[bEnd] );
 */
int MGC_CompleteRequest(struct usb_request *pRequest, int status) {
    struct MGC_UsbRequest* req=(struct MGC_UsbRequest*)pRequest;
            
    DBG(2, "<==\n");    
    if ( !req ) {    
	ERR("scheduling completition a null request\n");
	DBG(2, "==>\n");    
	return -EINVAL;
    }
    
    DBG(3, "scheduling completition of %p on bEnd=0x%x\n", req, 
	(req)?req->bEnd:-1);

#ifdef MUSB_PARANOID
    ASSERT_SPINLOCK_LOCKED(&pEnd->Lock); 
#endif
	        
    if ( is_corrupted_request( req ) ) {	    
	ERR("skipping completition of a corrupted request %p on bEnd=%d\n", req, 
	    req->bEnd); 
	return -EINVAL;
    }

    if ( is_queued_for_completition(req) ) {
	ERR("skipping completition of queued request %p on bEnd=%d\n", req, 
	    req->bEnd); 
	return -EINVAL;
    }

    pRequest->status=status;


    spin_lock( &SchedulerQueue.lock );    
    INIT_LIST_HEAD( &req->completition_list );	
    list_add_tail( &(req->completition_list), &(SchedulerQueue.req_list) );	
#if MUSB_DEBUG >0 		
    req->bQueued=1;
#endif	    
    spin_unlock( &SchedulerQueue.lock );
    DBG(3, "%p queued to completition list\n", req);	

    tasklet_schedule(&MGC_GadgetCompleteRequests);    
    DBG(2, "==>\n");
    return 0;
}

/*
 * @pre the end should be locked
 */
void
MGC_FlushEpRequests(
    MGC_GadgetLocalEnd* pEnd,
    const int status)
{
    unsigned long flags;
    struct usb_request* req;
    uint8_t bEnd = EP_NUMBER(pEnd);
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(-1, "<== ep%d, pEnd=%p\n", bEnd, pEnd);
    DEBUG_CODE(-1, dump_ep_queue(bEnd, 1); );

    spin_lock_irqsave(&pEnd->Lock, flags);

    req = MGC_CurrentRequest(pEnd);

    if (!bEnd) {
        if (req) {
            req->status = status;
            MGC_CompleteEp0Request();
        }
    }
    else {
        if (req) {
            stop_dma(pThis, bEnd);
        }
        list_for_each_entry(req, &pEnd->req_list, list) {
	    if(-EINPROGRESS == req->status){
                MGC_CompleteRequest( req, status );	
	    }
        }
    }

    spin_unlock_irqrestore(&pEnd->Lock, flags);

    MGC_GadgetFifoFlush_FlushOnly((struct usb_ep*)pEnd);

    DBG(2, "==>\n");
}


/**************************************************************************
Tx/Rx Data
**************************************************************************/

inline int get_ep_packet_size(MGC_LinuxCd* pThis, uint8_t bEnd) {
    int size=MGC_aGadgetLocalEnd[bEnd].wPacketSize;
    if ( (USB_ENDPOINT_XFER_BULK == MGC_aGadgetLocalEnd[bEnd].bTrafficType) && pThis->bBulkSplit) {
	size=pThis->aLocalEnd[bEnd].wMaxPacketSizeTx;
    }	
    return size;	
}		

/*
 * @pre spin_is_locked(MGC_aGadgetLocalEnd[bEnd].Lock)
 */
static void txstate(MGC_LinuxCd* pThis, struct MGC_UsbRequest* req) {
    unsigned long flags;
    uint32_t dwFifoCount = 0;
    uint16_t wCsrVal;
    
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    const uint8_t bEnd=req->bEnd;
    struct usb_request *pRequest=&req->request;
    MGC_GadgetLocalEnd* pEnd=&(MGC_aGadgetLocalEnd[bEnd]);
#ifdef MUSB_DMA
    MGC_DmaController* pDmaController = pThis->pDmaController;
    MGC_DmaChannel* pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    uint8_t bAllocChannel;
    uint8_t bDmaOk = FALSE;
    uint8_t bReqDmaState = USB_REQ_CPU;
    uint8_t bMode = 0 ;
    void* pData = NULL;
#endif

#ifdef MUSB_PARANOID 
    if ( !pRequest ) {
       ERR("cannot call txstate without request\n");
       return;
    }

   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
#endif

    DBG(2, "TestBulk txstate 1\n" );
#ifdef MUSB_DMA
	    /* usb_request¤dma != NULL 
	       ġpEnd->bTrafficType "ȥž"Ǥʤ 
	       ġžĹ0Ǥʤ 0 != pRequest->length */
	    if(((dma_addr_t)NULL != pRequest->dma) && 
	       (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType ) && 
	       (0 != pRequest->length) &&
	       (0 != (pRequest->length - pRequest->actual))) {
		/* pEnd->bTrafficType  "Bulkž"ξ硢ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
		if( (USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&   
		    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(pThis, bEnd))) ){
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1,"txstate[1]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* Forced CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[2]: bReqDmaState %d \n",bReqDmaState);
		     }
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[4]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
		}
		else {							/* ¾ξ Mode0 */
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"txstate[5]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* Forced CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[6]: bReqDmaState %d \n",bReqDmaState);
		     }


#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
		     if((pRequest->length - pRequest->actual) < (get_ep_packet_size(pThis, bEnd)) ){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[8]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[9]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
		}
		
	    }
	    else{
		bReqDmaState = USB_REQ_CPU;	/* DMAʤ CPUž */
		DBG(1,"txstate[10]: bReqDmaState %d \n",bReqDmaState);
	    }
#endif  /* MUSB_DMA */
        DBG(2, "TestBulk txstate 2\n" );

#ifdef MUSB_DMA
	switch( bReqDmaState ){
	    case USB_REQ_DMA_MODE1:	/* Mode1ǺǸΥ硼ȥѥåȰʳΥǡžԤ */
		DBG(1,"txstate[11]: bReqDmaState %d \n",bReqDmaState);
		dwFifoCount = (pRequest->length-pRequest->actual) - 
		    ( (pRequest->length-pRequest->actual) % get_ep_packet_size(pThis, bEnd) );
		DBG(2, "txstate: length=%d actual=%d dwFifoCount=%d \n",pRequest->length, pRequest->actual, dwFifoCount);
		DBG(2, "TestBulk USB_REQ_DMA_MODE1\n" );
		break;

	    case USB_REQ_DMA_MODE0:
		DBG(2, "TestBulk USB_REQ_DMA_MODE0\n" );
		DBG(1,"txstate[12]: bReqDmaState %d \n",bReqDmaState);

	    case USB_REQ_CPU:
		DBG(1,"txstate[13]: bReqDmaState %d \n",bReqDmaState);
                /* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
                dwFifoCount = (uint32_t)min(get_ep_packet_size(pThis, bEnd), 
                	(int)(pRequest->length-pRequest->actual)); 	
		break;
	    default:   	/* 顼 */
		break;
        }
#else
    dwFifoCount = (uint32_t)min(get_ep_packet_size(pThis, bEnd), 
	(int)(pRequest->length-pRequest->actual));
#endif  /* MUSB_DMA */
	
    if ( pRequest->length-pRequest->actual<0 ) {
	ERR("NEGATIVE FIFOCOUNT! pRequest->actual=%d, dwFifoCount=%d, pRequest->length=%d\n", \
		pRequest->actual, dwFifoCount, pRequest->length);
	dwFifoCount=0;
    }

    if ( pRequest->actual+dwFifoCount>pRequest->length ) {
	ERR("trying to write PAST length! pRequest->actual=%d, dwFifoCount=%d, pRequest->length=%d\n", \
		pRequest->actual, dwFifoCount, pRequest->length);
	dwFifoCount=pRequest->length-pRequest->actual;
    }
    DBG(3, "bEnd=0x%x, pEnd->wPacketSize=0x%x, dwFifoCount=%d, %s\n", \
	bEnd, pEnd->wPacketSize, dwFifoCount, 
	dump_usb_request(pRequest) );
    spin_lock_irqsave(&pThis->Lock, flags);   
    MGC_SelectEnd(pBase, bEnd);    
    wCsrVal=MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
    DBG(2, "TestBulk MGC_O_HDRC_TXCSR=%x \n", wCsrVal);
    if(pDmaChannel)  {							/* DMA꤬Ƥ */
	if(MGC_DMA_STATUS_FREE != 					/* DMAžνλǧ */
	    pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel) )  {
	    spin_unlock_irqrestore(&pThis->Lock, flags);   
            return;
        }
    }
    if( (MGC_M_TXCSR_FIFONOTEMPTY==(wCsrVal & MGC_M_TXCSR_FIFONOTEMPTY)) ||
        (MGC_M_TXCSR_TXPKTRDY==(wCsrVal & MGC_M_TXCSR_TXPKTRDY)) ){
	spin_unlock_irqrestore(&pThis->Lock, flags);   
        return;
    }
    if (USB_ENDPOINT_XFER_ISOC == pEnd->bTrafficType) {
	wCsrVal |= MGC_M_TXCSR_ISO;
    }

    DBG(2, "TestBulk txstate 3\n" );
    
#ifdef MUSB_DMA
	if( 0 != dwFifoCount ){
	    switch( bReqDmaState ){
		case USB_REQ_DMA_MODE1:
		DBG(1,"txstate[14]: bReqDmaState %d \n",bReqDmaState);
			bMode=1;
		case USB_REQ_DMA_MODE0:
		DBG(1,"txstate[15]: bReqDmaState %d \n",bReqDmaState);
			bAllocChannel = FALSE;
		
			if(!pDmaController) {
				bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[16]: bReqDmaState %d \n",bReqDmaState);
				break;
			}
			if(!pDmaChannel) {
			    pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel = 
				pDmaController->pfDmaAllocateChannel(
				    pDmaController->pPrivateData, bEnd, TRUE, 
				    pEnd->bTrafficType, pEnd->wPacketSize);
			    bAllocChannel = TRUE;
			}
			if(pDmaChannel) {
			    pEnd->dwRequestSize = min(pRequest->length, 
				pDmaChannel->dwMaxLength);
		 	    bDmaOk = pDmaController->pfDmaProgramChannel(pDmaChannel, MGC_TxDmaCallback, pData,
				pEnd->wPacketSize, bMode, (uint8_t *)(pRequest->dma + (void *)pRequest->actual), dwFifoCount);
			    DBG(2, "TestBulk dwFifoCount=%d \n", dwFifoCount );
			    if(bDmaOk) {
				pDmaChannel->dwActualLength = 0L;
				dwFifoCount = 0; 
				DBG(2, "TestBulk USB_REQ_DMA_MODE1/0 set success\n" );
			    } 
			    else{
                		/* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
				dwFifoCount = (uint32_t)min(get_ep_packet_size(pThis, bEnd), 
                			      (int)(pRequest->length-pRequest->actual)); 	
				bReqDmaState = USB_REQ_CPU;
				DBG(1,"txstate[17]: bReqDmaState %d \n",bReqDmaState);

				if(bAllocChannel) {
				    pDmaController->pfDmaReleaseChannel(pDmaChannel);
				    pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
				}
				break;
			    }
			}
			else {
                	     /* ǡĹDmaMode0/CPU žǡĹ(MAXPʲ)˥å */
			     dwFifoCount = (uint32_t)min(get_ep_packet_size(pThis, bEnd), 
                	          (int)(pRequest->length-pRequest->actual));
			     bReqDmaState = USB_REQ_CPU;
				 DBG(1,"txstate[18]: bReqDmaState %d \n",bReqDmaState);
			     break;
			}
		    default:
		        break;
		}
	}
	else{
	    bReqDmaState = USB_REQ_CPU;
		DBG(1,"txstate[19]: bReqDmaState %d \n",bReqDmaState);
	}
#endif /* MUSB_DMA */


#ifdef MUSB_DMA

    	wCsrVal |= MGC_M_TXCSR_MODE;

	switch( bReqDmaState ){
		case USB_REQ_DMA_MODE1:
				DBG(1,"txstate[20]: bReqDmaState %d \n",bReqDmaState);
	    	    wCsrVal |= (MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAENAB | MGC_M_TXCSR_DMAMODE );
		    MGC_WriteCsr8(pBase, MGC_O_HDRC_TXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
		    bDmaOk = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
		    if (!bDmaOk) {
    		        printk("DMA_TEST " "DmaEnableIntChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&pThis->Lock, flags);   
        	        return;
    		    }
		    /* MGC_GadgetEnableEnd()Ƚʣ(MAXP) */
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd, get_ep_packet_size(pThis, bEnd)); 
		    bDmaOk = pDmaController->pfDmaStartChannel(pDmaChannel);
		    if (!bDmaOk) {
    		        printk("DMA_TEST " "pfDmaStartChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&pThis->Lock, flags);   
        	        return;
    		    }
		    DBG(2, "TestBulk CSR=%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
		    DBG(2, "TestBulk USB_REQ_DMA_MODE1 set end\n" );
	    	    break;
		case USB_REQ_DMA_MODE0:
				DBG(1,"txstate[21]: bReqDmaState %d \n",bReqDmaState);
	            wCsrVal |= (MGC_M_TXCSR_DMAENAB);
	    	    wCsrVal &= (~(MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAMODE));
		    MGC_WriteCsr8(pBase, MGC_O_HDRC_TXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
		    bDmaOk = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
		    if (!bDmaOk) {
    		        printk("DMA_TEST " "DmaEnableIntChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&pThis->Lock, flags);   
        	        return;
    		    }
		    /* MGC_GadgetEnableEnd()Ƚʣ(MAXP) */
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd, get_ep_packet_size(pThis, bEnd));
		    bDmaOk = pDmaController->pfDmaStartChannel(pDmaChannel);
		    if (!bDmaOk) {
    		        printk("DMA_TEST " "pfDmaStartChannel Error %d\n", bDmaOk);
			spin_unlock_irqrestore(&pThis->Lock, flags);   
        	        return;
    		    }
		    DBG(2, "TestBulk pEnd->wPacketSize=%d \n", pEnd->wPacketSize );
		    DBG(2, "TestBulk get_ep_packet_size(pThis, bEnd)=%d \n", get_ep_packet_size(pThis, bEnd) );
		    DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t)pDmaChannel );
		    DBG(2, "TestBulk CSR=%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
		    DBG(2, "TestBulk USB_REQ_DMA_MODE0 set end\n" );
	    	    break;
		case USB_REQ_CPU:
				DBG(1,"txstate[22]: bReqDmaState %d \n",bReqDmaState);
	            wCsrVal |=  (MGC_M_TXCSR_TXPKTRDY);
	            wCsrVal &= ~(MGC_M_TXCSR_DMAENAB);
			if(dwFifoCount>MGC_FIFO_MAXSIZE){
			    /* check the upper 2Byte */
				/* length of packet err */
				DBG(2, "DMA_TEST " "TX dwFifoCount Error Size:%d\n",dwFifoCount);
				spin_unlock_irqrestore(&pThis->Lock, flags);   
        	    return;
			}
		    MGC_HdrcLoadFifo(pBase, bEnd, (uint16_t)dwFifoCount, 
			(uint8_t*)(pRequest->buf+pRequest->actual));	
		    pRequest->actual+=dwFifoCount;
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);
		    DBG(2, "TestBulk USB_REQ_CPU set end\n" );
		    break;
	}
#else

    wCsrVal = MGC_M_TXCSR_MODE | MGC_M_TXCSR_TXPKTRDY;

    MGC_HdrcLoadFifo(pBase, bEnd, (uint16_t)dwFifoCount, 
	(uint8_t*)(pRequest->buf+pRequest->actual));	
    pRequest->actual+=dwFifoCount;
    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);

#endif /* MUSB_DMA */

    DBG(3, "bEnd=0x%x, pEnd->wPacketSize=0x%x, dwFifoCount=%d, %s\n", \
	bEnd, pEnd->wPacketSize, dwFifoCount, \
	dump_usb_request(pRequest) );
    spin_unlock_irqrestore(&pThis->Lock, flags);   
}

/* Data ready for a request; called from IRQ; no need to disable the
 * IRQS bc they are already disabled.
 */
void MGC_HdrcServiceDeviceTxAvail(MGC_LinuxCd* pThis, uint8_t bEnd) {
    uint16_t wCsrVal;
    struct usb_request *pRequest;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[bEnd]);
#ifdef MUSB_DMA
    MGC_DmaChannel* pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    uint8_t bReqDmaState = USB_REQ_CPU;
    void* pData = NULL;
#endif /* MUSB_DMA */
    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
#endif

    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  1\n" );
    do {
	spin_lock(&pThis->Lock); 
	MGC_SelectEnd(pBase, bEnd);
	wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
	
	if(wCsrVal & MGC_M_TXCSR_P_SENTSTALL) {
	    wCsrVal &= ~MGC_M_TXCSR_P_SENTSTALL;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);
	
	    /* wait for clear! */
	    spin_unlock(&pThis->Lock); 
	    break;
	}
	DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  1.1\n" );
	
	if(wCsrVal & MGC_M_TXCSR_P_UNDERRUN) {
#ifdef MUSB_CONFIG_PROC_FS
	    pThis->aLocalEnd[bEnd].dwMissedTxPackets++;
#endif
	    wCsrVal &= ~MGC_M_TXCSR_P_UNDERRUN;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);
	    DBG(3, "underrun on ep%d\n", bEnd);
	}	
	spin_unlock(&pThis->Lock);

	DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  1.2\n" );
	spin_lock(&pEnd->Lock);	
	pRequest=MGC_CurrentRequest(pEnd);    
	DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  1.3\n" );
        if ( pRequest ) {		
	DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  1.4\n" );

	    if (pRequest->status != -EINPROGRESS) { 
	    	spin_unlock(&pEnd->Lock);	
		break;
	    }

#ifdef MUSB_DMA
	    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  2\n" );

	    /* usb_request¤dma != NULL 
	    ġpEnd->bTrafficType "ȥž"Ǥʤ 
	    ġžĹ0Ǥʤ 0 != pRequest->length */
	    if(((dma_addr_t)NULL != pRequest->dma) && 
	       (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType ) && 
	       (0 != pRequest->length) &&
	       (0 != (pRequest->length - pRequest->actual))) {
		/* pEnd->bTrafficType  "Bulkž"ξ硢ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
		if( (USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&
		    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(pThis, bEnd))) ){
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1,"TxAvail[1]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"TxAvail[2]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"TxAvail[4]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
		}
		else {							/* ¾ξ Mode0 */
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"TxAvail[5]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"TxAvail[6]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER		/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
		     if((pRequest->length - pRequest->actual) < (get_ep_packet_size(pThis, bEnd)) ){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"TxAvail[8]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"TxAvail[9]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
		}
		
	    }
	    else{
		bReqDmaState = USB_REQ_CPU;		/* DMAʤ CPUž */
		DBG(1,"TxAvail[10]: bReqDmaState %d \n",bReqDmaState);
	    }

	    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  3\n" );
	    if(pDmaChannel)  {				/* DMA꤬Ƥ */
		DBG(2, "TestBulk pfDmaGetChannelStatus  %d\n", (pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel)));
		DBG(2, "TestBulk MGC_DMA_STATUS_FREE=%d \n", MGC_DMA_STATUS_FREE );
		DBG(2, "TestBulk MGC_DMA_STATUS_BUSY=%d \n", MGC_DMA_STATUS_BUSY );
		DBG(2, "TestBulk MGC_DMA_STATUS_BUS_ABORT=%d \n", MGC_DMA_STATUS_BUS_ABORT );
		DBG(2, "TestBulk pDmaChannel->dwActualLength=%d \n", pDmaChannel->dwActualLength );
		if(MGC_DMA_STATUS_FREE == 		/* DMAžνλǧ */
		    pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel) )  {
			DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  3.1\n" );
			DBG(2, "TestBulk wCsrVal=%x  \n", wCsrVal );
			if( ( NULL != pThis->aLocalEnd[bEnd].pDmaChannel->pData ) &&
			    ( TRUE == *((uint8_t *)(pThis->aLocalEnd[bEnd].pDmaChannel->pData))) ){
			    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  3.2\n" );
			    switch( bReqDmaState ){
				case USB_REQ_DMA_MODE1:
					DBG(1,"TxAvail[11]: bReqDmaState %d \n",bReqDmaState);

					pRequest->actual+= pDmaChannel->dwActualLength;
				    	/* Mode1ξϡDMAλ */
				    	pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
					pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
			       		pThis->pDmaController->pfDmaProgramChannel(pDmaChannel, 
			       		    MGC_TxDmaCallback, pData,pEnd->wPacketSize, 0, 
			       		    (uint8_t *)(pRequest->dma + (void *)pRequest->actual), 0);
					wCsrVal &= ~(MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAENAB | 
					    MGC_M_TXCSR_DMAMODE);	/* AUTOSETΥꥢ */
					MGC_WriteCsr8(pBase, MGC_O_HDRC_TXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );

					wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal );
					if(MGC_M_TXCSR_FIFONOTEMPTY==(wCsrVal & MGC_M_TXCSR_FIFONOTEMPTY)){
					    spin_unlock(&pEnd->Lock);
					    return;
					}
					DBG(2, "TestBulk CSR=%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
					DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail 4 Mode1\n" );
				    break;
				case USB_REQ_DMA_MODE0:
				    DBG(1,"TxAvail[12]: bReqDmaState %d \n",bReqDmaState);

				    DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal );
				    DBG(2, "TestBulk CSR    =%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
				    pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
				    pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
				    wCsrVal &= ~MGC_M_TXCSR_DMAENAB;
				    /* Mode0ǤϡFIFOؤDMAžäƤUSB¹ԡ */
				    wCsrVal |= (MGC_M_TXCSR_TXPKTRDY);	
				    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);
				    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  4 Mode0\n" );
				    DBG(2, "TestBulk wCsrVal=%x \n", wCsrVal );
				    DBG(2, "TestBulk CSR    =%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
				    if( (USB_ENDPOINT_XFER_BULK != pEnd->bTrafficType) ||
				        ((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)&&
				        ((pRequest->length - pRequest->actual) < get_ep_packet_size(pThis, bEnd))) ) {
                                        spin_unlock(&pEnd->Lock);
				           return;
				    }
		    		}
		    	}
			else {
			    DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail  4.1 \n" );
			    DBG(2, "TestBulk CSR=%x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd) );
		            pRequest->actual+= pDmaChannel->dwActualLength;
			    DBG(2, "TestBulk pRequest->actual%d  \n", pRequest->actual );
			    if( pDmaChannel ){
			    	/* Mode1ξϡDMAλ */
			    	pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
				pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
		       		pThis->pDmaController->pfDmaProgramChannel(pDmaChannel, 
		       		    MGC_TxDmaCallback, pData,pEnd->wPacketSize, 0, 
		       		    (uint8_t *)(pRequest->dma + (void *)pRequest->actual), 0);
				wCsrVal &= ~(MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAENAB | 
				    MGC_M_TXCSR_DMAMODE);	/* AUTOSETΥꥢ */
				MGC_WriteCsr8(pBase, MGC_O_HDRC_TXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
			    }
			}
		}
		else{
		    spin_unlock(&pEnd->Lock);
		    return;
		}
	    }
	    else {
		    if( (MGC_M_TXCSR_FIFONOTEMPTY==(wCsrVal & MGC_M_TXCSR_FIFONOTEMPTY)) ||
		        (MGC_M_TXCSR_TXPKTRDY==(wCsrVal & MGC_M_TXCSR_TXPKTRDY)) ){
		        spin_unlock(&pEnd->Lock);
		        return;
		    }
		    /* ʤˤ⤷ʤ */
	    }

#endif /* MUSB_DMA */

	    if ( (pRequest->actual==pRequest->length) ) {     		/* žλ */
		DBG(3, "completing pRequest=%p on bEnd=0x%x\n", pRequest, bEnd); 
#ifdef MUSB_DMA 
	    	if(pDmaChannel) { 
			pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
			pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
			/* (pfDmaReleaseBuffer();) */
			pThis->pDmaController->pfDmaReleaseChannel(pDmaChannel);
			pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
	   	}
                /* use the zero flag to request the transmission setting it to
                 * zero, after the zero packet has been submitted */
                if (pRequest->zero
                    && pRequest->status == -EINPROGRESS
                    && pRequest->length
                    && pRequest->length % get_ep_packet_size(pThis, bEnd) == 0
                    && !(wCsrVal & MGC_M_TXCSR_P_SENTSTALL)) {
                    pRequest->zero = 0;
		    wCsrVal |= MGC_M_TXCSR_TXPKTRDY;
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal );
		    DBG(2, "TestBulk Zero set PktRdy\n");
                    spin_unlock(&pEnd->Lock);
		    break;
                }
#endif /* MUSB_DMA */
		wCsrVal &= ~(MGC_M_TXCSR_MODE);
		/* EP¦ˤǽˤƤ */
		MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsrVal);
		DBG(2, "TestBulk MGC_HdrcServiceDeviceTxAvail() 5\n" );
		if((NULL!=pRequest) && (pRequest->status == -EINPROGRESS)){
	            pRequest->status=0;  	/* ꥯȾ֤λˤ done */
		    /* λ ComplitionХå */
		    MGC_CompleteRequest(pRequest, pRequest->status);
		}
	    }
	    else{
		/* 󥻥뽪λƤʤСν */
		if( (NULL != pRequest) && (pRequest->status != -ECONNRESET) ){
		    spin_unlock(&pEnd->Lock);	
	    	    txstate(pThis, (struct MGC_UsbRequest*)pRequest);
		    spin_lock(&pEnd->Lock);	
	    	}
	    }
	}
	else {
	}
	spin_unlock(&pEnd->Lock);	


    } while(FALSE);

    DBG(2, "==>\n" );
}


/* ------------------------------------------------------------ */

/*
 * @pre spinlock_is_locked(&MGC_aGadgetLocalEnd[bEnd])
 */
static void rxstate(MGC_LinuxCd* pThis, struct MGC_UsbRequest* req) {
    uint16_t wCsrVal=0;
    unsigned long flags;
    const uint8_t bEnd=req->bEnd;
    struct usb_request *pRequest=&req->request;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[bEnd]);
    uint32_t dwFifoCount = 0;	

#ifdef MUSB_DMA
    MGC_DmaController* pDmaController = pThis->pDmaController;
    MGC_DmaChannel* pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    uint8_t bAllocChannel;
    uint8_t bDmaOk = FALSE;
    uint8_t bReqDmaState = USB_REQ_CPU;
    uint8_t bMode = 0 ;
    void* pData = NULL;
#endif /* MUSB_DMA */

#ifdef MUSB_PARANOID
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
#endif

    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, bEnd);
    wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);

    DBG(2, "TestBulk rxstate() 1\n" );
#ifdef MUSB_DMA
    /* usb_request¤dma != NULL 
    ġpEnd->bTrafficType "ȥž"Ǥʤ 
    ġžĹ0Ǥʤ 0 != pRequest->length */
    if(((dma_addr_t)NULL != pRequest->dma) && 
       (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType ) && 
       (0 != pRequest->length) &&
       (0 != (pRequest->length - pRequest->actual)) ){
        /* pEnd->bTrafficType  "Bulkž"ξ硢
	   ĺǸΥ硼ȥѥåȤǤʤ  Mode1 */
	if( (USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&
	    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(pThis, bEnd))) ){
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
	     bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1,"rxstate[1]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[2]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
	     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
     	         bReqDmaState = USB_REQ_CPU;
				 DBG(1,"rxstate[4]: bReqDmaState %d \n",bReqDmaState);
             }
#endif
	}
	else {								/* ¾ξ Mode0 */
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"rxstate[5]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[6]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER		/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
	     if((pRequest->length - pRequest->actual) < (get_ep_packet_size(pThis, bEnd)) ){
	     	bReqDmaState = USB_REQ_CPU;
			DBG(1,"rxstate[8]: bReqDmaState %d \n",bReqDmaState);
	     }
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
	     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
     	         bReqDmaState = USB_REQ_CPU;
				 DBG(1,"rxstate[9]: bReqDmaState %d \n",bReqDmaState);
             }
#endif
	}
	
    }
    else{
	bReqDmaState = USB_REQ_CPU;		/* DMAʤ CPUž */
	DBG(1,"rxstate[10]: bReqDmaState %d \n",bReqDmaState);
	
    }

    DBG(2, "TestBulk bReqDmaState=%d \n",bReqDmaState );
    DBG(2, "TestBulk RXCOUNT=%d \n",(MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd)));

    
    if( USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest ){
	   DBG(2, "TestBulk USB_INIT_REQ_PHASE1 \n" );
           if(USB_REQ_DMA_MODE1==bReqDmaState)	;
           else{
		wCsrVal &= (~MGC_M_RXCSR_DMAENAB);
		MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
               goto exit_rxstate;
	   }
    }
    else if( USB_INIT_REQ_PHASE2 == pEnd->bIsInitRequest ){
           if(pDmaChannel){
		wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
                dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
                if( (wCsrVal& MGC_M_RXCSR_RXPKTRDY) && 
		  (dwFifoCount < (uint32_t)get_ep_packet_size(pThis, bEnd))){
                    if( (1==pRequest->zero) && (0==pRequest->actual) && (0==dwFifoCount) ){
		        pDmaController->pfDmaStartChannel(pDmaChannel);
		    }
		    DBG(2, "TestBulk USB_INIT_REQ_PHASE2.1 \n" );
                	;
                }
                else{
		    DBG(2, "TestBulk USB_INIT_REQ_PHASE2.2 \n" );
		    pDmaController->pfDmaStartChannel(pDmaChannel);
                    goto exit_rxstate;
		}
	   }
	   else{
		DBG(2, "TestBulk USB_INIT_REQ_PHASE2.3 \n" );
           	if((uint16_t)NULL == (wCsrVal&MGC_M_RXCSR_RXPKTRDY))
                    goto exit_rxstate;
	   }
    }
    else if( (USB_REQ_DMA_MODE1 == bReqDmaState) && (NULL == pDmaChannel) ){
	 /* USB_REQ_DMA_MODE1 ν DMA ch ˶ʤǤCPUž륱 */
         bReqDmaState = USB_REQ_CPU;
    }

#endif /* MUSB_DMA */

#ifdef MUSB_DMA
    if ( (wCsrVal& MGC_M_RXCSR_RXPKTRDY) || (USB_REQ_DMA_MODE1==bReqDmaState)) {
#else
    if ( wCsrVal& MGC_M_RXCSR_RXPKTRDY ) { /* } */
#endif

	DBG(2, "TestBulk rxstate() 2\n" );
#ifdef MUSB_DMA
    	/* ⤷DMA¹ʤ (DMA Mode1硼ȥѥåȤˤǤξ) */
    	if( (MGC_M_RXCSR_DMAMODE == (wCsrVal&MGC_M_RXCSR_DMAMODE)) &&
    	    (FALSE != pDmaChannel ) && 
            (MGC_DMA_STATUS_FREE != pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel)) ) {

	    DBG(2, "TestBulk RX ShortPacket \n" );
	    DBG(2, "TestBulk pDmaChannel->dwActualLength = %d \n", pDmaChannel->dwActualLength);
	    DBG(2, "TestBulk wCsrVal = %x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));
	    DBG(2, "TestBulk DmaGetChannelStatus(pDmaChannel) = %d \n", pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel));
	    DBG(2, "TestBulk RXCOUNT = %d \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd));

	    pRequest->actual+= pDmaChannel->dwActualLength;	/* ޤǤDMAžʬ򥫥ȥå */

	    /* ǽ0ѥåȤ̵뤹ϡDMA Mode1ONΤޤޡ */
	    dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
	    if( !((1==pRequest->zero) && (0==pRequest->actual) && (0==dwFifoCount)) ){

		    pThis->pDmaController->pfDmaStopChannel(pDmaChannel);	/* DmaߡDma */
		    pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
	       	    pThis->pDmaController->pfDmaProgramChannel(pDmaChannel, MGC_RxDmaCallback, pData,
		        pEnd->wPacketSize, 0, (uint8_t *)(pRequest->dma + (void *)pRequest->actual), 0);
		    /* (pfDmaReleaseBuffer();) */
		    pThis->pDmaController->pfDmaReleaseChannel(pDmaChannel);
		    pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
		    DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t)pDmaChannel );
	
		    wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);	/* AUTOSETΥꥢ */
		    MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
		}
		
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"rxstate[11]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[12]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
                dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
		if( dwFifoCount < (uint32_t)(get_ep_packet_size(pThis, bEnd)) ){
	     	    bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[14]: bReqDmaState %d \n",bReqDmaState);
		}
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
                dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
        	if(MUSB_DMA_FEWBYTE >= dwFifoCount){
     	 	   bReqDmaState = USB_REQ_CPU;
			   DBG(1,"rxstate[15]: bReqDmaState %d \n",bReqDmaState);
         	}
#endif
        }

	switch( bReqDmaState ){
	    case USB_REQ_DMA_MODE1:
			DBG(1,"rxstate[16]: bReqDmaState %d \n",bReqDmaState);
		    dwFifoCount = (pRequest->length-pRequest->actual) - 
		        ( (pRequest->length-pRequest->actual) % get_ep_packet_size(pThis, bEnd) );
		    pEnd->dwRxOverrun =  0;
		    pEnd->bRxMaxpOverrun = FALSE;
	    
	    		DBG(2, "TestBulk USB_REQ_DMA_MODE1 dwFifoCount=%d\n", dwFifoCount );
		break;

	    case USB_REQ_DMA_MODE0:
			DBG(1,"rxstate[17]: bReqDmaState %d \n",bReqDmaState);

	    case USB_REQ_CPU:
			DBG(1,"rxstate[18]: bReqDmaState %d \n",bReqDmaState);
                dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd); 
                /* Mode0/CPURXCOUNTʬǡžԤ */
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
		if( dwFifoCount < (uint32_t)(get_ep_packet_size(pThis, bEnd)) ){
	     	    bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[19]: bReqDmaState %d \n",bReqDmaState);
		}
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
        	if(MUSB_DMA_FEWBYTE >= dwFifoCount){
     	 	   bReqDmaState = USB_REQ_CPU;
			   DBG(1,"rxstate[20]: bReqDmaState %d \n",bReqDmaState);
         	}
#endif
		if( dwFifoCount > (get_ep_packet_size(pThis, bEnd)) ){  
		    bReqDmaState = USB_REQ_CPU;
		    /* MAXPʾMAXPĹޤǡΤFlush롣*/
		    pEnd->dwRxOverrun =  dwFifoCount - ((uint32_t)min(get_ep_packet_size(pThis, bEnd), 
                			(int)(pRequest->length-pRequest->actual)));
		    pEnd->bRxMaxpOverrun = TRUE;
	
                    dwFifoCount = (uint32_t)min(get_ep_packet_size(pThis, bEnd), 
                			(int)(pRequest->length-pRequest->actual));
		}
		else if( (dwFifoCount > (pRequest->length - pRequest->actual)) ){
		    /* ׵Ĺʾ׵ĹޤǡΤFlush롣*/
		    /* OUT_BULK  СĹ򲱤Ƥ */
		    pEnd->dwRxOverrun =  dwFifoCount - (pRequest->length - pRequest->actual);
		    pEnd->bRxMaxpOverrun = FALSE;
	
		    dwFifoCount = pRequest->length - pRequest->actual;
		}
		DBG(2, "TestBulk USB_REQ_DMA_MODE0/CPU wFifoCount=%d\n", dwFifoCount );
		break;
	    default:
	    	/* 顼 */
		break;
        }
#else
        /* CPURXCOUNTʬΥǡžԤ */
        dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
#endif /* MUSB_DMA */

	DBG(2, "TestBulk rxstate() 3\n" );
	DBG(2, "TestBulk rxstate() 4\n" );
#ifdef MUSB_DMA
	if( 0 != dwFifoCount ){
	    switch( bReqDmaState ){
		case USB_REQ_DMA_MODE1:
			DBG(1,"rxstate[21]: bReqDmaState %d \n",bReqDmaState);
			bMode = 1;

		case USB_REQ_DMA_MODE0:
			DBG(1,"rxstate[22]: bReqDmaState %d \n",bReqDmaState);
			bAllocChannel = FALSE;
		
			if(!pDmaController) {
				DBG(1,"rxstate[23]: bReqDmaState %d \n",bReqDmaState);
				bReqDmaState = USB_REQ_CPU;
				break;
			}
			if(!pDmaChannel) {
			    pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel = 
				pDmaController->pfDmaAllocateChannel(
				    pDmaController->pPrivateData, bEnd, FALSE, 
				    pEnd->bTrafficType, pEnd->wPacketSize);
			    bAllocChannel = TRUE;
			}
			if(pDmaChannel) {
			    pEnd->dwRequestSize = min(pRequest->length, 
				pDmaChannel->dwMaxLength);
		 	    bDmaOk = pDmaController->pfDmaProgramChannel(pDmaChannel, 
		 	        MGC_RxDmaCallback, pData,pEnd->wPacketSize, bMode, 
		 	        (uint8_t *)(pRequest->dma + (void *)pRequest->actual), dwFifoCount);
			    if(bDmaOk) {
				pDmaChannel->dwActualLength = 0L;
				dwFifoCount = 0;
			    } 
			    else{
				/* USB_REQ_DMA_MODE1 ν DMA ch 
				   ʤǤCPUž륱 */
				if(bAllocChannel) {
				    pDmaController->pfDmaReleaseChannel(pDmaChannel);
				    pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
				}
				if( USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest ){
				    wCsrVal &= (~MGC_M_RXCSR_DMAENAB);
				    MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
				    goto exit_rxstate;
				}
				/* CPUǥѥåȤȤ˥ǡžԤ */
				dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
				bReqDmaState = USB_REQ_CPU;
				DBG(1,"rxstate[24]: bReqDmaState %d \n",bReqDmaState);

				break;
			    }
			}
			else {
			     /* USB_REQ_DMA_MODE1 ν DMA ch 
			        ʤǤCPUž륱 */
			     if( USB_INIT_REQ_PHASE1 == pEnd->bIsInitRequest ){
			        wCsrVal &= (~MGC_M_RXCSR_DMAENAB);
			        MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
				goto exit_rxstate;
			     }
			     /* CPUǥѥåȤȤ˥ǡžԤ */
			     dwFifoCount = (uint32_t)MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd); 
			     bReqDmaState = USB_REQ_CPU;
				 DBG(1,"rxstate[25]: bReqDmaState %d \n",bReqDmaState);
			     break;
			}

		default:			/* CPUžξϤʤˤ⤷ʤ */
			break;
	    }
	}
	else{
	    bReqDmaState = USB_REQ_CPU;		/* ǡĹ0ξϡCUPž */
		DBG(1,"rxstate[26]: bReqDmaState %d \n",bReqDmaState);
	}
#endif /* MUSB_DMA */

	DBG(2, "TestBulk rxstate() 5\n" );
#ifdef MUSB_DMA
	switch( bReqDmaState ){
		case USB_REQ_DMA_MODE1:
				DBG(1,"rxstate[27]: bReqDmaState %d \n",bReqDmaState);
	    	    wCsrVal |= (MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAENAB | MGC_M_RXCSR_DMAMODE);
		    MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
		    bDmaOk = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
		    if(!bDmaOk) {
			DBG(2, "TestBulk DmaEnableIntChannel() Error \n" );
		    } 
		    /* MGC_GadgetEnableEnd()Ƚʣ(MAXP) */
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd, get_ep_packet_size(pThis, bEnd));
		    DBG(2, "TestBulk RXMAXP = %d \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd));
		    if( (USB_INIT_REQ_PHASE1 != pEnd->bIsInitRequest)){
			DBG(2, "TestBulk DmaStartChannel() \n" );
			DBG(2, "TestBulk (pEnd->bIsInitRequest)=%d \n",(pEnd->bIsInitRequest) );
		        bDmaOk = pDmaController->pfDmaStartChannel(pDmaChannel);
		        if(!bDmaOk) {
			    DBG(2, "TestBulk DmaStartChannel() Error \n" );
		        } 
		    }
		    DBG(2, "TestBulk USB_REQ_DMA_MODE1\n" );
		    DBG(2, "TestBulk wCsrVal = %x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));
	    	    break;
		case USB_REQ_DMA_MODE0:
				DBG(1,"rxstate[28]: bReqDmaState %d \n",bReqDmaState);
	            wCsrVal |= (MGC_M_RXCSR_DMAENAB);
	    	    wCsrVal &= (~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE));
		    MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
		    bDmaOk = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
		    if(!bDmaOk) {
			DBG(2, "TestBulk DmaEnableIntChannel() Error \n" );
		    }
		    if( (USB_INIT_REQ_PHASE1 != pEnd->bIsInitRequest)){
			DBG(2, "TestBulk DmaStartChannel() \n" );
		        bDmaOk = pDmaController->pfDmaStartChannel(pDmaChannel);
		        if(!bDmaOk) {
			    DBG(2, "TestBulk DmaStartChannel() Error \n" );
		        } 
		    }
		    DBG(2, "TestBulk USB_REQ_DMA_MODE0\n" );
	    	    break;
		case USB_REQ_CPU:
				DBG(1,"rxstate[29]: bReqDmaState %d \n",bReqDmaState);
#endif /* MUSB_DMA */
		    DBG(2, "TestBulk USB_REQ_CPU 1\n" );
		    if( (NULL!=pRequest) && (pRequest->status != -ECONNRESET) &&
		        ( pRequest->actual <= pRequest->length )) { 
			if(dwFifoCount>MGC_FIFO_MAXSIZE){
			    /* check the upper 2Byte */
				/* length of packet err */
				DBG(2, "DMA_TEST " "RX dwFifoCount Error Size:%d\n",dwFifoCount);
				goto exit_rxstate;
			}
		   	MGC_HdrcUnloadFifo(pBase, bEnd, (uint16_t)dwFifoCount, 
				(uint8_t*)(pRequest->buf + pRequest->actual));
			DBG(2, "TestBulk USB_REQ_CPU 2\n" );
		    	pRequest->actual += dwFifoCount;
		    	/* DBG(3, "wFifoCount=%d, bEnd=0x%x, pEnd->wPacketSize=0x%x, wCount=0x%x, %s", \
			   wFifoCount, bEnd, pEnd->wPacketSize, wCount, dump_usb_request(pRequest));  */

			DBG(2, "TestBulk USB_REQ_CPU 3\n" );
/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
#if  defined(USB_DMA_SHORTPKT_USING_CPU_TRANSFER) || defined(MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER)
		   	if ( (pRequest->actual==pRequest->length) || 
		   	     (dwFifoCount < (uint32_t)get_ep_packet_size(pThis, bEnd)) || 
		   	     (TRUE == pEnd->bRxMaxpOverrun) ) {

		   		/* ǽ0ѥåȼξDMAENABӥåȤϤΤޤ */
		   		if((USB_ENDPOINT_XFER_BULK==pEnd->bTrafficType) && 
		   		    (!((pRequest->zero==1)&&(pRequest->actual==0))) ){	
		    	 	    if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
						/* DMA transfer */
						/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
						wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
						wCsrVal |=   MGC_M_RXCSR_DMAENAB ;
		     		    }
		   		}
				if ( 0 == pEnd->dwRxOverrun || TRUE == pEnd->bRxMaxpOverrun ){
				    wCsrVal &= ~MGC_M_RXCSR_RXPKTRDY;
				}
			}
			else{
		   		if((USB_ENDPOINT_XFER_BULK==pEnd->bTrafficType) && 
				   ((dma_addr_t)NULL!=pRequest->dma)){
		     		    if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
						/* DMA transfer */
						/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
						wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
		     		    }
				}
				wCsrVal &= ~MGC_M_RXCSR_RXPKTRDY;
			}
#endif
			MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);

			DBG(2, "TestBulk USB_REQ_CPU 4\n" );
			DBG(2, "TestBulk pRequest->actual=%u \n", pRequest->actual );
			DBG(2, "TestBulk pRequest->length=%u \n", pRequest->length );
			DBG(2, "TestBulk get_ep_packet_size(pThis, bEnd)=%u \n", get_ep_packet_size(pThis, bEnd) );
			DBG(2, "TestBulk dwFifoCount=%u \n", dwFifoCount );
		   	/* reach the end or short packet detected */
		   	if ( (pRequest->actual==pRequest->length) || 
		   	     (dwFifoCount < (uint32_t)get_ep_packet_size(pThis, bEnd)) || 
		   	     (TRUE == pEnd->bRxMaxpOverrun) ) {
		                if (pRequest->zero
		                    && pRequest->status == -EINPROGRESS
		                    && pRequest->actual == 0) {
		                    pRequest->zero = 0;
		                    goto exit_rxstate;
		                }
				spin_lock(&pEnd->Lock);			    
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
				if(pDmaChannel)  {		/* DMA꤬Ƥ */
					if( ( NULL != pThis->aLocalEnd[bEnd].pDmaChannel->pData ) &&
					    ( TRUE != *((uint8_t *)(pThis->aLocalEnd[bEnd].pDmaChannel->pData))) ){
                                            spin_unlock(&pEnd->Lock);
                                            goto exit_rxstate;
					}
					pThis->pDmaController->pfDmaStopChannel(pDmaChannel);
					pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
					/* (pfDmaReleaseBuffer();) */
					pThis->pDmaController->pfDmaReleaseChannel(pDmaChannel);
					pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
					DBG(2, "TestBulk pDmaChannel=%u \n", (uint32_t)pDmaChannel );
				}
#endif
				DBG(3, "completing pRequest=%p on bEnd=0x%x\n", pRequest, bEnd); 
				/* OUT_BULK ХξΥ顼ȡshort_not_okΥå */
				if((NULL!=pRequest) && (pRequest->status == -EINPROGRESS)){
				    if (0 == pEnd->dwRxOverrun){
					if(!((1==pRequest->short_not_ok) && 
					     ((0!=((pRequest->actual)%(get_ep_packet_size(pThis, bEnd))))
					     ||(0==dwFifoCount)))){
					    pRequest->status=0;
					}
					else{
					    pRequest->status=-EREMOTEIO;
					}
					/* async complete */
					MGC_CompleteRequest(pRequest, pRequest->status);
				    }
				    else{
				        if(TRUE == pEnd->bRxMaxpOverrun){
					    pRequest->status = -ECOMM;
					}
					else {
					    pRequest->status = -EOVERFLOW;
					}
					pRequest->actual = pRequest->actual + pEnd->dwRxOverrun;
					pEnd->dwRxOverrun =  0; 
					pEnd->bRxMaxpOverrun = FALSE;
					MGC_CompleteRequest(pRequest, pRequest->status);
				    }
			
				}
				spin_unlock(&pEnd->Lock);			    
	   	    	}
			DBG(2, "TestBulk USB_REQ_CPU\n" );
		    }
#ifdef MUSB_DMA
	    	    break;
	    	default:
		    /* 顼 */
	    	    break;
	}
#endif /* MUSB_DMA */

    }
    DBG(2, "TestBulk rxstate() 6\n" );
  exit_rxstate:
    spin_unlock_irqrestore(&pThis->Lock, flags);
}


/* Data ready for a request; called from IRQ
 */
void MGC_HdrcServiceDeviceRxReady(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t wCsrVal;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[bEnd]);
    struct usb_request* pRequest=NULL;
#ifdef MUSB_DMA
    MGC_DmaChannel* pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    uint8_t bReqDmaState = USB_REQ_CPU;
    void* pData = NULL;
    uint32_t dwFifoCount = 0;
#endif /* MUSB_DMA */
    
    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_LOCKED(&pEnd->Lock); 
#endif

    DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 1\n" );
    do {
	spin_lock(&pThis->Lock); /* executed from interrupt no need to disable them */
	MGC_SelectEnd(pBase, bEnd);
	wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
    
	if(wCsrVal & MGC_M_RXCSR_P_SENTSTALL) {
	    wCsrVal &= ~MGC_M_RXCSR_P_SENTSTALL;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, 
	    	wCsrVal);
	    DBG(3, "received stall on ep%d\n", bEnd);
	    spin_unlock(&pThis->Lock);
	    break;
	}
	
	if(wCsrVal & MGC_M_RXCSR_P_OVERRUN) {
#ifdef MUSB_CONFIG_PROC_FS
	    pThis->aLocalEnd[bEnd].dwMissedRxPackets++;
#endif
	    wCsrVal &= ~MGC_M_RXCSR_P_OVERRUN;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
	    DBG(3, "received overrun on ep%d\n", bEnd);
	}
	
#ifdef MUSB_CONFIG_PROC_FS
	if(wCsrVal & MGC_M_RXCSR_INCOMPRX) {
	    pThis->aLocalEnd[bEnd].dwErrorRxPackets++;
	}
#endif
        spin_unlock(&pThis->Lock);       

	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 2\n" );
	DBG(2, "TestBulk (wCsrVal&MGC_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal&MGC_M_RXCSR_RXPKTRDY) );
	/* analyze request if the ep is not inactive */
	spin_lock(&pEnd->Lock);
	pRequest=(pEnd->bInactive) ? NULL : MGC_CurrentRequest(pEnd);		    
	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 3 pRequest=%u \n", (uint32_t)pRequest );
	DBG(2, "TestBulk (wCsrVal&MGC_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal&MGC_M_RXCSR_RXPKTRDY) );

        if ( pRequest ) {

#ifdef MUSB_DMA
            if (pRequest->status != -EINPROGRESS) { 
                /* CurrentRequestꥯȥХåԤξ硢 */
                /* ߤ⤷;ʬߤäƤⲿ߽ʤ */
                spin_unlock(&pEnd->Lock);
                /* breakwhileȴƤǽ֥årxstate뤿return */
                return;
            }

	    /* usb_request¤dma != NULL 
	    ġpEnd->bTrafficType "ȥž"Ǥʤ 
	    ġžĹ0Ǥʤ 0 != pRequest->length */
	    if(((dma_addr_t)NULL != pRequest->dma) && 
	       (USB_ENDPOINT_XFER_CONTROL != pEnd->bTrafficType ) && 
	       (0 != pRequest->length) &&
	       (0 != (pRequest->length - pRequest->actual))) {
	        /* pEnd->bTrafficType  "Bulkž"ξ硢
		   ĺǸΥ硼ȥѥåȤǤʤ  Mode1,
		   ġǤ˥ѥåȤƤʤ */
		if( (USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType) &&
		    ((pRequest->length - pRequest->actual) >= (get_ep_packet_size(pThis, bEnd)))){
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE1;
				DBG(1,"RxReady[1]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* Forced CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"RxReady[2]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
	     	         bReqDmaState = USB_REQ_CPU;
					 DBG(1,"RxReady[4]: bReqDmaState %d \n",bReqDmaState);
	         }
#endif
		}
		else {							/* ¾ξ Mode0 */
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"RxReady[5]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"RxReady[6]: bReqDmaState %d \n",bReqDmaState);
		     }

#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER	/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
		     if((pRequest->length - pRequest->actual) < (get_ep_packet_size(pThis, bEnd)) ){
		         bReqDmaState = USB_REQ_CPU;
				 DBG(1,"RxReady[8]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
#ifdef MUSB_DMA_FEWBYTE_USING_CPU_TRANSFER
		     if(MUSB_DMA_FEWBYTE >= (pRequest->length - pRequest->actual)){
	     	         bReqDmaState = USB_REQ_CPU;
	 				 DBG(1,"RxReady[9]: bReqDmaState %d \n",bReqDmaState);
	             }
#endif
		}
		
	    }
	    else{
		bReqDmaState = USB_REQ_CPU;		/* DMAʤ CPUž */
		DBG(1,"RxReady[10]: bReqDmaState %d \n",bReqDmaState);
		
	    }

	    if(pDmaChannel)  {						/* DMA꤬Ƥ */
		if((MGC_DMA_STATUS_FREE == 				/* DMAžνλǧ */
		  pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel))){
		    DBG(2, "TestBulk RxReady pDmaChannel->dwActualLength=%d \n", pDmaChannel->dwActualLength );
		    if( 0 == pDmaChannel->dwActualLength )  {	/* ʤˤ⤷ʤ  (RxPktRdyΤ߼) */
		    }
		    else{
			    /* DMAžĹ򡢼žĹ­ */
			    pThis->aLocalEnd[bEnd].dwOffset += pDmaChannel->dwActualLength;
		            pRequest->actual+= pDmaChannel->dwActualLength;
			    dwFifoCount = pDmaChannel->dwActualLength;
			    pDmaChannel->dwActualLength = 0;

                            /* Mode1ξϡDMAλ */
                            if(MGC_M_RXCSR_DMAMODE == (wCsrVal&MGC_M_RXCSR_DMAMODE)){
			    	pThis->pDmaController->pfDmaStopChannel(pDmaChannel);	
				pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
 	        		pThis->pDmaController->pfDmaProgramChannel(pDmaChannel, 
 	        		    MGC_RxDmaCallback, pData,pEnd->wPacketSize, 0, 
 	        		    (uint8_t *)(pRequest->dma + (void *)pRequest->actual), 0);
				/* ž̤λĺǸ˥硼ȥѥåȤ */
 				if ( (NULL!=pRequest) && (((pRequest->actual)<(pRequest->length) ) && 
				     (0 == ((pRequest->actual)%(get_ep_packet_size(pThis, bEnd)))))) {
					wCsrVal &= ~(MGC_M_RXCSR_DMAENAB);
					MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
                                        spin_unlock(&pEnd->Lock);
					return;  /* Modyfied for AUTOCLEAR Clear */
				}
				else{
				    /* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
				    wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);
				}

				MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
			    }
			    /* Mode0ξϡRXPKTRDYΥꥢ  */
			    else if( MGC_M_RXCSR_DMAMODE != (wCsrVal&MGC_M_RXCSR_DMAMODE)){
			    	pThis->pDmaController->pfDmaStopChannel(pDmaChannel);	
				pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
 	        		pThis->pDmaController->pfDmaProgramChannel(pDmaChannel, 
 	        		    MGC_RxDmaCallback, pData,pEnd->wPacketSize, 0, 
 	        		    (uint8_t *)(pRequest->dma + (void *)pRequest->actual), 0);
				if((USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType)){
			    		wCsrVal |=   MGC_M_RXCSR_DMAENAB ;
					/* Modyfied for AUTOCLEAR Clear *//* AUTOSETΥꥢ */
					wCsrVal &= ~(MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_DMAMODE);
					MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR+1, bEnd, (0x00ff&(wCsrVal>>8)) );
					/* OverrunǤMAXPʲξϡ
					   MGC_M_RXCSR_RXPKTRDY򤪤Ȥʤ */
					if ( 0 == pEnd->dwRxOverrun || TRUE == pEnd->bRxMaxpOverrun ){
					
						wCsrVal &= ~(MGC_M_RXCSR_RXPKTRDY);
					}
				    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
				}
				else{
					wCsrVal &= ~(MGC_M_RXCSR_DMAENAB);
					/* OverrunǤMAXPʲξϡ
					   MGC_M_RXCSR_RXPKTRDY򤪤Ȥʤ */
					if ( 0 == pEnd->dwRxOverrun || TRUE == pEnd->bRxMaxpOverrun ){
					
						wCsrVal &= ~(MGC_M_RXCSR_RXPKTRDY);
					}
				    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
				}
			    }
		    }
		}
		else{
        	    /* DMAžλRxPtkRdyߤΤϡDmaMode1
        	       硼ȥѥåȽλξ */
        	    if( MGC_M_RXCSR_DMAMODE == (wCsrVal&MGC_M_RXCSR_DMAMODE) ){
		      if( 0 == pEnd->bIsInitRequest ){
                          /* Mode1ξϡ硼ȥѥåȤˤǤξν */
                          if( (pRequest->status == -ECONNRESET)) {
                              spin_unlock(&pEnd->Lock);
                              return;
                          }
			  DBG(2, "TestBulk Short Packet! \n");
			  DBG(2, "TestBulk wCsrVal = %x \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));
			  DBG(2, "TestBulk DmaGetChannelStatus(pDmaChannel) = %d \n", 
			      pThis->pDmaController->pfDmaGetChannelStatus(pDmaChannel));
			  DBG(2, "TestBulk RXMAXP = %d \n", MGC_ReadCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd));

                          spin_unlock(&pEnd->Lock);
                          
			if( (1==pRequest->zero) ){
			    rxstate(pThis, (struct MGC_UsbRequest*)pRequest); /* NULLѥåȼ */
			    if( (0==pRequest->zero) ){
				return;
			    }
			}
			else{
			    /* 硼ȥѥåȼ */
			    rxstate(pThis, (struct MGC_UsbRequest*)pRequest);
			}
			spin_lock(&pEnd->Lock);
			MGC_WriteCsr8(pBase, MGC_O_HDRC_RXCSR, bEnd, (0x00ff&wCsrVal) );
#ifdef USB_DMA_SHORTPKT_USING_CPU_TRANSFER		/* 4byte饤ȤΤ᥷硼ȥѥåȤCPUž */
			bReqDmaState = USB_REQ_CPU;
			DBG(1,"RxReady[11]: bReqDmaState %d \n",bReqDmaState);
#else
		     if(MGC_aGadgetLocalEnd[bEnd].bDma_ena==USB_MCD_DMA_SETON){
				/* DMA transfer */
		     	bReqDmaState = USB_REQ_DMA_MODE0;
				DBG(1,"RxReady[12]: bReqDmaState %d \n",bReqDmaState);
		     }else{
		     	/* CPU transfer */
		     	bReqDmaState = USB_REQ_CPU;
				DBG(1,"RxReady[13]: bReqDmaState %d \n",bReqDmaState);
		     }
#endif
		      }
		    }
		}
	    }
            spin_unlock(&pEnd->Lock);
#else
	    spin_unlock(&pEnd->Lock);
	    rxstate(pThis, (struct MGC_UsbRequest*)pRequest);
#endif /* MUSB_DMA */
	} else {	
	    spin_unlock(&pEnd->Lock);
	    DBG(3, "Rx: bytes waiting on ep=0x%x\n", bEnd); 
	    return;
	}	
    } while(FALSE);

#ifdef MUSB_DMA
    DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 4\n" );
    DBG(2, "TestBulk (wCsrVal&MGC_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal&MGC_M_RXCSR_RXPKTRDY) );
    /* žλޤϥ硼ȥѥåȤǴλ (DMAžξ) */
    /* MaxpOverData DmaMode0 is not used. */
    if ( (((NULL!=pRequest) && (((pRequest->actual)>=(pRequest->length) ) ||
          (0 != ((pRequest->actual)%(get_ep_packet_size(pThis, bEnd)))))) && 
          (pDmaChannel)) || (TRUE == pEnd->bRxMaxpOverrun) ) {
        if (pRequest->zero
          && pRequest->status == -EINPROGRESS
          && pRequest->actual == 0) {
            pRequest->zero = 0;
	    return;
        }
	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 5\n" );
	DBG(2, "TestBulk (wCsrVal&MGC_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal&MGC_M_RXCSR_RXPKTRDY) );
        spin_lock(&pEnd->Lock);
	if( (USB_REQ_DMA_MODE1 == bReqDmaState) ||
	    (USB_REQ_DMA_MODE0 == bReqDmaState) ) {

		DBG(1,"RxReady[15]: bReqDmaState %d \n",bReqDmaState);
	    pRequest=(pEnd->bInactive) ? NULL : MGC_CurrentRequest(pEnd);		    
	    pDmaChannel = pThis->aLocalEnd[bEnd].pDmaChannel;
    	    if(pDmaChannel)  {						/* DMA꤬Ƥ */
		if( ( NULL != pThis->aLocalEnd[bEnd].pDmaChannel->pData ) &&
		    ( TRUE != *((uint8_t *)(pThis->aLocalEnd[bEnd].pDmaChannel->pData))) ){
		      return;
		}
		pThis->pDmaController->pfDmaStopChannel(pDmaChannel);	/* ɲ */
		pThis->pDmaController->pfDmaDisableIntChannel(pDmaChannel);
		/* (pfDmaReleaseBuffer();) */
		pThis->pDmaController->pfDmaReleaseChannel(pDmaChannel);
		pThis->aLocalEnd[bEnd].pDmaChannel = pDmaChannel = FALSE;
	    }
	}
							/* ꥯȾ֤λˤ done */
	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 6\n" );
	/* OUT_BULK ХξΥ顼ȡshort_not_okΥå */
	if((NULL!=pRequest) && (pRequest->status == -EINPROGRESS)){
	    if (0 == pEnd->dwRxOverrun){
		if(!((1==pRequest->short_not_ok) && 
		     ((0!=((pRequest->actual)%(get_ep_packet_size(pThis, bEnd))))||(0==dwFifoCount)))){
		    pRequest->status=0;
		    MGC_CompleteRequest(pRequest, pRequest->status); /* async complete */
		}
		else{
		    pRequest->status=-EREMOTEIO;
		    MGC_CompleteRequest(pRequest, pRequest->status); 
		}
	    }
	    else{
		if(TRUE == pEnd->bRxMaxpOverrun){
		    pRequest->status = -ECOMM;
		}
		else {
		    pRequest->status = -EOVERFLOW;
		}
		pRequest->actual = pRequest->actual + pEnd->dwRxOverrun;
		pEnd->dwRxOverrun =  0;
		pEnd->bRxMaxpOverrun = FALSE;
		MGC_CompleteRequest(pRequest, pRequest->status);
	    }

	}
	spin_unlock(&pEnd->Lock);
	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 7\n" );
    }
    else{
	DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 5.1\n" );
	DBG(2, "TestBulk (wCsrVal&MGC_M_RXCSR_RXPKTRDY)=%x \n", (wCsrVal&MGC_M_RXCSR_RXPKTRDY) );
	/* 󥻥뽪λƤʤ,RXPKTRDYäƤ */
	if( (NULL!=pRequest) && (pRequest->status != -ECONNRESET)&&
	   ((wCsrVal&MGC_M_RXCSR_RXPKTRDY) || (USB_REQ_DMA_MODE1==bReqDmaState)) ){
	    DBG(1,"RxReady[16]: bReqDmaState %d \n",bReqDmaState);

	    spin_lock(&pEnd->Lock);
    	    rxstate(pThis, (struct MGC_UsbRequest*)pRequest);	/* μ */
	    spin_unlock(&pEnd->Lock);
	    DBG(2, "TestBulk pRequest->actual=%d \n", pRequest->actual );
    	}
    }
#endif /* MUSB_DMA */
    DBG(2, "TestBulk MGC_HdrcServiceDeviceRxReady() 8\n" );

    DBG(2, "==>\n" );
}


int MGC_RestartRequest(MGC_LinuxCd* pThis, uint8_t bEnd, struct usb_request *pRequest) {
    struct MGC_UsbRequest *req=(struct MGC_UsbRequest *)pRequest;

    DBG(2, "TestBulk MGC_RestartRequest 1\n" );

    if ( !pRequest ) {
    DBG(3, "nothing to execute on nEnd=0x%x, (tx=%d)\n", bEnd, \
    	MGC_aGadgetLocalEnd[bEnd].bIsTx);
     	return -EINVAL;
    }
   
    DBG(2, "TestBulk MGC_RestartRequest 2\n" );
    DBG(3, "executing pRequest=%p on nEnd=0x%x, (tx=%d), pRequest->length=%d, " \
	"pRequest->actual=0x%x\n", pRequest, bEnd, MGC_aGadgetLocalEnd[bEnd].bIsTx, \
	pRequest->length, pRequest->actual);

    if( MGC_aGadgetLocalEnd[bEnd].bIsTx ) {
	DBG(2, "TestBulk MGC_RestartRequest 31\n" );
	txstate(pThis, req);
	DBG(2, "TestBulk MGC_RestartRequest 32\n" );
    } else {
	rxstate(pThis, req);  
	DBG(2, "TestBulk MGC_RestartRequest 4\n" );
    }
    DBG(2, "TestBulk MGC_RestartRequest 5\n" );

    return 0;
}


/**************************************************************************
Gadget Functions
**************************************************************************/

/* new interface */
extern uint8_t MGC_HdrcServiceFunctionEp0(MGC_LinuxCd* pThis);

void MGC_HdrcServiceDeviceDefaultEnd(MGC_LinuxCd* pThis) {
	MGC_HdrcServiceFunctionEp0(pThis);
}


/* ------------------------------------------------------------ */


/**
 * The endpoint selection policy should have already selected
 * an appropriate usb_ep from us, since we follow the naming convention.
 */
STATIC int MGC_GadgetEnableEnd(struct usb_ep *ep,
	const struct usb_endpoint_descriptor *desc)
{
    unsigned long flags;
    uint8_t bEnd = EP_NUMBER(ep);
    uint8_t* pBase=MGC_pFirstDriver->pRegs;
    uint16_t wCsrVal;
    uint16_t wIntrTxE;
    uint16_t wIntrRxE;
    MGC_GadgetLocalEnd* pEnd=(MGC_GadgetLocalEnd*)ep;

    DBG(2, "<==\n ep=%p, desc=%p, nEnd=0x%x\n", ep, desc, 
    	EP_NUMBER(ep));
    
#ifdef MUSB_PARANOID	
    ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
    
    if ( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }

    if ( MGC_GadgetFindEnd(ep) < 0  ) {
    	ERR("invalid endpoint ep=%p\n", ep);
        return -EINVAL;
    }

    if( ( EP_NUMBER(ep) != (desc->bEndpointAddress & USB_ENDPOINT_NUMBER_MASK))) {
        return -EINVAL;
    }

    if ( !MGC_IsIdle(ep) ) {
	WARN("Enabling ep%d with request queued!\n", EP_NUMBER(ep));
    };

#endif

    spin_lock_irqsave(&pEnd->Lock, flags);

    ep->maxpacket = desc->wMaxPacketSize;
    pEnd->wPacketSize = desc->wMaxPacketSize;
    pEnd->bTrafficType = desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;
    pEnd->bIsTx = (desc->bEndpointAddress & USB_ENDPOINT_DIR_MASK) ? 
      TRUE : FALSE;
    pEnd->bInactive = 0;

    INIT_LIST_HEAD( &pEnd->req_list );

    /* register setting */
    MGC_SelectEnd(pBase, bEnd);

    /* USB is not configured with High-bandwidth or split transaction */
    if (pEnd->bIsTx) {
        /*  */
        MGC_WriteCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd, pEnd->wPacketSize);
	/* եåFIFO */
	MGC_GadgetFifoFlush_FlushOnly(ep);
	/* ǡȥ򥯥ꥢ */
	/* STALLꥢ */
	MGC_GadgetSetHalt(ep, 0);

        /* EPγߵ */
        wIntrTxE = MGC_Read16(pBase, MGC_O_HDRC_INTRTXE);
        MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, wIntrTxE | (1 << bEnd));
    }
    else {
        /*  */
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd, pEnd->wPacketSize);

        if(USB_ENDPOINT_XFER_BULK == pEnd->bTrafficType){
            wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
            wCsrVal |= (MGC_M_RXCSR_DMAENAB );
            MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsrVal);
        }
	/* եåFIFO */
	MGC_GadgetFifoFlush_FlushOnly(ep);
	/* ǡȥ򥯥ꥢ */
	/* STALLꥢ */
	MGC_GadgetSetHalt(ep, 0);

        /* EPγߵ */
        wIntrRxE = MGC_Read16(pBase, MGC_O_HDRC_INTRRXE);
        MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, wIntrRxE | (1 << bEnd));
    }
    
    spin_unlock_irqrestore(&pEnd->Lock, flags);
    
#if MUSB_DEBUG > 0
    memcpy(&pEnd->epDescriptor, desc, sizeof(struct usb_endpoint_descriptor));
#endif

    DBG(2, "==>\n");
    return 0;
}

int MGC_GadgetEnableEndPoint(uint8_t bEnd, const struct usb_endpoint_descriptor *desc) {
#if MUSB_DEBUG > 0
   if ( !desc ) {
	desc=&MGC_aGadgetLocalEnd[bEnd].epDescriptor;
   }   
#endif
   if ( !desc ) {
    	ERR("invalid description activating ep%d\n", bEnd);
   	return -EINVAL;
   }
   return MGC_GadgetEnableEnd( &MGC_aGadgetLocalEnd[bEnd].end_point, desc);
}

/* Disable an endpoint.
 * @param struct usb_ep *ep the endpoint to disable.
 */
STATIC int MGC_GadgetDisableEnd(struct usb_ep *ep)
{
    unsigned long flags;
    uint8_t bEnd = EP_NUMBER(ep);
    uint8_t* pBase=MGC_pFirstDriver->pRegs;
    uint16_t wIntrTxE;
    uint16_t wIntrRxE;
    MGC_GadgetLocalEnd* pEnd=(MGC_GadgetLocalEnd*)ep;
    struct usb_request* req;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    int status;

    DBG(2, "<==\n" );
    DBG(-1, "disabling ep %d\n", EP_NUMBER(ep));

#ifdef MUSB_PARANOID
    ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock);     

    if (!MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }

    if ( MGC_GadgetFindEnd(ep)<0  ) {
    	ERR("invalid endpoint ep=%p\n", ep);
        return -EINVAL;
    }
#endif
    
    spin_lock_irqsave(&pEnd->Lock, flags);

    ((MGC_GadgetLocalEnd*)ep)->bInactive=1;    

    /* EPγԵ */
    if (pEnd->bIsTx) {
        /*  */
        wIntrTxE = MGC_Read16(pBase, MGC_O_HDRC_INTRTXE);
        MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, wIntrTxE & ~(1 << bEnd));
    }
    else {
        /*  */
        wIntrRxE = MGC_Read16(pBase, MGC_O_HDRC_INTRRXE);
        MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, wIntrRxE & ~(1 << bEnd));
    }

    spin_unlock_irqrestore(&pEnd->Lock, flags);

    req = MGC_CurrentRequest(pEnd);

    spin_lock_irqsave(&pEnd->Lock, flags);

    status = -ESHUTDOWN;
    
    if (!bEnd) {
        if (req) {
	    req->status = status;
            MGC_CompleteEp0Request();
        }
    }
    else if (req && req->status != -ESHUTDOWN) {
        struct usb_request* tmp;
        stop_dma(pThis, bEnd);
        list_for_each_entry_safe(req, tmp, &pEnd->req_list, list) {
	    if(-EINPROGRESS == req->status){
	        req->status = status;
	        list_del(&req->list);
	        spin_unlock_irqrestore(&pEnd->Lock, flags);
	        if (req->complete) {
	            req->complete(&pEnd->end_point, req);
	        }
                spin_lock_irqsave(&pEnd->Lock, flags);
            }
        }
    }
    spin_unlock_irqrestore(&pEnd->Lock, flags);

    MGC_GadgetFifoFlush_FlushOnly(ep);

    DBG(2, "==>\n" );
    return 0;
}

int MGC_GadgetDisableEndPoint(uint8_t bEnd) {
    return MGC_GadgetDisableEnd( &MGC_aGadgetLocalEnd[bEnd].end_point);
}

/**
 */
STATIC struct usb_request*
	MGC_GadgetAllocRequest(struct usb_ep *ep, int gfp_flags)
{
    struct MGC_UsbRequest*  pRequest=NULL;
    
    DBG(2, "<==\n" );
    
#ifdef MUSB_PARANOID
    if ( MGC_GadgetFindEnd(ep)<0 ) {    	
    	ERR("invalid endpoint ep=%p\n", ep);
    	return NULL;
    }
#endif    
    
    KMALLOC(pRequest, sizeof(struct MGC_UsbRequest), gfp_flags);    	
    if ( pRequest ) {
	memset(pRequest, 0, sizeof(struct MGC_UsbRequest));
	
    	/* make sure it's fine! */
	INIT_LIST_HEAD(&pRequest->request.list);    
	pRequest->completition_list.prev=0;
	pRequest->completition_list.next=0;		
	pRequest->bEnd=EP_NUMBER(ep);

#if MUSB_DEBUG > 0
	pRequest->bQueued=0;
	pRequest->dwPadFront = MGC_PAD_FRONT;
	pRequest->dwPadBack = MGC_PAD_BACK;
#endif
    }

    DBG(2, "==> allocated request at %p for ep %d\n", pRequest, \
    	EP_NUMBER(ep));
    return (struct usb_request*)pRequest;
}


/**
 */
STATIC void MGC_GadgetFreeRequest(struct usb_ep *ep, struct usb_request *req)
{
#if MUSB_DEBUG > 0
    is_corrupted_request(req);
    if ( req )
#endif
    {
	KFREE(req);
    }
    DBG(2, "freed request at %p\n", req);
}


/**
 */
STATIC void* MGC_GadgetAllocBuffer(struct usb_ep *ep, unsigned bytes,
				   dma_addr_t *dma, int gfp_flags)
{
#ifdef MUSB_PARANOID    
    if (!MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return 0;
    }

    if ( MGC_GadgetFindEnd(ep) < 0 ) {
    	ERR("invalid endpoint ep=%p\n", ep);
    	return 0;
    } 

    if ( dma ) {
    	WARN("requested DMA for ep%d\n", EP_NUMBER(ep));
    }
#endif

    if ( !EP_NUMBER(ep) && dma ) {
        *dma = 0;
        dma = 0L;
    }

    /* PIO when DMA buffer pointer is NULL */
    if (!dma) {
        void *vadr;
        /* in case without DMAC */
        KMALLOC(vadr, bytes, gfp_flags);
        DBG(4, "malloced pAddr=%p, pDma=%p\n", vadr, dma); 
        return vadr;
    }

    /* DMA is controlled in the endpoint initialization; no DMA for now... */
    return MGC_AllocBufferMemory(MGC_pFirstDriver, bytes, gfp_flags, dma);
}



/**
 */
STATIC void MGC_GadgetFreeBuffer(struct usb_ep *ep, void *address, dma_addr_t dma,
				 unsigned bytes)
{
    if (!dma) {
        KFREE(address);
    }
    else {
        MGC_FreeBufferMemory(MGC_pFirstDriver, bytes, address, dma);
    }
}


/* Queue requests to endpoints for execution.
 * 
 * @param struct usb_ep *ep
 * @param struct usb_request *pRequest
 * @param int gfp_flags
 */
 
STATIC int MGC_GadgetQueue(struct usb_ep *ep, struct usb_request *pRequest,
    int gfp_flags)
{
    const uint8_t bEnd=EP_NUMBER(ep);
    MGC_GadgetLocalEnd* pEnd=(MGC_GadgetLocalEnd*)ep;
#ifdef MUSB_DMA
    unsigned long flags;
#endif /* MUSB_DMA */
#ifdef MUSB_PARANOID    
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
#endif
    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID    
    if (!pThis ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
    
    if ( !pRequest ) {
    	ERR("pRequest=%p\n", pRequest);
        return -EINVAL;    
    }

    if ( MGC_GadgetFindEnd(ep) < 0 ) {
    	ERR("invalid endpoint ep=%p\n", ep);
    	return -EINVAL;
    } 

   ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
   
   if ( is_queued_for_completition(pRequest) ) {
	ERR("pRequest=%p is in the completion list %s!\\n", pRequest, 
	    	dump_node( &((struct MGC_UsbRequest*)pRequest)->completition_list) ); 
	return -EINVAL;
   }
#endif

    DBG(2, "TestBulk MGC_GadgetQueue() 1 EP=%d \n", bEnd );
    pRequest->actual=0;    
    pRequest->status=-EINPROGRESS;    
    ((struct MGC_UsbRequest*)pRequest)->bEnd=bEnd;    
    /* Request was length=0,So it must be zero=0. */
    if( 0 == pRequest->length )  pRequest->zero = 0;  
    INIT_LIST_HEAD(&pRequest->list);

    DBG(2, "TestBulk MGC_GadgetQueue() 2\n" );
    /* list count==1 pRequest is the head of the list */
    EP_SPIN_LOCK_IRQSAVE( ep, flags );
    list_add_tail( &(pRequest->list), &(pEnd->req_list) );	
    if ( !pEnd->bInactive && (pRequest==MGC_CurrentRequest(pEnd)) ) {
        EP_SPIN_UNLOCK_IRQRESTORE( ep, flags );
	DBG(2, "TestBulk MGC_GadgetQueue() 3\n" );

	if ( bEnd==0 ) {
	    DBG(-6, "executing on bEnd=0x%x, (%s), pRequest->length=%d\n", \
		bEnd, pEnd->bIsTx?"tx":"rx", pRequest->length);     

	    DBG(2, "TestBulk MGC_GadgetQueue() 4\n" );
	    switch ( MGC_pFirstDriver->bEnd0Stage )  {		
		/* sequence #1, TX State */
    		case MGC_END0_STAGE_TX: 
		    ep0_txstate();
		    break;

		/* sequence #2, RX State */
    		case MGC_END0_STAGE_RX:
            ep0_rxstart();
		break;

	       /* certain gadged may keep ep0 busy; g_file_storage
		  does it after a set_config command.*/
		default: {
		    const uint8_t code=(pRequest->actual==pRequest->length)
			?0:-EINVAL;
		    pRequest->status=code;
		    MGC_CompleteEp0Request();
		} break;
	    }

	} else {
	    MGC_Call_RestartRequest( ep, pRequest, gfp_flags );
	}
    } else {
        EP_SPIN_UNLOCK_IRQRESTORE( ep, flags );
    }


    DBG(2, "==>\n");
    return 0;
}


STATIC int MGC_Call_RestartRequest(struct usb_ep *ep, struct usb_request *pRequest,
    int gfp_flags)		/* Modyfied for request queue restart */
{
    const uint8_t bEnd=EP_NUMBER(ep);
    MGC_GadgetLocalEnd* pEnd=(MGC_GadgetLocalEnd*)ep;
#ifdef MUSB_DMA
    unsigned long flags;
    uint16_t wIntrRxE;
    uint8_t *pBase = (uint8_t *)(MGC_pFirstDriver->pRegs);
#endif /* MUSB_DMA */

   if( pEnd->bIsTx ){
            MGC_RestartRequest(MGC_pFirstDriver, EP_NUMBER(ep), 
	    	pRequest);
    }
    else{

	DBG(2, "TestBulk MGC_GadgetQueue() 5\n" );

	spin_lock_irqsave(&(MGC_pFirstDriver->Lock), flags);
	pEnd->bIsInitRequest = USB_INIT_REQ_PHASE1;
	wIntrRxE = MGC_Read16(pBase, MGC_O_HDRC_INTRRXE);
        MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, wIntrRxE & ~(1 << bEnd));
	DBG(2, "TestBulk MGC_GadgetQueue() wIntrRxE=%x  wIntrRxE & ~(1 << bEnd)=%x \n",wIntrRxE, MGC_Read16(pBase, MGC_O_HDRC_INTRRXE) );
	DBG(2, "TestBulk EP%d CSR=%x 2 INTNG \n", bEnd, MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));

	MGC_SelectEnd(pBase, bEnd); 
        MGC_RestartRequest(MGC_pFirstDriver, EP_NUMBER(ep), 
	    pRequest);

	DBG(2, "TestBulk EP%d CSR=%x 3 INTNG \n", bEnd, MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));
	DBG(2, "TestBulk MGC_GadgetQueue() 5.1\n" );
        MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, wIntrRxE);
	DBG(2, "TestBulk MGC_GadgetQueue() wIntrRx =%x  \n",wIntrRxE );

	pEnd->bIsInitRequest = USB_INIT_REQ_PHASE2;

        MGC_RestartRequest(MGC_pFirstDriver, EP_NUMBER(ep), 
	    pRequest);

	pEnd->bIsInitRequest = 0;

	DBG(2, "TestBulk MGC_GadgetQueue() 5.2\n" );
	DBG(2, "TestBulk EP%d CSR=%x 4 INTE \n", bEnd, MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd));
	spin_unlock_irqrestore(&(MGC_pFirstDriver->Lock), flags);
    }

    return 0;
}

STATIC int MGC_GadgetDequeue(struct usb_ep *ep, struct usb_request *pRequest)
{
    DBG(2, "<==\n" );
    DBG(3, "dequeuing from nEnd=0x%x, pEnd=%p, pRequest=%p\n", \
    	EP_NUMBER((MGC_GadgetLocalEnd*)ep), (MGC_GadgetLocalEnd*)ep, pRequest);

#ifdef MUSB_PARANOID    
    if (!MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }

    if ( MGC_GadgetFindEnd(ep)<0 ) {
    	ERR("invalid endpoint ep=%p\n", ep);
        return -EINVAL;    
    }

    if ( !pRequest ) {
    	return -EINVAL; /* abort */
    }
    
    ASSERT_SPINLOCK_UNLOCKED(&((MGC_GadgetLocalEnd*)ep)->Lock); 
#endif

    return complete_request_immediate(ep, pRequest, -ECONNRESET);
}

int MGC_GadgetSetHalt(struct usb_ep *ep, int value)
{
    uint16_t wCsr;
    unsigned long flags;
    const uint8_t bEnd=EP_NUMBER(ep);
    uint8_t* pBase=MGC_pFirstDriver->pRegs;
    MGC_GadgetLocalEnd* pEnd =(MGC_GadgetLocalEnd*)ep;
    struct MGC_UsbRequest *req=(struct MGC_UsbRequest*)MGC_CurrentRequest(pEnd);

    DBG(3, "<== end=%d, value=%d\n", bEnd, value);
    
#ifdef MUSB_PARANOID
    if ( MGC_GadgetFindEnd(ep)<0 ) {
    	ERR("invalid endpoint ep=%p\n", ep);
        return -EINVAL;
    }
#endif    
    
    spin_lock_irqsave(&(MGC_pFirstDriver->Lock), flags);
    MGC_SelectEnd(pBase, bEnd );

    if ( 0==bEnd ) {	
	((MGC_GadgetLocalEnd*)ep)->bInactive=value;

	/* end 0 should never stall! */
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0);
	if ( value ) {
	    wCsr |= MGC_M_CSR0_P_SENDSTALL;
	} else {
	    wCsr &= ~MGC_M_CSR0_P_SENDSTALL;
	}
	wCsr &= ~MGC_M_CSR0_TXPKTRDY;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsr);        
	MGC_GadgetFifoFlush_FlushOnly(ep);
	spin_unlock_irqrestore( &(MGC_pFirstDriver->Lock), flags );
	return 0;
    }

    /* prevent further request to be executed */    
    pEnd->bInactive=value;

    /* abort the current request? */    
    if ( value && !pEnd->bInactive && req && ((MGC_GadgetLocalEnd*)ep)->bIsTx ) {    

	wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
	if ( wCsr & MGC_M_TXCSR_FIFONOTEMPTY ) {
	    DBG(-1, "ep=%d: %d request(s), %s cannot halt it\n",
		bEnd, queue_length( &pEnd->req_list ), 
		dump_usb_request( MGC_CurrentRequest(pEnd)) );	

	    spin_unlock_irqrestore(&(MGC_pFirstDriver->Lock), flags);    
	    return -EAGAIN;    
	}

    }
    
    if ( ((MGC_GadgetLocalEnd*)ep)->bIsTx ) {    
	wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
	if ( value ) {
	    wCsr |= MGC_M_TXCSR_P_SENDSTALL;
	} else {
	    wCsr &= ~MGC_M_TXCSR_P_SENDSTALL;
	    wCsr |=  MGC_M_TXCSR_CLRDATATOG;
	}
	
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
	MGC_GadgetFifoFlush_FlushOnly(ep);
    } else {
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
	if( value ) {
	    wCsr |= MGC_M_RXCSR_P_SENDSTALL;
	} else {
	    wCsr &= ~MGC_M_RXCSR_P_SENDSTALL;
	    wCsr |=  MGC_M_RXCSR_CLRDATATOG;
	}
	
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
	MGC_GadgetFifoFlush_FlushOnly(ep);
    }

    spin_unlock_irqrestore(&(MGC_pFirstDriver->Lock), flags);
    DBG(2, "==>\n" );    
    return 0;
}

STATIC int MGC_GadgetFifoStatus(struct usb_ep *ep)
{
    unsigned long flags;
    int nResult = 0;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    uint8_t* pBase=pThis->pRegs;
    int bEnd = EP_NUMBER(ep);

    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
    ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
    ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 

    if( MGC_GadgetFindEnd(ep)<0 ) {
    	ERR("invalid endpoint ep=%p\n", ep);
        return -EINVAL;
    }
#endif


    spin_lock_irqsave(&(pThis->Lock), flags);
    MGC_SelectEnd(pBase, bEnd);    
    nResult = MGC_ReadCsr16(pBase, (bEnd)? MGC_O_HDRC_RXCOUNT:MGC_O_HDRC_COUNT0, bEnd);
    spin_unlock_irqrestore(&(pThis->Lock), flags);

    DBG(2, "==>\n" );
    return nResult;
}

STATIC void MGC_GadgetFifoFlush(struct usb_ep *ep)
{
    uint16_t wCsr;
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    uint8_t* pBase = pThis->pRegs;
    int nEnd = EP_NUMBER(ep);
    struct usb_request *req;

    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
   ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
#endif


    spin_lock_irqsave(&(pThis->Lock), flags);
    MGC_SelectEnd(pBase, (uint8_t)nEnd);

    if(nEnd) {
        /* EP0ʳ */
        struct timeval tv;
        unsigned long susec;
        unsigned long check_time;
        /* TX */
        /* ϻּ */
        do_gettimeofday(&tv);
        susec = tv.tv_usec;
        /* bogoMIPSˤ5usȤлֳǧ */
        check_time = loops_per_jiffy / (200000 / HZ);
        do {
            wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, (uint8_t)nEnd);
            if (!(wCsr & MGC_M_TXCSR_FIFONOTEMPTY)) {
                break;
            }
            if (check_time == 0) {
                do_gettimeofday(&tv);
                if (tv.tv_usec - susec > 200) { /* Ϥ200ustimeout */
                    break;
                }
                check_time = loops_per_jiffy / (200000 / HZ); /* 5us */
            }
            check_time--;
            wCsr = (wCsr | MGC_M_TXCSR_FLUSHFIFO) & ~MGC_M_TXCSR_TXPKTRDY;
            MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, (uint8_t)nEnd, wCsr);
        } while (1);
        /* RX */
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd);
        wCsr |= MGC_M_RXCSR_FLUSHFIFO;
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd, wCsr);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd, wCsr);
        MGC_FlushEpRequests((MGC_GadgetLocalEnd*)ep, -ECONNRESET);
    } else {
        /* EP0 */
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, (uint8_t)nEnd);
        wCsr |= (MGC_M_CSR0_FLUSHFIFO | MGC_M_CSR0_P_SENTSTALL);
        wCsr &= ~MGC_M_CSR0_TXPKTRDY;
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsr);
        req = MGC_CurrentRequest(&MGC_aGadgetLocalEnd[0]);
        if (req) {
            req->status = -ECONNRESET;
            MGC_CompleteEp0Request();
        }
    }

    /* re-enable interrupt */
    spin_unlock_irqrestore(&(pThis->Lock), flags);
}

STATIC void MGC_GadgetFifoFlush_FlushOnly(struct usb_ep *ep)
{
    uint16_t wCsr;
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    uint8_t* pBase = pThis->pRegs;
    int nEnd = EP_NUMBER(ep);

    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
   ASSERT_SPINLOCK_UNLOCKED(&pEnd->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
#endif


    spin_lock_irqsave(&(pThis->Lock), flags);
    MGC_SelectEnd(pBase, (uint8_t)nEnd);

    if(nEnd) {
        /* EP0ʳ */
        struct timeval tv;
        unsigned long susec;
        unsigned long check_time;
        /* TX */
        /* ϻּ */
        do_gettimeofday(&tv);
        susec = tv.tv_usec;
        /* bogoMIPSˤ5usȤлֳǧ */
        check_time = loops_per_jiffy / (200000 / HZ);
        do {
            wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, (uint8_t)nEnd);
            if (!(wCsr & MGC_M_TXCSR_FIFONOTEMPTY)) {
                break;
            }
            if (check_time == 0) {
                do_gettimeofday(&tv);
                if (tv.tv_usec - susec > 200) { /* Ϥ200ustimeout */
                    break;
                }
                check_time = loops_per_jiffy / (200000 / HZ); /* 5us */
            }
            check_time--;
            wCsr = (wCsr | MGC_M_TXCSR_FLUSHFIFO) & ~MGC_M_TXCSR_TXPKTRDY;
            MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, (uint8_t)nEnd, wCsr);
        } while (1);
        /* RX */
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd);
        wCsr |= MGC_M_RXCSR_FLUSHFIFO;
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd, wCsr);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, (uint8_t)nEnd, wCsr);
    } else {
        /* EP0 */
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, (uint8_t)nEnd);
        wCsr |= (MGC_M_CSR0_FLUSHFIFO | MGC_M_CSR0_P_SENTSTALL);
        wCsr &= ~MGC_M_CSR0_TXPKTRDY;
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsr);
    }

    /* re-enable interrupt */
    spin_unlock_irqrestore(&(pThis->Lock), flags);
}

/** general operations */
STATIC int MGC_GadgetGetFrame(struct usb_gadget *gadget)
{
#ifdef MUSB_PARANOID
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif
    return (int)MGC_Read16(MGC_pFirstDriver->pRegs, MGC_O_HDRC_FRAME);
}

/** */
STATIC int MGC_GadgetWakeup(struct usb_gadget *gadget)
{
    uint8_t power;
    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif
    
    power = MGC_Read8(MGC_pFirstDriver->pRegs, MGC_O_HDRC_POWER);
    power |= MGC_M_POWER_RESUME;
    MGC_Write8(MGC_pFirstDriver->pRegs, MGC_O_HDRC_POWER, power);

    return 0;
}

/** */
STATIC int MGC_GadgetSetSelfPowered(struct usb_gadget *gadget, 
    int is_selfpowered)
{
    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif

    MGC_pFirstDriver->bIsSelfPowered = (uint8_t)is_selfpowered;
    return 0;
}

#if defined(MUSB_OTG) && defined(GADGET_FULL)

/** */
STATIC int MGC_GadgetVbusSession(struct usb_gadget *gadget, int is_active)
{
    DBG(2, "<==>\n" );
    return -EINVAL;
}

/** */
STATIC int MGC_GadgetVbusDraw(struct usb_gadget *gadget, unsigned mA)
{
#ifdef MUSB_PARANOID
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif
    MGC_pFirstDriver->bIsSelfPowered=(mA)?TRUE:FALSE;
    return 0;
}

/** */
STATIC int MGC_GadgetPullup(struct usb_gadget *gadget, int is_on)
{
    unsigned long flags;
    uint8_t power;
    uint8_t intrusbe;
    uint8_t* pBase;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif

    spin_lock_irqsave(&(pThis->Lock), flags);
    pBase = pThis->pRegs;
    power = MGC_Read8(pBase, MGC_O_HDRC_POWER);

    if(is_on) {
        MUSB_DEV_MODE(pThis);
        pThis->bSuspended = FALSE;
        /* Suspend/Resume/Resetߤε */
        intrusbe = MGC_M_INTR_SUSPEND | MGC_M_INTR_RESUME | MGC_M_INTR_RESET;
        MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, intrusbe);
        DBG(3, "gadget %s connecting to host\n", \
            pThis->pGadgetDriver->function);
        power |= MGC_M_POWER_SOFTCONN;
        MGC_Write8(pBase, MGC_O_HDRC_POWER, power);
#if defined(CONFIG_USB_MCD_GPIO_VBUS)
        /* Vbus ON */
        musb_otg_control_force_vbus(pThis->otg_control, USB_OTG_VBUS_ON);
#endif /* defined(CONFIG_USB_MCD_GPIO_VBUS) */
    } else {
        MUSB_OTG_MODE(pThis);
        pThis->bSuspended = TRUE;
        DBG(3, "gadget %s disconnecting from host\n", \
            pThis->pGadgetDriver->function);
        /* Suspend/Resume/ResetߤԵ */
        MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, 0);
#ifdef MUSB_AVAIL_WA_SOFTCON
        /* SoftConơȽ */
 #if !defined(CONFIG_USB_MCD_GPIO_VBUS)
        /* VbusΤ */
        musb_otg_control_stop_detector(pThis->otg_control);
 #endif /* !defined(CONFIG_USB_MCD_GPIO_VBUS) */
        /* Vbus OFF */
        musb_otg_control_force_vbus(pThis->otg_control, USB_OTG_VBUS_OFF);
        /* SoftConӥڥɵĥӥåȤΥꥢ */
        power &= ~MGC_M_POWER_ENSUSPEND;
#endif /* MUSB_AVAIL_WA_SOFTCON */
        power &= ~MGC_M_POWER_SOFTCONN;
        MGC_Write8(pBase, MGC_O_HDRC_POWER, power);
#ifdef MUSB_AVAIL_WA_SOFTCON
        spin_unlock_irqrestore(&(pThis->Lock), flags);
        msleep(25);             /* ChirpλԤ25msWAIT */
        spin_lock_irqsave(&(pThis->Lock), flags);
 #if !defined(CONFIG_USB_MCD_GPIO_VBUS)
        /* Vbus OFFβ */
        musb_otg_control_force_vbus(pThis->otg_control, USB_OTG_VBUS_THRU);
 #endif /* defined(CONFIG_USB_MCD_GPIO_VBUS) */
        power = MGC_Read8(pBase, MGC_O_HDRC_POWER);
        power |= MGC_M_POWER_ENSUSPEND;
        MGC_Write8(pBase, MGC_O_HDRC_POWER, power);
 #if !defined(CONFIG_USB_MCD_GPIO_VBUS)
        /* VbusΤκƳ */
        musb_otg_control_start_detector(pThis->otg_control);
 #endif /* !defined(CONFIG_USB_MCD_GPIO_VBUS) */
#else /* MUSB_AVAIL_WA_SOFTCON */
 #if defined(CONFIG_USB_MCD_GPIO_VBUS)
        /* Vbus OFF */
        musb_otg_control_force_vbus(pThis->otg_control, USB_OTG_VBUS_OFF);
 #endif /* defined(CONFIG_USB_MCD_GPIO_VBUS) */
#endif /* MUSB_AVAIL_WA_SOFTCON */
    }

    spin_unlock_irqrestore(&(pThis->Lock), flags);

    DBG(2, "==>\n" );
    return 0;
}

#endif

STATIC int MGC_GadgetIoctl(struct usb_gadget *gadget,
	unsigned code, unsigned long param)
{
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "<==>\n" );

    switch (code) {
      case USB_IOCTL_FIFO_RESET: {
          musb_fifo_reset(pThis);
		  MGC_GadgetDmaSetDefault();
          return 0;
      }
      case USB_IOCTL_FIFO_MAP: {
          void* adr;
          uint8_t bEnd;
          struct usb_ioctl_fifo_param* fifo_param
              = (struct usb_ioctl_fifo_param*)param;
          adr = musb_fifo_map(pThis,
                              fifo_param->bEpAdr,
                              fifo_param->wSize,
                              fifo_param->bType, /* Double:2, Single:1 */
                              FALSE); /* Gadget */
          if (adr) {
              bEnd = fifo_param->bEpAdr & USB_ENDPOINT_NUMBER_MASK;
              MGC_aGadgetLocalEnd[bEnd].end_point.maxpacket = fifo_param->wSize;
          }
          return (adr ? 0 : -EINVAL);
      }
      case USB_IOCTL_FIFO_DUMP: {
          musb_fifo_dump(pThis);
          return 0;
      }
      case USB_IOCTL_TESTMODE: {
	    if( NULL == pThis )  return -EAGAIN;
	    pThis->bTestMode = TRUE;
	    switch(param) {
		case USB_IOCTL_TESTMODE_NORMAL:
		    DBG(3, "TESTMODE_NORMAL\n");
#ifdef DUMMY_USB_IOCTL_TESTMODE_NORMAL
		    pThis->bTestMode = FALSE;
#else
		    /* TESTMODE_CLEAR */
		    pThis->bTestModeValue = (uint8_t)NULL;
#endif /* DUMMY_USB_IOCTL_TESTMODE_NORMAL */
		    break;
		case USB_IOCTL_TESTMODE_TEST_J:
		    DBG(3, "TEST_J\n");
		    /* TEST_J */
		    pThis->bTestModeValue = MGC_M_TEST_J;
		    break;
		case USB_IOCTL_TESTMODE_TEST_K:
		    /* TEST_K */
		    DBG(3, "TEST_K\n");
		    pThis->bTestModeValue = MGC_M_TEST_K;
		    break;
		case USB_IOCTL_TESTMODE_SE0_NAK:
		    /* TEST_SE0_NAK */
		    DBG(3, "TEST_SE0_NAK\n");
		    pThis->bTestModeValue = MGC_M_TEST_SE0_NAK;
		    break;
		case USB_IOCTL_TESTMODE_TEST_PACKET:
		    /* TEST_PACKET */
		    DBG(3, "TEST_PACKET\n");
		    pThis->bTestModeValue = MGC_M_TEST_PACKET;
		    break;
		default:
		    /* my gadget might know what to do with it */
		    return -EINVAL;
		}
          return 0;
      }
      case USB_IOCTL_DMA_SET:{
      		struct usb_ioctl_dmaset_param *dmaset_param
      			= (struct usb_ioctl_dmaset_param *)param;
      		uint8_t bEnd = dmaset_param->bEp_num & USB_ENDPOINT_NUMBER_MASK;
#if defined(CONFIG_ARCH_CXD4108)
            unsigned short uidtype0;
            extern unsigned short musb_get_uidtype0(void);

            uidtype0 = musb_get_uidtype0();
#endif /* defined(CONFIG_ARCH_CXD4108) */

          	if(dmaset_param->bDma_ena == USB_MCD_DMA_SETON){
          		/* DMA transfer enable */
          	  	DBG(1,"DMA_SET ON bDma_ena:%d \n",dmaset_param->bDma_ena);
#if defined(CONFIG_ARCH_CXD4108)
                if (uidtype0) {
                    MGC_aGadgetLocalEnd[bEnd].bDma_ena = USB_MCD_DMA_SETON;
                }
                else {
                    /* always DMA OFF */
                    MGC_aGadgetLocalEnd[bEnd].bDma_ena = USB_MCD_DMA_SETOFF;
                }
#else
				MGC_aGadgetLocalEnd[bEnd].bDma_ena = USB_MCD_DMA_SETON;
#endif /* defined(CONFIG_ARCH_CXD4108) */
          	}else if(dmaset_param->bDma_ena == USB_MCD_DMA_SETOFF){
				/* DMA transfer disable */
          	  	DBG(1,"DMA_SET OFF bDma_ena:%d \n",dmaset_param->bDma_ena);
	          	MGC_aGadgetLocalEnd[bEnd].bDma_ena = USB_MCD_DMA_SETOFF;
          	}else{
          	  	DBG(1,"DMA_SET error INVALID param bDma_ena:%d \n",dmaset_param->bDma_ena);
          		return -EINVAL;
          	}
          	return 0;
     	}
    }
    
    return -EINVAL;
}

/**
 * Start Gadget HW
 * @param pThis controller driver
 */
void
MGC_GadgetStart(
    MGC_LinuxCd* pThis)
{
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, 1);
    MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_TESTMODE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_POWER, 
               MGC_M_POWER_HSENAB | MGC_M_POWER_ENSUSPEND);
}

/**
 * Stop Gadget HW
 * @param pThis controller driver
 */
void
MGC_GadgetStop(
        MGC_LinuxCd* pThis)
{
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, 0);
    MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_TESTMODE, 0);
    MGC_Write8(pBase, MGC_O_HDRC_POWER, MGC_M_POWER_ENSUSPEND);
}

/****************************************************************
 * Registration operations 
 ****************************************************************/

/**
 *
 * -EINVAL
 * -EBUSY
 * -ENOMEM
 */  
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
    int ret;
    uint8_t bEnd, power;
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "<==\n");
    	
#ifdef MUSB_PARANOID    
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif

    /* this one is checked to verify that no Gadget is registered */
    if ( pThis->pGadgetDriver ) {
    	ERR("musb-gcd is alredy used by %s\n", 
		pThis->pGadget->name);
        return -EBUSY;
    }
	
    INFO("registering gadget %s\n", driver->function);

    spin_lock_irqsave(&pThis->Lock, flags);	
    /* allocate memory for gadget structure */    
    KMALLOC(pThis->pGadget, sizeof(struct usb_gadget), GFP_KERNEL);
    if ( !pThis->pGadget ) {
	spin_unlock_irqrestore(&pThis->Lock, flags);
	ERR("out of memory\n");	
	return -ENOMEM;	    
    }    
    memset(pThis->pGadget, 0, sizeof(struct usb_gadget));

    /* allocate real memory for ep0 buffer, when MUSB_TEST_GADGET 
     * is set, I'll be using the static buffer up there */    
    KMALLOC(pThis->pEnd0Buffer, sizeof(MGC_End0Buffer), GFP_KERNEL);
    if ( !pThis->pEnd0Buffer) {
	KFREE( pThis->pGadget );    
	spin_unlock_irqrestore(&pThis->Lock, flags);
	ERR("out of memory\n");	
	return -ENOMEM;
    }    
    memset(pThis->pEnd0Buffer, 0, sizeof(MGC_End0Buffer));
    
    INFO("Discovered %d end points\n", min(pThis->bEndCount, 
    	(uint8_t)MUSB_C_NUM_EPS) );

    /** Init completition scheduler */        	
    spin_lock_init( &SchedulerQueue.lock );
    INIT_LIST_HEAD( &SchedulerQueue.req_list );

    /** Init local, and gadget endpoints */        	
    INIT_LIST_HEAD( &(pThis->pGadget->ep_list) );        
    for(bEnd = 0; bEnd < pThis->bEndCount; bEnd++) {

    	memset(&MGC_aGadgetLocalEnd[bEnd], 0, sizeof( MGC_aGadgetLocalEnd[bEnd] ));
#if MUSB_DEBUG>0	
	MGC_aGadgetLocalEnd[bEnd].wPadFront=MGC_LOCAL_PAD;
#endif
	MGC_aGadgetLocalEnd[bEnd].bEndNumber=bEnd;
    	spin_lock_init( &MGC_aGadgetLocalEnd[bEnd].Lock );    
	INIT_LIST_HEAD( &MGC_aGadgetLocalEnd[bEnd].req_list );	
	
	if ( bEnd<iGadgetEndNamesLen || !aGadgetEndNames[bEnd] ) {
		sprintf(aGadgetDefaultEndNames[bEnd], "ep%d", bEnd);
		aGadgetEndNames[bEnd]=aGadgetDefaultEndNames[bEnd];
	}
	
	MGC_aGadgetLocalEnd[bEnd].end_point.name=aGadgetEndNames[bEnd];	
	MGC_aGadgetLocalEnd[bEnd].end_point.ops = &MGC_GadgetEndpointOperations;
	INIT_LIST_HEAD( &MGC_aGadgetLocalEnd[bEnd].end_point.ep_list );
		
	if ( bEnd ) {
	    MGC_aGadgetLocalEnd[bEnd].end_point.maxpacket = max(pThis->aLocalEnd[bEnd].wMaxPacketSizeTx, 
		pThis->aLocalEnd[bEnd].wMaxPacketSizeRx);		
	    list_add_tail(&MGC_aGadgetLocalEnd[bEnd].end_point.ep_list, 
	    	&(pThis->pGadget->ep_list));
	} else {
	    MGC_aGadgetLocalEnd[0].end_point.maxpacket=MGC_END0_FIFOSIZE;
	    MGC_aGadgetLocalEnd[0].wPacketSize=MGC_END0_FIFOSIZE;
	}		

    }

    pThis->pGadget->ops = &MGC_GadgetOperations;
    pThis->pGadget->ep0 = &MGC_aGadgetLocalEnd[0].end_point;

#ifdef MUSB_OTG    
    GADGET_SET_IS_DUALSPEED(pThis->pGadget, 1);
    GADGET_SET_IS_OTG(pThis->pGadget, 1);
    GADGET_SET_A_HNP_SUPPORT(pThis->pGadget, 1);
#endif
	
    /* it takes the name of the driver */	
    pThis->pGadget->name = driver->function; 
	
    pThis->pGadget->speed = USB_SPEED_FULL;
    power = MGC_Read8(pThis->pRegs, MGC_O_HDRC_POWER);

    if(power & MGC_M_POWER_HSMODE)
    {
	pThis->pGadget->speed = USB_SPEED_HIGH;
    }

    /* this one is checked to verify that a Gadget registered */
    pThis->pGadgetDriver = driver;

#ifdef MUSB_V26
    pThis->pGadget->dev.driver=&driver->driver;
    driver->driver.bus=NULL;
    
    /* KLUDGE for weird gadgets */
    kobject_init( &driver->driver.kobj );
    //kobject_set_name(&driver->driver.kobj, "kludge");
    //kobject_register(&driver->driver.kobj);
#endif
  
    spin_unlock_irqrestore(&pThis->Lock, flags);
    if ((ret = driver->bind( pThis->pGadget ))) {
        ERR("gadget %s not registered. fail bind\n", pThis->pGadget->name);

        spin_lock_irqsave(&pThis->Lock, flags);
        KFREE(pThis->pGadget);
        KFREE(pThis->pEnd0Buffer);
        pThis->pGadgetDriver = NULL;
        pThis->pGadget = NULL;
        pThis->pEnd0Buffer = NULL;
        spin_unlock_irqrestore(&pThis->Lock, flags);

        return ret;
    }

    INFO("gadget %s registered\n", pThis->pGadget->name);

    return 0;
}

#ifdef MUSB_V26
EXPORT_SYMBOL (usb_gadget_register_driver);
#endif

/**
 *
 *
 */
int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
#if !defined(CONFIG_ARCH_INTEGRATOR)
    extern void phy_reset(void);
#endif

    DBG(2, "<==\n" );

#ifdef MUSB_PARANOID    
    if( !MGC_pFirstDriver ) {
    	ERR("Gadget not initialized, pThis=NULL\n");
        return -EINVAL;
    }
#endif

    INFO("unregistering gadget %s\n", driver->function); 

    spin_lock_irqsave(&pThis->Lock, flags);

   /* shall I flush/abort the pending requests?? */
    if ( pThis->pGadgetDriver ) {
	driver->unbind( pThis->pGadget );    

	if ( pThis->pGadget ) {
	    KFREE( pThis->pGadget );

	    if ( pThis->pEnd0Buffer ) {
	        KFREE( pThis->pEnd0Buffer );    
	    }
       }
    }
    
    pThis->pGadgetDriver=NULL;
    pThis->pGadget=NULL;
    pThis->pEnd0Buffer=NULL;
    
#if !defined(CONFIG_ARCH_INTEGRATOR)
    phy_reset();
#endif
    spin_unlock_irqrestore(&pThis->Lock, flags);

    return 0;
}

#ifdef MUSB_V26
EXPORT_SYMBOL (usb_gadget_unregister_driver);
#endif

#ifdef MUSB_DMA

static uint8_t MGC_IsDmaInterrupt[MUSB_C_NUM_EPS];

void MGC_TxDmaCallback(MGC_DmaChannel* pChannel,
                                  uint8_t bEnd, 
                                  uint8_t bTransmit,
                                  uint8_t bProtocol,
				  void* pData)
{
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "TestBulk DMA CallBack MGC_HdrcServiceDeviceTxAvail\n" );
    MGC_IsDmaInterrupt[bEnd] = TRUE;
    if(NULL != pThis->aLocalEnd[bEnd].pDmaChannel)
    	pThis->aLocalEnd[bEnd].pDmaChannel->pData = &(MGC_IsDmaInterrupt[bEnd]);

    MGC_HdrcServiceDeviceTxAvail( pThis,  bEnd );

    MGC_IsDmaInterrupt[bEnd] = FALSE;
    DBG(2, "TestBulk DMA CallBack MGC_HdrcServiceDeviceTxAvail end \n" );
}


void MGC_RxDmaCallback(MGC_DmaChannel* pChannel,
                                  uint8_t bEnd, 
                                  uint8_t bTransmit,
                                  uint8_t bProtocol,
				  void* pData)
{
    MGC_LinuxCd* pThis = MGC_pFirstDriver;

    DBG(2, "TestBulk DMA CallBack MGC_HdrcServiceDeviceRxReady\n" );
    MGC_IsDmaInterrupt[bEnd] = TRUE;
    if(NULL != pThis->aLocalEnd[bEnd].pDmaChannel)
    	pThis->aLocalEnd[bEnd].pDmaChannel->pData = &(MGC_IsDmaInterrupt[bEnd]);

    MGC_HdrcServiceDeviceRxReady( pThis,  bEnd );

    MGC_IsDmaInterrupt[bEnd] = FALSE;
    DBG(2, "TestBulk DMA CallBack MGC_HdrcServiceDeviceRxReady end \n" );
}
#endif /* MUSB_DMA */

/*----------------------------------------------------------
 name:MGC_GadgetDmaSetDefault
 param: non
 content: Set a Default Value for bDma_ena 
 			in MGC_aGadgetLocalEnd
----------------------------------------------------------*/
STATIC void MGC_GadgetDmaSetDefault(void)
{
	uint8_t	i;
#if defined(CONFIG_ARCH_CXD4108)
    unsigned short uidtype0;
    extern unsigned short musb_get_uidtype0(void);

    uidtype0 = musb_get_uidtype0();
#endif /* defined(CONFIG_ARCH_CXD4108) */
	
	for(i=0; i<MUSB_C_NUM_EPS; i++){
		/* set a Default value(DMA enable) for bDma_ena*/
#if defined(CONFIG_ARCH_CXD4108)
        if (uidtype0) {
            MGC_aGadgetLocalEnd[i].bDma_ena = USB_MCD_DMA_SETON;
        }
        else {
            MGC_aGadgetLocalEnd[i].bDma_ena = USB_MCD_DMA_SETOFF;
        }
#else
		MGC_aGadgetLocalEnd[i].bDma_ena = USB_MCD_DMA_SETON;
#endif /* defined(CONFIG_ARCH_CXD4108) */
		DBG(3,"Init MGC_aGadgetLocalEnd[%d].bDma_ena:%d \n",i,MGC_aGadgetLocalEnd[i].bDma_ena);
	}
}
