/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_dma.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This code is based on musbhsdma.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include "usb_mcd_defs.h"

#if defined(DMA_HOST_CODE)||defined(MUSB_DMA)
 #define MUSB_HSDMA
#else
 #ifdef MUSB_DMA
 #define MUSB_HSDMA
 #endif
#endif	/* DMA_HOST_CODE end */
/****************************** CONSTANTS ********************************/

#define MGC_O_HSDMA_BASE    0x200
#define MGC_O_HSDMA_INTR    0x200

#define MGC_O_HSDMA_CONTROL 4
#define MGC_O_HSDMA_ADDRESS 8
#define MGC_O_HSDMA_COUNT   0xc

#define MGC_HSDMA_CHANNEL_OFFSET(_bChannel, _bOffset) (MGC_O_HSDMA_BASE + (_bChannel << 4) + _bOffset)

/* control register (16-bit): */
#define MGC_S_HSDMA_ENABLE	0
#define MGC_S_HSDMA_TRANSMIT	1
#define MGC_S_HSDMA_MODE1	2
#define MGC_S_HSDMA_IRQENABLE	3
#define MGC_S_HSDMA_ENDPOINT	4
#define MGC_S_HSDMA_BUSERROR	8
#define MGC_S_HSDMA_BURSTMODE	9

/******************************** TYPES **********************************/

typedef struct __MGC_HsDmaController MGC_HsDmaController;
typedef struct __MGC_HsDmaChannel MGC_HsDmaChannel;

/****************************** STRUCTURES *******************************/

struct __MGC_HsDmaChannel
{
    MGC_DmaChannel Channel;
    MGC_HsDmaController* pController;
    uint32_t dwStartAddress;
    uint32_t dwCount;
    uint8_t bIndex;
    uint8_t bEnd;
    uint8_t bProtocol;
    uint8_t bTransmit;
};

struct __MGC_HsDmaController
{
    MGC_DmaController Controller;
    MGC_HsDmaChannel* paChannel;
    uint8_t* pCoreBase;
    uint8_t bChannelCount;
    uint8_t bmUsedChannels;
};

/******************************* FORWARDS ********************************/

static uint8_t MGC_HsDmaStartController(
    MGC_DmaController* pController);

static uint8_t MGC_HsDmaStopController(
    MGC_DmaController* pController);

static MGC_DmaChannel* MGC_HsDmaAllocateChannel(
    MGC_DmaController* pController,
    uint8_t bLocalEnd,
    uint8_t bTransmit,
    uint8_t bProtocol,
    uint16_t wMaxPacketSize);

static void MGC_HsDmaReleaseChannel(
    MGC_DmaChannel* pChannel);

static uint8_t* MGC_HsDmaAllocateBuffer(
    MGC_DmaChannel* pChannel,
    uint32_t dwLength);

static uint8_t MGC_HsDmaReleaseBuffer(
    MGC_DmaChannel* pChannel,
    uint8_t* pBuffer);

static uint8_t MGC_HsDmaProgramChannel(
    MGC_DmaChannel* pChannel,
    MGC_pfDmaCallback pfDmaCallback,
    void* pData,
    uint16_t wPacketSize,
    uint8_t bMode,
    const uint8_t* pBuffer, 
    uint32_t dwLength);

static uint8_t MGC_HsDmaEnableIntChannel(
    MGC_DmaChannel* pChannel);

static uint8_t MGC_HsDmaDisableIntChannel(
    MGC_DmaChannel* pChannel);

static uint8_t MGC_HsDmaStartChannel(
    MGC_DmaChannel* pChannel);

static uint8_t MGC_HsDmaStopChannel(
    MGC_DmaChannel* pChannel);

static MGC_DmaChannelStatus MGC_HsDmaGetChannelStatus(
    MGC_DmaChannel* pChannel);

static irqreturn_t MGC_HsDmaControllerIsr(
    int irq,
    void *__pController,
    struct pt_regs *r);

static MGC_DmaController* MGC_HsNewDmaController(
    uint8_t* pCoreBase,
    uint8_t bChannelCount,
    int nIrq);

static void MGC_HsDestroyDmaController(
    MGC_DmaController* pController);

/******************************* GLOBALS *********************************/

MGC_DmaControllerFactory MGC_HdrcDmaControllerFactory =
{
    0x300,
    MGC_HsNewDmaController,
    MGC_HsDestroyDmaController
};

/****************************** FUNCTIONS ********************************/

#ifdef MUSB_HSDMA

static uint8_t
MGC_HsDmaStartController(
    MGC_DmaController* pController)
{
    MGC_HsDmaController* pHsController
        = (MGC_HsDmaController*)pController->pPrivateData;
    MGC_HsDmaStopController(pController);
    pHsController->bmUsedChannels = 0; /* all channels are not used */

    return TRUE;
}

static uint8_t
MGC_HsDmaStopController(
    MGC_DmaController* pController)
{
    uint8_t bBit;
    MGC_HsDmaController* pHsController
        = (MGC_HsDmaController*)pController->pPrivateData;

    for(bBit = 0; bBit < pHsController->bChannelCount; bBit++) {
        MGC_HsDmaDisableIntChannel(&pHsController->paChannel[bBit].Channel);
        MGC_HsDmaStopChannel(&pHsController->paChannel[bBit].Channel);
    }

    return TRUE;
}

static MGC_DmaChannel*
MGC_HsDmaAllocateChannel(
    MGC_DmaController* pController,
    uint8_t bLocalEnd,
    uint8_t bTransmit,
    uint8_t bProtocol, 
    uint16_t wMaxPacketSize)
{
    uint8_t bBit;
    MGC_DmaChannel* pChannel = NULL;
    MGC_HsDmaChannel* pImplChannel = NULL;
    MGC_HsDmaController* pHsController
        = (MGC_HsDmaController*)pController->pPrivateData;

    for(bBit = 0; bBit < pHsController->bChannelCount; bBit++)
    {
	if(!(pHsController->bmUsedChannels & (1 << bBit)))
	{
	    pHsController->bmUsedChannels |= (1 << bBit);
	    pImplChannel = &(pHsController->paChannel[bBit]);
	    pImplChannel->bEnd = bLocalEnd;
	    pImplChannel->bProtocol = bProtocol;
	    pImplChannel->bTransmit = bTransmit;
	    pChannel = &(pImplChannel->Channel);
	    pChannel->bStatus = MGC_DMA_STATUS_FREE;
	    pChannel->dwMaxLength = 0x10000;
	    break;
	}
    }
    return pChannel;
}

static void
MGC_HsDmaReleaseChannel(
    MGC_DmaChannel* pChannel)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;

    pImplChannel->pController->bmUsedChannels &= ~(1 << pImplChannel->bIndex);
    pImplChannel->Channel.bStatus = MGC_DMA_STATUS_FREE;

    if (pImplChannel->Channel.pfDmaCallback) {
	pImplChannel->Channel.pfDmaCallback = NULL;
    }
}

static uint8_t*
MGC_HsDmaAllocateBuffer(
    MGC_DmaChannel* pChannel,
    uint32_t dwLength)
{
    /* nothing to do */
    return (uint8_t*)0;
}

static uint8_t
MGC_HsDmaReleaseBuffer(
    MGC_DmaChannel* pChannel,
    uint8_t* pBuffer)
{
    /* nothing to do */
    return FALSE;
}

static uint8_t
MGC_HsDmaProgramChannel(

    MGC_DmaChannel* pChannel, 
    MGC_pfDmaCallback pfDmaCallback,
    void* pData,
    uint16_t wPacketSize,
    uint8_t bMode,
    const uint8_t* pBuffer, 
    uint32_t dwLength)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint8_t bChannel = pImplChannel->bIndex;
    uint16_t wCsr;

    /* EP number */
    wCsr = pImplChannel->bEnd << MGC_S_HSDMA_ENDPOINT;
    /* DMA mode */
    if(bMode)
    {
        wCsr |= 1 << MGC_S_HSDMA_MODE1;
    }
#if defined(CONFIG_ARCH_CXD4108)
    /* BURST mode3 */
    wCsr |= 3 << MGC_S_HSDMA_BURSTMODE;
#endif /* defined(CONFIG_ARCH_CXD4108) */
    /* Direction */
    if(pImplChannel->bTransmit)
    {
        wCsr |= 1 << MGC_S_HSDMA_TRANSMIT;
    }
    /* control */
    MGC_Write16(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL), wCsr);

    /* address/count */
    MGC_Write32(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_ADDRESS), 
                (uint32_t)pBuffer /* pBuffer is bus address */);
    MGC_Write32(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_COUNT), dwLength);


    /* Channel parameters */
    pChannel->dwActualLength = 0L;
    pChannel->pfDmaCallback = pfDmaCallback;
    pChannel->pData = pData;
    pImplChannel->dwStartAddress = (uint32_t)pBuffer;
    pImplChannel->dwCount = dwLength;

    return TRUE;
}

static uint8_t
MGC_HsDmaEnableIntChannel(
    MGC_DmaChannel* pChannel)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint16_t wCsr;
    uint8_t bChannel = pImplChannel->bIndex;

    /* read control register */
    wCsr = MGC_Read16(pBase, 
                      MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL));

    /* set IRQ Enable */
    wCsr |= 1 << MGC_S_HSDMA_IRQENABLE;

    /* control (this should start things) */
    MGC_Write16(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL), wCsr);

    return TRUE;
}

static uint8_t
MGC_HsDmaDisableIntChannel(
    MGC_DmaChannel* pChannel)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint16_t wCsr;
    uint8_t bChannel = pImplChannel->bIndex;

    /* read control register */
    wCsr = MGC_Read16(pBase, 
                      MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL));

    /* clear IRQ Enable */
    wCsr &= ~(1 << MGC_S_HSDMA_IRQENABLE);

    /* control (this should start things) */
    MGC_Write16(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL), wCsr);
    return TRUE;
}

static uint8_t
MGC_HsDmaStartChannel(
    MGC_DmaChannel* pChannel)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint16_t wCsr;
    uint8_t bChannel = pImplChannel->bIndex;

    /* read control register */
    wCsr = MGC_Read16(pBase, 
                      MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL));

    /* set DMA Enable */
    wCsr |= 1 << MGC_S_HSDMA_ENABLE;

    /* control (this should start things) */
    MGC_Write16(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL), wCsr);

    return TRUE;
}

static uint8_t
MGC_HsDmaStopChannel(
    MGC_DmaChannel* pChannel)
{
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint16_t wCsr;
    uint8_t bChannel = pImplChannel->bIndex;

    /* read control register */
    wCsr = MGC_Read16(pBase, 
                      MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL));

    /* clear DMA Enable */
    wCsr &= ~(1 << MGC_S_HSDMA_ENABLE);

    /* control (this should start things) */
    MGC_Write16(pBase, 
                MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL), wCsr);

    return TRUE;
}

static MGC_DmaChannelStatus
MGC_HsDmaGetChannelStatus(
    MGC_DmaChannel* pChannel)
{
    uint32_t dwAddress;
    MGC_HsDmaChannel* pImplChannel = (MGC_HsDmaChannel*)pChannel->pPrivateData;
    MGC_HsDmaController* pController = pImplChannel->pController;
    uint8_t* pBase = pController->pCoreBase;
    uint8_t bChannel = pImplChannel->bIndex;
    uint16_t wCsr;

    /* Read Control register */
    wCsr
        = MGC_Read16(pBase, 
                     MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_CONTROL));

    /* most DMA controllers would update the count register for simplicity... */
    dwAddress
        = MGC_Read32(pBase,
                     MGC_HSDMA_CHANNEL_OFFSET(bChannel, MGC_O_HSDMA_ADDRESS));
    pImplChannel->Channel.dwActualLength
        = dwAddress - pImplChannel->dwStartAddress;

    /* Bus Error */
    if(wCsr & (1 << MGC_S_HSDMA_BUSERROR))
    {
        return MGC_DMA_STATUS_BUS_ABORT;
    }
    /* Now DMA working */
    if(dwAddress < (pImplChannel->dwStartAddress + pImplChannel->dwCount))
    {
        return MGC_DMA_STATUS_BUSY;
    }
    /* DMA is IDLE */
    return MGC_DMA_STATUS_FREE;
}

static irqreturn_t
MGC_HsDmaControllerIsr(
    int irq,
    void *__pController,
    struct pt_regs *r)
{
    uint8_t bChannel;
    uint16_t wCsr;
    uint32_t dwAddress;
    MGC_HsDmaChannel* pImplChannel;
    MGC_HsDmaController* pHsController = (MGC_HsDmaController*)__pController;
    uint8_t* pBase = pHsController->pCoreBase;
    uint8_t bIntr = MGC_Read8(pBase, MGC_O_HSDMA_INTR);

    if(!bIntr) {
        return IRQ_NONE;
    }
    for(bChannel = 0; bChannel < pHsController->bChannelCount; bChannel++) {
        if(bIntr & (1 << bChannel)) {
            pImplChannel
                = (MGC_HsDmaChannel*)&(pHsController->paChannel[bChannel]);
            /* Read Control register */
            wCsr
                = MGC_Read16(pBase, 
                             MGC_HSDMA_CHANNEL_OFFSET(bChannel,
                                                      MGC_O_HSDMA_CONTROL));
            /* most DMA controllers would update the count register
               for simplicity... */
            dwAddress
                = MGC_Read32(pBase,
                             MGC_HSDMA_CHANNEL_OFFSET(bChannel,
                                                      MGC_O_HSDMA_ADDRESS));
            pImplChannel->Channel.dwActualLength
                = dwAddress - pImplChannel->dwStartAddress;
            /* Bus Error */
            if(wCsr & (1 << MGC_S_HSDMA_BUSERROR)) {
                pImplChannel->Channel.bStatus = MGC_DMA_STATUS_BUS_ABORT;
            }
            /* DMA is completed */
            else {
                pImplChannel->Channel.bStatus = MGC_DMA_STATUS_FREE;
            }
            /* If callback exists, call it and set NULL */
            if (pImplChannel->Channel.pfDmaCallback) {
                (pImplChannel->Channel.pfDmaCallback)(
                    &pImplChannel->Channel,
                    pImplChannel->bEnd,
                    pImplChannel->bTransmit,
                    pImplChannel->bProtocol,
                    pImplChannel->Channel.pData);
            }
        }
    }
    return IRQ_HANDLED;
}

#endif	/* MUSB_HSDMA */

static MGC_DmaController*
MGC_HsNewDmaController(
    uint8_t* pCoreBase,
    uint8_t bChannelCount,
    int nIrq)
{
    uint8_t bBit;
    MGC_DmaController* pResult = NULL;
#ifdef MUSB_HSDMA
    MGC_DmaController* pController;
    MGC_HsDmaController* pHsController;

    if (bChannelCount == 0) {
        return pResult;
    }

    pHsController = (MGC_HsDmaController*)kmalloc(sizeof(MGC_HsDmaController),
                                                  GFP_KERNEL);

    if(pHsController)
    {
        /* Set params to MGC_HsDmaController */
        memset(pHsController, 0, sizeof(MGC_HsDmaController));
        pHsController->paChannel
            = (MGC_HsDmaChannel*)kmalloc(sizeof(MGC_HsDmaChannel)*bChannelCount,
                                         GFP_KERNEL);
        if (!pHsController->paChannel) {
            kfree(pHsController);
            return pResult;
        }
        memset(pHsController->paChannel, 0,
               sizeof(MGC_HsDmaChannel)*bChannelCount);
        
        pHsController->bChannelCount = bChannelCount;
        pHsController->pCoreBase = pCoreBase;
        /* Set params to MGC_DmaController */
        pController = &pHsController->Controller;
        pController->pPrivateData = (void*)pHsController;
        pController->nIrq = nIrq;
        pController->pfDmaStartController = MGC_HsDmaStartController;
        pController->pfDmaStopController = MGC_HsDmaStopController;
        pController->pfDmaAllocateChannel = MGC_HsDmaAllocateChannel;
        pController->pfDmaReleaseChannel = MGC_HsDmaReleaseChannel;
        pController->pfDmaAllocateBuffer = MGC_HsDmaAllocateBuffer;
        pController->pfDmaReleaseBuffer = MGC_HsDmaReleaseBuffer;
        pController->pfDmaProgramChannel = MGC_HsDmaProgramChannel;
        pController->pfDmaEnableIntChannel = MGC_HsDmaEnableIntChannel;
        pController->pfDmaDisableIntChannel = MGC_HsDmaDisableIntChannel;
        pController->pfDmaStartChannel = MGC_HsDmaStartChannel;
        pController->pfDmaStopChannel = MGC_HsDmaStopChannel;
        pController->pfDmaGetChannelStatus = MGC_HsDmaGetChannelStatus;
        pController->pfDmaControllerIsr = MGC_HsDmaControllerIsr;
        pResult = &(pHsController->Controller);

        for(bBit = 0; bBit < bChannelCount; bBit++) {
            pHsController->paChannel[bBit].bIndex = bBit;
            pHsController->paChannel[bBit].Channel.pPrivateData
                = (void*)&pHsController->paChannel[bBit];
            pHsController->paChannel[bBit].Channel.pController = pController;
            pHsController->paChannel[bBit].pController = pHsController;
        }
    }
#endif
    return pResult;
}

static void
MGC_HsDestroyDmaController(
    MGC_DmaController* pController)
{
#ifdef MUSB_HSDMA
    MGC_HsDmaController* pHsController = 
        (MGC_HsDmaController*)pController->pPrivateData;
    
    if(pHsController)
    {
        kfree(pHsController->paChannel);
        kfree(pHsController);
    }
#endif
}

