/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_ep0.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This code is based on g_ep0.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/interrupt.h>

#if defined(MUSB_V24) && !defined(MUSB_LINUX_MV21)
/* dealing with Linux headers */
struct usb_tt {
        struct usb_device       *hub;   /* upstream highspeed hub */
        int                     multi;  /* true means one TT per port */
}; 
#include "hcd.h"
#endif

#include <linux/init.h>
#include <linux/usb_ch9.h>
#include "usb_mcd_defs.h"
#include "usb_mcd_gadgetdefs.h"

#ifdef MUSB_MONITOR_DATA
static int monitor_loglevel;
#endif

const uint8_t MGC_aTestPacket[] = {
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
        0xaa, 0xee, 0xee, 0xee, 0xee, 0xee, 0xee, 0xee,
        0xee, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
        0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xbf, 0xdf,
        0xef, 0xf7, 0xfb, 0xfd, 0xfc, 0x7e, 0xbf, 0xdf,
        0xef, 0xf7, 0xfb, 0xfd, 0x7e };

/* FROM: usb_gadget.h
 *
 * Accordingly, the driver's setup() callback must always implement all
 * get_descriptor requests, returning at least a device descriptor and
 * a configuration descriptor.  Drivers must make sure the endpoint
 * descriptors match any hardware constraints. Some hardware also constrains
 * other descriptors. (The pxa250 allows only configurations 1, 2, or 3).
 *
 * The driver's setup() callback must also implement set_configuration,
 * and should also implement set_interface, get_configuration, and
 * get_interface.  Setting a configuration (or interface) is where
 * endpoints should be activated or (config 0) shut down.
 */


/**
 * Read a FULL header packet from the hardware. The buffer starts with 
 * struct usb_ctrlrequest (fields are converted to device specific
 * byte order).
 *
 * @param wCount>0
 * @return 0 when the packet is complete, a negative number when an error 
 * occurred, a positive number when still there are bytes to read.
 */
static int read_control_request(MGC_LinuxCd* pThis, uint16_t wCount) {
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_End0Buffer* pEnd0Buffer=(MGC_End0Buffer*)pThis->pEnd0Buffer;
    struct usb_ctrlrequest* pControlRequest=(struct usb_ctrlrequest*)pEnd0Buffer;

    if (!pEnd0Buffer) {
        return -EINVAL;
    }

    DBG(3, "<==\n");
    DBG(4,"wCount=%u, pEnd0Buffer->wEnd0Count=%u\n", wCount, 
    	pEnd0Buffer->wEnd0Count);

   if (!wCount) {
   	return -EINVAL;
   }

    /* buffer overrun, it should never happen */
    if ( wCount>(MUSB_MAX_END0_PACKET-sizeof(struct usb_ctrlrequest)) ) {
	ERR("buffer overrun! wCount=%d\n", wCount ); 
	return -EINVAL; 
    }	

    /* need to have at least enough bytes for the control request
     * comment this out to enable fifo size < 8 bytes */
    if ( wCount<sizeof(struct usb_ctrlrequest) ) {
	ERR("wCount=%u<sizeof(struct usb_ctrlrequest)=%u\n",
    	    wCount, sizeof(struct usb_ctrlrequest)); 
	return -EINVAL; 
    }

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
#endif

    spin_lock(&pThis->Lock);
    MGC_SelectEnd(pBase, 0); /* select ep0 */
	
    /* wEnd0Count=0 means that I'm reading the USB standard header:
     * that's the first thing I need to read from the FIFO (8 bytes). */
    if ( 0==pEnd0Buffer->wEnd0Count ) {
	DBG(4, "reading header\n" );
	MGC_HdrcUnloadFifo(pBase, 0, sizeof(struct usb_ctrlrequest), 
		(uint8_t*)pControlRequest);
	wCount-=sizeof(struct usb_ctrlrequest);
	DBG(6, "header read\n");
	
	/* data from the USB bus must be converted from LSB to 
	 * host-specific byte ordering; the control request header 
	 * tell me the payload length etc. */
	le16_to_cpus( pControlRequest->wLength );
	le16_to_cpus( pControlRequest->wIndex );
	le16_to_cpus( pControlRequest->wValue );

	DBG(4, "bRequest=%02x, kind=%s, wValue=%04x, wIndex=%04x, wLength=%04x\n", 
	    pControlRequest->bRequest, decode_request(pControlRequest), 
	    pControlRequest->wValue, pControlRequest->wIndex, 
	    pControlRequest->wLength);

	/* the header will set me in the right stage:
		pThis->bEnd0Stage=MGC_END0_STAGE_TX
		pThis->bEnd0Stage=MGC_END0_STAGE_RX
	*/
	if( pControlRequest->bRequestType & USB_DIR_IN ) { 
	    /* write to host: up to wLength bytes */
	    pEnd0Buffer->wEnd0Count=0;
	    pThis->bEnd0Stage = MGC_END0_STAGE_TX;
		DBG(2,"[1] Ep0Stage %d \n",pThis->bEnd0Stage);
		
	} else if( pControlRequest->bRequestType & USB_DIR_OUT ) { 
	    /* out to function: wLength to go for the payload */
	    pEnd0Buffer->wEnd0Count=pControlRequest->wLength;
	    pThis->bEnd0Stage = MGC_END0_STAGE_RX;
		DBG(2,"[2] Ep0Stage %d \n",pThis->bEnd0Stage);
	} 
    }

    if ( wCount>0 ) {    
	/* now Im reading the rest of it, this will never be executed I guess */
	uint16_t offset=sizeof(struct usb_ctrlrequest)+ /* read past the header */
		(pControlRequest->wLength)-(pEnd0Buffer->wEnd0Count);	
	MGC_HdrcUnloadFifo(pBase, 0, wCount, &pEnd0Buffer->aEnd0Data[offset]);
	pEnd0Buffer->wEnd0Count-=wCount;
    }

    DBG(5, "pEnd0Buffer->wEnd0Count=%d, ep0stage=%s, %s\n", 
	pEnd0Buffer->wEnd0Count, decode_ep0stage(pThis->bEnd0Stage), 
	(pEnd0Buffer->wEnd0Count)?"still to go":"header completed"); 
    DBG(3, "==>\n");

    spin_unlock(&pThis->Lock);
    
    /* 0 header completed, <0 error, >0 bytes to go*/
    return (pEnd0Buffer->wEnd0Count);
}



static int forward_to_driver(const struct usb_ctrlrequest *pControlRequest) 
{
    int handled=-EOPNOTSUPP;

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&MGC_aGadgetLocalEnd[0]); 
#endif
	
    DBG(-1002, "<==\n");	
    if ( MGC_pFirstDriver->pGadgetDriver ) {
	handled=MGC_pFirstDriver->pGadgetDriver->setup(MGC_pFirstDriver->pGadget, 
		pControlRequest);    
    }
    DBG(-1002, "==>\n");
			
    return handled;
}
    

static int service_rx_request(struct usb_ctrlrequest *pControlRequest) 
{
#ifdef MUSB_PARANOID 
    MGC_LinuxCd* pThis=MGC_pFirstDriver;
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&MGC_aGadgetLocalEnd[0]); 
#endif

    return forward_to_driver(pControlRequest);
}

/**
 * handle a request, the end0 buffer contains the current request
 * that is supposed to be a standard control request. Assumes the 
 * fifo to be at least 2 bytes long.
 *
 
	case USB_REQ_GET_CONFIGURATION:
	case USB_REQ_GET_INTERFACE:
	case USB_REQ_GET_DESCRIPTOR:
	case USB_REQ_GET_STATUS:
	case USB_REQ_SYNC_FRAME:
 
 
 * @return 0 if the request was NOT HANDLED,
 * < 0 when error (ENOSUPP not supproted)
 * > 0 when the request is processed 
 */
static int service_tx_request(const struct usb_ctrlrequest *pControlRequest) 
{
    int handled=0; /* not handled */

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&MGC_aGadgetLocalEnd[0]); 
#endif

    if ( USB_TYPE_STANDARD!=(pControlRequest->bRequestType&USB_TYPE_MASK )) {
	return forward_to_driver(pControlRequest);
    }
    
    switch(pControlRequest->bRequest) {
	case USB_REQ_GET_CONFIGURATION:
	    DBG(3, "USB_REQ_GET_CONFIGURATION()\n");
	    break;

	case USB_REQ_GET_INTERFACE:
	    DBG(3, "USB_REQ_GET_INTERFACE()\n");
	    break;

	case USB_REQ_GET_DESCRIPTOR:
	    DBG(3, "USB_REQ_GET_DESCRIPTOR()\n");
	    break;

	case USB_REQ_GET_STATUS: {
	    DBG(3, "USB_REQ_GET_STATUS()\n");
#if 0
	    service_tx_status_request(pControlRequest);
#endif
	    } break;
	
	/* case USB_REQ_SYNC_FRAME:
	    break; */

	default: 
	    break;
    }

    if ( !handled ) {
    	handled=forward_to_driver(pControlRequest);
    }

     /* now tx! */
    return handled;
}

/* Executed @ interrupt time...
 */
void MGC_CompleteEp0Request(void) {
    struct usb_request *pRequest;

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_LOCKED(&MGC_aGadgetLocalEnd[0]); 

   
   
#endif

    spin_lock( &MGC_aGadgetLocalEnd[0].Lock );
    pRequest=MGC_CurrentRequest( &MGC_aGadgetLocalEnd[0] );

    DBG(5, "completing request pRequest=%p\n", pRequest);     	    
    /* completing on ep0 is from interrupt */
    if ( pRequest ) {
	/* there's only one request here */
     	list_del( &pRequest->list ); 	
	spin_unlock( &MGC_aGadgetLocalEnd[0].Lock );
    	if ( pRequest->complete ) {	
	    pRequest->complete(&MGC_aGadgetLocalEnd[0].end_point, pRequest);
	}
    } else {
	spin_unlock( &MGC_aGadgetLocalEnd[0].Lock );
    }

    if(((uint8_t)NULL == MGC_pFirstDriver->bSetAddress) &&
       ((uint8_t)NULL == MGC_pFirstDriver->bTestMode) ) {
        MGC_pFirstDriver->bEnd0Stage = MGC_END0_STAGE_SETUP;
		DBG(2,"[4] Ep0Stage %d \n",MGC_pFirstDriver->bEnd0Stage);
    }
}

/* called when the completiotion interrupt occours.. 
  hopefully */
static uint8_t handle_ep0_completition_irq(void) 
{
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    uint16_t wCsrVal;
	uint8_t	status = MCD_EP0_COMP_END;
#if (MUSB_DEBUG>0)
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[0]);
    struct usb_request* pRequest = MGC_CurrentRequest(pEnd);
#endif    
    
    DBG(3, "<==\n");
    DBG(4, "post event interrupts ep0stage=%s\n", 
    	decode_ep0stage(pThis->bEnd0Stage));

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&MGC_aGadgetLocalEnd[0]); 
#endif

 		DBG(2,"bEp0Stage %d \n",pThis->bEnd0Stage);
    switch (pThis->bEnd0Stage) {	
	/* end of sequence #2 (RX state) or #3 (no data) */
	case MGC_END0_STAGE_STATUSIN: {
	    struct usb_request *pRequest;
	    DBG(-1001, "MGC_END0_STAGE_STATUSIN request\n");

	    if(pThis->bSetAddress) {	    
		pThis->bSetAddress = FALSE;
		MGC_Write8(pBase, MGC_O_HDRC_FADDR, pThis->bAddress);		
		pThis->bEnd0Stage = MGC_END0_STAGE_SETUP;
		DBG(2,"[5] Ep0Stage %d \n",pThis->bEnd0Stage);
		DBG(-1002,"address configured... pThis->bAddress=0x%x\n", 
			pThis->bAddress);	
#ifdef MUSB_MONITOR_DATA
		MGC_SetDebugLevel(monitor_loglevel);
#endif		    
	    }

	    /* enter test mode if needed */
	    if(pThis->bTestMode) {
		DBG(-1001, "entering TESTMODE\n");

		if (MGC_M_TEST_PACKET == pThis->bTestModeValue) {
		    MGC_HdrcLoadFifo(pBase, 0, sizeof(MGC_aTestPacket), 
			MGC_aTestPacket);
		    spin_lock(&pThis->Lock);
		    wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, bEnd );
		    wCsrVal |=  (MGC_M_CSR0_TXPKTRDY);
		    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, bEnd, wCsrVal);
		    spin_unlock(&pThis->Lock);
		}
		if ( (uint8_t)NULL == pThis->bTestModeValue) {
		    pThis->bTestMode = (uint8_t)NULL;
		}
		spin_lock(&pThis->Lock);
		MGC_SelectEnd(pBase, 0); /* select ep0 */
		MGC_Write8(pBase, MGC_O_HDRC_TESTMODE, 
	    	    pThis->bTestModeValue);
		pThis->bEnd0Stage = MGC_END0_STAGE_SETUP;
		DBG(2,"[6] Ep0Stage %d \n",pThis->bEnd0Stage);
		DBG(2, "Read  MGC_O_HDRC_TESTMODE = %x \n", MGC_Read8(pBase, MGC_O_HDRC_TESTMODE ) );
		spin_unlock(&pThis->Lock);
	    }
	    
	    DBG(-1001, "completing posted request (if any)\n");	    
	    /* short_not_ok ե饰ǧ */
	    pRequest = MGC_CurrentRequest( &MGC_aGadgetLocalEnd[0] );
	    if((NULL!=pRequest) && (-EINPROGRESS == pRequest->status)){
	        if((1==pRequest->short_not_ok) &&
	           (0!=((pRequest->actual)%(MGC_aGadgetLocalEnd[0].wPacketSize))) ){
	            pRequest->status = -EREMOTEIO;
	        }
	        else{
	            pRequest->status = 0;
	        }
	    }
	    MGC_CompleteEp0Request(); /* async complete */
	  } break;
	
	/* sequence #1: write to host (TX state)  */
	case MGC_END0_STAGE_STATUSOUT: {
	    struct usb_request *pRequest;
	    DBG(-1001, "completing posted request (if any)\n");	    
	    /* short_not_ok ե饰ǧ */
	    pRequest = MGC_CurrentRequest( &MGC_aGadgetLocalEnd[0] );
	    if((NULL!=pRequest) && (-EINPROGRESS == pRequest->status)){
	        if((1==pRequest->short_not_ok) &&
	           (0!=((pRequest->actual)%(MGC_aGadgetLocalEnd[0].wPacketSize))) ){
	            pRequest->status = -EREMOTEIO;
	        }
	        else{
	            pRequest->status = 0;
	        }
	    }
	    MGC_CompleteEp0Request(); /* async complete */
	  } break;

	case MGC_END0_STAGE_TX:
		DBG(-1001, "TX changeing ep status\n");
		DBG(2," STAGE_TX  status:%d\n", MGC_CurrentRequest(&MGC_aGadgetLocalEnd[0])->status);
	    if ( MGC_CurrentRequest(&MGC_aGadgetLocalEnd[0])->status!=-EINPROGRESS ) 		{
	        struct usb_request *pRequest;
	        pRequest = MGC_CurrentRequest( &MGC_aGadgetLocalEnd[0] );
		DBG(3,"length=%d actual=%d\n",pRequest->length, pRequest->actual);
		/* completing on ep0 is from interrupt */
		if ( pRequest ) {
		        /* there's only one request here */
		     	list_del( &pRequest->list ); 	
		    	if ( pRequest->complete ) {	
			    pRequest->complete(&MGC_aGadgetLocalEnd[0].end_point, pRequest);
			}
		}
		/* change state in case taht tx ended */
		pThis->bEnd0Stage=MGC_END0_STAGE_STATUSOUT;
		DBG(2,"[7] Ep0Stage %d \n",pThis->bEnd0Stage);
	    }
	    else{
		DBG(2," MCD_EP0_COMP_NON \n");
		status = MCD_EP0_COMP_NON;
	    }
	    break;
	case MGC_END0_STAGE_RX:
		DBG(-1001, "RX changeing ep status\n");
	    if ( MGC_CurrentRequest(&MGC_aGadgetLocalEnd[0])->status!=-EINPROGRESS ) {
		pThis->bEnd0Stage=MGC_END0_STAGE_STATUSIN;
		DBG(2,"[8] Ep0Stage %d \n",pThis->bEnd0Stage);
	    }
	    break;

	default: /* IT WAS STALLED */
	    DBG(-1002, "recovering from stall? ep0stage=%s\n", 
	    	decode_ep0stage(pThis->bEnd0Stage));
	    pThis->bEnd0Stage = MGC_END0_STAGE_SETUP;		    
		DBG(2,"[8] Ep0Stage %d \n",pThis->bEnd0Stage);
	break;
    }

    /* weird core! clear CSR0 */
    //spin_lock(&pThis->Lock);
    //MGC_SelectEnd(pBase, 0);
    //MGC_WriteCsr8(pBase, MGC_O_HDRC_CSR0, 0, 0);
    //spin_unlock(&pThis->Lock);
	if(status == MCD_EP0_COMP_END){
		return TRUE;
	}else{
		return FALSE;
	}
    DBG(3, "==>\n");
}

/*
 *	USB_REQ_SET_INTERFACE: 
 *	USB_REQ_SET_CONFIGURATION:
 *	USB_REQ_SET_ADDRESS:
 *	USB_REQ_CLEAR_FEATURE:
 *	USB_REQ_SET_FEATURE:
 */
static int service_zero_data_request(MGC_LinuxCd* pThis, 
	struct usb_ctrlrequest *pControlRequest) 
{
    int handled=1; /* handled, DO NOT not pass down */

    DBG(-1002, "<==\n");

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_UNLOCKED(&MGC_aGadgetLocalEnd[0]); 
#endif

    /* non standard requests are piped to the gadget */
    if ( USB_TYPE_STANDARD!=(pControlRequest->bRequestType&USB_TYPE_MASK )) {
    	return forward_to_driver(pControlRequest);
    }

    /* zero data phase */	
    switch (pControlRequest->bRequest) {

#if 0
	case USB_REQ_SET_INTERFACE:
	    DBG(3, "USB_REQ_SET_INTERFACE()\n");
	    handled=0;/* pass it to the gadget */
	    break;	     

	case USB_REQ_SET_CONFIGURATION:
	    /* remember state & handle on the end status stage interrupt */
	    DBG(3, "USB_REQ_SET_CONFIGURATION()\n");
	    pThis->bDeviceState = (pControlRequest->wValue & 0xff)
		    ? MGC_STATE_CONFIGURED : MGC_STATE_ADDRESS;
	    handled=0; /* pass it to the gadget */
	    break;
#endif

	case USB_REQ_SET_ADDRESS:
	    /* remember state & handle on the end status stage interrupt */
	    DBG(3, "USB_REQ_SET_ADDRESS(0x%x)\n",(uint8_t)
	    	(pControlRequest->wValue & 0x7f));

		if(pControlRequest->wValue < 0x80){
		    pThis->bSetAddress = TRUE;
		    pThis->bAddress = (uint8_t)(pControlRequest->wValue & 0x7f);
		    pThis->bDeviceState = MGC_STATE_ADDRESS;
		}
		else {
		    handled=-EINVAL;
		}

	    break;

#if 0
	case USB_REQ_CLEAR_FEATURE:
	    DBG(3, "USB_REQ_CLEAR_FEATURE()\n");

	    switch(bRecip) {
		case USB_RECIP_DEVICE:
		    DBG(3, "USB_RECIP_DEVICE()\n");
		    break;
		case USB_RECIP_INTERFACE:
		    DBG(3, "USB_RECIP_INTERFACE()\n");
		    break;
		case USB_RECIP_ENDPOINT: {
			struct usb_request *pRequest;
			const uint8_t bEnd = (uint8_t)pControlRequest->wIndex & 0x7f ;
			MGC_GadgetLocalEnd* pEnd=&MGC_aGadgetLocalEnd[ bEnd ];

			DBG(4, "CLEAR_FEATURE: USB_RECIP_ENDPOINT() %d\n", bEnd );			
			MGC_GadgetSetHalt( &pEnd->end_point, 0);
			EP_SPIN_LOCK( pEnd );
			pRequest=MGC_CurrentRequest( pEnd );
			EP_SPIN_UNLOCK( pEnd );
			MGC_RestartRequest(pThis, bEnd, pRequest);  
			
			/* select ep0 again */
			MGC_SelectEnd(pBase, 0);			
		    } break;
		default:
		    break;		
	    }
	    break; /* END: CLEAR_FEATURE */

	case USB_REQ_SET_FEATURE:
	    DBG(3, "USB_REQ_SET_FEATURE()\n");		
	    switch(bRecip) {
		case USB_RECIP_DEVICE:
		    DBG(3, "USB_RECIP_DEVICE()\n");
		    switch(pControlRequest->wValue) {
			case 1:
			    DBG(3, "REMOTE_WAKEUP()\n");
			    while (0) { } /* remote wakeup */
			break;
			case 2:
			    if (pControlRequest->wIndex & 0xff) {
			        handled=-EINVAL;
 			    } else {
			    uint16_t wTest;

			    DBG(3, "ENTERING TESTMODE\n");
			    pThis->bTestMode = TRUE;
			    wTest = (uint8_t)pControlRequest->wIndex >> 8;
			    switch(wTest) {
				case 1:
				    DBG(3, "TEST_J\n");
				    /* TEST_J */
				    pThis->bTestModeValue = MGC_M_TEST_J;
				    break;
				case 2:
				    /* TEST_K */
				    DBG(3, "TEST_K\n");
				    pThis->bTestModeValue = MGC_M_TEST_K;
				    break;
				case 3:
				    /* TEST_SE0_NAK */
				    DBG(3, "TEST_SE0_NAK\n");
				    pThis->bTestModeValue = MGC_M_TEST_SE0_NAK;
				    break;
				case 4:
				    /* TEST_PACKET */
				    DBG(3, "TEST_PACKET\n");
				    pThis->bTestModeValue = MGC_M_TEST_PACKET;
				    break;
				default:
				    /* my gadget might know what to do with it */
				    break;
			    }
			}
			break;
#if 0
#ifdef MUSB_OTG
		    	case 3:
			    GADGET_SET_B_HNP_ENABLE(pThis->pGadget, 1);			
			    MGC_OtgMachineSetFeature(&(pThis->OtgMachine), 
			        pControlRequest->wValue);
			break;
			case 4:
			    GADGET_SET_A_HNP_SUPPORT(pThis->pGadget, 1);			
			    MGC_OtgMachineSetFeature(&(pThis->OtgMachine), 
				pControlRequest->wValue);
			break;
			case 5:
			    GADGET_SET_A_ALT_HNP_SUPPORT(pThis->pGadget, 1);
			    MGC_OtgMachineSetFeature(&(pThis->OtgMachine), 
				pControlRequest->wValue);
			break;
#endif
#endif /* 0 */
		    }
		    break;

		    case USB_RECIP_INTERFACE:
			DBG(3, "USB_RECIP_INTERFACE()\n");
			break;

		    case USB_RECIP_ENDPOINT: {
			const uint8_t bEnd = (uint8_t)pControlRequest->wIndex & 0x7f ;
			MGC_GadgetLocalEnd* pEnd=&MGC_aGadgetLocalEnd[ bEnd ];

			DBG(3, "SET_FEATURE: USB_RECIP_ENDPOINT() %d\n", bEnd );			
			MGC_GadgetSetHalt(&pEnd->end_point, 1);

			/* select ep0 again */
			MGC_SelectEnd(pBase, 0);
		    } break;
		
		}
	    break; /* END: SET_FEATURE */
#endif
	    
	default:
	    handled=0;
	    break;
    }

    /* standard request not handed by this code go to the gadget */
    if ( !handled ) {
    	handled=forward_to_driver(pControlRequest);
    }

    DBG(-1002, "==>\n");
    return handled;
}

/* ep0_rxstart - enable receive data in ControlOUT */
int
ep0_rxstart(
    void)
{
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    uint16_t wCount;            /* COUNT0 - received bytes */

    spin_lock_irqsave(&pThis->Lock, flags);

    MGC_SelectEnd(pBase, 0);
    wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);
    
    if (wCount != 0) {
        spin_unlock_irqrestore(&pThis->Lock, flags);
        return -ECONNRESET;
    }
    /* ack the request first */
    DBG(3, "delayed acking request %s\n",
        decode_csr0(MGC_M_CSR0_P_SVDRXPKTRDY) );
    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0,  MGC_M_CSR0_P_SVDRXPKTRDY);

    spin_unlock_irqrestore(&pThis->Lock, flags);

    return 0;
}

/*
 * ep0_rxstate is receive data from FIFO to buffer.
 * it should be called on interrupt context.
 */
int
ep0_rxstate(
    void)
{
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[0]);
    struct usb_request* pRequest = MGC_CurrentRequest(pEnd);
    uint16_t wCount;            /* COUNT0 - received bytes */
    uint16_t wCountChk;
    uint32_t dwRemain;           /* Remain size of required data */
    uint8_t* pFifoDist;
    uint8_t wCsrVal = 0;
    uint8_t wCsrValChk;
    uint8_t wFifoCount; 

    DBG(-1002, "<==\n");

    /* check EP0 registers */
    
    spin_lock(&pEnd->Lock);

    spin_lock(&pThis->Lock);

    MGC_SelectEnd(pBase, 0); /* select ep0 */
    wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);

    spin_unlock(&pThis->Lock);

    pFifoDist = (uint8_t*)pRequest->buf + pRequest->actual;
    dwRemain = (uint32_t)(pRequest->length - pRequest->actual);
    wFifoCount = (uint16_t)min((int)wCount, (int)dwRemain);

    /* update the ep0 hardware flags and software state*/

    if (wCount > MUSB_MAX_END0_PACKET) {
        uint8_t bFifoOffset = MGC_FIFO_OFFSET(0);
        /* OUT_CTRL  DATAEND  wCsrVal˥åȡ 
           ʹOUTSTALL֤*/
        wCsrVal |= MGC_M_CSR0_P_DATAEND;
        pRequest->actual += wCount;
        pRequest->status = -ECOMM; /* USB_ST_BUFFEROVERRUN */
        while (wCount / 4) {
            MGC_Read32(pBase, bFifoOffset);
            wCount -= 4;
        }
        while(wCount) {
            MGC_Read8(pBase, bFifoOffset);
            wCount -= 1;
        }
        goto just_comp;
    }
    else if (wCount < MUSB_MAX_END0_PACKET) {
        wCsrVal |= MGC_M_CSR0_P_DATAEND;
        if (pRequest->short_not_ok) {
            pRequest->status = -EREMOTEIO; /* USB_ST_SHORT_PACKET */
        }
        else {
            pRequest->status = 0;
        }
    }
    /* wCount == MUSB_MAX_END0_PACKET */
    /* OUT_CTRL  pRequest->zero ΩäƤС
       1Ǥޤ路NULLѥåȤԤ롣 */
    else if ((dwRemain == wCount)) {
        if(    pRequest->zero 
            && pRequest->status == -EINPROGRESS
            && pRequest->length
            && ((pRequest->length % MGC_aGadgetLocalEnd[0].wPacketSize) == 0)) {
            pRequest->zero = 0;
        }
        else {
            wCsrVal |= MGC_M_CSR0_P_DATAEND;
            pRequest->status = 0;
        }
    }

    spin_lock(&pThis->Lock);

    /* ľControlžǤΥåθSETUPEND */
    MGC_SelectEnd(pBase, 0); /* select ep0 */
    wCsrValChk = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0);
    if (wCsrValChk & MGC_M_CSR0_P_SETUPEND) {
        spin_unlock(&pThis->Lock);
        return -ECONNRESET;
    }

    /* Unload data */

    MGC_HdrcUnloadFifo(pBase, 0, wFifoCount, pFifoDist);

    if (wCount > wFifoCount) {
        pRequest->actual += wCount;
        pRequest->status = -EOVERFLOW; /* USB_ST_DATAOVERRUN */
        /* ;FIFOǡFlush */
        wCsrVal |= MGC_M_CSR0_FLUSHFIFO;
        wCsrVal |= MGC_M_CSR0_P_SENDSTALL;
    }
    else {
        pRequest->actual += wFifoCount;
    }

    wCsrVal |= MGC_M_CSR0_P_SVDRXPKTRDY | MGC_M_CSR0_P_SENTSTALL;  
    wCountChk = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);
    if (!wCountChk || wCountChk == wCount - wFifoCount) {
        DBG(3, "delayed acking request %s\n", decode_csr0(wCsrVal));
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsrVal);
        if (wCsrVal & MGC_M_CSR0_P_DATAEND) {
            DBG(2,"[10] Ep0Stage %d \n",pThis->bEnd0Stage);
            pThis->bEnd0Stage=MGC_END0_STAGE_STATUSOUT;
        }
    }

    spin_unlock(&pThis->Lock);

  just_comp:
    spin_unlock(&pEnd->Lock);

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_LOCKED(&pEnd->Lock); 
#endif

    DBG(-1002, "==>\n");

    return 0;
}

/*
 * im transmitting to the host, this code might be called from IRQ 
 * and from kernel thread.
 *
 * @pre spin_is_locked(&pEnd->Lock)
 */
int ep0_txstate(void) {
    unsigned long flags;
    MGC_LinuxCd* pThis = MGC_pFirstDriver;
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_GadgetLocalEnd* pEnd = &(MGC_aGadgetLocalEnd[0]);
    struct usb_request *pRequest=MGC_CurrentRequest(pEnd);
    uint16_t wCount;            /* COUNT0 - received bytes */
    uint16_t wCsrVal;
    uint8_t* pFifoSource; 
    uint8_t wFifoCount; 

    DBG(-1002, "<==\n"); 

#ifdef MUSB_PARANOID
    if ( !pThis || !pRequest )  {
        ERR("pThis=%p, pRequest=%p", pThis, pRequest); 
        return -EINVAL;
    }
#endif

#ifdef MUSB_PARANOID 
   ASSERT_SPINLOCK_UNLOCKED(&pThis->Lock); 
   ASSERT_SPINLOCK_LOCKED(&pEnd->Lock); 
#endif

    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, 0);
    wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);

    if (wCount != 0) {
        spin_unlock_irqrestore(&pThis->Lock, flags);
        return -ECONNRESET;
    }
    if (pRequest->actual == 0) {
        /* ack the request first */
        DBG(3, "delayed acking request %s\n",
            decode_csr0(MGC_M_CSR0_P_SVDRXPKTRDY) );
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0,  MGC_M_CSR0_P_SVDRXPKTRDY | MGC_M_CSR0_P_SENTSTALL);
    }

    /* load the data */		
    pFifoSource = (uint8_t*)pRequest->buf+pRequest->actual;
    wFifoCount =min((int)MGC_END0_FIFOSIZE,
                    (int)(pRequest->length-pRequest->actual));
    /* ľControlžǤΥåθSETUPEND */
    wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0);
    if (wCsrVal & MGC_M_CSR0_P_SETUPEND) {
        spin_unlock_irqrestore(&pThis->Lock, flags);
        return -ECONNRESET;
    }
    MGC_HdrcLoadFifo(pBase, 0, wFifoCount, pFifoSource);
    pRequest->actual+=wFifoCount; /* done */
    wCsrVal = MGC_M_CSR0_TXPKTRDY;
   
    /* update the flags */
    if ( wFifoCount <= MUSB_MAX_END0_PACKET ) {
        /* IN_CTRL pRequest->zero ΩäƤС
           1Ǥޤ路NULLѥåȤ롣 */
        if ( (pRequest->actual==pRequest->length) ) {     		/* žλ */
         /* use the zero flag to request the transmission setting it to
          * zero, after the zero packet has been submitted */
               if (pRequest->zero
                    && pRequest->status == -EINPROGRESS
                    && pRequest->length
                    && ((pRequest->length % MGC_aGadgetLocalEnd[0].wPacketSize) == 0)) {
                        pRequest->zero = 0;
                }
	        else {
		       wCsrVal |= MGC_M_CSR0_P_DATAEND;
		       pRequest->status=0; /* done */
	        }
        }
    }	
    
    /* send it out! (this will trigger the ep0 completition IRQ) 
     * serviced in interrupt_complete()  */
    DBG(-1001, "wrote wFifoCount=%d bytes, wCsrVal=%s\n", wFifoCount, 
    	decode_csr0(wCsrVal) );
    wCsrVal |= MGC_M_CSR0_P_SENTSTALL;
    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsrVal);
    spin_unlock_irqrestore(&pThis->Lock, flags);

    DBG(-1002, "==>\n"); 
    return 0;
}


/**
 * Handle ep0 interrupt of a device, lock & release pThis.
 * @param pThis this
 */
uint8_t MGC_HdrcServiceFunctionEp0(MGC_LinuxCd* pThis)
{
    uint16_t wCsrVal;
    uint16_t wCount; /* bytes available */
    const uint8_t* pBase = (uint8_t*)pThis->pRegs;
    struct usb_request* pRequest = MGC_CurrentRequest(&MGC_aGadgetLocalEnd[0]);
    
    DBG(-1002, "<==\n");

    spin_lock(&pThis->Lock);
    MGC_SelectEnd(pBase, 0); /* select ep0 */
    wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0);
    wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);

    DEBUG_CODE(4, { uint8_t myaddr=MGC_Read8(pBase, MGC_O_HDRC_FADDR); \
        uint8_t devctl = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL); \
	    printk(KERN_INFO "%s: wCrsVal=0x%x, wCount=%d, myaddr=%0x, mode=%s, ep0stage=%s\n", \
	        __FUNCTION__, wCsrVal, wCount, myaddr, decode_devctl(devctl), \
	    decode_ep0stage(pThis->bEnd0Stage) ); } );

    /* I sent a stall.. need to acknowledge it now.. */
    if(wCsrVal & MGC_M_CSR0_P_SENTSTALL) {    
	DBG(-1002, "acking stall while in ep0stage=%s\n",
	    decode_ep0stage(pThis->bEnd0Stage));

        /* STALLˤǤξstatus-EPROTO */ 
        if ( pRequest && 
            ((-EINPROGRESS == pRequest->status) || (0 == pRequest->status)) ) {
            pRequest->status = -EPROTO; /* -EPROTO */
        }

        (MGC_aGadgetLocalEnd[0]).bInactive = 0;  

        MGC_CompleteEp0Request();
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0,
                       wCsrVal & ~(MGC_M_CSR0_P_SENTSTALL | MGC_M_CSR0_TXPKTRDY));
    }

    /* setup ended prematurely, abort it  */
    if(wCsrVal & MGC_M_CSR0_P_SETUPEND) {	
	DBG(-1002, "acking setupend while in ep0stage=%s\n",
	    decode_ep0stage(pThis->bEnd0Stage));

        /* ControlIN FIFO񤭹SETUPν */
        /* SETUPSTALL */
        if (wCsrVal & MGC_M_CSR0_TXPKTRDY) {
            MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0,
                           MGC_M_CSR0_P_SVDSETUPEND
                           | MGC_M_CSR0_FLUSHFIFO
                           | MGC_M_CSR0_P_SENDSTALL);
            spin_unlock(&pThis->Lock);
            return TRUE;
        }

	/* IN_CTRL  DataStage,StatusStageˤ
	   Ǥξstatus */
	if((NULL!=pRequest) && (pThis->bEnd0Stage == MGC_END0_STAGE_RX)){
	    if(wCsrVal & MGC_M_CSR0_RXPKTRDY ){
                pRequest->status = -EPROTO;
            }
        }
	if((NULL!=pRequest) && (pThis->bEnd0Stage == MGC_END0_STAGE_TX)){
	    if(wCsrVal & MGC_M_CSR0_RXPKTRDY ){
                pRequest->status = 0;
            }
	}

        if (!wCount && pRequest) {
	    /* short_not_ok ե饰ǧ */
	    if((NULL!=pRequest) && (-EINPROGRESS == pRequest->status)){
	        if((1==pRequest->short_not_ok) &&
	           (0!=((pRequest->actual)%(MGC_aGadgetLocalEnd[0].wPacketSize))) ){
	            pRequest->status = -EREMOTEIO;
	        }
	        else{
	            pRequest->status = 0;
	        }
	    }
	}
        MGC_CompleteEp0Request();
        /* clearing it */
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0,  MGC_M_CSR0_P_SVDSETUPEND | MGC_M_CSR0_P_SENTSTALL);
    }

    spin_unlock(&pThis->Lock);    

    /* handle completition interrupt */
    if ( !wCsrVal && !wCount ) {
		DBG(2,"handle_ep0_completition_irq called [!wCount] \n");
		if(handle_ep0_completition_irq()==TRUE){
			return TRUE;
		}
    }

    switch( pThis->bEnd0Stage ) {
      case MGC_END0_STAGE_TX:
        if ((NULL!=pRequest) && (wCsrVal & MGC_M_CSR0_RXPKTRDY)) {  
            /* ǽǡqueue塢ǡߤȥơߤ */
            /* ǤSETUP */
            /* θEP0ơMGC_END0_STAGE_STATUSOUTȤʤ괰λ */
            /* 롣 */
            handle_ep0_completition_irq();
	    MGC_CompleteEp0Request();
        }
	    break;
	/* done transmitting */
	case MGC_END0_STAGE_STATUSOUT:
	case MGC_END0_STAGE_STATUSIN:
	    /* short_not_ok ե饰ǧ */
	    if((NULL!=pRequest) && (-EINPROGRESS == pRequest->status)){
	        if((1==pRequest->short_not_ok) &&
	           (0!=((pRequest->actual)%(MGC_aGadgetLocalEnd[0].wPacketSize))) ){
	            pRequest->status = -EREMOTEIO;
	        }
	        else{
	            pRequest->status = 0;
	        }
	    }
	    MGC_CompleteEp0Request(); /* async complete */
	    break;
    }    
    
    switch( pThis->bEnd0Stage ) {

	/* im alrewady writing to host, TX state, 
	 sequence #1 initiated during the setup */
	case MGC_END0_STAGE_TX: 
		DBG(2,"STAGE_TX to ep0  \n");
	    if ( (NULL!=pRequest) && ((wCsrVal & MGC_M_CSR0_TXPKTRDY)== 0) ) {
			DBG(2,"STAGE_TX to ep0_txstate \n");
	    	DBG(-1001, "MGC_END0_STAGE_TX\n");
		ep0_txstate();
	    } break;
	    
	/* im alrewady receiving from host, RX state, 
	 sequence #2 initiated during the setup */
	case MGC_END0_STAGE_RX: 
	    if (  (NULL!=pRequest) && (wCsrVal & MGC_M_CSR0_RXPKTRDY)  ) {
		DBG(-1001, "MGC_END0_STAGE_RX\n");
		ep0_rxstate();	    
	    }
	    break;
	    	    
	/* received from host, RX State, header */
	case MGC_END0_STAGE_SETUP:	
	    if ( wCsrVal & MGC_M_CSR0_RXPKTRDY ) { 
		int count=0;
		int handled=0;

		count=read_control_request(pThis, wCount);
		if ( count<0 ) { 
		    /* ack the request */
		    ERR("error reading the control request\n");
		} else if ( 0==count ) { /* I got the full packet, GREAT! */
		    struct usb_ctrlrequest *pControlRequest=(struct usb_ctrlrequest*)
				pThis->pEnd0Buffer;
            uint8_t wCsrValSet = 0;

		    DBG(-1002, "%s\n", decode_request(pControlRequest)); 

		    (MGC_aGadgetLocalEnd[0]).bInactive = 0;  

		    /* sequence #3 */
		    if ( is_zerodata_request(pControlRequest) ) {
                wCsrValSet= MGC_M_CSR0_P_SVDRXPKTRDY | MGC_M_CSR0_P_DATAEND;
                pThis->bEnd0Stage = MGC_END0_STAGE_STATUSIN;
			DBG(2,"[11] Ep0Stage %d \n",pThis->bEnd0Stage);
			handled=service_zero_data_request(pThis, 
				pControlRequest);
			if ( handled<0 ) {			
                wCsrValSet |= MGC_M_CSR0_P_SENDSTALL;    
			}
		    } else {
			/* sequence #1 */
			if ( is_tx_request(pControlRequest) ) {
			    /* write to host, a request is posted on ep0 */			    			    
			    pThis->bEnd0Stage=MGC_END0_STAGE_TX;
				DBG(2,"[12] Ep0Stage %d \n",pThis->bEnd0Stage);
			    handled=service_tx_request(pControlRequest);
			/* sequence #2, a request is posted on ep0 */
			} else if ( is_rx_request(pControlRequest) ) {
			    pThis->bEnd0Stage=MGC_END0_STAGE_RX;
				DBG(2,"[13] Ep0Stage %d \n",pThis->bEnd0Stage);
			    handled=service_rx_request(pControlRequest);
			}

                /* application stall */
                if ( handled<0 ) {
                    wCsrValSet = MGC_M_CSR0_P_SVDRXPKTRDY
                        | MGC_M_CSR0_P_SENDSTALL;
                }
            }
            /* ack the request */
            DBG(3, "handled=%d, wCsrVal=%s, ep0stage=%s\n", handled, 
                decode_csr0(wCsrValSet), 
                decode_ep0stage(pThis->bEnd0Stage) );
            spin_lock(&pThis->Lock);
            MGC_SelectEnd(pBase, 0);
            /* SETUPǻRxPktRdy򥯥ꥢʤ */
            wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);
            if (!wCount && wCsrValSet) {
		wCsrValSet |= MGC_M_CSR0_P_SENTSTALL;
                MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsrValSet);
            }
            spin_unlock(&pThis->Lock);
		}
		
	    } break;

	
	/* handle the application stall */
	default: {	    
	    uint16_t wCsrVal = MGC_M_CSR0_P_SENDSTALL;
	    switch ( pThis->bEnd0Stage & ~MGC_END0_STAGE_STALL_BIT ) {
		case MGC_END0_STAGE_TX:
	    	    wCsrVal|=MGC_M_CSR0_TXPKTRDY;
		break;
		case MGC_END0_STAGE_RX:
		    wCsrVal|=MGC_M_CSR0_RXPKTRDY;
		break;
	    }

	    WARN("Application stall from ep0stage=%s\n", 
	    	decode_ep0stage(pThis->bEnd0Stage));
	    spin_lock(&pThis->Lock);
	    MGC_SelectEnd(pBase, 0);
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsrVal);   

	    spin_unlock(&pThis->Lock);		 
	    
	    pThis->bEnd0Stage = MGC_END0_STAGE_SETUP;	
		DBG(2,"[14] Ep0Stage %d \n",pThis->bEnd0Stage);
	} break;
    } 

    return 1;
}
