/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_fifo.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/usb.h>
#include "usb_mcd_defs.h"
#include "usb_mcd_fifo.h"

#define FIFO_END 4224

/**
 * Reset dynamic FIFO size
 *
 * @param pThis controller driver instance pointer
 */
void
musb_fifo_reset(
    MGC_LinuxCd* pThis)
{
    uint8_t bEnd;

    pThis->wFifoHead = FIFO_END - 1; /* safe region */
    for(bEnd = 1; bEnd < pThis->bEndCount; bEnd++) {
        musb_fifo_map(pThis, 0x80|bEnd, 0, 1, FALSE); /* IN */
        musb_fifo_map(pThis, 0x00|bEnd, 0, 1, TRUE); /* OUT */
    }
    pThis->wFifoHead = 64;      /* 64byte offset for EP0 */
}

/**
 * map dynamic FIFO for specified ep
 *
 * @param pThis controller driver instance pointer
 * @param bEpAdr endpoint address described in endpoint descriptor
 * @param bSize specify FIFO size
 * @param bType specify 2 when double buffer or 1 when single buffer
 * @param bIsHost specify 1 when use as a host, 0 as a gadget.
 * @return FIFO address mapped in FIFO map
 */
void*
musb_fifo_map(
    MGC_LinuxCd* pThis,
    uint8_t bEpAdr,
    uint16_t wSize,
    uint8_t bType,
    uint8_t bIsHost)
{
    uint8_t i;
    uint16_t wCalSize;
    uint8_t bSizeVal;
    uint16_t wAdrVal;
    uint16_t wAdrRet;
    uint8_t bEnd = bEpAdr & USB_ENDPOINT_NUMBER_MASK;
    uint8_t bTx;
    void* pBase = pThis->pRegs;

    MGC_SelectEnd(pBase, bEnd);

    if (bType < 1 || bType > 2) {
        ERR("Unknown FIFO type\n");
        return NULL;
    }

    if (pThis->wFifoHead + wSize * bType >= FIFO_END) {
        ERR("FIFO overflow\n");
        return NULL;
    }

    bTx = (bEpAdr & USB_ENDPOINT_DIR_MASK) ? !bIsHost : bIsHost;

    i = 0;
    /* wSizeVal = ROUNDUP(log2(wSize)) - 3 */
    wCalSize = wSize >> 3;
    while (1) {
        wCalSize >>= 1;
        if (!wCalSize) {
            bSizeVal = i;
            break;
        }
        i++;
        if (i > 9) {
            ERR("FIFO size is too big\n");
            return NULL;
        }
    }
    bSizeVal |= (bType - 1) << 4; /* Double buffe or not? */

    /* wAdrVal unit is 8Byte */
    wAdrVal = pThis->wFifoHead >> 3;

    /* EP0ξTXRXɤFIFOꤹ */
    if (!bEnd || bTx) {
        MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, bSizeVal);
        MGC_Write16(pBase, MGC_O_HDRC_TXFIFOADD, wAdrVal);
    }
    if (!bEnd || !bTx) {
        MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, bSizeVal);
        MGC_Write16(pBase, MGC_O_HDRC_RXFIFOADD, wAdrVal);
    }

    wAdrRet = pThis->wFifoHead;
    pThis->wFifoHead += wSize * bType;
    DBG(3, "FifoHead is 0x%x\n", pThis->wFifoHead);

    return (void*)(0x0+wAdrRet);
}

/**
 * set default dynamic FIFO map
 *
 * @param pThis controller driver instance pointer
 */
void
musb_fifo_set_default_map(
    MGC_LinuxCd* pThis)
{
    void* pBase = pThis->pRegs;
    MGC_LinuxLocalEnd* pEnd;
    uint8_t bEnd;
    uint16_t wFifoOffset;

    /* Dynamic FIFO sizing: use pre-computed values */
    MGC_SelectEnd(pBase, 0);
    MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, 3);
    MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, 3);
    MGC_Write16(pBase, MGC_O_HDRC_TXFIFOADD, 0);
    MGC_Write16(pBase, MGC_O_HDRC_RXFIFOADD, 0);

    wFifoOffset = MGC_END0_FIFOSIZE;
    bEnd = 1;
    MGC_SelectEnd(pBase, bEnd);
    pEnd = &(pThis->aLocalEnd[bEnd]);
#if MGC_DFIFO_ISO_TX >= 0
    /* reserve ISO Tx */
    MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, MGC_DFIFO_ISO_TX_VAL);
    pEnd->wMaxPacketSizeTx = 1 << ((MGC_DFIFO_ISO_TX_VAL & 0xf)+3+(MGC_DFIFO_ISO_TX_VAL>>4));
    pEnd->wMaxPacketSizeRx = 0;
    pEnd->bIsSharedFifo = FALSE;
    MGC_Write16(pBase, MGC_O_HDRC_TXFIFOADD, wFifoOffset >> 3);
    /* move to next */
    wFifoOffset += pEnd->wMaxPacketSizeTx;
    bEnd++;
    MGC_SelectEnd(pBase, bEnd);
    pEnd = &(pThis->aLocalEnd[bEnd]);
#endif

#if MGC_DFIFO_ISO_RX >= 0
    /* reserve ISO Rx */
    MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, MGC_DFIFO_ISO_RX_VAL);
    pEnd->wMaxPacketSizeTx = 0;
    pEnd->wMaxPacketSizeRx = 1 << ((MGC_DFIFO_ISO_RX_VAL & 0xf)+3+(MGC_DFIFO_ISO_RX_VAL>>4));
    pEnd->bIsSharedFifo = FALSE;
    MGC_Write16(pBase, MGC_O_HDRC_RXFIFOADD, wFifoOffset >> 3);
    /* move to next */
    wFifoOffset += pEnd->wMaxPacketSizeRx;
    bEnd++;
    MGC_SelectEnd(pBase, bEnd);
    pEnd = &(pThis->aLocalEnd[bEnd]);
#endif
    /* reserve bulk */
    pThis->bBulkTxEnd = bEnd;
    pEnd->wMaxPacketSizeTx = 1 << ((MGC_DFIFO_BLK_VAL & 0xf)+3+(MGC_DFIFO_BLK_VAL>>4));
    MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, MGC_DFIFO_BLK_VAL);
    MGC_Write16(pBase, MGC_O_HDRC_TXFIFOADD, wFifoOffset >> 3);
    wFifoOffset += pEnd->wMaxPacketSizeTx;
    bEnd++;
    MGC_SelectEnd(pBase, bEnd);
    pEnd = &(pThis->aLocalEnd[bEnd]);
    pEnd->wMaxPacketSizeRx = 1 << ((MGC_DFIFO_BLK_VAL & 0xf)+3+(MGC_DFIFO_BLK_VAL>>4));
    MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, MGC_DFIFO_BLK_VAL);
    MGC_Write16(pBase, MGC_O_HDRC_RXFIFOADD, wFifoOffset >> 3);
    pThis->bBulkRxEnd = bEnd;
    /* move to next */
    wFifoOffset += pEnd->wMaxPacketSizeRx;
    bEnd++;
    /* finish */
    for(; bEnd < MUSB_C_NUM_EPS; bEnd++)
    {
        MGC_SelectEnd(pBase, bEnd);
	pEnd = &(pThis->aLocalEnd[bEnd]);
	MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, MGC_DFIFO_ALL_VAL);
	MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, MGC_DFIFO_ALL_VAL);
	pEnd->wMaxPacketSizeTx = pEnd->wMaxPacketSizeRx = 1 << (MGC_DFIFO_ALL_VAL+3);
	pEnd->bIsSharedFifo = TRUE;
	MGC_Write16(pBase, MGC_O_HDRC_TXFIFOADD, wFifoOffset >> 3);
	MGC_Write16(pBase, MGC_O_HDRC_RXFIFOADD, wFifoOffset >> 3);
	wFifoOffset += pEnd->wMaxPacketSizeRx;
    }

}

void
musb_fifo_dump(
    MGC_LinuxCd* pThis)
{
    void* pBase = pThis->pRegs;
    uint8_t bEnd;
    uint8_t bTxSize;
    uint8_t bRxSize;
    uint16_t wTxAdr;
    uint16_t wRxAdr;
    const char *sType[] = { "S", "D" };

    /* print config */
    for(bEnd = 0; bEnd < pThis->bEndCount; bEnd++)
    {
        MGC_SelectEnd(pBase, bEnd);
        bTxSize = MGC_Read8(pBase, MGC_O_HDRC_TXFIFOSZ);
        wTxAdr = MGC_Read16(pBase, MGC_O_HDRC_TXFIFOADD);
        bRxSize = MGC_Read8(pBase, MGC_O_HDRC_RXFIFOSZ);
        wRxAdr = MGC_Read16(pBase, MGC_O_HDRC_RXFIFOADD);
        printk("FIFO DUMP : EP=%d TX(0x%x,%s,Sz=%d,Adr=0x%x) RX(0x%x,%s,Sz=%d,Adr=0x%x)\n",
               bEnd,
               bTxSize, sType[(0x10&bTxSize)>>4], 1<<((0x0f&bTxSize)+3), wTxAdr,
               bRxSize, sType[(0x10&bRxSize)>>4], 1<<((0x0f&bRxSize)+3), wRxAdr);
    }
}
