/*-----------------------------------------------------------------------------
 * drivers/usb/mcd/usb_mcd_otg.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/interrupt.h>

#include <linux/usb.h>

#include <linux/version.h>

#include <linux/usb/mcd/usb_otg_control.h>
#include "usb_mcd_defs.h"

/******************************* INTERNALS *******************************/

enum {
    MUSB_OTGC_S_STOP = 1,
    MUSB_OTGC_S_AB_IDLE,
    MUSB_OTGC_S_B_PERIPH,
    MUSB_OTGC_S_B_WAIT_ACON,
    MUSB_OTGC_S_B_HOST,
    MUSB_OTGC_S_A_HOST,
    MUSB_OTGC_S_A_PERIPH,
};

#define MUSB_OTGC_MAX_NUM 2
static int otgc_num;
static struct usb_otg_control* otgc_list[MUSB_OTGC_MAX_NUM];

#define MUSB_OTGC_T_A_AIDL_BDIS 250
#define MUSB_OTGC_T_B_ASE0_BRST 4

/******************************* FORWARDS ********************************/

/******************************* GLOBALS *********************************/

/****************************** FUNCTIONS ********************************/

/*************************** INTERNAL FUNCTIONS **************************/

static void
expire_a_aidl_bdis_timer(
    unsigned long ptr)
{
    struct usb_otg_control* otg_control;
    struct musb_otgc_intdat* otgc_intdat;
    union usb_otg_event otg_event;

    otg_control = (struct usb_otg_control*)ptr;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    if (otgc_intdat->timer.data) {
        otgc_intdat->timer.data = (unsigned long)NULL;
    }

    otg_event.type = USB_OTG_EVENT_TYPE_RCGADGET;
    otg_event.rcgadget.status = USB_OTG_FAIL_T_A_AIDL_BDIS;
    musb_otg_control_signal(otg_control, &otg_event);
}

static void
expire_b_ase0_brst_timer(
    unsigned long ptr)
{
    struct usb_otg_control* otg_control;
    struct musb_otgc_intdat* otgc_intdat;
    union usb_otg_event otg_event;

    otg_control = (struct usb_otg_control*)ptr;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    if (otgc_intdat->timer.data) {
        otgc_intdat->timer.data = (unsigned long)NULL;
    }

    otg_event.type = USB_OTG_EVENT_TYPE_RCHOST;
    otg_event.rchost.status = USB_OTG_FAIL_T_B_ASE0_BRST;
    musb_otg_control_signal(otg_control, &otg_event);
}

static void
restart_timer(
    struct usb_otg_control* otg_control,
    void (*function)(unsigned long),
    unsigned long timeout)
{
    struct musb_otgc_intdat* otgc_intdat;

    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    if (otgc_intdat->timer.data) {
        /* Error - ¿ŸƤӽФƤ */
        return ;
    }    
    otgc_intdat->timer.function = function;
    otgc_intdat->timer.data = (unsigned long)otg_control;
    otgc_intdat->timer.expires = jiffies + timeout * HZ / 1000;
    add_timer(&otgc_intdat->timer);
}

inline void
restart_a_aidl_bdis_timer(
    struct usb_otg_control* otg_control)
{
    restart_timer(otg_control, expire_a_aidl_bdis_timer, MUSB_OTGC_T_A_AIDL_BDIS);
}

inline void
restart_b_ase0_brst_timer(
    struct usb_otg_control* otg_control)
{
    restart_timer(otg_control, expire_b_ase0_brst_timer, MUSB_OTGC_T_B_ASE0_BRST);
}

static void
stop_timer(
    struct usb_otg_control* otg_control)
{
    struct musb_otgc_intdat* otgc_intdat;

    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    if (otgc_intdat->timer.data) {
        otgc_intdat->timer.data = (unsigned long)NULL;
        del_timer(&otgc_intdat->timer);
    }
}

unsigned int
notify_cid(
    struct usb_otg_control* otg_control)
{
    int current_cid;
    union usb_otg_event otg_event;
    struct musb_otgc_intdat* otgc_intdat;
    unsigned int notified;

    notified = FALSE;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;

    if (otg_control->check_cid) {
        current_cid = otg_control->check_cid(otg_control);
        if (current_cid) {
            /* A-Plugǻ¨ */
            if (current_cid != otgc_intdat->last_cid) {
                otgc_intdat->last_cid = current_cid;
                /* CID OTG٥ȥʥ */
                otg_event.type = USB_OTG_EVENT_TYPE_CID;
                otg_event.cid.value = current_cid;
                musb_otg_control_signal(otg_control, &otg_event);
                notified = TRUE;
            }
            otgc_intdat->cid_check_count = 0;
        }
        else if (!current_cid
                 && current_cid == otgc_intdat->ref_cid
                 && current_cid != otgc_intdat->last_cid) {
            /* A-Plug³ե륿 */
            otgc_intdat->cid_check_count++;
            if (otgc_intdat->cid_check_count >= otg_control->cid_filter_count) {
                otgc_intdat->last_cid = current_cid;
                otgc_intdat->cid_check_count = 0;
                /* CID OTG٥ȥʥ */
                otg_event.type = USB_OTG_EVENT_TYPE_CID;
                otg_event.cid.value = current_cid;
                musb_otg_control_signal(otg_control, &otg_event);
                notified = TRUE;
            }
        }
        else {
            otgc_intdat->cid_check_count = 0;
        }
        otgc_intdat->ref_cid = current_cid;
    }

    return notified;
}

unsigned int
notify_vbus(
    struct usb_otg_control* otg_control)
{
    struct musb_otgc_intdat* otgc_intdat;
    union usb_otg_event otg_event;
    enum usb_otg_vbus_stat current_vbus;
    enum usb_otg_vbus_stat notify_vbus;
    unsigned int notified;

    notified = FALSE;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;

    if (otg_control->check_vbus) {
        current_vbus = otg_control->check_vbus(otg_control);
    }
    else {
        current_vbus = musb_bus_check_vbus(otgc_intdat->regs);
    }
    if (otgc_intdat->last_vbus > current_vbus) {
        /* ǻ¨ */
        /* Vbus OTG٥ȥʥ */
        otg_event.type = USB_OTG_EVENT_TYPE_VBUS;
        notify_vbus = otgc_intdat->last_vbus;
        do {
            notify_vbus--;
            otg_event.vbus.vbus_stat = notify_vbus;
            musb_otg_control_signal(otg_control, &otg_event);
        } while (notify_vbus > current_vbus);
        /* musb_otg_control_signallast_vbusȤ */
        otgc_intdat->last_vbus = notify_vbus;
        otgc_intdat->vbus_check_count = 0;
        notified = TRUE;
    }
    else if (otgc_intdat->last_vbus < current_vbus) {
        if (current_vbus == otgc_intdat->ref_vbus) {
            /* ³ե륿 */
            otgc_intdat->vbus_check_count++;
            if (otgc_intdat->vbus_check_count
                >= otg_control->vbus_filter_count) {
                /* Vbus OTG٥ȥʥ */
                otg_event.type = USB_OTG_EVENT_TYPE_VBUS;
                notify_vbus = otgc_intdat->last_vbus;
                do {
                    notify_vbus++;
                    otg_event.vbus.vbus_stat = notify_vbus;
                    musb_otg_control_signal(otg_control, &otg_event);
                } while (notify_vbus < current_vbus);
                /* musb_otg_control_signallast_vbusȤ */
                otgc_intdat->last_vbus = notify_vbus;
                otgc_intdat->vbus_check_count = 0;
                notified = TRUE;
            }
        }
        else {
            otgc_intdat->vbus_check_count = 0;
        }
    }
    else {
        /* VbusѲ̵ϥ󥿥ꥻå */
        otgc_intdat->vbus_check_count = 0;
        notified = TRUE;
    }
    otgc_intdat->ref_vbus = current_vbus;

    return notified;
}

/*
 * ؿdetector
 * CID/Vbusƻؿ(ޥϥɥ)
 */

static void
detector(
    unsigned long ptr)
{
    unsigned long flags;
    unsigned int notified_vbus;
    struct usb_otg_control* otg_control;
    struct musb_otgc_intdat* otgc_intdat;

    otg_control = (struct usb_otg_control*)ptr;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;

#if defined(CONFIG_USB_MCD_HOST)
    notify_cid(otg_control);
#endif /* defined(CONFIG_USB_MCD_HOST) */
    notified_vbus = notify_vbus(otg_control);

#if defined(CONFIG_USB_MCD_VBUSINT)
    if (notified_vbus) {
        return;
    }
#endif /* defined(CONFIG_USB_MCD_VBUSINT) */

    /* ¾泫 */
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    otgc_intdat->detector.expires
        = jiffies + otg_control->periodic_delay * HZ / 1000;
    add_timer(&otgc_intdat->detector);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);
}

/********************************* APIs **********************************/

/*
 * ؿmusb_otg_start_control
 * Controller򳫻Ϥ롣
 */

static int
start_control(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;
#if !defined(CONFIG_USB_MCD_HOST)
    union usb_otg_event otg_event;
#endif /* !defined(CONFIG_USB_MCD_HOST) */

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_STOP) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    otgc_intdat->state = MUSB_OTGC_S_AB_IDLE;

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    /* be sure interrupts are disabled before connecting ISR */
#ifdef MCD_PM_CODE
	if(musb_bus_start_hdrc()){
		printk("error in musb_bus_start_hdrc!!  \n");
	}
#endif

    /* CIDͳǧ */
    otgc_intdat->cid_check_count = 0;
    otgc_intdat->last_cid = otgc_intdat->ref_cid = -1;

    /* Vbusͳǧ */
    otgc_intdat->vbus_check_count = 0;
    otgc_intdat->last_vbus = otgc_intdat->ref_vbus = USB_OTG_VBUS_STAT_UNKNOWN;

    otgc_intdat->detector.function = detector;
    otgc_intdat->detector.data = (unsigned long)otg_control;

    if (otg_control->control_io) {
        int retval;

        retval = otg_control->control_io(otg_control, USB_OTG_IO_START);
        if (retval) {
            return retval;
        }
    }

#if !defined(CONFIG_USB_MCD_HOST)
    /* CID OTG٥ȥʥ */
    otg_event.type = USB_OTG_EVENT_TYPE_CID;
    otg_event.cid.value = 1;
    musb_otg_control_signal(otg_control, &otg_event);
#endif /* defined(CONFIG_USB_MCD_HOST) */

    /* ؿ */
    otgc_intdat->detector.expires = jiffies;
    add_timer(&otgc_intdat->detector);

    return 0;
}

/*
 * ؿstop_control
 * Controllerߤ롣
 */

static int
stop_control(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_AB_IDLE) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

#ifdef MCD_PM_CODE
	if(musb_bus_stop_hdrc()){
		printk("error in musb_bus_stop_hdrc!!  \n");
	}
#endif

    /* ؿ */
    del_timer(&otgc_intdat->detector);

    if (otg_control->control_io) {
        int retval;

        retval = otg_control->control_io(otg_control, USB_OTG_IO_STOP);
        if (retval) {
            /* ¾潪λ */
            spin_unlock_irqrestore(&otgc_intdat->lock, flags);
            return retval;
        }
    }

    otgc_intdat->state = MUSB_OTGC_S_STOP;

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstart_gadget
 * ControllerGADGET⡼ɤˤ롣
 */

static int
start_gadget(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_AB_IDLE) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    otgc_intdat->state = MUSB_OTGC_S_B_PERIPH;

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

#ifdef MCD_PM_CODE
	if(musb_bus_start_hdrc_int()){
		printk("error in musb_bus_start_hdrc_int!!  \n");
	}
#endif

    return 0;
}

/*
 * ؿstop_gadget
 * ControllerIDLE⡼ɤˤ롣
 */

static int
stop_gadget(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_B_PERIPH &&
        otgc_intdat->state != MUSB_OTGC_S_B_WAIT_ACON) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_B_ASE0_BRST */
    stop_timer(otg_control);

#ifdef MCD_PM_CODE
	if(musb_bus_stop_hdrc_int()){
		printk("error in musb_bus_stop_hdrc_int!!  \n");
	}
#endif
    otgc_intdat->state = MUSB_OTGC_S_AB_IDLE;

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿenable_rchost
 * ControllerRCHOST⡼ɤˤʤ뤳ȤĤ롣
 */

static int
enable_rchost(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_B_PERIPH) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    otgc_intdat->state = MUSB_OTGC_S_B_WAIT_ACON;

    musb_bus_enable_role_change(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿdisable_rchost
 * ControllerRCHOST⡼ɤˤʤ뤳ȤԵĤȤ롣
 */

static int
disable_rchost(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_B_PERIPH) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_B_ASE0_BRST */
    restart_b_ase0_brst_timer(otg_control);

    musb_bus_disable_role_change(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstop_rchost
 * ControllerIDLE⡼ɤˤ롣
 */

static int
stop_rchost(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_B_HOST &&
        otgc_intdat->state != MUSB_OTGC_S_B_WAIT_ACON) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_B_ASE0_BRST */
    stop_timer(otg_control);

    otgc_intdat->state = MUSB_OTGC_S_AB_IDLE;

    musb_bus_suspend(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstart_host
 * ControllerHOST⡼ɤˤ롣
 */

static int
start_host(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_AB_IDLE) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    if (otg_control->control_vbus) {
        int retval;

        retval = otg_control->control_vbus(otg_control, USB_OTG_VBUS_ON);
        if (retval) {
            /* ¾潪λ */
            spin_unlock_irqrestore(&otgc_intdat->lock, flags);
            return retval;
        }
    }

    otgc_intdat->state = MUSB_OTGC_S_A_HOST;

    musb_bus_start_session(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstop_host
 * ControllerIDLE⡼ɤˤ롣
 */

static int
stop_host(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_A_HOST) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_A_AIDL_BDIS */
    stop_timer(otg_control);

    if (otg_control->control_vbus) {
        int retval;

        retval = otg_control->control_vbus(otg_control, USB_OTG_VBUS_OFF);
        if (retval) {
            /* ¾潪λ */
            spin_unlock_irqrestore(&otgc_intdat->lock, flags);
            return retval;
        }
    }

    otgc_intdat->state = MUSB_OTGC_S_AB_IDLE;

    musb_bus_end_session(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstart_rcgadget
 * ControllerRCGADGET⡼ɤˤ롣
 */

static int
start_rcgadget(
    struct usb_otg_control* otg_control,
    unsigned int port)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_A_HOST) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_A_AIDL_BDIS */
    restart_a_aidl_bdis_timer(otg_control);

    otgc_intdat->state = MUSB_OTGC_S_A_PERIPH;

    musb_bus_suspend(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿstop_rcgadget
 * ControllerIDLE⡼ɤˤ롣
 */

static int
stop_rcgadget(
    struct usb_otg_control* otg_control,
    unsigned int port)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_A_PERIPH) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_A_AIDL_BDIS */
    stop_timer(otg_control);

    if (otg_control->control_vbus) {
        int retval;

        retval = otg_control->control_vbus(otg_control, USB_OTG_VBUS_OFF);
        if (retval) {
            /* ¾潪λ */
            spin_unlock_irqrestore(&otgc_intdat->lock, flags);
            return retval;
        }
    }

    otgc_intdat->state = MUSB_OTGC_S_AB_IDLE;

    musb_bus_end_session(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿget_mode
 * OTG⡼ɤ롣
 */

static int
get_mode(
    struct usb_otg_control* otg_control)
{
    int state;
    struct musb_otgc_intdat* otgc_intdat;

    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    switch (otgc_intdat->state) {
      case MUSB_OTGC_S_STOP:
        state = USB_OTG_CONTROL_STOP;
        break;
      case MUSB_OTGC_S_AB_IDLE:
        state = USB_OTG_CONTROL_IDLE;
        break;
      case MUSB_OTGC_S_B_PERIPH:
        state = USB_OTG_CONTROL_GADGET;
        break;
      case MUSB_OTGC_S_B_WAIT_ACON:
        state = USB_OTG_CONTROL_GADGET;
        break;
      case MUSB_OTGC_S_B_HOST:
        state = USB_OTG_CONTROL_RCHOST;
        break;
      case MUSB_OTGC_S_A_HOST:
        state = USB_OTG_CONTROL_HOST;
        break;
      case MUSB_OTGC_S_A_PERIPH:
        state = USB_OTG_CONTROL_RCGADGET;
        break;
      default:
        /* error */
        state = 0;
        break;
    }

    return state;
}

/*
 * ؿrequest_session
 * ControllerSRPѥ륷󥰤׵᤹롣
 */

static int
request_session(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_AB_IDLE) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    musb_bus_request_session(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿset_speed
 * ̿ԡɤꤹ롣
 */

static int
set_speed(
    struct usb_otg_control* otg_control,
    unsigned int speed)
{
    struct musb_otgc_intdat* otgc_intdat;

    if (speed != USB_OTG_HS_ENABLE &&
        speed != USB_OTG_HS_DISABLE) {
        return -EINVAL;
    }

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock(&otgc_intdat->lock);

    musb_bus_set_speed(otgc_intdat->regs, speed);

    /* ¾潪λ */
    spin_unlock(&otgc_intdat->lock);

    return 0;
}

/*
 * ؿget_speed
 * ̿ԡɤ롣
 */

static enum usb_otg_control_speed
get_speed(
    struct usb_otg_control* otg_control)
{
    enum usb_otg_control_speed speed;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock(&otgc_intdat->lock);

    speed = musb_bus_get_speed(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock(&otgc_intdat->lock);

    return speed;
}

/*
 * ؿset_test_mode
 * ƥȥ⡼ɤꤹ롣
 */

static int
set_test_mode(
    struct usb_otg_control* otg_control,
    enum usb_otg_control_test_mode test_mode)
{
    int retval;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock(&otgc_intdat->lock);

    if (otg_control->control_test) {
        otg_control->control_test(otg_control, test_mode);
    }
    retval = musb_bus_set_test_mode(otgc_intdat->regs, test_mode);

    /* ¾潪λ */
    spin_unlock(&otgc_intdat->lock);

    return retval;
}

/*
 * ؿget_test_mode
 * ƥȥ⡼ɤ롣
 */

static enum usb_otg_control_test_mode
get_test_mode(
    struct usb_otg_control* otg_control)
{
    enum usb_otg_control_test_mode test_mode;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock(&otgc_intdat->lock);

    test_mode = musb_bus_get_test_mode(otgc_intdat->regs);

    /* ¾潪λ */
    spin_unlock(&otgc_intdat->lock);

    return test_mode;
}

static struct usb_otg_control_ops otg_control_ops = {
    start_control,
    stop_control,
    start_gadget,
    stop_gadget,
    enable_rchost,
    disable_rchost,
    stop_rchost,
    start_host,
    stop_host,
    start_rcgadget,
    stop_rcgadget,
    get_mode,
    request_session,
    set_speed,
    get_speed,
    set_test_mode,
    get_test_mode
};

/*
 * ؿmusb_otg_control_init
 * MUSB OTGȥɥ饤Фν
 */

int
musb_otg_control_init(void)
{
    int i;

    DBG(3, ">> musb_otg_control_init\n");
    
    for (i = 0; i < MUSB_OTGC_MAX_NUM; i++) {
        otgc_list[i] = 0;
    }
    otgc_num = 0;

    return musb_otg_user_init();
}

/*
 * ؿmusb_otg_control_exit
 * MUSB OTGȥɥ饤Фνλ
 */

int
musb_otg_control_exit(void)
{
    int i;
    int retval;
    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, ">> musb_otg_control_exit\n");

    retval = musb_otg_user_exit();

    for (i = 0; i < otgc_num; i++) {
        otgc_intdat = (struct musb_otgc_intdat*)otgc_list[i]->data;
        del_timer(&otgc_intdat->detector);
        del_timer(&otgc_intdat->timer);
        kfree(otgc_list[i]->name);
        kfree(otgc_list[i]->data); // free otgc_intdat
        kfree(otgc_list[i]);
        otgc_list[i] = 0;
    }
    otgc_num = 0;

    return retval;
}

/*
 * ؿmusb_otg_control_map
 * OTGȥΥޥåԥ󥰡쥸ɥ쥹Ͽ
 */

struct usb_otg_control*
musb_otg_control_map(
    const char* name,
    void* regs)
{
    int i;
    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, ">> musb_otg_control_map\n");
    
    if (!name || !regs) {
        DBG(2, "musb_otg_control_map - arg error %p %p\n", name, regs);
        return 0;
    }

    for (i = 0; i < otgc_num; i++) {
        if (!strcmp(otgc_list[i]->name, name)) {
            otgc_intdat = (struct musb_otgc_intdat*)otgc_list[i]->data;
            otgc_intdat->regs = regs;
            DBG(3, "found\n");
            return otgc_list[i];
        }
    }
    
    return 0;
}

/*
 * ؿmusb_otg_control_map
 * OTGȥΥޥåԥ
 */

int
musb_otg_control_unmap(
    struct usb_otg_control* otg_control)
{
    DBG(3, ">> musb_otg_control_unmap\n");

    return 0;
}

/*
 * ؿmusb_otg_control_signal
 * CIDVbusѲConnect/DisconnectSRPVbusErrorߤ
 * OTGơȥޥϤ
 */

int
musb_otg_control_signal(
    struct usb_otg_control* otg_control,
    union usb_otg_event* otg_event)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, ">> musb_otg_control_signal\n");

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    switch (otg_event->type) {
      case USB_OTG_EVENT_TYPE_CID:
        goto event_notify;
        break;
      case USB_OTG_EVENT_TYPE_VBUS:
        musb_bus_notify_vbus_change(
            otgc_intdat->regs,
            otgc_intdat->last_vbus, otg_event->vbus.vbus_stat);
        goto event_notify;
        break;
      case USB_OTG_EVENT_TYPE_CON:
        if (otgc_intdat->state == MUSB_OTGC_S_B_HOST &&
            otg_event->con.value == 0) {
            goto event_notify;
        }
        else if (otgc_intdat->state == MUSB_OTGC_S_A_PERIPH &&
            otg_event->con.value == 0) {
            /* ॢT_A_AIDL_BDIS */
            stop_timer(otg_control);
            /* ٥ȥפؤ */
            otg_event->type = USB_OTG_EVENT_TYPE_RCGADGET;
            otg_event->rcgadget.status = USB_OTG_SUCCESS;
            goto event_notify;
        }
        else if (otgc_intdat->state == MUSB_OTGC_S_B_WAIT_ACON &&
            otg_event->con.value == 1) {
            /* ॢT_B_ASE0_BRST */
            stop_timer(otg_control);
            /* ơȤ */
            otgc_intdat->state = MUSB_OTGC_S_B_HOST;
            /* ٥ȥפؤ */
            otg_event->type = USB_OTG_EVENT_TYPE_RCHOST;
            otg_event->rchost.status = USB_OTG_SUCCESS;
            goto event_notify;
        }
        else {
            goto signal_ignore;
        }
        break;
      case USB_OTG_EVENT_TYPE_SRP:
        if (otgc_intdat->state == MUSB_OTGC_S_AB_IDLE) {
            goto event_notify;
        }
        else {
            goto signal_error;
        }
        break;
      case USB_OTG_EVENT_TYPE_RCHOST:
        if (otgc_intdat->state == MUSB_OTGC_S_B_WAIT_ACON) {
            if (otg_event->rchost.status == 0) {
                otgc_intdat->state = MUSB_OTGC_S_B_HOST;
            }
            else {
                otgc_intdat->state = MUSB_OTGC_S_B_PERIPH;
            }
            goto event_notify;
        }
        else {
            goto signal_error;
        }
        break;
      case USB_OTG_EVENT_TYPE_RCGADGET:
        if (otgc_intdat->state == MUSB_OTGC_S_A_PERIPH) {
            goto event_notify;
        }
        else {
            goto signal_error;
        }
        break;
      case USB_OTG_EVENT_TYPE_VBUSERROR:
        goto event_notify;
        break;
      default:
        goto signal_error;
        break;
    }

    if (otgc_intdat->state != MUSB_OTGC_S_AB_IDLE) {
        goto signal_error;
    }

  event_notify:
    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    if (!otgc_intdat->otg_core) {
        return -EINVAL;
    }

    usb_otg_core_notify(otgc_intdat->otg_core, otg_event);

    return 0;

  signal_error:
    /* 顼 */
  signal_ignore:
    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return -EBUSY;
}

/*
 * ؿevent_signal_tasklet
 * Connect/DisconnectSRPVbusErrorߤTaskletȤ
 * OTGơȥޥϤ
 */

static void
event_signal_tasklet(
    unsigned long data)
{
    unsigned long flags;
    struct usb_otg_control* otg_control;
    struct musb_otgc_intdat* otgc_intdat;
    struct musb_otgc_event_container* event_container;

    DBG(3, ">> event_signal_tasklet\n");

    /* ¾泫 */
    otg_control = (struct usb_otg_control*)data;
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    /* ƤӽФϿѤߤΤ٤ƤΥ٥ȤˤĤƽ */
    while (!list_empty(&otgc_intdat->event_list)) {
        event_container = list_entry(otgc_intdat->event_list.next,
                                     struct musb_otgc_event_container, list);
        /* ƥʥꥹȤκ */
        list_del(&event_container->list);
        /* ¾ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);
        /* ơȥޥƤӽФ */
        musb_otg_control_signal(otg_control, &event_container->event);
        /* ƥʤ˴ */
        kfree(event_container);
        /* ¾Ƴ */
        spin_lock_irqsave(&otgc_intdat->lock, flags);
    }

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);
}

/* tasklet */
DECLARE_TASKLET(event_signal, event_signal_tasklet, 0);

/*
 * ؿmusb_otg_control_signal_interrupt
 * Connect/DisconnectSRPVbusErrorߤTasklet𤷤
 * OTGơȥޥϤ
 */

void
musb_otg_control_signal_interrupt(
    struct usb_otg_control* otg_control,
    union usb_otg_event* otg_event)
{
    struct musb_otgc_intdat* otgc_intdat;
    struct musb_otgc_event_container* event_container;

    DBG(3, ">> musb_otg_control_signal_interrupt\n");

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock(&otgc_intdat->lock);

    /* ٥ȥƥʤγ */
    event_container
        = (struct musb_otgc_event_container*)
            kmalloc(sizeof(struct musb_otgc_event_container), GFP_ATOMIC);
    if (!event_container) {
        /* kmallocǤʤä硢ߤμꤳܤˤʤ */
        return;
    }

    /* ٥ȤΥԡ */
    memcpy(&event_container->event, otg_event, sizeof(union usb_otg_event));

    INIT_LIST_HEAD(&event_container->list);
    list_add_tail(&event_container->list, &otgc_intdat->event_list);

    /* ¾潪λ */
    spin_unlock(&otgc_intdat->lock);

    /* taskletΥ塼 */
    event_signal.data = (unsigned long)otg_control;
    tasklet_schedule(&event_signal);
}

/*
 * ؿmusb_otg_control_start_rchost
 * SuspendȥꥬȤControllerRCHOST⡼ɤˤ롣
 */

int
musb_otg_control_start_rchost(
    struct usb_otg_control* otg_control)
{
    unsigned long flags;
    struct musb_otgc_intdat* otgc_intdat;

    /* ¾泫 */
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_irqsave(&otgc_intdat->lock, flags);

    if (otgc_intdat->state != MUSB_OTGC_S_B_PERIPH) {
        /* ¾潪λ */
        spin_unlock_irqrestore(&otgc_intdat->lock, flags);

        return -EBUSY;
    }

    /* ॢT_B_ASE0_BRST */
    restart_b_ase0_brst_timer(otg_control);

    otgc_intdat->state = MUSB_OTGC_S_B_WAIT_ACON;

    /* ¾潪λ */
    spin_unlock_irqrestore(&otgc_intdat->lock, flags);

    return 0;
}

/*
 * ؿmusb_otg_control_create_object
 * usb_otg_control¤ΤΥ󥹥󥹥֥Ȥ
 */

struct usb_otg_control*
musb_otg_control_create_object(
    unsigned int user_data,
    unsigned int periodic_delay,
    unsigned int cid_filter_count,
    unsigned int vbus_filter_count,
    int (*check_cid)(struct usb_otg_control* otg_control),
    enum usb_otg_vbus_stat (*check_vbus)(struct usb_otg_control* otg_control),
    int (*check_suspend)(struct usb_otg_control* otg_control),
    int (*control_io)(struct usb_otg_control* otg_control, unsigned int control),
    int (*control_vbus)(struct usb_otg_control* otg_control,
                        unsigned int control),
    int (*control_test)(struct usb_otg_control* otg_control,
                        enum usb_otg_control_test_mode test_mode),
    const char* name)
{
    struct usb_otg_control* otg_control;
    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, "musb_otg_control_create_object\n");

    /* ȥΰ */
    otgc_list[otgc_num]
        = (struct usb_otg_control*)kmalloc(sizeof(struct usb_otg_control),
                                           GFP_KERNEL);
    otg_control = otgc_list[otgc_num];
    if (!otg_control) {
        DBG(2, "musb_otg_control_create_object - no memory for otg_control\n");
        return 0;
    }

    /* ȥѿ */
    otg_control->ops = &otg_control_ops;
    otg_control->data =
        (struct musb_otgc_intdat*)kmalloc(sizeof(struct musb_otgc_intdat),
                                          GFP_KERNEL);
    if (!otg_control->data) {
        DBG(2, "musb_otg_control_create_object - no memory for musb_otgc_intdat\n");
        kfree(otg_control);
        return 0;
    }
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    spin_lock_init(&otgc_intdat->lock);
    otgc_intdat->state = MUSB_OTGC_S_STOP;
    init_timer(&otgc_intdat->detector);
    init_timer(&otgc_intdat->timer);
    otgc_intdat->timer.data = (unsigned long)NULL;
    otgc_intdat->regs = 0;
    otgc_intdat->otg_core = 0;
    INIT_LIST_HEAD(&otgc_intdat->event_list);

    otg_control->user_data = user_data;
    otg_control->periodic_delay = periodic_delay;
    otg_control->cid_filter_count = cid_filter_count;
    otg_control->vbus_filter_count = vbus_filter_count;
    otg_control->check_cid = check_cid;
    otg_control->check_vbus = check_vbus;
    otg_control->check_suspend = check_suspend;
    otg_control->control_io = control_io;
    otg_control->control_vbus = control_vbus;
    otg_control->control_test = control_test;
    otg_control->name = (char *)kmalloc(strlen(name)+1, GFP_KERNEL);
    if (!otg_control->name) {
        DBG(2, "musb_otg_control_create_object - no memory for name\n");
        kfree(otg_control);
        kfree(otg_control->data);
        return 0;
    }
    strcpy((char*)otg_control->name, name);

    otgc_num++;

    DBG(3, "finished otg_control allocate %s %d\n", name, otgc_num);
    
    return otg_control;
}
EXPORT_SYMBOL(musb_otg_control_create_object);

/*
 * ؿmusb_otg_control_notify_vbuserror
 * Vbus顼Τ롣
 */

int
musb_otg_control_notify_vbuserror(
    struct usb_otg_control* otg_control)
{
    int i;
    union usb_otg_event otg_event;
    struct musb_otgc_intdat* otgc_intdat;

    if (!otg_control) {
        return -EINVAL;
    }

    i = 0;
#ifdef DEBUG
    if (otgc_num != 1) {
        return -EINVAL;
    }
#endif
    otgc_intdat = (struct musb_otgc_intdat*)otgc_list[i]->data;

    if (!otgc_intdat->otg_core) {
        return -EINVAL;
    }

    otg_event.type = USB_OTG_EVENT_TYPE_VBUSERROR;
    musb_otg_control_signal_interrupt(otg_control, &otg_event);

    return 0;
}
EXPORT_SYMBOL(musb_otg_control_notify_vbuserror);

/*
 * ؿmusb_otg_control_force_vbus
 * VbusϤ椹롣
 */
int
musb_otg_control_force_vbus(
    struct usb_otg_control* otg_control,
    unsigned int control)
{
    return (otg_control->control_vbus
            ? otg_control->control_vbus(otg_control, control)
            : 0);
}

/*
 * ؿmusb_otg_control_check_suspend
 * ǥХSuspend֤å
 */
int
musb_otg_control_check_suspend(
    struct usb_otg_control* otg_control,
    int expected)
{
    return (otg_control->check_suspend
            ? otg_control->check_suspend(otg_control)
            : expected);
}

/*
 * ؿmusb_otg_control_start_detecotr
 * Vbus/CIDΥޡdetector򳫻Ϥ롣
 */
void
musb_otg_control_start_detector(
    struct usb_otg_control* otg_control)
{
    struct musb_otgc_intdat* otgc_intdat;

    DBG(2, "<==\n" );
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    otgc_intdat->detector.expires
        = jiffies + otg_control->periodic_delay * HZ / 1000;
    add_timer(&otgc_intdat->detector);
    DBG(2, "==>\n" );
}

/*
 * ؿmusb_otg_control_stop_detecotr
 * Vbus/CIDΥޡdetectorߤ롣
 */
void
musb_otg_control_stop_detector(
    struct usb_otg_control* otg_control)
{
    struct musb_otgc_intdat* otgc_intdat;

    DBG(2, "<==\n" );
    otgc_intdat = (struct musb_otgc_intdat*)otg_control->data;
    del_timer(&otgc_intdat->detector);
    DBG(2, "==>\n" );
}

/*
 * ؿusb_otg_control_register_core
 * OTG Core򥳥ȥϿ롣
 */

int
usb_otg_control_register_core(
    struct usb_otg_core* otg_core)
{
    int i;
    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, ">> musb_otg_control_register_core\n");

    i = 0;
#ifdef DEBUG
    if (otgc_num != 1) {
        return -EINVAL;
    }
#endif

    otgc_intdat = (struct musb_otgc_intdat*)otgc_list[i]->data;
    otgc_intdat->otg_core = otg_core;

    return usb_otg_core_bind(otg_core, otgc_list[i]);
}
EXPORT_SYMBOL(usb_otg_control_register_core);

/*
 * ؿusb_otg_control_unregister_core
 * OTG Core򥳥ȥϿ롣
 */

int
usb_otg_control_unregister_core(
    struct usb_otg_core* otg_core)
{
    int i;

    struct musb_otgc_intdat* otgc_intdat;

    DBG(3, ">> musb_otg_control_unregister_core\n");

    i = 0;
#ifdef DEBUG
    if (otgc_num != 1) {
        return -EINVAL;
    }
#endif

    otgc_intdat = (struct musb_otgc_intdat*)otgc_list[i]->data;
    if (otgc_intdat->otg_core != otg_core) {
        return -EINVAL;
    }
    otgc_intdat->otg_core = 0;

    return usb_otg_core_unbind(otg_core, otgc_list[i]);
}
EXPORT_SYMBOL(usb_otg_control_unregister_core);

