/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_procfs.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This code is based on musb_procfs.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/usb_gadget.h>
#include "usb_mcd_defs.h"

int musb_proc_cid = 1;

#if MUSB_DEBUG > 0
static int atoi(char* buffer, int base, int len) {
    int result=0, digit=0;
	
    while ( len-->0 && (*buffer) ) {    
	digit=((*buffer>='0') && (*buffer<='9'))
	    ? *buffer-'0'
	    : ((*buffer>='a') && (*buffer<='f'))
	    ? *buffer-'a'
	    : -1;

	if ( digit<0 ) {	
	    break;
	}
	
	buffer++; 
	result=result*base+digit;
    }
    
    return result;
}


static const char* decode_address(int index) {
    static const char* COMMON_REGISTER_MAP[] = {
	"FAddr", "Power", "IntrTx", "IntrRx", 
	"IntrTxE", "IntrRxE", "IntrUSB", "IntrUSBE", 
	"Frame", "Index","TestMode" };
    return (index<11) ? COMMON_REGISTER_MAP[index]:NULL;
}
#endif



static int
uname(
    void* tgt,
    int (*func)(void* tgt, const char* fmt, ...),
    MGC_LinuxCd* pThis)
{
    const uint8_t* pBase=pThis->pRegs;

    return func(
        tgt, 
        "%sDRC, Mode=%s [debug=%d][dma=%s][gadget=%s][otg=%s][eps=%d(%p/%s/%d)] (Power=%02x, DevCtl=%02x)",
        ( pThis->bIsMultipoint ? "MH" : "H"),
        MUSB_MODE(pThis),
#if MUSB_DEBUG>0 
        MGC_DebugLevel,
#else
        -1,
#endif    
#if defined(MUSB_DMA)||defined(DMA_HOST_CODE)
        "yes",
#else
        "no",
#endif
#ifdef MUSB_GADGET
        "yes",
#else
        "no",
#endif
#ifdef MUSB_OTG
        "yes",
#else
        "no",
#endif    	
        pThis->bEndCount, pThis, pThis->aName,pThis->bRootSpeed,
        MGC_Read8(pBase, MGC_O_HDRC_POWER),
        MGC_Read8(pBase, MGC_O_HDRC_DEVCTL));
}


/* Read the registers and generate the report.
 */
static int
report(
    void* tgt,
    int (*func)(void* tgt, const char* fmt, ...),
    MGC_LinuxCd* pThis)
{
    int code;
    char* pProto;
    uint8_t bEnd;

    code = uname(tgt, func, pThis);
    if (code < 0) {
        goto report_error;
    }

    for(bEnd = 0; bEnd < pThis->bEndCount; bEnd++) {
        MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
        spin_lock(&pEnd->Lock);

        if (pThis->bIsDevice) {
            switch (pEnd->bTrafficType) {
              case USB_ENDPOINT_XFER_CONTROL:
                pProto = "Ctrl";
                break;
              case USB_ENDPOINT_XFER_ISOC:
                pProto = "Isoc";
                break;
              case USB_ENDPOINT_XFER_BULK:
                pProto = "Bulk";
                break;
              case USB_ENDPOINT_XFER_INT:
                pProto = "Intr";
                break;
              default:
                pProto = "Err ";
            }
        }
        else {
            switch(pEnd->bTrafficType) {
              case PIPE_ISOCHRONOUS:
                pProto = "Isoc";
                break;
              case PIPE_INTERRUPT:
                pProto = "Intr";
                break;
              case PIPE_CONTROL:
                pProto = "Ctrl";
                break;
              case PIPE_BULK:
                pProto = "Bulk";
                break;
              default:
                pProto = "Err ";
            }
        }

        code = func(
            tgt,
            "End-%01x: %s, %s, %s, proto=%s, pkt 	size=%04x, address=%02x, end=%02x\n", 
            bEnd, 
            (pEnd->pUrb ? "Busy" : "Idle"),
            ( list_empty(&(pEnd->urb_list)) ? "Q Empty" : " Q Full" ),
            ( pEnd->bIsTx ? "Tx" : "Rx" ),
            pProto, 
            pEnd->wPacketSize, 
            pEnd->bAddress, 
            pEnd->bEnd);
        if (code < 0) {
            spin_unlock(&pEnd->Lock);
            goto report_error;
        }

        code = func(
            tgt,
            "  %10ld bytes Rx in %10ld pkts; %10ld errs, %10ld overruns\n",
            pEnd->dwTotalRxBytes, 
            pEnd->dwTotalRxPackets,
            pEnd->dwErrorRxPackets,
            pEnd->dwMissedRxPackets);
        if (code < 0) {
            spin_unlock(&pEnd->Lock);
            goto report_error;
        }

        code= func(
            tgt,
            "  %10ld bytes Tx in %10ld pkts; %10ld errs, %10ld underruns\n",
            pEnd->dwTotalTxBytes, 
            pEnd->dwTotalTxPackets,
            pEnd->dwErrorTxPackets, pEnd->dwMissedTxPackets);
        if (code < 0) {
            spin_unlock(&pEnd->Lock);
            goto report_error;
        }
        spin_unlock(&pEnd->Lock);
    }
    
    return 0;

  report_error:
    WARN("A problem generating the report\n");
    return code;
}

struct print_tgt {
    int size;
};

static int
print_func(
    void* tgt,
    const char* fmt,
    ...)
{
    int code;
    va_list args;

    va_start(args, fmt);
    code = vprintk(fmt, args);
    va_end(args);
    if (code >= 0) {
        ((struct print_tgt*)tgt)->size += code;
    }

    return code;
}

struct sprint_tgt {
    int size;
    int limit;
    int off;
    char* buf;
};

static int
sprint_func(
    void* tgt,
    const char* fmt,
    ...)
{
    int code;
    struct sprint_tgt* stgt = (struct sprint_tgt*)tgt;
    va_list args;

    if (!stgt->buf) {
        return -1;
    }

    va_start(args, fmt);
    code = vsnprintf(stgt->buf, stgt->limit, fmt, args);
    va_end(args);
    if (code >= 0) {
        /* not support offset */
        stgt->buf += code;
        stgt->size += code;
        stgt->limit -= code;
    }

    return code;
}

/* Write to ProcFS
 *
 * a dummy input for a-device as CID=0
 * b dummy input for b-device as CID=1
 * C soft-connect
 * c soft-disconnect
 * I enable HS
 * i disable HS
 * R resume bus
 * S start session (OTG-friendly when OTG-compiled)
 * s stop session
 * F force session (OTG-unfriendly)
 * H request host mode
 * P disable the low-power mode that kills us in peripheral mode
 * D<num> set/query the debug level
 * Z zap
 */
static int MGC_ProcWrite(struct file *file, const char *buffer,
	unsigned long count, void *data)
{
    char cmd;
    uint8_t bReg;    
    uint8_t* pBase=((MGC_LinuxCd*)data)->pRegs;

    /* MOD_INC_USE_COUNT; */

    copy_from_user(&cmd, buffer, 1);
    switch(cmd)
    {
      case ':':
        if ( pBase ) {
            MGC_SelectEnd(pBase, 0);
            printk("" " HWVERS   %x\n", MGC_Read16(pBase, 0x6c));
            printk("" " CFGDATA  %x\n", MGC_Read8(pBase, 0x1f));
            printk("" " FADDR    %x\n", MGC_Read8(pBase, MGC_O_HDRC_FADDR));
            printk("" " INTRUSBE %x\n", MGC_Read8(pBase, MGC_O_HDRC_INTRUSBE));
            printk("" " POWER    %x\n", MGC_Read8(pBase, MGC_O_HDRC_POWER));
            printk("" " DEVCTL   %x\n", MGC_Read8(pBase, MGC_O_HDRC_DEVCTL));
            printk("" " TESTMODE %x\n", MGC_Read8(pBase, MGC_O_HDRC_TESTMODE));
            printk("" " FRAME    %x\n", MGC_Read8(pBase, MGC_O_HDRC_FRAME));
#if !defined(CONFIG_ARCH_INTEGRATOR)
            {
                unsigned long r = (unsigned long)ioremap_nocache(0x76700300, 100);
                printk("" " GPIOEASY3/Data   %x\n", MGC_Read32(r, 0x4));
                iounmap((void*)r);
            }
#endif
        }
        break;
      case 'a':
        musb_proc_cid = 0;
        break;
      case 'b':
        musb_proc_cid = 1;
        break;
      case 'v':
        {
#if defined(CONFIG_ARCH_INTEGRATOR)
            extern void musb_otg_integrator_notify_vbuserror(void);
            musb_otg_integrator_notify_vbuserror();
#endif
        }
        break;

      case 'C':
        /* D+ Pullup ON */
        if ( pBase ) {
            MGC_LinuxCd* pThis=(MGC_LinuxCd*)data;
            struct usb_gadget* gadget = pThis->pGadget;
            usb_gadget_connect(gadget);
        }
        break;
      case 'c':
        /* D+ Pullup OFF */
        if ( pBase ) {
            MGC_LinuxCd* pThis=(MGC_LinuxCd*)data;
            struct usb_gadget* gadget = pThis->pGadget;
            usb_gadget_disconnect(gadget);
        }
        break;

    case 'I':
    	if ( pBase ) {
	    bReg = MGC_Read8(pBase, MGC_O_HDRC_POWER) | MGC_M_POWER_HSENAB;
	    MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg);
	}
	break;

    case 'i':
    	if ( pBase ) {
	    bReg = MGC_Read8(pBase, MGC_O_HDRC_POWER) & ~MGC_M_POWER_HSENAB;
	    MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg);
	}
	break;

    case 'R':
    	if ( pBase ) {
        	bReg = MGC_Read8(pBase, MGC_O_HDRC_POWER);
		MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg | MGC_M_POWER_RESUME);
		WAIT_MS(10);
		MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg);
		WARN("Power Resumed\n");
	} break;
    case 'S':
    	if ( pBase ) {
	    /* might not b necessary */
	    MUSB_DEV_MODE( ((MGC_LinuxCd*)data) );
	    
	    bReg = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);
	    MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, bReg | MGC_M_DEVCTL_SESSION);
	    WARN("SESSION MODE\n");
	    MGC_OtgUpdate((MGC_LinuxCd*)data, FALSE, FALSE);
	}
	break;

    case 's':
	bReg = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);
	bReg &= ~MGC_M_DEVCTL_SESSION;
	MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, bReg);
	break;

    case 'F':
	bReg = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);
	bReg |= MGC_M_DEVCTL_SESSION;
	MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, bReg);
	break;

    case 'U':
        {
            int code;
            struct print_tgt tgt;
            tgt.size = 0;
            code = uname((void*)&tgt, print_func, (MGC_LinuxCd*)data);
            if (code < 0) {
                WARN("A problem generating the report\n");
            } 
        }
        break;

    case 'H':
    	if ( pBase ) {
        	bReg = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);
		bReg |= MGC_M_DEVCTL_HR;
		MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, bReg);
		MUSB_HST_MODE( ((MGC_LinuxCd*)data) );
		WARN("Host Mode\n");
	}
	break;
		
      case 'p':
        /* D+ Pullup ON (direct) */
        bReg = MGC_Read8(pBase, MGC_O_HDRC_POWER) | MGC_M_POWER_SOFTCONN;
        MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg);
        break;
      case 'P':
        /* D+ Pullup OFF (direct) */
        bReg = MGC_Read8(pBase, MGC_O_HDRC_POWER) & ~MGC_M_POWER_SOFTCONN;
        MGC_Write8(pBase, MGC_O_HDRC_POWER, bReg);
        break;
        
#if !defined(CONFIG_ARCH_INTEGRATOR)
      case '@':
        /* PHY reset */
        {
            extern void phy_reset(void);
            phy_reset();
        }
        break;
#endif /* !defined(CONFIG_ARCH_INTEGRATOR) */

#if (MUSB_DEBUG>0)	
    /* read & write registers */
    case 'r':
    case 'w': {
	uint8_t index=0;
    	uint32_t value=0;
	char command[64];    	
	
	memset(command, 0, sizeof(command));
	copy_from_user(command, buffer, min(count, (unsigned long)63));
	
	/* detrermine the index, 
	 * only the adrress now */
	index=atoi(&command[2], 16, count-2);
	if ( index>0 && pBase ) {
	    const char *address=decode_address(index);

	    if ( buffer[0]=='r' ) {
		value=(command[1]=='8')
        	    ? MGC_Read8(pBase, index)
		    : (command[1]=='f')
        	    ? MGC_Read16(pBase, index)
		    : 0;		
	    } else {
	    	// not write, not yet...
		index=-1;
	    }

	    if ( address ) {
		printk(KERN_INFO "%s=0x%x\n", address, value); 
	    } else {
		printk(KERN_INFO "0x%x=0x%x\n", index, value); 
	    }
	}	
    } break;

    /* Xap the controller */
    case 'Z':
    	if ( pBase ) {
		MGC_LinuxCd* pThis=data;
	
		MGC_HdrcStop(pThis);
 #if defined(CONFIG_USB_MCD_HOST)
		MGC_VirtualHubPortDisconnected(&(pThis->RootHub), 0);
 #endif
		pThis->pRootDevice = NULL;
		MGC_OtgUpdate(pThis, TRUE, FALSE);

		WARN("Stopped\n");		    
		WAIT_MS(1000);
		
		MGC_HdrcStart( pThis );
		WARN("Restarted\n");		    
	}
	break;
		
    /* set/read debug level */
    case 'D': {		
	    if ( count>1 ) {
		    char digits[8], *p=digits;
	    	    int i=0, level=0, sign=1, len=min(count-1, (unsigned long)8);
		    
		    copy_from_user(&digits, &buffer[1], len);

		    /* optional sign */
		    if ( *p=='-' ) {
			len-=1; sign=-sign; p++;
		    }
		    
		    /* read it */
		    while ( i++<len && *p>'0' && *p<'9') {
		    	level=level*10 + (*p -'0');
			p++;
		    }		    
		    
		    MGC_DebugLevel=sign*level;		    
		    INFO("MGC_DebugLevel=%d\n", MGC_DebugLevel); 
            } else {
		INFO("MGC_DebugLevel=%d\n", MGC_DebugLevel); 
	    }
		 
	} break;

    /* display queue status */
    case 'Q':
        {
            MGC_LinuxCd* pThis=(MGC_LinuxCd*)data;
            struct print_tgt tgt;
            tgt.size = 0;
    
            if( MUSB_IS_HST(pThis) ) {
                report((void*)&tgt, print_func, pThis);
            }
 #ifdef MUSB_GADGET
            else if( MUSB_IS_OTG(pThis) ) {
                report((void*)&tgt, print_func, pThis);
                dump_ep_status( pThis );
            } else if( MUSB_IS_DEV(pThis) ) {
                int index=-1;
                if (count > 2) {
                    char digits[8];
                    int len = min(count, (unsigned long)8);
                    copy_from_user(&digits, &buffer[1], len);
                    index = atoi(digits, 10, len);
                }

                if (index < 0) {
                    int code;
                    struct print_tgt tgt;
                    tgt.size = 0;
                    code = uname((void*)&tgt, print_func, (MGC_LinuxCd*)data);
                    if (code < 0) {
                        WARN("A problem generating the report\n");
                    } 
                    dump_ep_status(pThis);
                }
                else {
                    dump_ep_queue(index, 0);
                }
            }
 #endif
        }
        break;

    /* flush */
 #if defined(MUSB_GADGET) && !defined(MUSB_OBSOLETE_GADGET)
    case 'd': 
    case 'e': {
	int index=0;
	char command[64];    	
	MGC_LinuxCd* pThis=(MGC_LinuxCd*)data;
	
	memset(command, 0, sizeof(command));
	copy_from_user(command, buffer, min(count, (unsigned long)63));
	
	/* detrermine the index, 
	 * only the adrress now */
	index=atoi(&command[1], 10, count-1);
	if ( index==0 ) {
	    ERR("you can't oprate on ep0\n");
	    break;
	}

	if ( index<0 || index>pThis->bEndCount ) {
	    ERR("ep%d does not exist\n", index);
	    break;
	}
	
	if ( cmd=='d' ) {
	    MGC_GadgetDisableEndPoint(index);
	} else {
	    MGC_GadgetEnableEndPoint(index, (struct usb_endpoint_descriptor*)NULL);
	}
	} break;
 #endif /* defined(MUSB_GADGET) && !defined(MUSB_OBSOLETE_GADGET) */

	case '?': 
  	    INFO("?: D'ho\n");
            INFO(":: show registers\n");
  	    INFO("C/c: soft connect enable/disable\n");
  	    INFO("I/i: hispeed enable/disable\n");
  	    INFO("S/s: session set/clear\n");
  	    INFO("F: \n");
  	    INFO("H: host mode\n");
  	    INFO("r/w: read write register\n");
  	    INFO("Z: zap\n");
  	    INFO("U: uname\n");
  	    INFO("D: set/read dbug level\n");
  	    INFO("Q: show queue status\n");
 #if defined(MUSB_GADGET) && !defined(MUSB_OBSOLETE_GADGET)
  	    INFO("d/e: disable/enable end point (only new gadgets)\n");
            INFO("p: force D+ Pullup ON\n");
            INFO("P: force D+ Pullup OFF\n");
 #endif /* defined(MUSB_GADGET) && !defined(MUSB_OBSOLETE_GADGET) */
	 break;
#endif /* (MUSB_DEBUG>0) */

	default:
	    ERR("Command %c not implemented\n", cmd);
	break;
    }

    /* MOD_DEC_USE_COUNT; */
    return count;
} 


/*
 *
 *
 *
 *
 */
static int
MGC_ProcRead(
    char *page,
    char **start, 
    off_t off,
    int count,
    int *eof,
    void *data)
{
    struct sprint_tgt tgt;
    unsigned long flags;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)data;

    tgt.buf = page;
    tgt.size = 0;
    tgt.off = off;
    tgt.limit = count;

    spin_lock_irqsave(&pThis->Lock, flags);

    /* off0ʳξưʤmore/lessˤ̤б(catOK) */
    report((void*)&tgt, sprint_func, pThis);

    spin_unlock_irqrestore(&pThis->Lock, flags);

    *eof = 1;
    
    return tgt.size;
} 


/**
 * TODO: 2.4 and 2.6 create the pseudo-device in 2 different points
 */
void MGC_LinuxDeleteProcFs(MGC_LinuxCd* data) {    
    remove_proc_entry(data->pProcEntry->name, NULL);
}

void MGC_LinuxRemoveProcFs(MGC_LinuxCd* data) {
    remove_proc_entry(data->pProcEntry->name, NULL);
}

/**
 * TODO: 2.4 and 2.6 create the pseudo-device in 2 different points
 */
struct proc_dir_entry* MGC_LinuxCreateProcFs(char *name, MGC_LinuxCd* data) {
    if ( !name ) {
    	name=data->aName;
    }

    data->pProcEntry=create_proc_entry(name, 
	S_IFREG | S_IRUGO | S_IWUSR, NULL);
    if( data->pProcEntry )
    {
	data->pProcEntry->data = data;
#ifdef MUSB_V26
	data->pProcEntry->owner=THIS_MODULE;
#endif
	// pProcEntry->proc_fops = &MGC_ProcFileOperations;
	
	data->pProcEntry->read_proc = MGC_ProcRead;
	data->pProcEntry->write_proc = MGC_ProcWrite;

	data->pProcEntry->size = 0;

    	dbg("Registered /proc/%s\n", name);
    } else {
        dbg ("Cannot create a valid proc file entry");    
    }

    musb_proc_cid = 1;

    return data->pProcEntry;  
}
