/*
 * Copyright (C) 2004, OGAWA Hirofumi
 * Released under GPL v2.
 */

#include <linux/module.h>
#include <linux/backing-dev.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>

struct fatent_operations {
	void (*ent_blocknr)(struct super_block *, int, int *, sector_t *);
	void (*ent_set_ptr)(struct fat_entry *, int);
	int (*ent_bread)(struct super_block *, struct fat_entry *,
			 int, sector_t);
	int (*ent_get)(struct fat_entry *, struct inode *inode);
	void (*ent_put)(struct fat_entry *, int, int);
	int (*ent_next)(struct fat_entry *);
};

static void fat12_ent_blocknr(struct super_block *sb, int entry,
			      int *offset, sector_t *blocknr)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int bytes = entry + (entry >> 1);
	WARN_ON(entry < FAT_START_ENT || sbi->max_cluster <= entry);
	*offset = bytes & (sb->s_blocksize - 1);
	*blocknr = sbi->fat_start + (bytes >> sb->s_blocksize_bits);
}

static void fat_ent_blocknr(struct super_block *sb, int entry,
			    int *offset, sector_t *blocknr)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int bytes = (entry << sbi->fatent_shift);
	WARN_ON(entry < FAT_START_ENT || sbi->max_cluster <= entry);
	*offset = bytes & (sb->s_blocksize - 1);
	*blocknr = sbi->fat_start + (bytes >> sb->s_blocksize_bits);
}

static void fat12_ent_set_ptr(struct fat_entry *fatent, int offset)
{
	struct buffer_head **bhs = fatent->bhs;
	if (fatent->nr_bhs == 1) {
		WARN_ON(offset >= (bhs[0]->b_size - 1));
		fatent->u.ent12_p[0] = bhs[0]->b_data + offset;
		fatent->u.ent12_p[1] = bhs[0]->b_data + (offset + 1);
	} else {
		WARN_ON(offset != (bhs[0]->b_size - 1));
		fatent->u.ent12_p[0] = bhs[0]->b_data + offset;
		fatent->u.ent12_p[1] = bhs[1]->b_data;
	}
}

static void fat16_ent_set_ptr(struct fat_entry *fatent, int offset)
{
	WARN_ON(offset & (2 - 1));
	fatent->u.ent16_p = (__le16 *)(fatent->bhs[0]->b_data + offset);
}

static void fat32_ent_set_ptr(struct fat_entry *fatent, int offset)
{
	WARN_ON(offset & (4 - 1));
	fatent->u.ent32_p = (__le32 *)(fatent->bhs[0]->b_data + offset);
}

static int fat12_ent_bread(struct super_block *sb, struct fat_entry *fatent,
			   int offset, sector_t blocknr)
{
	struct buffer_head **bhs = fatent->bhs;

	WARN_ON(blocknr < MSDOS_SB(sb)->fat_start);
	bhs[0] = sb_bread(sb, blocknr);
	if (!bhs[0])
		goto err;

	if ((offset + 1) < sb->s_blocksize)
		fatent->nr_bhs = 1;
	else {
		/* This entry is block boundary, it needs the next block */
		blocknr++;
		bhs[1] = sb_bread(sb, blocknr);
		if (!bhs[1])
			goto err_brelse;
		fatent->nr_bhs = 2;
	}
	fat12_ent_set_ptr(fatent, offset);
	return 0;

err_brelse:
	brelse(bhs[0]);
err:
	printk(KERN_ERR "FAT: FAT read failed (blocknr %llu)\n",
	       (unsigned long long)blocknr);
	return -EIO;
}

static int fat_ent_bread(struct super_block *sb, struct fat_entry *fatent,
			 int offset, sector_t blocknr)
{
	struct fatent_operations *ops = MSDOS_SB(sb)->fatent_ops;

	WARN_ON(blocknr < MSDOS_SB(sb)->fat_start);
	fatent->bhs[0] = sb_bread(sb, blocknr);
	if (!fatent->bhs[0]) {
		printk(KERN_ERR "FAT: FAT read failed (blocknr %llu)\n",
		       (unsigned long long)blocknr);
		return -EIO;
	}
	fatent->nr_bhs = 1;
	ops->ent_set_ptr(fatent, offset);
	return 0;
}

static void ent_get_fixup(struct fat_entry *fatent, struct inode *inode, int *next)
{
	if (unlikely(inode && MSDOS_I(inode)->i_last_dclus)) {
		struct super_block *sb = inode->i_sb;
		struct msdos_sb_info *sbi = MSDOS_SB(sb);

		if (sbi->options.batch_sync &&
		    fatent->entry == MSDOS_I(inode)->i_last_dclus) {
			pr_debug("%s :%d, %d - %d, %d\n",
				 __func__,
				 fatent->entry,
				 MSDOS_I(inode)->i_last_dclus,
				 MSDOS_I(inode)->i_new_dclus,
				 *next);
			
			*next = MSDOS_I(inode)->i_new_dclus;
		}
	}
}

static int fat12_ent_get(struct fat_entry *fatent, struct inode *inode)
{
	u8 **ent12_p = fatent->u.ent12_p;
	int next;

	if (fatent->entry & 1)
		next = (*ent12_p[0] >> 4) | (*ent12_p[1] << 4);
	else
		next = (*ent12_p[1] << 8) | *ent12_p[0];
	next &= 0x0fff;
	if (next >= BAD_FAT12)
		next = FAT_ENT_EOF;

	ent_get_fixup(fatent, inode, &next);
	
	return next;
}

static int fat16_ent_get(struct fat_entry *fatent, struct inode *inode)
{
	int next = le16_to_cpu(*fatent->u.ent16_p);
	WARN_ON((unsigned long)fatent->u.ent16_p & (2 - 1));
	if (next >= BAD_FAT16)
		next = FAT_ENT_EOF;

	ent_get_fixup(fatent, inode, &next);
	
	return next;
}

static int fat32_ent_get(struct fat_entry *fatent, struct inode *inode)
{
	int next = le32_to_cpu(*fatent->u.ent32_p) & 0x0fffffff;
	WARN_ON((unsigned long)fatent->u.ent32_p & (4 - 1));
	if (next >= BAD_FAT32)
		next = FAT_ENT_EOF;

	ent_get_fixup(fatent, inode, &next);
	
	return next;
}

static void fat12_ent_put(struct fat_entry *fatent, int new, int lock)
{
	u8 **ent12_p = fatent->u.ent12_p;

        if(lock){
                lock_buffer(fatent->bhs[0]);
                if(fatent->nr_bhs == 2){
                        lock_buffer(fatent->bhs[1]);
                }
        }
	if (new == FAT_ENT_EOF)
		new = EOF_FAT12;

	if (fatent->entry & 1) {
		*ent12_p[0] = (new << 4) | (*ent12_p[0] & 0x0f);
		*ent12_p[1] = new >> 4;
	} else {
		*ent12_p[0] = new & 0xff;
		*ent12_p[1] = (*ent12_p[1] & 0xf0) | (new >> 8);
	}

	mark_buffer_dirty(fatent->bhs[0]);
	if (fatent->nr_bhs == 2)
		mark_buffer_dirty(fatent->bhs[1]);

        if(lock){
                unlock_buffer(fatent->bhs[0]);
                if(fatent->nr_bhs == 2){
                        unlock_buffer(fatent->bhs[1]);
                }
        }
}

static void fat16_ent_put(struct fat_entry *fatent, int new, int lock)
{
	if (new == FAT_ENT_EOF)
		new = EOF_FAT16;

        if(lock)
                lock_buffer(fatent->bhs[0]);

	*fatent->u.ent16_p = cpu_to_le16(new);
	mark_buffer_dirty(fatent->bhs[0]);

        if(lock)
                unlock_buffer(fatent->bhs[0]);
}

static void fat32_ent_put(struct fat_entry *fatent, int new, int lock)
{
	if (new == FAT_ENT_EOF)
		new = EOF_FAT32;

        if(lock)
                lock_buffer(fatent->bhs[0]);

	WARN_ON(new & 0xf0000000);
	new |= le32_to_cpu(*fatent->u.ent32_p) & ~0x0fffffff;
	*fatent->u.ent32_p = cpu_to_le32(new);
	mark_buffer_dirty(fatent->bhs[0]);

        if(lock)
                unlock_buffer(fatent->bhs[0]);
}

static int fat12_ent_next(struct fat_entry *fatent)
{
	u8 **ent12_p = fatent->u.ent12_p;
	struct buffer_head **bhs = fatent->bhs;
	u8 *nextp = ent12_p[1] + 1 + (fatent->entry & 1);

	fatent->entry++;
	if (fatent->nr_bhs == 1) {
		WARN_ON(ent12_p[0] > (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 2)));
		WARN_ON(ent12_p[1] > (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1)));
		if (nextp < (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1))) {
			ent12_p[0] = nextp - 1;
			ent12_p[1] = nextp;
			return 1;
		}
	} else {
		WARN_ON(ent12_p[0] != (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1)));
		WARN_ON(ent12_p[1] != (u8 *)bhs[1]->b_data);
		ent12_p[0] = nextp - 1;
		ent12_p[1] = nextp;
		brelse(bhs[0]);
		bhs[0] = bhs[1];
		fatent->nr_bhs = 1;
		return 1;
	}
	ent12_p[0] = NULL;
	ent12_p[1] = NULL;
	return 0;
}

static int fat12_ent_next_no_sect_bndry(struct fat_entry *fatent)
{
	u8 **ent12_p = fatent->u.ent12_p;
	struct buffer_head **bhs = fatent->bhs;
	u8 *nextp = ent12_p[1] + 1 + (fatent->entry & 1);

	fatent->entry++;
	if (fatent->nr_bhs == 1) {
		WARN_ON(ent12_p[0] > (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 2)));
		WARN_ON(ent12_p[1] > (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1)));
		if (nextp < (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1))) {
			ent12_p[0] = nextp - 1;
			ent12_p[1] = nextp;
			return 1;
		} else if (nextp == (u8 *)(bhs[0]->b_data + bhs[0]->b_size)) {
			/* To avoid acrossing sector boundary */
			fatent->entry++;
		}
	} else {
		WARN_ON(ent12_p[0] != (u8 *)(bhs[0]->b_data + (bhs[0]->b_size - 1)));
		WARN_ON(ent12_p[1] != (u8 *)bhs[1]->b_data);
		ent12_p[0] = nextp - 1;
		ent12_p[1] = nextp;
		brelse(bhs[0]);
		bhs[0] = bhs[1];
		fatent->nr_bhs = 1;
		return 1;
	}
	ent12_p[0] = NULL;
	ent12_p[1] = NULL;
	return 0;
}

static int fat16_ent_next(struct fat_entry *fatent)
{
	const struct buffer_head *bh = fatent->bhs[0];
	fatent->entry++;
	if (fatent->u.ent16_p < (__le16 *)(bh->b_data + (bh->b_size - 2))) {
		fatent->u.ent16_p++;
		return 1;
	}
	fatent->u.ent16_p = NULL;
	return 0;
}

static int fat32_ent_next(struct fat_entry *fatent)
{
	const struct buffer_head *bh = fatent->bhs[0];
	fatent->entry++;
	if (fatent->u.ent32_p < (__le32 *)(bh->b_data + (bh->b_size - 4))) {
		fatent->u.ent32_p++;
		return 1;
	}
	fatent->u.ent32_p = NULL;
	return 0;
}

static struct fatent_operations fat12_ops = {
	.ent_blocknr	= fat12_ent_blocknr,
	.ent_set_ptr	= fat12_ent_set_ptr,
	.ent_bread	= fat12_ent_bread,
	.ent_get	= fat12_ent_get,
	.ent_put	= fat12_ent_put,
	.ent_next	= fat12_ent_next,
};

static struct fatent_operations fat12_ops_no_sect_bndry = {
	.ent_blocknr	= fat12_ent_blocknr,
	.ent_set_ptr	= fat12_ent_set_ptr,
	.ent_bread	= fat12_ent_bread,
	.ent_get	= fat12_ent_get,
	.ent_put	= fat12_ent_put,
	.ent_next	= fat12_ent_next_no_sect_bndry,
};

static struct fatent_operations fat16_ops = {
	.ent_blocknr	= fat_ent_blocknr,
	.ent_set_ptr	= fat16_ent_set_ptr,
	.ent_bread	= fat_ent_bread,
	.ent_get	= fat16_ent_get,
	.ent_put	= fat16_ent_put,
	.ent_next	= fat16_ent_next,
};

static struct fatent_operations fat32_ops = {
	.ent_blocknr	= fat_ent_blocknr,
	.ent_set_ptr	= fat32_ent_set_ptr,
	.ent_bread	= fat_ent_bread,
	.ent_get	= fat32_ent_get,
	.ent_put	= fat32_ent_put,
	.ent_next	= fat32_ent_next,
};

static inline void lock_fat(struct msdos_sb_info *sbi)
{
	down(&sbi->fat_lock);
}

static inline void unlock_fat(struct msdos_sb_info *sbi)
{
	up(&sbi->fat_lock);
}

void fat_ent_access_init(struct super_block *sb)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);

	init_MUTEX(&sbi->fat_lock);

	switch (sbi->fat_bits) {
	case 32:
		sbi->fatent_shift = 2;
		sbi->fatent_ops = &fat32_ops;
		break;
	case 16:
		sbi->fatent_shift = 1;
		sbi->fatent_ops = &fat16_ops;
		break;
	case 12:
		sbi->fatent_shift = -1;
		if (sbi->options.no_sect_bndry) {
			sbi->fatent_ops = &fat12_ops_no_sect_bndry;
		} else {
			sbi->fatent_ops = &fat12_ops;
		}
		break;
	}
}

static inline int fat_ent_update_ptr(struct super_block *sb,
				     struct fat_entry *fatent,
				     int offset, sector_t blocknr)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct fatent_operations *ops = sbi->fatent_ops;
	struct buffer_head **bhs = fatent->bhs;

	/* Is this fatent's blocks including this entry? */
	if (!fatent->nr_bhs || bhs[0]->b_blocknr != blocknr)
		return 0;
	/* Does this entry need the next block? */
	if (sbi->fat_bits == 12 && (offset + 1) >= sb->s_blocksize) {
		if (fatent->nr_bhs != 2 || bhs[1]->b_blocknr != (blocknr + 1))
			return 0;
	}
	ops->ent_set_ptr(fatent, offset);
	return 1;
}

int fat_ent_read(struct inode *inode, struct fat_entry *fatent, int entry)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(inode->i_sb);
	struct fatent_operations *ops = sbi->fatent_ops;
	int err, offset;
	sector_t blocknr;

	if (entry < FAT_START_ENT || sbi->max_cluster <= entry) {
		fatent_brelse(fatent);
		fat_fs_panic(sb, "invalid access to FAT (entry 0x%08x)", entry);
		return -EIO;
	}

	fatent_set_entry(fatent, entry);
	ops->ent_blocknr(sb, entry, &offset, &blocknr);

	if (!fat_ent_update_ptr(sb, fatent, offset, blocknr)) {
		fatent_brelse(fatent);
		err = ops->ent_bread(sb, fatent, offset, blocknr);
		if (err)
			return err;
	}
	return ops->ent_get(fatent, inode);
}

/* FIXME: We can write the blocks as more big chunk. */
static int fat_mirror_bhs(struct super_block *sb, struct buffer_head **bhs,
			  int nr_bhs)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct buffer_head *c_bh;
	int err, n, copy;

	err = 0;
	for (copy = 1; copy < sbi->fats; copy++) {
		sector_t backup_fat = sbi->fat_length * copy;

		for (n = 0; n < nr_bhs; n++) {
			c_bh = sb_getblk(sb, backup_fat + bhs[n]->b_blocknr);
			if (!c_bh) {
				err = -ENOMEM;
				goto error;
			}
			memcpy(c_bh->b_data, bhs[n]->b_data, sb->s_blocksize);
			set_buffer_uptodate(c_bh);
			mark_buffer_dirty(c_bh);
			if (sb->s_flags & MS_SYNCHRONOUS)
				if (!sbi->options.relax_sync)
					err = sync_dirty_buffer(c_bh);
			brelse(c_bh);
			if (err)
				goto error;
		}
	}
error:
	return err;
}

int fat_ent_write(struct inode *inode, struct fat_entry *fatent,
		  int new, int wait)
{
	struct super_block *sb = inode->i_sb;
	struct fatent_operations *ops = MSDOS_SB(sb)->fatent_ops;
	int err;

	ops->ent_put(fatent, new, 1);
	if (wait) {
		struct msdos_sb_info *sbi = MSDOS_SB(sb);
		err = sbi->options.relax_sync ?
			fat_flush_bhs(fatent->bhs, fatent->nr_bhs) 
			: fat_sync_bhs(fatent->bhs, fatent->nr_bhs);
		if (err)
			return err;
	}
	return fat_mirror_bhs(sb, fatent->bhs, fatent->nr_bhs);
}

static inline int fat_ent_next(struct msdos_sb_info *sbi,
			       struct fat_entry *fatent)
{
	if (sbi->fatent_ops->ent_next(fatent)) {
		if (fatent->entry < sbi->max_cluster)
			return 1;
	}
	return 0;
}

static inline int fat_ent_read_block(struct super_block *sb,
				     struct fat_entry *fatent)
{
	struct fatent_operations *ops = MSDOS_SB(sb)->fatent_ops;
	sector_t blocknr;
	int offset;

	fatent_brelse(fatent);
	ops->ent_blocknr(sb, fatent->entry, &offset, &blocknr);
	return ops->ent_bread(sb, fatent, offset, blocknr);
}

static void fat_collect_bhs(struct buffer_head **bhs, int *nr_bhs,
			    struct fat_entry *fatent)
{
	int n, i;

	for (n = 0; n < fatent->nr_bhs; n++) {
		for (i = 0; i < *nr_bhs; i++) {
			if (fatent->bhs[n] == bhs[i])
				break;
		}
		if (i == *nr_bhs) {
			get_bh(fatent->bhs[n]);
			bhs[i] = fatent->bhs[n];
			(*nr_bhs)++;
		}
	}
}

#ifdef FAT_SOFTUPDATE
int fat_softupdate_make_dependency(struct inode *inode, struct fat_entry *target_ent, struct fat_entry *depend_ent)
{
	int ret = 0;
	struct super_block *sb = inode->i_sb;
	
	DPRINT("[%s]\n", __FUNCTION__);
	/* judge */
	if(inode_needs_sync(inode) && 
           MSDOS_SB(sb)->options.softupdate_with_sync &&
	   !S_ISDIR(inode->i_mode)){
		ret = fat_softupdate_add_dependency(inode, target_ent->bhs[0], target_ent->entry, 
                                                    depend_ent->bhs[0], depend_ent->entry);
	}
	
	return ret;
}
#endif /* FAT_SOFTUPDATE */
#ifdef FAT_SOFTUPDATE
#define FAT_IN_CASE_NO_SYNC (99)
#endif
int fat_alloc_clusters(struct inode *inode, int *cluster, int nr_cluster)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct fatent_operations *ops = sbi->fatent_ops;
	struct fat_entry fatent, prev_ent;
	struct buffer_head *bhs[MAX_BUF_PER_PAGE];
	int i, count, err, nr_bhs, idx_clus;
	int start_from_prev = ((sbi->fat_bits == 12) && 
				sbi->options.no_sect_bndry);
#ifdef FAT_SOFTUPDATE
	int ret, flag_over_block = 0;
#endif /* FAT_SOFTUPDATE */

	BUG_ON(nr_cluster > (MAX_BUF_PER_PAGE / 2));	/* fixed limit */

	lock_fat(sbi);
	if (sbi->free_clusters != -1 && sbi->free_clusters < nr_cluster) {
		unlock_fat(sbi);
		return -ENOSPC;
	}

	err = nr_bhs = idx_clus = 0;
	count = FAT_START_ENT;
	fatent_init(&prev_ent);
	fatent_init(&fatent);
	/* if START_FROM_PREV is set, find free cluster from PREV_FREE, 
	   because PREV_FREE+1 may across sector boundary */
	fatent_set_entry(&fatent, sbi->prev_free + (start_from_prev ? 0 : 1));
	while (count < sbi->max_cluster) {
		if (fatent.entry >= sbi->max_cluster)
			fatent.entry = FAT_START_ENT;
		fatent_set_entry(&fatent, fatent.entry);
		err = fat_ent_read_block(sb, &fatent);
		if (err)
			goto out;

		/* Find the free entries in a block */
		do {
			if (ops->ent_get(&fatent, inode) == FAT_ENT_FREE) {
				int entry = fatent.entry;

				fat_gc_mark_cluster_valid(sbi, entry);
				
				/* make the cluster chain */
				ops->ent_put(&fatent, FAT_ENT_EOF, 1);
				if (prev_ent.nr_bhs)
					ops->ent_put(&prev_ent, entry, 1);

#ifdef FAT_SOFTUPDATE
				if(prev_ent.bhs[0] && flag_over_block){ /* !IS_DIR(i_mode) only called */
					DPRINT("[%s]:call fat_softupdate_make_dependency\n", __FUNCTION__);
					DTRACE("[%s]:Make UD bh=0x%p\n", __FUNCTION__, prev_ent.bhs[0]);
					ret = fat_softupdate_make_dependency(inode, &prev_ent, &fatent);
					if(ret) printk("Cannot make dependency\n");
					flag_over_block = 0;
				}
#endif /* FAT_SOFTUPDATE */
				
				fat_collect_bhs(bhs, &nr_bhs, &fatent);

				sbi->prev_free = entry;
				if (sbi->free_clusters != -1)
					sbi->free_clusters--;

				cluster[idx_clus] = entry;
				idx_clus++;
				if (idx_clus == nr_cluster)
					goto out;

				/*
				 * fat_collect_bhs() gets ref-count of bhs,
				 * so we can still use the prev_ent.
				 */
				prev_ent = fatent;
			}
			count++;
			if (count == sbi->max_cluster)
				break;
		} while (fat_ent_next(sbi, &fatent));
#ifdef FAT_SOFTUPDATE
		flag_over_block = 1;
#endif /* FAT_SOFTUPDATE */
	}

	/* Couldn't allocate the free entries */
	sbi->free_clusters = 0;
	err = -ENOSPC;

out:
	unlock_fat(sbi);
	fatent_brelse(&fatent);
	if (!err) {
#ifdef FAT_SOFTUPDATE
		if (inode_needs_sync(inode)){
                        err = FAT_IN_CASE_NO_SYNC; /* When No sync, to skip fat_mirror_bhs() */
                        if(S_ISDIR(inode->i_mode)){
                                err = sbi->options.relax_sync ?
                                        fat_flush_bhs(bhs, nr_bhs)
                                        : fat_sync_bhs(bhs, nr_bhs);
                        }else if(sbi->options.relax_sync){
                                if(!sbi->options.softupdate_with_sync){
                                        err = fat_flush_bhs(bhs, nr_bhs);
                                }
                        }else{
                                err = fat_sync_bhs(bhs, nr_bhs);
                        }
                }
#else
		if (inode_needs_sync(inode))
			err = sbi->options.relax_sync ?
				fat_flush_bhs(bhs, nr_bhs)
				: fat_sync_bhs(bhs, nr_bhs);
#endif
		if (!err)
			err = fat_mirror_bhs(sb, bhs, nr_bhs);
#ifdef FAT_SOFTUPDATE
                if(err == FAT_IN_CASE_NO_SYNC) err = 0;
#endif
	}
	for (i = 0; i < nr_bhs; i++)
		brelse(bhs[i]);
	fat_clusters_flush(sb);

	if (err && idx_clus)
		fat_free_clusters(inode, cluster[0]);

	return err;
}

int fat_free_clusters(struct inode *inode, int cluster)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct fatent_operations *ops = sbi->fatent_ops;
	struct fat_entry fatent;
	struct buffer_head *bhs[MAX_BUF_PER_PAGE];
#ifdef FAT_SOFTUPDATE
	struct buffer_head *prev_bhs;
	int rm_entry_target = 0, rm_entry_depend = 0;
        int first_link_check = 1;
#endif /* FAT_SOFTUPDATE */
	int i, err, nr_bhs;
        DTRACE("[%s]:Start\n", __FUNCTION__);
#ifdef FAT_SOFTUPDATE
        if(!cluster) return 0;
#endif /* FAT_SOFTUPDATE */
	nr_bhs = 0;
	fatent_init(&fatent);
	lock_fat(sbi);
	do {
		fat_gc_mark_cluster_valid(sbi, cluster);
		
#ifdef FAT_SOFTUPDATE
	        prev_bhs = fatent.bhs[0];
		rm_entry_target = rm_entry_depend;
		rm_entry_depend = cluster;
#endif /* FAT_SOFTUPDATE */
		cluster = fat_ent_read(inode, &fatent, cluster);
		if (cluster < 0) {
			err = cluster;
			goto error;
		} else if (cluster == FAT_ENT_FREE) {
#ifdef FAT_SOFTUPDATE
                        printk("[%s]:cluster=%d\n", __FUNCTION__, rm_entry_depend);
#endif /* FAT_SOFTUPDATE */
			fat_fs_panic(sb, "%s: deleting FAT entry beyond EOF",
				     __FUNCTION__);
			err = -EIO;
			goto error;
		}
#ifdef FAT_SOFTUPDATE
                if(first_link_check){
                        DTRACE("[%s]:Check first link\n", __FUNCTION__);
                        fat_softupdate_test_rm_dependency(fatent.bhs[0], rm_entry_depend);
                        first_link_check = 0;
                }
		DPRINT("[%s]:Check fatent.bhs[0]=0x%p\n", __FUNCTION__, fatent.bhs[0]);
		if((prev_bhs != fatent.bhs[0]) && prev_bhs && prev_bhs->b_private && (sbi->options.softupdate_with_sync)){
			DPRINT("[%s]:call fat_softupdate_rm_dependency\n", __FUNCTION__);
			DTRACE("[%s]:call fat_softupdate_rm_dependency\n", __FUNCTION__);
			if(fat_softupdate_rm_dependency(prev_bhs, rm_entry_target, rm_entry_depend)){
				printk("[%s]: Cannot search depend list\n",
				       __FUNCTION__);
			}
			/* Check target ud list exist */
			fat_softupdate_test_free_private(fatent.bhs[0]);
		}
#endif /* FAT_SOFTUPDATE */

                DPRINT("[%s]:rm_entry_target(%d)->rm_entry_depend(%d)=>FAT_ENT_FREE\n", __FUNCTION__, rm_entry_target, rm_entry_depend);
		ops->ent_put(&fatent, FAT_ENT_FREE, 1);
		if (sbi->free_clusters != -1)
			sbi->free_clusters++;

		if (nr_bhs + fatent.nr_bhs > MAX_BUF_PER_PAGE) {
			if (sb->s_flags & MS_SYNCHRONOUS) {
				err = sbi->options.relax_sync ?
					fat_flush_bhs(bhs, nr_bhs)
					: fat_sync_bhs(bhs, nr_bhs);
				if (err)
					goto error;
			}
			err = fat_mirror_bhs(sb, bhs, nr_bhs);
			if (err)
				goto error;
			for (i = 0; i < nr_bhs; i++)
				brelse(bhs[i]);
			nr_bhs = 0;
		}
		fat_collect_bhs(bhs, &nr_bhs, &fatent);
	} while (cluster != FAT_ENT_EOF);

#ifdef FAT_SOFTUPDATE
	/* When EOF link removed */
	if(!prev_bhs && (sbi->options.softupdate_with_sync)){
		DPRINT("[%s]:call fat_softupdate_test_rm_dependency\n", __FUNCTION__);
		if(fat_softupdate_test_rm_dependency(fatent.bhs[0], rm_entry_depend)){
			printk("[%s]: Cannot search depend list\n",
			       __FUNCTION__);
		}
	}
#endif /* FAT_SOFTUPDATE */

	if (sb->s_flags & MS_SYNCHRONOUS) {
		err = sbi->options.relax_sync ?
			fat_flush_bhs(bhs, nr_bhs)
			: fat_sync_bhs(bhs, nr_bhs);
		if (err)
			goto error;
	}
	err = fat_mirror_bhs(sb, bhs, nr_bhs);
error:
	fatent_brelse(&fatent);
	for (i = 0; i < nr_bhs; i++)
		brelse(bhs[i]);
	unlock_fat(sbi);

	fat_clusters_flush(sb);

	return err;
}

EXPORT_SYMBOL(fat_free_clusters);

int fat_count_free_clusters(struct super_block *sb)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct fatent_operations *ops = sbi->fatent_ops;
	struct fat_entry fatent;
	int err = 0, free;

	lock_fat(sbi);
	if (sbi->free_clusters != -1)
		goto out;

	free = 0;
	fatent_init(&fatent);
	fatent_set_entry(&fatent, FAT_START_ENT);
	while (fatent.entry < sbi->max_cluster) {
		err = fat_ent_read_block(sb, &fatent);
		if (err)
			goto out;

		do {
			if (ops->ent_get(&fatent, NULL) == FAT_ENT_FREE)
				free++;
		} while (fat_ent_next(sbi, &fatent));
	}
	sbi->free_clusters = free;
	fatent_brelse(&fatent);
out:
	unlock_fat(sbi);
	return err;
}
#define FAT_DEBUG_ON
#ifdef FAT_SOFTUPDATE
#define FAT_SU_WRITE_DUMMY         (0x1000)
#define FAT_SU_REWRITE_ORG_LINK    (0x1001)
#include <linux/pagevec.h>
static int fat_softupdate_is_all_UD_updated(struct msdos_inode_info *inode_info)
{
	unsigned long flags;
        DTRACE("[%s]:Is_list_empty=%d, inode_info->i_ud=0x%p\n", __FUNCTION__, 
               list_empty(&inode_info->i_depend_private), inode_info->i_ud);
        spin_lock_irqsave(&inode_info->softupdate_ud_lock, flags);
        if(list_empty(&inode_info->i_depend_private) && (inode_info->i_ud == NULL)){
                spin_unlock_irqrestore(&inode_info->softupdate_ud_lock, flags);
                DTRACE("[%s]:All UD updated\n", __FUNCTION__);
                return 1;
        }
        spin_unlock_irqrestore(&inode_info->softupdate_ud_lock, flags);
        DTRACE("[%s]:remain some UDs\n", __FUNCTION__);
        return 0;
}
static int fat_softupdate_replace_link(struct buffer_head *bhp, int cmd)
{
	struct super_block *sb;
        struct msdos_dir_entry *raw_entry;
	struct fatent_operations *ops;
	struct list_head *p_lh, *p_lh_start;
	struct fat_softupdate_ud *p_replace_ud;
	struct fat_entry *fatent;
	int entry_target, entry_depend;
        loff_t i_pos;
        loff_t file_size;
	
	DPRINT("[%s]:cmd=0x%x, bhp=0x%p\n", __FUNCTION__, cmd, bhp);
	/* search target list */
	if(bhp->b_private == NULL) {
		DTRACE("[%s]:No UD link in this bh(=0x%x)\n", __FUNCTION__, bhp);
		return 0;
	}
	sb = ((struct fat_bh_private_data *)bhp->b_private)->ud.sb;
	p_lh_start = &((struct fat_bh_private_data *)bhp->b_private)->ud.d_target_head;
	
	/* Check target list exist */
	if(!list_empty(p_lh_start)) p_lh = p_lh_start->next;
	else return 0;

	/* replace entry by all target list */
	do{
		p_replace_ud = list_entry(p_lh, struct fat_softupdate_ud, d_target);
		
                entry_target = p_replace_ud->entry_target;
		entry_depend = p_replace_ud->entry_depend;
                file_size = p_replace_ud->size;
		DPRINT("[%s]:target=0x%x, depend=0x%x\n", __FUNCTION__, entry_target, entry_depend);
		fatent = &(p_replace_ud->fat_ent_ud);
		
#ifdef FAT_DEBUG_ON
		/* FOR debug */
		if(p_replace_ud->fat_ent_ud.bhs[0] != bhp) BUG_ON(1);
#endif /* FAT_DEBUG_ON */
		
		switch(cmd){
                case FAT_SU_WRITE_DUMMY:
                        if(p_replace_ud->flag_in_replace){
                                printk("ERR!!, This UD is in replace\n");
                                BUG_ON(1);
                                return -1;
                        }
                        /* Set in_replace flag */
                        p_replace_ud->flag_in_replace = 1;

                        if(p_replace_ud->ud_disc == FAT_SU_UD_DISC_CLSLINK){
                                DTRACE("[%s]:FAT_SU_WRITE_DUMMY, Clustor link\n", __FUNCTION__);
                                ops = MSDOS_SB(sb)->fatent_ops;
                                /* read fat entry */
                                fatent_init(fatent);
                                fatent_set_entry(fatent, entry_target);
                                DPRINT("[%s]:1\n", __FUNCTION__);
                                /* Only FAT16 or FAT32, FAT12 with NO boundary */
                                {
                                        sector_t blocknr;
                                        int offset;
                                        fatent_brelse(fatent);
                                        ops->ent_blocknr(sb, fatent->entry, &offset, &blocknr);
                                        fatent->bhs[0] = bhp;
                                        fatent->nr_bhs = 1;
                                        ops->ent_set_ptr(fatent, offset);
                                }
                                DPRINT("[%s]:2\n", __FUNCTION__);
                                if(entry_depend != ops->ent_get(fatent, NULL)){
                                        printk("[%s]:ERR!!, Wrong depend entry[0x%x]<-> fatent entry[0x%x]\n", __FUNCTION__,
                                               entry_depend, ops->ent_get(fatent, NULL));
                                        return -1;
                                }
			
                                /* replace (entry_target -> entry_depend) -> (entry_target -> EOF) */
                                DPRINT("[%s]:Rollback bh->b_blocknr[%d]\n", __FUNCTION__, bhp->b_blocknr);
                                DPRINT("[%s]:Rollback entry fatent entry[0x%x] -> FAT_ENT_EOF\n", __FUNCTION__, ops->ent_get(fatent, NULL));
                                DTRACE("[%s]:Rollback entry fatent entry[0x%x] -> FAT_ENT_EOF(0x%x)\n", __FUNCTION__, ops->ent_get(fatent, NULL), FAT_ENT_EOF);
                                ops->ent_put(fatent, FAT_ENT_EOF, 0);
                                DPRINT("[%s]:Rollbacked entry fatent entry[0x%x] -> FAT_ENT_EOF\n", __FUNCTION__, ops->ent_get(fatent, NULL));
			
                        }else if( p_replace_ud->ud_disc == FAT_SU_UD_DISC_DIRENT ){
                                DTRACE("[%s]:FAT_SU_WRITE_DUMMY, Dirent\n", __FUNCTION__);
#if 1
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [p_replace_ud->i_pos_and_mask];
#else
                                i_pos = MSDOS_I(p_replace_ud->inode)->i_pos;
                                DTRACE("[%s]:i_pos=%d, start=%d\n", __FUNCTION__, (int)i_pos, (int)cpu_to_le16(entry_depend));
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [i_pos & (sbi->dir_per_block - 1)];
#endif
                                if(raw_entry->start != cpu_to_le16(entry_depend)){
                                        printk("[%s]:Wrong entry, raw_entry->start=%d, entry_depend=%d\n", __FUNCTION__, 
                                               (int)raw_entry->start, (int)cpu_to_le16(entry_depend));
                                        break;
                                }
                                if(raw_entry->starthi != cpu_to_le16(entry_depend >> 16)){
                                        printk("[%s]:Wrong entry hi, raw_entry->starthi=%d, entry_dependhi=%d\n", __FUNCTION__, 
                                               (int)raw_entry->starthi, (int)cpu_to_le16(entry_depend >> 16));
                                        break;
                                }
                                
                                /* Set dummy data */
                                DTRACE("[%s]:WB set raw_entry start=%d -> 0\n", __FUNCTION__, (int)raw_entry->start);
                                DTRACE("[%s]:WB set raw_entry starthi=%d -> 0\n", __FUNCTION__, (int)raw_entry->starthi);
                                raw_entry->start = 0;
                                raw_entry->starthi = 0;
                        }else if( p_replace_ud->ud_disc == FAT_SU_UD_DISC_DIRENT_SIZE ){
                                DTRACE("[%s]:FAT_SU_WRITE_DUMMY, Dirent for size\n", __FUNCTION__);
#if 1
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [p_replace_ud->i_pos_and_mask];
#else
                                i_pos = MSDOS_I(p_replace_ud->inode)->i_pos;
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [i_pos & (sbi->dir_per_block - 1)];
#endif
                                DTRACE("[%s]:UD=(0x%p), raw_entry->size=%d, file_size=%d\n", __FUNCTION__, 
                                       p_replace_ud, (int)raw_entry->size, (int)cpu_to_le32(file_size));
{
                                int nr_pages;
                                struct pagevec pvec;
                                pgoff_t index;
                                pgoff_t end = -1;		/* Inclusive */
                                pagevec_init(&pvec, 0);
                                index = 0;			  /* whole-file sweep */
                                nr_pages = pagevec_lookup_tag(&pvec, p_replace_ud->inode->i_mapping, &index,
                                                             PAGECACHE_TAG_DIRTY,
                                                             min(end - index, (pgoff_t)PAGEVEC_SIZE-1) + 1);
                                DTRACE("[%s]:nr_pages=%d\n", __FUNCTION__, nr_pages);
                                if (nr_pages == 0 && 
                                    fat_softupdate_is_all_UD_updated(MSDOS_I(p_replace_ud->inode))){
                                        DTRACE("[%s]:Already body data and UDs updated, UD(0x%p)\n", __FUNCTION__, p_replace_ud);
                                        p_replace_ud->flag_ud_remove = 1;
                                        DTRACE("[%s]:Write Size=%d \n", __FUNCTION__, (int)cpu_to_le32(file_size));
                                        raw_entry->size = cpu_to_le32(file_size); /* Write size Now */
                                        break;
                                }
 }
                        }else{
                                DTRACE("[%s]:Strenge UD found\n", __FUNCTION__);
                                p_replace_ud->flag_in_replace = 0;
                        }
                        break;
                case FAT_SU_REWRITE_ORG_LINK:
			if(p_replace_ud->flag_in_replace != 1){
				printk("Caution!!, This UD is not in replace\n");
//                                BUG_ON(1);
//				return -1;
                                break;
			}
			
                        if(p_replace_ud->ud_disc == FAT_SU_UD_DISC_CLSLINK){
                                DTRACE("[%s]:FAT_SU_REWRITE_ORG_LINK, Clustor link\n", __FUNCTION__);
                                ops = MSDOS_SB(sb)->fatent_ops;
                                if(FAT_ENT_EOF != ops->ent_get(fatent, NULL)){
                                        printk("ERR!!, Not EOF[0x%x]<-> fatent entry[0x%x]\n", FAT_ENT_EOF, ops->ent_get(fatent, NULL));
                                        return -1;
                                }
			
                                /* replace (entry_target -> entry_depend) -> (entry_target -> EOF) */
                                DPRINT("[%s]:Rollforward bh->b_blocknr[%d]\n", __FUNCTION__, bhp->b_blocknr);
                                DPRINT("[%s]:Rollforward entry fatent entry[0x%x] <- FAT_ENT_EOF\n", __FUNCTION__, ops->ent_get(fatent, NULL));
                                ops->ent_put(fatent, entry_depend, 0);
                                DTRACE("[%s]:RF t_entry(%d) fatent entry[%d] entget(%d)\n", __FUNCTION__, fatent->entry, entry_depend, ops->ent_get(fatent, NULL));
                                DPRINT("[%s]:Rollforwarded entry fatent entry[0x%x] <- FAT_ENT_EOF\n", __FUNCTION__, ops->ent_get(fatent, NULL));
                                DTRACE("[%s]:Rollforwarded entry fatent entry[0x%x] <- FAT_ENT_EOF\n", __FUNCTION__, ops->ent_get(fatent, NULL));
                        }else if( p_replace_ud->ud_disc == FAT_SU_UD_DISC_DIRENT ){
                                DTRACE("[%s]:FAT_SU_REWRITE_ORG_LINK, Dirent\n", __FUNCTION__);
#if 1
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [p_replace_ud->i_pos_and_mask];
#else
                                i_pos = MSDOS_I(p_replace_ud->inode)->i_pos;
                                DTRACE("[%s]:i_pos=%d, filesize=%d\n", __FUNCTION__, (int)i_pos, (int)cpu_to_le32(file_size));
                                raw_entry = &((struct msdos_dir_entry *) (bhp->b_data))
                                        [i_pos & (sbi->dir_per_block - 1)];
#endif
                                if(raw_entry->start != 0){
                                        printk("[%s]:Wrong entry, raw_entry->start=%d, Must be 0\n", __FUNCTION__, 
                                               (int)raw_entry->start);
                                        break;
                                }
                                if(raw_entry->starthi != 0){
                                        printk("[%s]:Wrong entry hi, raw_entry->starthi=%d, Must be 0\n", __FUNCTION__, 
                                               (int)raw_entry->starthi);
                                        break;
                                }

                                /* Set dummy data */
                                DTRACE("[%s]:WF reset raw_entry size=%d -> %d\n", __FUNCTION__, 
                                       (int)raw_entry->size, (int)cpu_to_le32(file_size));
                                DTRACE("[%s]:Wf reset raw_entry start=%d -> %d\n", __FUNCTION__, 
                                       (int)raw_entry->start, (int)cpu_to_le16(entry_depend));
                                DTRACE("[%s]:Wf reset raw_entry starthi=%d -> %d\n", __FUNCTION__, 
                                       (int)raw_entry->starthi, (int)cpu_to_le16(entry_depend >> 16));
//                                raw_entry->size = cpu_to_le32(file_size);
                                raw_entry->start = cpu_to_le16(entry_depend);
                                raw_entry->starthi = cpu_to_le16(entry_depend >> 16);
                                mark_buffer_dirty(bhp);
                        }else if( p_replace_ud->ud_disc == FAT_SU_UD_DISC_DIRENT_SIZE ){
                                DTRACE("[%s]:FAT_SU_REWRITE_ORG_LINK, Dirent for size\n", __FUNCTION__);
                                if(p_replace_ud->flag_ud_remove){
                                        /* Body data is already updated */
                                        DTRACE("[%s]:Body data is already updated, remove UD(0x%p)\n", __FUNCTION__, p_replace_ud);
                                        p_replace_ud->flag_in_replace = 0;
                                        break;
                                }

                                mark_buffer_dirty(bhp);
                        }else{
                                printk("[%s]:Strenge UD found\n", __FUNCTION__);
                        }
			
			/* Set in_replace flag */
			p_replace_ud->flag_in_replace = 0;
                        break;
                default:
			printk("ERR!!, Unknown command\n");
			return -1;
		}
		p_lh = p_lh->next;
                if((p_replace_ud->flag_in_replace == 0) && p_replace_ud->flag_ud_remove){
                        struct buffer_head *tmp_bhp = NULL;
                        if( p_replace_ud->ud_disc == FAT_SU_UD_DISC_DIRENT_SIZE ){
                                tmp_bhp = bhp;  /* remove private data */
                                DTRACE("[%s]:Size UD(0x%p) removed\n", __FUNCTION__, p_replace_ud);
                        }
                        if(fat_softupdate_del_list_free_mem(tmp_bhp, p_replace_ud)){
                                DTRACE("[%s]:Cannot Removed in replaced UD list\n", __FUNCTION__);
                        }else{
                                DTRACE("[%s]:Removed in replaced UD list\n", __FUNCTION__);
                        }
                }
	}while(p_lh_start != p_lh);
	return 0;
}

int fat_softupdate_replace_ud_link(struct buffer_head *bhp)
{
	int ret;
	DTRACE("[%s]\n", __FUNCTION__);
	set_buffer_rollback(bhp);
	ret = fat_softupdate_replace_link(bhp, FAT_SU_WRITE_DUMMY);
	return ret;
}
static int fat_softupdate_replace_org_link(struct buffer_head *bhp)
{
	int ret;
	DTRACE("[%s]\n", __FUNCTION__);
	ret = fat_softupdate_replace_link(bhp, FAT_SU_REWRITE_ORG_LINK);
	clear_buffer_rollback(bhp);
	return ret;
}
/* Copied from end_buffer_io_error() in buffer.c */
static void buffer_io_error(struct buffer_head *bh)
{
	char b[BDEVNAME_SIZE];
	
	printk(KERN_ERR "Buffer I/O error on device %s, logical block %Lu\n",
	       bdevname(bh->b_bdev, b),
	       (unsigned long long)bh->b_blocknr);
}

/* Copied from end_buffer_async_write() in buffer.c */
/*
 * Completion handler for block_write_full_page() - pages which are unlocked
 * during I/O, and which have PageWriteback cleared upon I/O completion.
 */
void fat_end_buffer_async_write(struct buffer_head *bh, int uptodate)
{
	char b[BDEVNAME_SIZE];
	static DEFINE_SPINLOCK(page_uptodate_lock);
	unsigned long flags;
	struct buffer_head *tmp;
	struct page *page;

	BUG_ON(!buffer_async_write(bh));
	DPRINT("[%s]:Start\n", __FUNCTION__);
	DTRACE("[%s]:Start\n", __FUNCTION__);
	
	page = bh->b_page;
	if (uptodate) {
		set_buffer_uptodate(bh);
	} else {
		if (printk_ratelimit()) {
			buffer_io_error(bh);
			printk(KERN_WARNING "lost page write due to "
			       "I/O error on %s\n",
			       bdevname(bh->b_bdev, b));
		}
		set_bit(AS_EIO, &page->mapping->flags);
		clear_buffer_uptodate(bh);
		SetPageError(page);
	}
	
	spin_lock_irqsave(&page_uptodate_lock, flags);
	clear_buffer_async_write(bh);
	/* Clear UD list */
	if(bh->b_private != NULL) {
#if 0
                int erase_private = list_empty(&((struct fat_bh_private_data *)bh->b_private)->ud.d_target_head);
#else
                int erase_private = (list_empty(&((struct fat_bh_private_data *)bh->b_private)->ud.d_target_head) &&
                                     list_empty(&((struct fat_bh_private_data *)bh->b_private)->ud.d_depend_head))? 1:0 ;
#endif
		DTRACE("[%s]:Replace to org UD link, bh=0x%x\n", __FUNCTION__, bh);
		fat_softupdate_replace_org_link(bh);
		fat_softupdate_rm_all_dependency(bh);
                if(erase_private) {
                        DTRACE("[%s]:Free private, bh=0x%p\n", __FUNCTION__, bh);
                        fat_softupdate_test_free_private(bh);
                }
                if(bh->b_private != NULL){
                        DTRACE("[%s]:mark_buffer_dirty, bh=0x%p\n", __FUNCTION__, bh);
                        mark_buffer_dirty(bh);
                }
	}
	unlock_buffer(bh);
	tmp = bh->b_this_page;
	while (tmp != bh) {
                DPRINT("[%s]:call buffer_async_write(bh=0x%p)\n", __FUNCTION__, tmp);
		if (buffer_async_write(tmp)) { /* check test bit */
			BUG_ON(!buffer_locked(tmp));
			goto still_busy;
		}
		tmp = tmp->b_this_page;
	}
	spin_unlock_irqrestore(&page_uptodate_lock, flags);
	end_page_writeback(page);
	return;
	
still_busy:
	spin_unlock_irqrestore(&page_uptodate_lock, flags);
	return;
}
#endif /* FAT_SOFTUPDATE */
