/*
 *  linux/fs/fat/file.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *
 *  Copyright 2005,2006 Sony Corporation
 *
 *  regular file handling primitives for fat-based filesystems
 */

#include <linux/module.h>
#include <linux/time.h>
#include <linux/msdos_fs.h>
#include <linux/smp_lock.h>
#include <linux/buffer_head.h>
#include <linux/uio.h>
#include <linux/writeback.h>

#include <linux/fs.h>

int fat_generic_ioctl(struct inode *inode, struct file *filp,
		      unsigned int cmd, unsigned long arg)
{
	struct msdos_sb_info *sbi = MSDOS_SB(inode->i_sb);
	u32 __user *user_attr = (u32 __user *)arg;

	switch (cmd) {
	case FAT_IOCTL_CLOSE_NOSYNC:
	{
		filp->f_op = &fat_file_operations_nosync;
		return 0;
	}
	case FAT_IOCTL_GET_ATTRIBUTES:
	{
		u32 attr;

		if (inode->i_ino == MSDOS_ROOT_INO)
			attr = ATTR_DIR;
		else
			attr = fat_attr(inode);

		return put_user(attr, user_attr);
	}
	case FAT_IOCTL_SET_ATTRIBUTES:
	{
		u32 attr, oldattr;
		int err, is_dir = S_ISDIR(inode->i_mode);
		struct iattr ia;

		err = get_user(attr, user_attr);
		if (err)
			return err;

		down(&inode->i_sem);

		if (IS_RDONLY(inode)) {
			err = -EROFS;
			goto up;
		}

		/*
		 * ATTR_VOLUME and ATTR_DIR cannot be changed; this also
		 * prevents the user from turning us into a VFAT
		 * longname entry.  Also, we obviously can't set
		 * any of the NTFS attributes in the high 24 bits.
		 */
		attr &= 0xff & ~(ATTR_VOLUME | ATTR_DIR);
		/* Merge in ATTR_VOLUME and ATTR_DIR */
		attr |= (MSDOS_I(inode)->i_attrs & ATTR_VOLUME) |
			(is_dir ? ATTR_DIR : 0);
		oldattr = fat_attr(inode);

		/* Equivalent to a chmod() */
		ia.ia_valid = ATTR_MODE | ATTR_CTIME;
		if (is_dir) {
			ia.ia_mode = MSDOS_MKMODE(attr,
				S_IRWXUGO & ~sbi->options.fs_dmask)
				| S_IFDIR;
		} else {
			ia.ia_mode = MSDOS_MKMODE(attr,
				(S_IRUGO | S_IWUGO | (inode->i_mode & S_IXUGO))
				& ~sbi->options.fs_fmask)
				| S_IFREG;
		}

		/* The root directory has no attributes */
		if (inode->i_ino == MSDOS_ROOT_INO && attr != ATTR_DIR) {
			err = -EINVAL;
			goto up;
		}

		if (sbi->options.sys_immutable) {
			if ((attr | oldattr) & ATTR_SYS) {
				if (!capable(CAP_LINUX_IMMUTABLE)) {
					err = -EPERM;
					goto up;
				}
			}
		}

		/* This MUST be done before doing anything irreversible... */
		err = notify_change(filp->f_dentry, &ia);
		if (err)
			goto up;

		if (sbi->options.sys_immutable) {
			if (attr & ATTR_SYS)
				inode->i_flags |= S_IMMUTABLE;
			else
				inode->i_flags &= S_IMMUTABLE;
		}

		MSDOS_I(inode)->i_attrs = attr & ATTR_UNUSED;
		mark_inode_dirty(inode);
	up:
		up(&inode->i_sem);
		return err;
	}
	default:
		return -ENOTTY;	/* Inappropriate ioctl for device */
	}
}

static int fat_clean_inode(struct inode *inode)
{
	int ret = 0;

	if (!(inode->i_state & I_DIRTY))
		return 0;
	
	ret = fat_sync_inode(inode);
	if (!ret) {
		pr_debug("Sync inode ok, make it clean!\n");
		mark_inode_clean(inode);
	}

	return ret;
}

static int fat_release(struct inode *inode, struct file *filp)
{
	struct msdos_sb_info *sbi = MSDOS_SB(filp->f_mapping->host->i_sb);
	int ret = fs_flush_sync_fdata(inode, filp);
	
	if (sbi->options.batch_sync && !ret) {
		down(&inode->i_sem);
		ret = fat_clean_inode(inode);
		up(&inode->i_sem);
	}
	return ret;
}

static int fat_release_nosync(struct inode *inode, struct file *filp)
{
	int ret = fs_flush_sync_fdata(inode, filp);

	return ret;
}

static ssize_t fat_sync_write(struct file *filp, const char __user *buf, size_t len, loff_t *ppos)
{
	struct inode *inode = filp->f_mapping->host;
	struct msdos_sb_info *sbi = MSDOS_SB(inode->i_sb);
	ssize_t ret = do_sync_write(filp, buf, len, ppos);
	if (sbi->options.batch_sync && ret > 0) {
		filemap_flush(filp->f_mapping);
	}
	
	return ret;
}

static int fat_fsync(struct file *filp, struct dentry *dentry, int datasync)
{
	struct inode *inode = filp->f_mapping->host;
	struct msdos_sb_info *sbi = MSDOS_SB(inode->i_sb);

	if (sbi->options.batch_sync)
		return fat_clean_inode(inode);
	else
		return file_fsync(filp, dentry, datasync);
}

#ifdef FAT_SOFTUPDATE
/* Copied from mm/filemap.c */
/* Original is  generic_file_aio_write() */
extern ssize_t __generic_file_aio_write_nolock(struct kiocb *iocb, const struct iovec *iov, unsigned long nr_segs, loff_t *ppos);
static ssize_t fat_file_aio_write(struct kiocb *iocb, const char __user *buf,
			       size_t count, loff_t pos)
{
	struct file *file = iocb->ki_filp;
	struct address_space *mapping = file->f_mapping;
	struct inode *inode = mapping->host;
	ssize_t ret;
	struct iovec local_iov = { .iov_base = (void __user *)buf,
					.iov_len = count };

	BUG_ON(iocb->ki_pos != pos);

	down(&inode->i_sem);
	ret = __generic_file_aio_write_nolock(iocb, &local_iov, 1,
						&iocb->ki_pos);
	up(&inode->i_sem);

	if (ret > 0 && ((file->f_flags & O_SYNC) || IS_SYNC(inode))
	    && !(MSDOS_SB(inode->i_sb)->options.relax_sync && MSDOS_SB(inode->i_sb)->options.softupdate_with_sync)) {
                ssize_t err;
			
                err = sync_page_range(inode, mapping, pos, ret);
                if (err < 0)
                        ret = err;
        }
        return ret;
}
#endif /* FAT_SOFTUPDATE */
struct file_operations fat_file_operations = {
	.llseek		= generic_file_llseek,
	.read		= do_sync_read,
	.write		= fat_sync_write,
	.readv		= generic_file_readv,
	.writev		= generic_file_writev,
	.aio_read	= generic_file_aio_read,
#ifdef FAT_SOFTUPDATE
 	.aio_write	= fat_file_aio_write,
#else  /* FAT_SOFTUPDATE */
  	.aio_write	= generic_file_aio_write,
#endif /* FAT_SOFTUPDATE */
	.open           = generic_file_open,
	.release	= fat_release,
	.mmap		= generic_file_mmap,
	.ioctl		= fat_generic_ioctl,
	.fsync		= fat_fsync,
	.sendfile	= generic_file_sendfile,
};

struct file_operations fat_file_operations_nosync = {
	.llseek		= generic_file_llseek,
	.read		= do_sync_read,
	.write		= fat_sync_write,
	.readv		= generic_file_readv,
	.writev		= generic_file_writev,
	.aio_read	= generic_file_aio_read,
#ifdef FAT_SOFTUPDATE
 	.aio_write	= fat_file_aio_write,
#else  /* FAT_SOFTUPDATE */
  	.aio_write	= generic_file_aio_write,
#endif /* FAT_SOFTUPDATE */
	.open           = generic_file_open,
	.release	= fat_release_nosync,
	.mmap		= generic_file_mmap,
	.ioctl		= fat_generic_ioctl,
	.fsync		= fat_fsync,
	.sendfile	= generic_file_sendfile,
};

int fat_notify_change(struct dentry *dentry, struct iattr *attr)
{
	struct msdos_sb_info *sbi = MSDOS_SB(dentry->d_sb);
	struct inode *inode = dentry->d_inode;
 	int need_size_change = 0;
	int mask, error = 0;
	int px_uid = sbi->options.fs_uid ? sbi->options.fs_uid : (u16)-1;
	int px_gid = sbi->options.fs_gid ? sbi->options.fs_gid : (u16)-1;

	lock_kernel();

	/* FAT cannot truncate to a longer file */
	if (attr->ia_valid & ATTR_SIZE) {
		if (attr->ia_size > inode->i_size) {
			error = -EPERM;
			goto out;
		}
	}

	if (sbi->options.posix_attr) {
		if (attr->ia_valid & ATTR_UID) {
			/* root can change uid to any value */
			if (capable(CAP_CHOWN) && 
			    (attr->ia_uid && attr->ia_uid != px_uid))
				attr->ia_uid = px_uid;
			/* change-uid affects gid */
			attr->ia_valid |= ATTR_GID;
			attr->ia_gid = attr->ia_uid ? px_gid : 0;
		} else {
			/* chown syscall sets both uid and gid */
			if (attr->ia_valid & ATTR_GID) {
				/* root can change gid to any value */
				if (capable(CAP_CHOWN) && 
				    (attr->ia_gid && attr->ia_gid != px_gid))
					attr->ia_gid = px_gid;
				/* change-gid affects uid */
				attr->ia_valid |= ATTR_UID;
				attr->ia_uid = attr->ia_gid ? px_uid : 0;
			}
		}
		/* change-group-mode affects on others-mode */
		if (attr->ia_valid & ATTR_MODE) { 
			int others_mode = (attr->ia_mode & S_IRWXG) >> 3;
			attr->ia_mode &=  ~S_IRWXO;
			attr->ia_mode |=  others_mode;
		}
	}

	error = inode_change_ok(inode, attr);
	if (error) {
		goto error_out;
	}

	if (sbi->options.posix_attr) {
		if (((attr->ia_valid & ATTR_UID) &&
		       ((attr->ia_uid) && (attr->ia_uid != px_uid))) ||
		    ((attr->ia_valid & ATTR_GID) &&
		       ((attr->ia_gid) && (attr->ia_gid != px_gid))) ||
		    ((attr->ia_valid & ATTR_MODE) &&
		     (attr->ia_mode & ~VFAT_POSIX_ATTR_VALID_MODE)))
			error = -EPERM;
	} else if (((attr->ia_valid & ATTR_UID) &&
	     (attr->ia_uid != sbi->options.fs_uid)) ||
	    ((attr->ia_valid & ATTR_GID) &&
	     (attr->ia_gid != sbi->options.fs_gid)) ||
	    ((attr->ia_valid & ATTR_MODE) &&
	     (attr->ia_mode & ~MSDOS_VALID_MODE))) {
		error = -EPERM;
	}

	if (error)
		goto error_out;

	if (attr->ia_valid & ATTR_SIZE) {
		/* 
		 * Update i_size here, not in inode_setattr(),
		 * because inode_setattr() calls vmtruncate() and  
		 * fat_truncate() consequently. This may release 
		 * cluster chain, before updating size field of 
		 * the dir entry on disk. 
		 */
		if (IS_SWAPFILE(inode)){
			error = -ETXTBSY;
			goto out; 
		}
		attr->ia_valid &= ~(ATTR_SIZE);
		if (attr->ia_size == inode->i_size) {
			attr->ia_valid |= ATTR_MTIME|ATTR_CTIME;
		} else {
			need_size_change = 1;
			i_size_write(inode, attr->ia_size);
		}
	}
	/*
	 * Here, no error returns from inode_setattr().
	 * Because ATTR_SIZE is cleared before calling it.
	 */
	(void) inode_setattr(inode, attr);

	if (!sbi->options.posix_attr) {
		if (S_ISDIR(inode->i_mode))
			mask = sbi->options.fs_dmask;
		else
			mask = sbi->options.fs_fmask;
		inode->i_mode &= S_IFMT | (S_IRWXUGO & ~mask);
	}

	if (need_size_change) {
		/* inode is already marked dirty in inode_setattr() */
		if (IS_SYNC(inode))
			generic_osync_inode(inode, NULL, OSYNC_INODE);
		error = vmtruncate(inode, inode->i_size);
	}
out:
	unlock_kernel();
	return error;
error_out:
	if (sbi->options.quiet)
		error = 0;
	unlock_kernel();
	return error;
}

EXPORT_SYMBOL(fat_notify_change);

/* Free all clusters after the skip'th cluster. */
static int fat_free(struct inode *inode, int skip)
{
	struct super_block *sb = inode->i_sb;
	int err, wait, free_start, i_start, i_logstart;

	if (MSDOS_I(inode)->i_start == 0)
		return 0;

	/*
	 * Write a new EOF, and get the remaining cluster chain for freeing.
	 */
	wait = IS_DIRSYNC(inode);
	if (skip) {
		struct fat_entry fatent;
		int ret, fclus, dclus;

		ret = fat_get_cluster(inode, skip - 1, &fclus, &dclus);
		if (ret < 0)
			return ret;
		else if (ret == FAT_ENT_EOF)
			return 0;

		fatent_init(&fatent);
		ret = fat_ent_read(inode, &fatent, dclus);
		if (ret == FAT_ENT_EOF) {
			fatent_brelse(&fatent);
			return 0;
		} else if (ret == FAT_ENT_FREE) {
                        printk("[%s]:dclus=%d\n", __FUNCTION__, dclus);
			fat_fs_panic(sb,
				     "%s: invalid cluster chain (i_pos %lld)",
				     __FUNCTION__, MSDOS_I(inode)->i_pos);
			ret = -EIO;
		} else if (ret > 0) {
			err = fat_ent_write(inode, &fatent, FAT_ENT_EOF, wait);
			if (err)
				ret = err;
		}
		fatent_brelse(&fatent);
		if (ret < 0)
			return ret;

		free_start = ret;
		i_start = i_logstart = 0;
		fat_cache_inval_inode(inode);
	} else {
		fat_cache_inval_inode(inode);

		i_start = free_start = MSDOS_I(inode)->i_start;
		i_logstart = MSDOS_I(inode)->i_logstart;
		MSDOS_I(inode)->i_start = 0;
		MSDOS_I(inode)->i_logstart = 0;
	}
	MSDOS_I(inode)->i_attrs |= ATTR_ARCH;
	inode->i_ctime = inode->i_mtime = CURRENT_TIME_SEC;
	if (wait) {
		err = fat_sync_inode(inode);
		if (err)
			goto error;
	} else
		mark_inode_dirty(inode);
	inode->i_blocks = skip << (MSDOS_SB(sb)->cluster_bits - 9);

	/* Freeing the remained cluster chain */
	return fat_free_clusters(inode, free_start);

error:
	if (i_start) {
		MSDOS_I(inode)->i_start = i_start;
		MSDOS_I(inode)->i_logstart = i_logstart;
	}
	return err;
}

void fat_truncate(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	const unsigned int cluster_size = sbi->cluster_size;
	int nr_clusters;

	/*
	 * This protects against truncating a file bigger than it was then
	 * trying to write into the hole.
	 */
	if (MSDOS_I(inode)->mmu_private > inode->i_size)
		MSDOS_I(inode)->mmu_private = inode->i_size;

	nr_clusters = (inode->i_size + (cluster_size - 1)) >> sbi->cluster_bits;

	/* wirte inode first to update file size*/
	fat_sync_inode(inode);
	lock_kernel();
	fat_free(inode, nr_clusters);
	unlock_kernel();
	fat_sync_inode(inode);
	fs_mark_flush(sb);
}

struct inode_operations fat_file_inode_operations = {
	.truncate	= fat_truncate,
	.setattr	= fat_notify_change,
};

struct inode_operations fat_symlink_inode_operations = {
       .readlink       = page_readlink,
       .follow_link    = page_follow_link_light,
       .setattr        = fat_notify_change,
};

EXPORT_SYMBOL(fat_symlink_inode_operations);
