/*
 *  linux/fs/fat/misc.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *  22/11/2000 - Fixed fat_date_unix2dos for dates earlier than 01/01/1980
 *		 and date_dos2unix for date==0 by Igor Zhbanov(bsg@uniyar.ac.ru)
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include <linux/buffer_head.h>

/*
 * fat_fs_panic reports a severe file system problem and sets the file system
 * read-only. The file system can be made writable again by remounting it.
 */
void fat_fs_panic(struct super_block *s, const char *fmt, ...)
{
	va_list args;

	printk(KERN_ERR "FAT: Filesystem panic (dev %s)\n", s->s_id);

	WARN_ON(1);
	
	printk(KERN_ERR "    ");
	va_start(args, fmt);
	vprintk(fmt, args);
	va_end(args);
	printk("\n");

	if (!(s->s_flags & MS_RDONLY)) {
		s->s_flags |= MS_RDONLY;
		printk(KERN_ERR "    File system has been set read-only\n");
	}
}

EXPORT_SYMBOL(fat_fs_panic);

/* Flushes the number of free clusters on FAT32 */
/* XXX: Need to write one per FSINFO block.  Currently only writes 1 */
void fat_clusters_flush(struct super_block *sb)
{
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	struct buffer_head *bh;
	struct fat_boot_fsinfo *fsinfo;

	if (sbi->fat_bits != 32)
		return;

	bh = sb_bread(sb, sbi->fsinfo_sector);
	if (bh == NULL) {
		printk(KERN_ERR "FAT: bread failed in fat_clusters_flush\n");
		return;
	}

	fsinfo = (struct fat_boot_fsinfo *)bh->b_data;
	/* Sanity check */
	if (!IS_FSINFO(fsinfo)) {
		printk(KERN_ERR "FAT: Did not find valid FSINFO signature.\n"
		       "     Found signature1 0x%08x signature2 0x%08x"
		       " (sector = %lu)\n",
		       le32_to_cpu(fsinfo->signature1),
		       le32_to_cpu(fsinfo->signature2),
		       sbi->fsinfo_sector);
	} else {
		if (sbi->free_clusters != -1)
			fsinfo->free_clusters = cpu_to_le32(sbi->free_clusters);
		if (sbi->prev_free != -1)
			fsinfo->next_cluster = cpu_to_le32(sbi->prev_free);
		mark_buffer_dirty(bh);
		if (sb->s_flags & MS_SYNCHRONOUS)
			if (!sbi->options.relax_sync)
				sync_dirty_buffer(bh);
	}
	brelse(bh);
}

/*
 * fat_chain_add() adds a new cluster to the chain of clusters represented
 * by inode.
 */
int fat_chain_add(struct inode *inode, int new_dclus, int nr_cluster)
{
	struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int ret, new_fclus, last;

	/*
	 * We must locate the last cluster of the file to add this new
	 * one (new_dclus) to the end of the link list (the FAT).
	 */
	last = new_fclus = 0;
	if (MSDOS_I(inode)->i_start) {
		int fclus, dclus;

		ret = fat_get_cluster(inode, FAT_ENT_EOF, &fclus, &dclus);
		if (ret < 0)
			return ret;
		new_fclus = fclus + 1;
		last = dclus;
	}

	/* add new one to the last of the cluster chain */
	if (last) {
		/*the FAT will be updated until right before write_inode*/
		if (sbi->options.batch_sync &&
		    !MSDOS_I(inode)->i_new &&
		    !MSDOS_I(inode)->i_last_dclus) {
			MSDOS_I(inode)->i_last_dclus = last;
			MSDOS_I(inode)->i_new_dclus = new_dclus;

			pr_debug("%s: %d %d\n", __func__, last, new_dclus);
		}
		else {
			struct fat_entry fatent;
			
			fatent_init(&fatent);
			ret = fat_ent_read(inode, &fatent, last);
			if (ret >= 0) {
				int wait = inode_needs_sync(inode);
#ifdef FAT_SOFTUPDATE
                                if(sbi->options.softupdate_with_sync && sbi->options.relax_sync){
                                        struct fat_entry fatent_next;
                                        int next_new_cls;
                                        wait = 0;
                                        fatent_init(&fatent_next);
                                        next_new_cls = fat_ent_read(inode, &fatent_next, new_dclus);
                                        if(next_new_cls >= 0){
                                                if(fatent_next.bhs[0] != fatent.bhs[0]){
                                                        DPRINT("[%s]:call fat_softupdate_make_dependency\n", __FUNCTION__);
                                                        DTRACE("[%s]:Make UD ta-bh=0x%p, dep-bh=0x%p\n", __FUNCTION__, 
                                                               fatent.bhs[0], fatent_next.bhs[0]);
                                                        ret = fat_softupdate_make_dependency(inode, &fatent, &fatent_next);
                                                        if(ret) printk("Cannot make dependency\n");
                                                }
                                        }
                                }
                                DPRINT("[%s]:link, last=%d -> new_dclus=%d\n", __FUNCTION__, last, new_dclus);
#endif /* FAT_SOFTUPDATE */
				ret = fat_ent_write(inode, &fatent, new_dclus, wait);
				fatent_brelse(&fatent);
			}
			if (ret < 0)
				return ret;
		}
//		fat_cache_add(inode, new_fclus, new_dclus);
	} else {
                DTRACE("[%s]:new link, new_dclus=%d\n", __FUNCTION__, new_dclus);
		MSDOS_I(inode)->i_start = new_dclus;
		MSDOS_I(inode)->i_logstart = new_dclus;
		MSDOS_I(inode)->i_new = 1;
#ifdef FAT_SOFTUPDATE
                if(sbi->options.softupdate_with_sync && sbi->options.relax_sync){
                        struct fat_entry fatent_next;
                        int next_new_cls;
                        fatent_init(&fatent_next);
                        next_new_cls = fat_ent_read(inode, &fatent_next, new_dclus);
                        if(!S_ISDIR(inode->i_mode) && 
                           (next_new_cls > 0) && 
                           buffer_dirty(fatent_next.bhs[0])){ /* include (next_new_cls != FAT_ENT_FREE(=0))  */
                                DPRINT("[%s]:call fat_softupdate_make_entry_UD\n", __FUNCTION__);
                                DTRACE("[%s]:Make Entry UD First dclus=%d\n", __FUNCTION__, new_dclus);
                                ret = fat_softupdate_make_entry_UD(inode, fatent_next.bhs[0], new_dclus);
                                if(ret) printk("Cannot make entry UD\n");
                        }
                }
#endif /* FAT_SOFTUPDATE */
		/*
		 * Since generic_osync_inode() synchronize later if
		 * this is not directory, we don't here.
		 */
		if (S_ISDIR(inode->i_mode) && IS_DIRSYNC(inode)) {
			ret = fat_sync_inode(inode);
                        DTRACE("[%s]:ret(%d)=fat_sync_inode()\n", __FUNCTION__, ret);
			if (ret)
				return ret;
		} else
			mark_inode_dirty(inode);
	}
	if (new_fclus != (inode->i_blocks >> (sbi->cluster_bits - 9))) {
		fat_fs_panic(sb, "clusters badly computed (%d != %lu)",
			new_fclus, inode->i_blocks >> (sbi->cluster_bits - 9));
		fat_cache_inval_inode(inode);
	}
	inode->i_blocks += nr_cluster << (sbi->cluster_bits - 9);

	return 0;
}

extern struct timezone sys_tz;

/* Linear day numbers of the respective 1sts in non-leap years. */
static int day_n[] = {
   /* Jan  Feb  Mar  Apr   May  Jun  Jul  Aug  Sep  Oct  Nov  Dec */
	0,  31,  59,  90,  120, 151, 181, 212, 243, 273, 304, 334, 0, 0, 0, 0
};

/* Convert a MS-DOS time/date pair to a UNIX date (seconds since 1 1 70). */
time_t date_dos2unix(unsigned short time, unsigned short date)
{
	int month, year;
	time_t secs;

	/*
	 * first subtract and mask after that... Otherwise, if
	 * date == 0, bad things happen
	 */
	month = ((date >> 5) - 1) & 15;
	year = date >> 9;
	secs = (time & 31)*2+60*((time >> 5) & 63)+(time >> 11)*3600+86400*
	    ((date & 31)-1+day_n[month]+(year/4)+year*365-((year & 3) == 0 &&
	    month < 2 ? 1 : 0)+3653);
			/* days since 1.1.70 plus 80's leap day */
	secs += sys_tz.tz_minuteswest*60;
	if (secs < 0)
		secs = LONG_MAX;
	return secs;
}

/* Convert linear UNIX date to a MS-DOS time/date pair. */
void fat_date_unix2dos(time_t unix_date, __le16 *time, __le16 *date)
{
	int day, year, nl_day, month;

	unix_date -= sys_tz.tz_minuteswest*60;

	/* Jan 1 GMT 00:00:00 1980. But what about another time zone? */
	if (unix_date < 315532800)
		unix_date = 315532800;

	*time = cpu_to_le16((unix_date % 60)/2+(((unix_date/60) % 60) << 5)+
	    (((unix_date/3600) % 24) << 11));
	day = unix_date/86400-3652;
	year = day/365;
	if ((year+3)/4+365*year > day)
		year--;
	day -= (year+3)/4+365*year;
	if (day == 59 && !(year & 3)) {
		nl_day = day;
		month = 2;
	} else {
		nl_day = (year & 3) || day <= 59 ? day : day-1;
		for (month = 0; month < 12; month++) {
			if (day_n[month] > nl_day)
				break;
		}
	}
	*date = cpu_to_le16(nl_day-day_n[month-1]+1+(month << 5)+(year << 9));
}

EXPORT_SYMBOL(fat_date_unix2dos);

int fat_sync_bhs(struct buffer_head **bhs, int nr_bhs)
{
	int i, e, err = 0;

	for (i = 0; i < nr_bhs; i++) {
		lock_buffer(bhs[i]);
		if (test_clear_buffer_dirty(bhs[i])) {
			get_bh(bhs[i]);
			bhs[i]->b_end_io = end_buffer_write_sync;
			e = submit_bh(WRITE, bhs[i]);
			if (!err && e)
				err = e;
		} else
			unlock_buffer(bhs[i]);
	}
	for (i = 0; i < nr_bhs; i++) {
		wait_on_buffer(bhs[i]);
		if (buffer_eopnotsupp(bhs[i])) {
			clear_buffer_eopnotsupp(bhs[i]);
			err = -EOPNOTSUPP;
		} else if (!err && !buffer_uptodate(bhs[i]))
			err = -EIO;
	}
	return err;
}

int fat_flush_bhs(struct buffer_head **bhs, int nr_bhs)
{
	int i, e, err = 0;

	for (i = 0; i < nr_bhs; i++) {
		lock_buffer(bhs[i]);
		if (test_clear_buffer_dirty(bhs[i])) {
			get_bh(bhs[i]);
			bhs[i]->b_end_io = end_buffer_write_sync;
			e = submit_bh(WRITE, bhs[i]);
			if (!err && e)
				err = e;
		} else
			unlock_buffer(bhs[i]);

	}
	return err;
}

EXPORT_SYMBOL(fat_sync_bhs);
EXPORT_SYMBOL_GPL(fat_flush_bhs);
#ifdef FAT_SOFTUPDATE
static int init_fat_softupdate_ud_head(struct super_block *sb, struct buffer_head *bh)
{
	struct fat_softupdate_ud_head *p_ud = &((struct fat_bh_private_data *)bh->b_private)->ud;
	
        memset(bh->b_private, 0, sizeof(struct fat_bh_private_data));
	p_ud->sb = sb;
	list_add(&p_ud->d_sb_head, &(MSDOS_SB(sb)->d_ud_head));
	INIT_LIST_HEAD(&p_ud->d_deppriv_head);
	INIT_LIST_HEAD(&p_ud->d_target_head);
	INIT_LIST_HEAD(&p_ud->d_depend_head);
	p_ud->p_bh = bh;
	DTRACE("[%s]:get_bh(), bh=0x%p\n", __FUNCTION__, bh);
	get_bh(bh);
	
	return 0;
}
#define PRIVATE_DATA_FOR_DEPEND 1
#define PRIVATE_DATA_FOR_TARGET 0
static int fat_softupdate_check_alloc_private(struct inode *inode, struct buffer_head *bh, int make_depend_private)
{
	if(!bh->b_private){
		bh->b_private = kmalloc(sizeof(struct fat_bh_private_data), GFP_KERNEL);
		if(!bh->b_private) {
                        printk("Cannot alloc mem\n");
                        return -ENOMEM;
                }
		init_fat_softupdate_ud_head(inode->i_sb, bh);
		DTRACE("[%s]:Made b_private, bh(0x%p)->bprivate=0x%p\n", __FUNCTION__, bh, bh->b_private);
		DTRACE("[%s]:Made t private, bh(0x%p)->bprivate=0x%p\n", __FUNCTION__, bh, bh->b_private);
	}
        if(make_depend_private){
                struct fat_softupdate_depend_private *dep_pri;
                unsigned long flags;
                dep_pri = kmalloc(sizeof(struct fat_softupdate_depend_private), GFP_KERNEL);
                dep_pri->inode = inode;
                list_add(&dep_pri->l_sb, &(MSDOS_SB(inode->i_sb)->d_depend_private_head));
                spin_lock_irqsave(&MSDOS_I(dep_pri->inode)->softupdate_ud_lock, flags);
                list_add(&dep_pri->l_inode, &(MSDOS_I(inode)->i_depend_private));
                spin_unlock_irqrestore(&MSDOS_I(dep_pri->inode)->softupdate_ud_lock, flags);
                list_add(&dep_pri->l_private, &(((struct fat_bh_private_data *)bh->b_private)->ud.d_deppriv_head));
		DTRACE("[%s]:Make depend private(0x%p) list\n", __FUNCTION__, dep_pri);
        }
        return 0;
}
static int fat_softupdate_remove_all_depend_private(struct list_head *lh_private_head)
{
        struct list_head *p_lh = lh_private_head;
        struct fat_softupdate_depend_private *p_dep;
        unsigned long flags;
	/* remove depend ud list */
        DTRACE("[%s]:Start\n", __FUNCTION__);
	while(!list_empty(p_lh)){
                if(p_lh->next == lh_private_head){
                        DTRACE("[%s]:end of depend private list\n", __FUNCTION__);
                        break;
                }
                p_dep = list_entry(p_lh->next, struct fat_softupdate_depend_private, l_private);
		DTRACE("[%s]:Free depend private(0x%p) list\n", __FUNCTION__, p_dep);
                list_del(&p_dep->l_sb);
                spin_lock_irqsave(&MSDOS_I(p_dep->inode)->softupdate_ud_lock, flags);
                list_del(&p_dep->l_inode);
                spin_unlock_irqrestore(&MSDOS_I(p_dep->inode)->softupdate_ud_lock, flags);
                list_del(&p_dep->l_private);
                kfree(p_dep);
        }
	
        return 0;
}
int fat_softupdate_test_free_private(struct buffer_head *bh)
{
	struct fat_softupdate_ud_head *p_ud;
        void *p_private;
	/* Check target ud list exist */
        DTRACE("[%s]:remove private=0x%p\n", __FUNCTION__, bh);
        if(bh == NULL) {
                DTRACE("[%s]:bh=NULL\n",  __FUNCTION__);
                return 0;
        }
        DTRACE("[%s]:remove private, bh(0x%p)->bprivate=0x%p\n", __FUNCTION__, bh, bh->b_private);
        p_private = bh->b_private;
	if(p_private){
		p_ud = &((struct fat_bh_private_data *)p_private)->ud;
		if(list_empty(&p_ud->d_target_head) && list_empty(&p_ud->d_depend_head)){
			/* Free private mem */
			bh->b_private = NULL;
			DPRINT("[%s]:Free private mem\n", __FUNCTION__);
			DTRACE("[%s]:Free private, bh->b_private=0x%p\n", __FUNCTION__, p_private);
			DTRACE("[%s]:put_bh(), bh=0x%p\n", __FUNCTION__, bh);
			list_del(&p_ud->d_sb_head);
                        fat_softupdate_remove_all_depend_private(&p_ud->d_deppriv_head);
			kfree(p_private);
			put_bh(bh);
			DTRACE("[%s]:Free private, bh(0x%p)->b_private=0x%p\n", __FUNCTION__, bh, p_private);
		}
	}
	return 0;
}
int fat_softupdate_del_list_free_mem(struct buffer_head *bh, struct fat_softupdate_ud *p_clear_ud)
{
        DTRACE("[%s]:remove p_clear_ud=0x%p\n", __FUNCTION__, p_clear_ud);
        if(p_clear_ud->flag_in_replace == 1){
                DTRACE("This UD(0x%p) is in replace, so cannot rm\n", p_clear_ud);
                p_clear_ud->flag_ud_remove = 1;
                return -1;
        }
        if(p_clear_ud->inode){
                unsigned long flags;
               spin_lock_irqsave(&MSDOS_I(p_clear_ud->inode)->softupdate_ud_lock, flags);
                if(MSDOS_I(p_clear_ud->inode)->i_ud == p_clear_ud){
                        MSDOS_I(p_clear_ud->inode)->i_ud = NULL;
                        DTRACE("[%s]:clear inode(0x%p)->i_ud(0x00)\n", __FUNCTION__, p_clear_ud->inode);
                        DTRACE("[%s]:inode->i_ud cleared, ud=0x%p\n", __FUNCTION__, p_clear_ud);
                }
                if(MSDOS_I(p_clear_ud->inode)->i_ud_size == p_clear_ud){
                        MSDOS_I(p_clear_ud->inode)->i_ud_size = NULL;
                        DTRACE("[%s]:inode->i_ud_size cleared, ud=0x%p\n", __FUNCTION__, p_clear_ud);
                }
                spin_unlock_irqrestore(&MSDOS_I(p_clear_ud->inode)->softupdate_ud_lock, flags);
        }
	/* Find dependency */
	list_del(&p_clear_ud->d_sb); /* del sb list */
        if(p_clear_ud->d_depend.next != NULL)
                list_del(&p_clear_ud->d_depend);
        if(p_clear_ud->d_target.next != NULL)
                list_del(&p_clear_ud->d_target);
	kfree(p_clear_ud);
  
	DTRACE("[%s]:Remove UD=0x%p\n", __FUNCTION__, p_clear_ud);

	/* Check target ud list exist */
        if(bh) fat_softupdate_test_free_private(bh);

	return 0;
}
/**
 * fat_softupdate_rm_dependency  -- remove one dependency
 * @target_lh:	the list head which be removed ud
 * @rm_entry_target: the target entry number which be removed ud
 * @rm_entry_dep:    the depend entry number which be removed ud
 *
 * Remove one UD from target link list
 */
int fat_softupdate_rm_dependency(struct buffer_head *target_bh, int rm_entry_target, int rm_entry_dep)
{
	struct list_head *target_lh;
	struct list_head *p_lh = NULL;
	struct fat_softupdate_ud *p_clear_ud;
	
	DPRINT("[%s]:Start\n", __FUNCTION__);
	DPRINT("[%s]:target_bh=0x%p\n", __FUNCTION__, target_bh);
	/* Check target list head */
	if(target_bh->b_private == NULL){
		DPRINT("[%s]:No UD link in this bh\n", __FUNCTION__);
		return 0;
	}
	target_lh = &((struct fat_bh_private_data *)(target_bh->b_private))->ud.d_target_head;
	if(list_empty(target_lh)) {
		DTRACE("No Link UD list!!\n");
		return -EINVAL;
	}
	
	p_lh = target_lh->next;
	while(p_lh != target_lh){
		p_clear_ud = list_entry(p_lh, struct fat_softupdate_ud, d_target);
		DPRINT("rm_entry_target=0x%x, p_clear_ud->entry_target=0x%x\n", rm_entry_target, p_clear_ud->entry_target);
		DPRINT("rm_entry_dep=0x%x, p_clear_ud->entry_depend=0x%x\n", rm_entry_dep, p_clear_ud->entry_depend);
		if((rm_entry_target == p_clear_ud->entry_target) && 
		   (rm_entry_dep == p_clear_ud->entry_depend)){
			/* Find dependency */
			/* delete UD list */
			if(p_clear_ud->fat_ent_ud.bhs[0] != target_bh) BUG_ON(1);
			if(fat_softupdate_del_list_free_mem(target_bh, p_clear_ud)){
                                printk("Maybe UD remain\n");
                        }
                        return 0;
		}
		p_lh = p_lh->next;
	}
	
	printk("No dependency!!\n");
	return -EINVAL;
}
/**
 * fat_softupdate_test_rm_dependency  -- test and remove dependency
 * @bh:	buffer head which want to remove UD depend list  
 * @rm_entry_dep: dependency entry which be removed
 *
 * Remove UD which link to depend (Not target)
 */
int fat_softupdate_test_rm_dependency(struct buffer_head *bh, int rm_entry_dep)
{
	struct fat_softupdate_ud *p_ud;
	struct list_head *p_lh, *p_lh_start;
	DTRACE("[%s]:Start\n", __FUNCTION__);
	if(bh->b_private){
		p_lh = p_lh_start = &((struct fat_bh_private_data *)bh->b_private)->ud.d_depend_head;
		do{
			p_lh = p_lh->next;
			p_ud = list_entry(p_lh, struct fat_softupdate_ud, d_depend);
			DPRINT("[%s]:Searching\n", __FUNCTION__);
			if(p_ud->entry_depend == rm_entry_dep){
				DTRACE("[%s]:Find Remove UD List, dep_bh=0x%p, entry=%d\n", __FUNCTION__, 
                                       bh, rm_entry_dep);
				fat_softupdate_del_list_free_mem(p_ud->fat_ent_ud.bhs[0], p_ud);
				break;
			}
		}while(p_lh != p_lh_start);
	}
#if 1 /* Only called by fat_free_clusters() */
	/* Check target ud list exist */
	fat_softupdate_test_free_private(bh);
#endif
	return 0;
}
/**
 * fat_softupdate_rm_all_dependency  -- remove all dependency
 * @bh:	buffer head which want to remove all UD depend list  
 *
 * Remove all UD which link to depend (Not target)
 */
int fat_softupdate_rm_all_dependency(struct buffer_head *bh)
{
	struct list_head *p_lh, *p_lh_start;
	struct fat_softupdate_ud *p_clear_ud;
	
	DPRINT("[%s]:Start\n", __FUNCTION__);
        DPRINT("[%s]:Is buffer locked(0x%p)=%d\n", __FUNCTION__, bh, buffer_locked(bh));
        if(bh->b_private == NULL){
                DTRACE("[%s]:No private data\n", __FUNCTION__);
                return 0;
        }
	/* remove depend ud list */
	p_lh = p_lh_start = &((struct fat_bh_private_data *)bh->b_private)->ud.d_depend_head;
	while(!list_empty(p_lh)){
                if(p_lh->next == p_lh_start){
                        DTRACE("[%s]:end of target UD list\n", __FUNCTION__);
                        break;
                }
		p_clear_ud = list_entry(p_lh->next, struct fat_softupdate_ud, d_depend);
		if(fat_softupdate_del_list_free_mem(/* p_clear_ud->fat_ent_ud.bhs[0], */NULL, p_clear_ud)){
                        p_lh = p_lh->next;
			if(p_lh == p_lh_start){
				printk("****Caution!!, UD remain*****\n");
				return 0;
			}
                }else{
                        DPRINT("[%s]:Free UD(0x%p) list\n", __FUNCTION__, p_clear_ud);
                }
        }
	
#if 0
	/* Check target ud list exist */
	fat_softupdate_test_free_private(bh);
#endif	
	return 0;
}
/**
 * fat_softupdate_add_dependency  -- add dependency
 * @bh_target:	  buffer head which want to add target list 
 * @entry_target: entry number which want to add target list
 * @bh_depend:    buffer head which want to add depend list
 * @entry_depend: entry number which want to add depend list
 *
 * Add one UD to target link list, and depend link list
 */
int fat_softupdate_add_dependency(struct inode *inode, 
				  struct buffer_head *bh_target, int entry_target, 
				  struct buffer_head *bh_depend, int entry_depend)
{
	struct fat_softupdate_ud *p_ud;
        struct super_block *sb = inode->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	
	DPRINT("[%s]:Start\n", __FUNCTION__);
	DTRACE("[%s]:sb=0x%p, bh_target=0x%p, entry_target=0x%p\n", __FUNCTION__, sb, bh_target, entry_target);
	DTRACE("[%s]:         bh_depend=0x%p, entry_depend=0x%p\n", __FUNCTION__, bh_depend, entry_depend);
	/* bh private initialize, if needed */
        if(fat_softupdate_check_alloc_private(inode, bh_target, PRIVATE_DATA_FOR_TARGET)){
                printk("[%s]:Cannot alloc mem for bh_target\n", __FUNCTION__);
                return -ENOMEM;
        }
        if(fat_softupdate_check_alloc_private(inode, bh_depend, PRIVATE_DATA_FOR_DEPEND)){
                printk("[%s]:Cannot alloc mem for bh_depend\n", __FUNCTION__);
                return -ENOMEM;
        }
	/* alloc mem */
	p_ud = kmalloc(sizeof(struct fat_softupdate_ud), GFP_KERNEL);
	if(!p_ud){
		return -ENOMEM;
	}
        memset(p_ud, 0, sizeof(struct fat_softupdate_ud));
	/* set data */
	p_ud->flag_in_replace = 0;
	p_ud->entry_target = entry_target;
	p_ud->entry_depend = entry_depend;
	p_ud->fat_ent_ud.bhs[0] = bh_target;
        p_ud->ud_disc = FAT_SU_UD_DISC_CLSLINK;
	
	/* set link */
	list_add(&p_ud->d_target, &((struct fat_bh_private_data *)bh_target->b_private)->ud.d_target_head);
	list_add(&p_ud->d_depend, &((struct fat_bh_private_data *)bh_depend->b_private)->ud.d_depend_head);
	list_add(&p_ud->d_sb, &sbi->d_ud);
	DTRACE("[%s]:Made UD List, UD=0x%p\n", __FUNCTION__, p_ud);

	return 0;
}

int fat_softupdate_make_entry_UD(struct inode *inode, struct buffer_head *depend_bh, int depend_entry)
{
	struct fat_softupdate_ud *p_ud;
	unsigned long flags;

        DTRACE("[%s]:Start\n", __FUNCTION__);

        if(depend_bh == NULL){
                printk("[%s]:ERR, bh=NULL\n", __FUNCTION__);
                return -1;
        }
        /* Make depend private */
        if(fat_softupdate_check_alloc_private(inode, depend_bh, PRIVATE_DATA_FOR_DEPEND)){
                printk("[%s]:Cannot alloc mem for depend_bh\n", __FUNCTION__);
                return -ENOMEM;
        }
        /* Make dirent UD */
        p_ud = kmalloc(sizeof(struct fat_softupdate_ud), GFP_KERNEL);
        if(!p_ud){
                printk("[%s]:ERR, Cannot alloc mem\n", __FUNCTION__);
                return -ENOMEM;
        }
        memset(p_ud, 0, sizeof(struct fat_softupdate_ud));

        /* Set data */
        DPRINT("[%s]:depend_entry=%d, i_logstart=%d, Make entUD\n", __FUNCTION__, depend_entry, MSDOS_I(inode)->i_logstart);
	p_ud->flag_in_replace = 0;
        p_ud->size = 0;
        p_ud->ud_disc = FAT_SU_UD_DISC_DIRENT;
	p_ud->entry_target = 0;
	p_ud->entry_depend = MSDOS_I(inode)->i_logstart;
	p_ud->fat_ent_ud.bhs[0] = NULL;
        p_ud->inode = inode;
	/* set link */
//	list_add(&p_ud->d_target, &((struct fat_bh_private_data *)bh->b_private)->ud.d_target_head);
	list_add(&p_ud->d_depend, &((struct fat_bh_private_data *)depend_bh->b_private)->ud.d_depend_head);
	list_add(&p_ud->d_sb, &MSDOS_SB(inode->i_sb)->d_ud);
        /* set inode */
        spin_lock_irqsave(&MSDOS_I(inode)->softupdate_ud_lock, flags);
        MSDOS_I(inode)->i_ud = p_ud;
        spin_unlock_irqrestore(&MSDOS_I(inode)->softupdate_ud_lock, flags);
        DPRINT("[%s]:Set inode(0x%p)->i_ud(0x%p)\n", __FUNCTION__, inode, p_ud);
	DTRACE("[%s]:Made Entry UD List, UD=0x%p\n", __FUNCTION__, p_ud);
	DTRACE("[%s]:Made Entry UD List, bh(0x%p)->b_private=0x%p\n", __FUNCTION__, depend_bh, depend_bh->b_private);
        return 0;
}
int fat_softupdate_list_entry_dependency(struct buffer_head *dirent_bh, struct inode *inode, struct msdos_dir_entry *raw_entry, int *wait)
{
	struct fat_softupdate_ud *p_ud = MSDOS_I(inode)->i_ud;
	struct fat_softupdate_ud *p_ud_size = MSDOS_I(inode)->i_ud_size;

	DTRACE("[%s]:Start\n", __FUNCTION__);
        DPRINT("[%s]:inode(0x%p)->i_ud(0x%p)\n", __FUNCTION__, inode, p_ud);

        *wait = 0;
        if(S_ISDIR(inode->i_mode) || 
           raw_entry->size > cpu_to_le32(inode->i_size) || 
           (cpu_to_le16(MSDOS_I(inode)->i_logstart) == 0)){
                /* File truncate */
                DTRACE("[%s]:rm entry UD and set wait=1\n", __FUNCTION__);
                *wait = 1;
                if(p_ud){
                        DTRACE("[%s]:remove entry UD list\n", __FUNCTION__);
                        DTRACE("[%s]:S_ISDIR(inode->i_mode)=%d\n", __FUNCTION__, S_ISDIR(inode->i_mode));
                        DTRACE("[%s]:raw_entry->size=%d >? cpu_to_le32(inode->i_size)=%d\n", __FUNCTION__,
                               (int)raw_entry->size, (int)cpu_to_le32(inode->i_size));
                        DTRACE("[%s]:(cpu_to_le16(MSDOS_I(inode)->i_logstart)=%d\n", __FUNCTION__, 
                               (int)(cpu_to_le16(MSDOS_I(inode)->i_logstart)));
                        fat_softupdate_del_list_free_mem(dirent_bh, p_ud);
                }
                if(p_ud_size){
                        DTRACE("[%s]:remove entry UD(0x%p) list for size\n", __FUNCTION__, p_ud_size);
                        fat_softupdate_del_list_free_mem(dirent_bh, p_ud_size);
                }
                return 0;
        }
        if(cpu_to_le32(inode->i_size) == 0){
                *wait = 1;
                if(p_ud_size){
                        DTRACE("[%s]:remove entry UD(0x%p) list for size, i_size=0\n", __FUNCTION__, p_ud_size);
                        fat_softupdate_del_list_free_mem(dirent_bh, p_ud_size);
                }
                if(p_ud){
                        printk("[%s]:entry UD(0x%p) alive, but write inode now\n", __FUNCTION__, p_ud);
                }
                return 0;
        }
        if(p_ud){
                if(p_ud->entry_depend != MSDOS_I(inode)->i_logstart){
                        printk("[%s]:Caution!!, Wrong entry UD list\n", __FUNCTION__);
                        fat_softupdate_del_list_free_mem(dirent_bh, p_ud);
                }else if(p_ud->fat_ent_ud.bhs[0] == NULL){
                        /* Lock this UD */
                        p_ud->flag_in_replace = 1;

                        /* Make dirent dependency */
                        if(fat_softupdate_check_alloc_private(inode, dirent_bh, PRIVATE_DATA_FOR_TARGET)){
                                printk("[%s]:Cannot alloc mem for depend_bh\n", __FUNCTION__);
                                return -ENOMEM;
                        }
                        /* Set data */
                        p_ud->ud_disc = FAT_SU_UD_DISC_DIRENT;
                        p_ud->entry_target = 0;
                        p_ud->fat_ent_ud.bhs[0] = dirent_bh;
                        p_ud->i_pos_and_mask = (MSDOS_I(inode)->i_pos & (loff_t)((MSDOS_SB(inode->i_sb)->dir_per_block) - 1));
                        /* set link */
                        if((p_ud->d_target.next == NULL) && (p_ud->d_target.prev == NULL)){
                                DTRACE("[%s]:link to dirent private \n", __FUNCTION__);
                                list_add(&p_ud->d_target, &((struct fat_bh_private_data *)dirent_bh->b_private)->ud.d_target_head);
                        }
                        /* Unlock this UD */
                        p_ud->flag_in_replace = 0;
                        if(p_ud->flag_ud_remove){
                                DTRACE("[%s]:remove entry UD(0x%p) list because remove this UD while set \n", __FUNCTION__, p_ud);
                                fat_softupdate_del_list_free_mem(dirent_bh, p_ud);
                        }
                        DTRACE("[%s]:List Entry UD List, UD=0x%p\n", __FUNCTION__, p_ud);
                }
        }
        if (!S_ISDIR(inode->i_mode)){
                if(!p_ud_size){
                        unsigned long flags;
                        /* Make dirent size UD */
                        p_ud_size = kmalloc(sizeof(struct fat_softupdate_ud), GFP_KERNEL);
                        if(!p_ud_size){
                                printk("[%s]:ERR, Cannot alloc mem size UD\n", __FUNCTION__);
                                return -ENOMEM;
                        }
                        memset(p_ud_size, 0, sizeof(struct fat_softupdate_ud));
                        /* Set data */
                        p_ud_size->flag_in_replace = 0;
                        p_ud_size->i_pos_and_mask = (MSDOS_I(inode)->i_pos & (loff_t)((MSDOS_SB(inode->i_sb)->dir_per_block) - 1));
                        p_ud_size->size = inode->i_size;
                        DPRINT("[%s]:Set size =%d\n", __FUNCTION__, (int)cpu_to_le32(inode->i_size));
                        p_ud_size->ud_disc = FAT_SU_UD_DISC_DIRENT_SIZE;
                        p_ud_size->entry_target = 0;
                        p_ud_size->entry_depend = 0;
                        p_ud_size->fat_ent_ud.bhs[0] = dirent_bh;
                        p_ud_size->inode = inode;
                        
                        /* Make dirent dependency */
                        if(fat_softupdate_check_alloc_private(inode, dirent_bh, PRIVATE_DATA_FOR_TARGET)){
                                printk("[%s]:Cannot alloc mem for depend_bh\n", __FUNCTION__);
                                return -ENOMEM;
                        }
                        /* set link */
                        list_add(&p_ud_size->d_target, &((struct fat_bh_private_data *)dirent_bh->b_private)->ud.d_target_head);
                        list_add(&p_ud_size->d_sb, &MSDOS_SB(inode->i_sb)->d_ud);
                        /* set inode */
                        spin_lock_irqsave(&MSDOS_I(inode)->softupdate_ud_lock, flags);
                        MSDOS_I(inode)->i_ud_size = p_ud_size;
                        spin_unlock_irqrestore(&MSDOS_I(inode)->softupdate_ud_lock, flags);
                        DTRACE("[%s]:Made Entry UD List for size, UD=0x%p\n", __FUNCTION__, p_ud_size);
                        DTRACE("[%s]:Made Entry UD List for size, bh(0x%p)->b_private=0x%p\n", __FUNCTION__, 
                               dirent_bh, dirent_bh->b_private);
                        DTRACE("[%s]:Entry UD(0x%p) List for size, Size set, (%d)->(%d)\n", __FUNCTION__, 
                               p_ud_size, (int)raw_entry->size, (int)cpu_to_le32(inode->i_size));
                        DTRACE("[%s]:inode(0x%p) raw_entry(0x%p)\n", __FUNCTION__, inode, raw_entry);
                }else{
                        DTRACE("[%s]:Entry UD(0x%p) List for size, Size updated, (%d)->(%d)\n", __FUNCTION__, 
                               p_ud_size, (int)raw_entry->size, (int)cpu_to_le32(inode->i_size));
                        DTRACE("[%s]:inode(0x%p) raw_entry(0x%p)\n", __FUNCTION__, inode, raw_entry);
                        DTRACE("[%s]:flag_in_replace=%d, flag_ud_remove=%d\n", __FUNCTION__,
                               p_ud_size->flag_in_replace, p_ud_size->flag_ud_remove);
                        if(p_ud_size->flag_in_replace){
                                printk("[%s]:Now rollbacking!! (%d) -> (%d) inode write want change\n", __FUNCTION__,
                                       (int)p_ud_size->size, (int)inode->i_size);
                                dump_stack();
//                                while(p_ud_size->flag_in_replace) usleep(1);
                        }
                        if(p_ud_size->flag_ud_remove){
                                printk("[%s]:Removing this UD!! (%d) -> (%d) inode write want change\n", __FUNCTION__,
                                       (int)p_ud_size->size, (int)inode->i_size);
                                dump_stack();
                        }
                        p_ud_size->size = inode->i_size;
                }
                DTRACE("[%s]:list Entry UD List for size, UD_size=0x%p\n", __FUNCTION__, p_ud_size);
        }

	DTRACE("[%s]:list Entry UD List, UD=0x%p\n", __FUNCTION__, p_ud);
	DTRACE("[%s]:list Entry UD List, bh(0x%p)->b_private=0x%p\n", __FUNCTION__, dirent_bh, dirent_bh->b_private);
        return 0;
}
#endif /* FAT_SOFTUPDATE */
