/*
 * Copyright (C) 2005, OGAWA Hirofumi
 *
 * Copyright 2005,2006 Sony Corporation
 *
 * Released under GPL v2.
 */

#include <linux/fs.h>
#include <linux/module.h>
#include <linux/writeback.h>
#include <linux/buffer_head.h>

#define FLUSH_INITAL_DELAY	HZ
#define FLUSH_DELAY		(HZ / 2)

int fs_flush_sync_fdata(struct inode *inode, struct file *filp)
{
	int err = 0;

	if (IS_FLUSH(inode) && filp->f_mode & FMODE_WRITE) {
		current->flags |= PF_SYNCWRITE;
		err = filemap_write_and_wait(filp->f_mapping);
		current->flags &= ~PF_SYNCWRITE;
	}
	return err;
}
EXPORT_SYMBOL_GPL(fs_flush_sync_fdata);

static void fs_flush_pdflush_handler(unsigned long arg)
{
	struct super_block *sb = (struct super_block *)arg;
	fsync_super(sb);
	up_read(&sb->s_umount);
}

static void fs_flush_timer(unsigned long data)
{
	struct super_block *sb = (struct super_block *)data;
	unsigned long last_flush_jiff;
	int err;

	last_flush_jiff = sb->last_flush_jiff;

	if (!time_after_eq(jiffies, last_flush_jiff + FLUSH_DELAY)) {
		mod_timer(&sb->flush_timer, last_flush_jiff + FLUSH_DELAY);
		return;
	}

	if (down_read_trylock(&sb->s_umount)) {
		if (sb->s_root) {
			err = pdflush_operation(fs_flush_pdflush_handler, data);
			if (!err)
				return;
			mod_timer(&sb->flush_timer, jiffies + FLUSH_DELAY);
		}
		up_read(&sb->s_umount);
	}
}

void __fs_mark_flush(struct super_block *sb)
{
	sb->last_flush_jiff = jiffies;
	/*
	 * make sure by smb_wmb() that dirty buffers before here is
	 * processed at the timer routine.
	 */
	smp_wmb();

	if (!timer_pending(&sb->flush_timer))
		mod_timer(&sb->flush_timer, jiffies + FLUSH_INITAL_DELAY);
}
EXPORT_SYMBOL_GPL(__fs_mark_flush);

/*
 * caller must take down_write(sb->s_umount), otherwise pdflush
 * handler may be run after this del_timer_sync.
 */
void fs_flush_stop(struct super_block *sb)
{
	sb->s_flags &= ~MS_FLUSH;
	del_timer_sync(&sb->flush_timer);
}

void fs_flush_init(struct super_block *sb)
{
	init_timer(&sb->flush_timer);
	sb->flush_timer.data = (unsigned long)sb;
	sb->flush_timer.function = fs_flush_timer;
	sb->last_flush_jiff = 0;
}
