/*-----------------------------------------------------------------------------
 * gpio-core.h
 *
 * a header file for gpio
 *
 * Copyright 2005,2006 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/
#ifndef __GPIO_CORE_H__
#define __GPIO_CORE_H__

#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/pm.h>
#include <asm/hardware/amba.h>

#undef IN
#undef OUT
#undef INOUT

#define IN
#define OUT
#define INOUT

struct gpio_callback{
    unsigned char	sense;
    unsigned char	timing;
    void		(*function)(IN unsigned int nr, IN void *param);
    void		*param;
};

/*=============================================================================
 *
 * gpio-core's public method. 
 * 
 * Scripts from "HEAD-START" to "HEAD-END" are created as a specification
 * sheet automatically by generator.
 *
 *===========================================================================*/
typedef struct{
/*[HEAD-START] (for "spec sheet generator")*/
/*=============================================================================
 * GPIO Core API. 
 *
 *   GPIO Dev or other driver call these functions
 *
 *===========================================================================*/
/*-----------------------------------------------------------------------------
 * gpio_open()
 *-----------------------------------------------------------------------------
 * Summary		: open
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: hndl	: >=0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int (*open)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * gpio_close()
 *-----------------------------------------------------------------------------
 * Summary		: close
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int (*close)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_set_io_func()
 *-----------------------------------------------------------------------------
 * Summary		: set IO sel setting
 *     IN param	: hndl	: GPIO hndl
 *     	        : func	: 0: output port
 *     	                  1: input port
 *     	                  2: input port with pull up(input enable)
 *     	                  3: input port with pull down(input enable)
 *     	                  4: input port with bus holder(input enable)
 *                        5: pull up(input disable)
 *                        6: pull down(input disable)
 *                        7: bus holder(input disable)
 *     	                  8: no set(nothing)
 *     	                  >=9: other function
 *     	        : val	: write value (if select func==0)
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*set_io_func)
    (
        IN int hndl,
        IN unsigned char func,
        IN unsigned int val
    );

/*-----------------------------------------------------------------------------
 * gpio_get_io_func()
 *-----------------------------------------------------------------------------
 * Summary		: read IO sel setting
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: func	: 0: output port
 *                        1: input port
 *                        2: input port with pull up(input enable)
 *                        3: input port with pull down(input enable)
 *                        4: input port with bus holder(input enable)
 *                        5: pull up(input disable)
 *                        6: pull down(input disable)
 *                        7: bus holder(input disable)
 *                        8: no set(nothing)
 *                        >=9: other function
 *  Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*get_io_func)
    (
        IN int hndl,
        OUT unsigned char *func
    );

/*-----------------------------------------------------------------------------
 * gpio_write()
 *-----------------------------------------------------------------------------
 * Summary		: write value to port
 *     IN param	: hndl	: GPIO hndl
 *              : val	: write value
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*write)
    (
        IN int hndl,
        IN unsigned int val
    );

/*-----------------------------------------------------------------------------
 * gpio_read()
 *-----------------------------------------------------------------------------
 * Summary		: read port value
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: val	: 0:low / 1: high / <0:err
 *---------------------------------------------------------------------------*/
    int	(*read)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_set_callback_func()
 *-----------------------------------------------------------------------------
 * Summary		: Set setting how detect interrupt
 *     IN param	: hndl	: GPIO hndl
 *     	        : sense	: 0: down edge
 *     	                  1: up edge
 *     	                  2: up down edge
 *     	                  3: low level
 *     	                  4: high level
 *     	                  5: no set(nothing)
 *              : timing: 0: soon(in irq) / 1: in tasklet
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*set_callback_func)
    (
        IN int hndl,
        IN unsigned char sense,
        IN unsigned char timing
    );

/*-----------------------------------------------------------------------------
 * gpio_get_callback_func()
 *-----------------------------------------------------------------------------
 * Summary		: get setting how detect interrupt
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: sense	: 0: down edge
 *     	                  1: up edge
 *     	                  2: up down edge
 *     	                  3: low level
 *     	                  4: high level
 *     	                  5: no set(nothing)
 *              : timing: 0: soon(in irq) / 1: in tasklet
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*get_callback_func)
    (
        IN int hndl,
        OUT unsigned char *sense,
        OUT unsigned char *timing
    );

/*-----------------------------------------------------------------------------
 * gpio_enable_callback()
 *-----------------------------------------------------------------------------
 * Summary		: enable callback
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*enable_callback)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_disable_callback()
 *-----------------------------------------------------------------------------
 * Summary		: disable callback
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*disable_callback)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_is_enable_callback()
 *-----------------------------------------------------------------------------
 * Summary		: check enable/disable callback
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: status: 1:enable / 0:disble / <0:err
 *---------------------------------------------------------------------------*/
    int	(*is_enable_callback)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_register_callback()
 *-----------------------------------------------------------------------------
 * Summary		: register callback
 *     IN param	: hndl	: GPIO hndl
 *              : callback: callback parameter
 *                    sense	: 0: down edge
 *     	                          1: up edge
 *     	                          2: up down edge
 *     	                          3: low level
 *     	                          4: high level
 *     	                          5: no set(nothing)
 *                    timing	: 0: soon(in irq) / 1: in tasklet
 *                    (*function)(IN unsigned int nr, IN void *param);
 *                    *param	: callback parameter
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*register_callback)
    (
        IN int hndl, 
        IN struct gpio_callback *callback
    );

/*-----------------------------------------------------------------------------
 * gpio_unregister_callback()
 *-----------------------------------------------------------------------------
 * Summary	: unregister callback
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*unregister_callback)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_enable_wakeup()
 *-----------------------------------------------------------------------------
 * Summary	: enable wakeup
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err 
 *---------------------------------------------------------------------------*/
    int	(*enable_wakeup)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_disable_wakeup()
 *-----------------------------------------------------------------------------
 * Summary	: disble wakeup
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*disable_wakeup)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------------------
 * gpio_is_enable_wakeup()
 *-----------------------------------------------------------------------------
 * Summary	: check enable/disable wakeup
 *     IN param	: hndl	: GPIO hndl
 *     OUT param: -	: -
 *     Return	: status: 1:enable / 0:disble / <0:err
 *---------------------------------------------------------------------------*/
    int	(*is_enable_wakeup)
    (
        IN int hndl
    );

/*-----------------------------------------------------------------[HEAD-END]*/
}TGpioPublicApi;

/*[HEAD-START] (for "spec sheet generator")*/
/*=============================================================================
 * GPIO Core API parameter
 *===========================================================================*/
/*-----------------------------------------------------------------------------
 * IO sel select
 *---------------------------------------------------------------------------*/
enum{
    GPIO_OUTPUT,
    GPIO_INPUT,
    GPIO_INPUT_PU,
    GPIO_INPUT_PD,
    GPIO_INPUT_BH,
    GPIO_PULLUP,
    GPIO_PULLDN,
    GPIO_BUSHD,
    GPIO_NOSET,
    GPIO_FUNC_ARCH,	/* arch depend */
};

/*-----------------------------------------------------------------------------
 * Interrupt sense select
 *---------------------------------------------------------------------------*/
enum{
    GPIO_INT_DOWN_EDGE,
    GPIO_INT_UP_EDGE,
    GPIO_INT_UPDOWN_EDGE,
    GPIO_INT_L_LEVEL,
    GPIO_INT_H_LEVEL,
    GPIO_INT_NOSET,
};

/*-----------------------------------------------------------------------------
 * Callback timing
 *---------------------------------------------------------------------------*/
enum{
    GPIO_CALLBACK_IRQ,
    GPIO_CALLBACK_TASKLET
};
/*[HEAD-END]*/

/*=============================================================================
 * Pulgin Operations. Core call these functions
 *===========================================================================*/
typedef struct{
/*-----------------------------------------------------------------------------
 * open
 *-----------------------------------------------------------------------------
 * Summary		: open
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int (*open)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * close
 *-----------------------------------------------------------------------------
 * Summary		: close
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int (*close)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * set_sel_func
 *-----------------------------------------------------------------------------
 * Summary		: set IO sel setting
 *     IN param	: nr	: GPIO number
 *     	        : func	: 0: output port
 *     	                  1: input port
 *     	                  2: input port with pull up(input enable)
 *     	                  3: input port with pull down(input enable)
 *     	                  4: input port with bus holder(input enable)
 *                        5: pull up(input disable)
 *                        6: pull down(input disable)
 *                        7: bus holder(input disable)
 *     	                  8: no set(nothing)
 *     	                  >=9: other function
 *     	        : val	: write value (if select func==0)
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*set_sel_func)
    (
        IN unsigned int nr,
        IN unsigned char func,
        IN unsigned int val
    );

/*-----------------------------------------------------------------------------
 * get_sel_func
 *-----------------------------------------------------------------------------
 * Summary		: read IO sel sreadetting
 *    IN param	: nr	: GPIO number
 *    OUT param	: func	: 0: output port
 *                        1: input port
 *                        2: input port with pull up(input enable)
 *                        3: input port with pull down(input enable)
 *     	                  4: input port with bus holder(input enable)
 *                        5: pull up(input disable)
 *                        6: pull down(input disable)
 *                        7: bus holder(input disable)
 *     	                  8: no set(nothing)
 *     	                  >=9: other function
 *  Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*get_sel_func)
    (
        IN unsigned int nr,
        OUT unsigned char *func
    );

/*-----------------------------------------------------------------------------
 * write
 *-----------------------------------------------------------------------------
 * Summary		: write value to port
 *     IN param	: nr	: GPIO number
 *              : val	: write value
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*write)
    (
        IN unsigned int nr,
        IN unsigned int val
    );

/*-----------------------------------------------------------------------------
 * read
 *-----------------------------------------------------------------------------
 * Summary		: read port value
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: val	: 1: high / 0:low / <0:err
 *---------------------------------------------------------------------------*/
    int	(*read)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * set_irq_sense
 *-----------------------------------------------------------------------------
 * Summary		: Set setting how detect interrupt
 *     IN param	: nr	: GPIO number
 *     	        : sense	: 0: down edge
 *     	                  1: up edge
 *     	                  2: up down edge
 *     	                  3: low level
 *     	                  4: high level
 *     	                  5: no set(nothing)
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*set_irq_sense)
    (
        IN unsigned int nr,
        IN unsigned char sense
    );

/*-----------------------------------------------------------------------------
 * get_irq_sense
 *-----------------------------------------------------------------------------
 * Summary		: get setting how detect interrupt
 *     IN param	: nr	: GPIO number
 *     OUT param: sense	: 0: down edge
 *     	                  1: up edge
 *     	                  2: up down edge
 *     	                  3: low level
 *     	                  4: high level
 *     	                  5: no set(nothing)
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*get_irq_sense)
    (
        IN unsigned int nr,
        OUT unsigned char *sense
    );

/*-----------------------------------------------------------------------------
 * enable_irq
 *-----------------------------------------------------------------------------
 * Summary		: enable interrupt
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*enable_irq)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * disable_irq
 *-----------------------------------------------------------------------------
 * Summary		: disable interrupt
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*disable_irq)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * is_enable_irq
 *-----------------------------------------------------------------------------
 * Summary		: check enable/disable interrupt
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: status: 1:enable / 0:disble / <0:err
 *---------------------------------------------------------------------------*/
    int	(*is_enable_irq)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * is_irq
 *-----------------------------------------------------------------------------
 * Summary		: check irq request
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: status: 1:request / 0:no request / <0:err
 *---------------------------------------------------------------------------*/
    int	(*is_irq)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * clear_irq
 *-----------------------------------------------------------------------------
 * Summary		: clear irq request
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*clear_irq)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * register_handler
 *-----------------------------------------------------------------------------
 * Summary		: register handler
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*register_handler)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * unregister_handler
 *-----------------------------------------------------------------------------
 * Summary	: unregister handler
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*unregister_handler)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * enable_wakeup
 *-----------------------------------------------------------------------------
 * Summary	: enable wakeup
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err 
 *---------------------------------------------------------------------------*/
    int	(*enable_wakeup)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * disable_wakeup
 *-----------------------------------------------------------------------------
 * Summary	: disble wakeup
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*disable_wakeup)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * is_enable_wakeup
 *-----------------------------------------------------------------------------
 * Summary	: check enable/disable wakeup 
 *     IN param	: nr	: GPIO number
 *     OUT param: -	: -
 *     Return	: status: 1:enable / 0:disble / <0:err
 *---------------------------------------------------------------------------*/
    int	(*is_enable_wakeup)
    (
        IN unsigned int nr
    );

/*-----------------------------------------------------------------------------
 * suspend
 *-----------------------------------------------------------------------------
 * Summary	: suspend callback routine
 *     IN param	: *dev	: amba_device structure
 *     	        : state	: device state
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*suspend)
    (
        IN struct amba_device *dev,
        IN u32 state
    );

/*-----------------------------------------------------------------------------
 * resume
 *-----------------------------------------------------------------------------
 * Summary	: resume callback routine
 *     IN param	: *dev	: amba_device structure
 *     OUT param: -	: -
 *     Return	: err	: 0:succeed / <0:err
 *---------------------------------------------------------------------------*/
    int	(*resume)
    (
        IN struct amba_device *dev
    );

}TGpioPluginOps;

// Plugin structure
typedef struct{
    const char 		*name;
    const char 		*version;
    unsigned int 	portNum;
    TGpioPluginOps	*ops;
    unsigned long	start;
    unsigned long	end;
}CGpioPlugin;

typedef struct __CGpioSel CGpioSel;

struct __CGpioSel{
    CGpioSel			*next; /* Not use now */
    unsigned int		openCnt;
    unsigned int		hndl;
    struct gpio_callback	callback;
    struct tasklet_struct	tasklet;
    unsigned char		registCb;
};

// Templare Core structure
typedef struct{
    const char 		*name;
    const char 		*version;
    TGpioPublicApi	*api;
    CGpioPlugin		*plugin;
    CGpioSel		**gpioSel;
    int			(*attach_plugin)(IN CGpioPlugin *plugin);
    int			(*detach_plugin)(IN CGpioPlugin *plugin);
    void		(*handler)(IN unsigned int nr);
    int			(*get_hndl)(IN unsigned int nr);
    int			(*probe)(IN struct amba_device *dev, IN void *id);
    int			(*suspend)(IN struct amba_device *dev, IN u32 state);
    int			(*resume)(IN struct amba_device *dev);
}CGpioCore;

#ifndef __GPIO_CORE_C__
extern CGpioCore *pGpioCore;

#define gpio_open(a)			(pGpioCore->api->open(a))
#define gpio_close(a)			(pGpioCore->api->close(a))
#define gpio_set_io_func(a,b,c)		(pGpioCore->api->set_io_func(a,b,c))
#define gpio_get_io_func(a,b)		(pGpioCore->api->get_io_func(a,b))
#define gpio_write(a,b)			(pGpioCore->api->write(a,b))
#define gpio_read(a)			(pGpioCore->api->read(a))
#define gpio_set_callback_func(a,b,c)	(pGpioCore->api->set_callback_func(a,b,c))
#define gpio_get_callback_func(a,b,c)	(pGpioCore->api->get_callback_func(a,b,c))
#define gpio_enable_callback(a)		(pGpioCore->api->enable_callback(a))
#define gpio_disable_callback(a)	(pGpioCore->api->disable_callback(a))
#define gpio_is_enable_callback(a)	(pGpioCore->api->is_enable_callback(a))
#define gpio_register_callback(a,b)	(pGpioCore->api->register_callback(a,b))
#define gpio_unregister_callback(a)	(pGpioCore->api->unregister_callback(a))
#define gpio_enable_wakeup(a)		(pGpioCore->api->enable_wakeup(a))
#define gpio_disable_wakeup(a)		(pGpioCore->api->disable_wakeup(a))
#define gpio_is_enable_wakeup(a)	(pGpioCore->api->is_enable_wakeup(a))
#define gpio_attach_plugin		(pGpioCore->attach_plugin)
#define gpio_detach_plugin		(pGpioCore->detach_plugin)
#define gpio_handler(a)			(pGpioCore->handler(a))
#define gpio_get_hndl(a)		(pGpioCore->get_hndl(a))
#define gpio_probe(a,b)			(pGpioCore->probe(a,b))
#define gpio_suspend(a,b)		(pGpioCore->suspend(a,b))
#define gpio_resume(a)			(pGpioCore->resume(a))

#endif /* __GPIO_CORE_C__ */
#endif /*__GPIO_CORE_H__*/
