/*
 * include/linux/msdos_fs.h
 * 
 * common header file for fat
 * 
 * Copyright 2005,2006 Sony Corporation
 */
#ifndef _LINUX_MSDOS_FS_H
#define _LINUX_MSDOS_FS_H

/*
 * The MS-DOS filesystem constants/structures
 */
#include <asm/byteorder.h>

#include <linux/fs.h>

#define SECTOR_SIZE	512		/* sector size (bytes) */
#define SECTOR_BITS	9		/* log2(SECTOR_SIZE) */
#define MSDOS_DPB	(MSDOS_DPS)	/* dir entries per block */
#define MSDOS_DPB_BITS	4		/* log2(MSDOS_DPB) */
#define MSDOS_DPS	(SECTOR_SIZE / sizeof(struct msdos_dir_entry))
#define MSDOS_DPS_BITS	4		/* log2(MSDOS_DPS) */
#define CF_LE_W(v)	le16_to_cpu(v)
#define CF_LE_L(v)	le32_to_cpu(v)
#define CT_LE_W(v)	cpu_to_le16(v)
#define CT_LE_L(v)	cpu_to_le32(v)


#define MSDOS_SUPER_MAGIC 0x4d44 /* MD */

#define MSDOS_ROOT_INO	1	/* == MINIX_ROOT_INO */
#define MSDOS_DIR_BITS	5	/* log2(sizeof(struct msdos_dir_entry)) */

/* directory limit */
#define FAT_MAX_DIR_ENTRIES	(65536)
#define FAT_MAX_DIR_SIZE	(FAT_MAX_DIR_ENTRIES << MSDOS_DIR_BITS)

#define ATTR_NONE	0	/* no attribute bits */
#define ATTR_RO		1	/* read-only */
#define ATTR_HIDDEN	2	/* hidden */
#define ATTR_SYS	4	/* system */
#define ATTR_VOLUME	8	/* volume label */
#define ATTR_DIR	16	/* directory */
#define ATTR_ARCH	32	/* archived */

/* attribute bits that are copied "as is" */
#define ATTR_UNUSED	(ATTR_VOLUME | ATTR_ARCH | ATTR_SYS | ATTR_HIDDEN)
/* bits that are used by the Windows 95/Windows NT extended FAT */
#define ATTR_EXT	(ATTR_RO | ATTR_HIDDEN | ATTR_SYS | ATTR_VOLUME)

#define CASE_LOWER_BASE	8	/* base is lower case */
#define CASE_LOWER_EXT	16	/* extension is lower case */

#define DELETED_FLAG	0xe5	/* marks file as deleted when in name[0] */
#define IS_FREE(n)	(!*(n) || *(n) == DELETED_FLAG)

/* valid file mode bits */
#define MSDOS_VALID_MODE (S_IFREG | S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO)
/* Convert attribute bits and a mask to the UNIX mode. */
#define MSDOS_MKMODE(a, m) (m & (a & ATTR_RO ? S_IRUGO|S_IXUGO : S_IRWXUGO))

#define MSDOS_NAME	11	/* maximum name length */
#define MSDOS_LONGNAME	256	/* maximum name length */
#define MSDOS_SLOTS	21	/* max # of slots for short and long names */
#define MSDOS_DOT	".          "	/* ".", padded to MSDOS_NAME chars */
#define MSDOS_DOTDOT	"..         "	/* "..", padded to MSDOS_NAME chars */

/* media of boot sector */
#define FAT_VALID_MEDIA(x)	((0xF8 <= (x) && (x) <= 0xFF) || (x) == 0xF0)
#define FAT_FIRST_ENT(s, x)	((MSDOS_SB(s)->fat_bits == 32 ? 0x0FFFFF00 : \
	MSDOS_SB(s)->fat_bits == 16 ? 0xFF00 : 0xF00) | (x))

/* start of data cluster's entry (number of reserved clusters) */
#define FAT_START_ENT	2

/* maximum number of clusters */
#define MAX_FAT12	0xFF4
#define MAX_FAT16	0xFFF4
#define MAX_FAT32	0x0FFFFFF6
#define MAX_FAT(s)	(MSDOS_SB(s)->fat_bits == 32 ? MAX_FAT32 : \
	MSDOS_SB(s)->fat_bits == 16 ? MAX_FAT16 : MAX_FAT12)

/* bad cluster mark */
#define BAD_FAT12	0xFF7
#define BAD_FAT16	0xFFF7
#define BAD_FAT32	0x0FFFFFF7

/* standard EOF */
#define EOF_FAT12	0xFFF
#define EOF_FAT16	0xFFFF
#define EOF_FAT32	0x0FFFFFFF

#define FAT_ENT_FREE	(0)
#define FAT_ENT_BAD	(BAD_FAT32)
#define FAT_ENT_EOF	(EOF_FAT32)

#define FAT_FSINFO_SIG1	0x41615252
#define FAT_FSINFO_SIG2	0x61417272
#define IS_FSINFO(x)	(le32_to_cpu((x)->signature1) == FAT_FSINFO_SIG1 \
			 && le32_to_cpu((x)->signature2) == FAT_FSINFO_SIG2)

/*
 * ioctl commands
 */
#define VFAT_IOCTL_READDIR_BOTH		_IOR('r', 1, struct dirent [2])
#define VFAT_IOCTL_READDIR_SHORT	_IOR('r', 2, struct dirent [2])
/* <linux/videotext.h> has used 0x72 ('r') in collision, so skip a few */
#define FAT_IOCTL_GET_ATTRIBUTES	_IOR('r', 0x10, __u32)
#define FAT_IOCTL_SET_ATTRIBUTES	_IOW('r', 0x11, __u32)
#define FAT_IOCTL_CLOSE_NOSYNC	        _IOW('r', 0x12, __u32)

/*
 * vfat shortname flags
 */
#define VFAT_SFN_DISPLAY_LOWER	0x0001 /* convert to lowercase for display */
#define VFAT_SFN_DISPLAY_WIN95	0x0002 /* emulate win95 rule for display */
#define VFAT_SFN_DISPLAY_WINNT	0x0004 /* emulate winnt rule for display */
#define VFAT_SFN_CREATE_WIN95	0x0100 /* emulate win95 rule for create */
#define VFAT_SFN_CREATE_WINNT	0x0200 /* emulate winnt rule for create */

struct fat_boot_sector {
	__u8	ignored[3];	/* Boot strap short or near jump */
	__u8	system_id[8];	/* Name - can be used to special case
				   partition manager volumes */
	__u8	sector_size[2];	/* bytes per logical sector */
	__u8	sec_per_clus;	/* sectors/cluster */
	__le16	reserved;	/* reserved sectors */
	__u8	fats;		/* number of FATs */
	__u8	dir_entries[2];	/* root directory entries */
	__u8	sectors[2];	/* number of sectors */
	__u8	media;		/* media code */
	__le16	fat_length;	/* sectors/FAT */
	__le16	secs_track;	/* sectors per track */
	__le16	heads;		/* number of heads */
	__le32	hidden;		/* hidden sectors (unused) */
	__le32	total_sect;	/* number of sectors (if sectors == 0) */

	/* The following fields are only used by FAT32 */
	__le32	fat32_length;	/* sectors/FAT */
	__le16	flags;		/* bit 8: fat mirroring, low 4: active fat */
	__u8	version[2];	/* major, minor filesystem version */
	__le32	root_cluster;	/* first cluster in root directory */
	__le16	info_sector;	/* filesystem info sector */
	__le16	backup_boot;	/* backup boot sector */
	__le16	reserved2[6];	/* Unused */
};

struct fat_boot_fsinfo {
	__le32   signature1;	/* 0x41615252L */
	__le32   reserved1[120];	/* Nothing as far as I can tell */
	__le32   signature2;	/* 0x61417272L */
	__le32   free_clusters;	/* Free cluster count.  -1 if unknown */
	__le32   next_cluster;	/* Most recently allocated cluster */
	__le32   reserved2[4];
};

struct msdos_dir_entry {
	__u8	name[8],ext[3];	/* name and extension */
	__u8	attr;		/* attribute bits */
	__u8    lcase;		/* Case for base and extension */
	__u8	ctime_cs;	/* Creation time, centiseconds (0-199) */
	__le16	ctime;		/* Creation time */
	__le16	cdate;		/* Creation date */
	__le16	adate;		/* Last access date */
	__le16	starthi;	/* High 16 bits of cluster in FAT32 */
	__le16	time,date,start;/* time, date and first cluster */
	__le32	size;		/* file size (in bytes) */
};

/* Up to 13 characters of the name */
struct msdos_dir_slot {
	__u8    id;		/* sequence number for slot */
	__u8    name0_4[10];	/* first 5 characters in name */
	__u8    attr;		/* attribute byte */
	__u8    reserved;	/* always 0 */
	__u8    alias_checksum;	/* checksum for 8.3 alias */
	__u8    name5_10[12];	/* 6 more characters in name */
	__le16   start;		/* starting cluster number, 0 in long slots */
	__u8    name11_12[4];	/* last 2 characters in name */
};

struct fat_slot_info {
	loff_t i_pos;		/* on-disk position of directory entry */
	loff_t slot_off;	/* offset for slot or de start */
	int nr_slots;		/* number of slots + 1(de) in filename */
	struct msdos_dir_entry *de;
	struct buffer_head *bh;
};

#ifdef __KERNEL__

#include <linux/buffer_head.h>
#include <linux/string.h>
#include <linux/nls.h>
#include <linux/fs.h>

#define FAT_LKUP_HINT_MAX     16
/* assume FAT_LKUP_HINT_MAX <= sizeof(long)*8 */

struct fat_lookup_hint {
	struct list_head lru_list;
	pid_t pid;
	struct inode *dir;
	loff_t last_pos;
#define FAT_LKUP_ST_RANDOM       1
#define FAT_LKUP_ST_LFN_RANDOM       2
	unsigned int state;
};

struct fat_mount_options {
	uid_t fs_uid;
	gid_t fs_gid;
	unsigned short fs_fmask;
	unsigned short fs_dmask;
	unsigned short codepage;  /* Codepage for shortname conversions */
	char *iocharset;          /* Charset used for filename input/display */
	unsigned short shortname; /* flags for shortname display/create rule */
	unsigned char name_check; /* r = relaxed, n = normal, s = strict */
	unsigned quiet:1,         /* set = fake successful chmods and chowns */
		 showexec:1,      /* set = only set x bit for com/exe/bat */
		 sys_immutable:1, /* set = system files are immutable */
		 brkn_frclstrs:1, /* calc. free_clusters's vaule */
		 dotsOK:1,        /* set = hidden and system files are named '.filename' */
		 no_sect_bndry:1, /* set = avoid acrossing sector boundary on FAT12 alloc */
		 isvfat:1,        /* 0=no vfat long filename support, 1=vfat support */
		 posix_attr:1,       /* 1= posix attribute mapping support */
		 utf8:1,	  /* Use of UTF8 character set (Default) */
		 unicode_xlate:1, /* create escape sequences for unhandled Unicode */
		 numtail:1,       /* Does first alias have a numeric '~1' type tail? */
		 atari:1,         /* Use Atari GEMDOS variation of MS-DOS fs */
		 nocase:1,	  /* Does this need case conversion? 0=need case conversion*/
		 gc:1,            /* enable Garbage Collector */
		 relax_sync:1,	  /* set = make sync behavior looser */
		 batch_sync:1,    /* set = make sync in BATCH mode */
		 dirsync:1	    
#ifdef FAT_SOFTUPDATE
		 ,softupdate_with_sync:1  /* set = softupdate when sync mode */
#endif /* FAT_SOFTUPDATE */
		;
};

#define FAT_HASH_BITS	8
#define FAT_HASH_SIZE	(1UL << FAT_HASH_BITS)
#define FAT_HASH_MASK	(FAT_HASH_SIZE-1)

struct fat_posix_ops {
	int (*is_symlink)(struct inode *, struct msdos_dir_entry *);
	int (*set_attr)(struct msdos_dir_entry *, struct inode *);
};

/*
 * MS-DOS file system in-core superblock data
 */
struct msdos_sb_info {
	unsigned short sec_per_clus; /* sectors/cluster */
	unsigned short cluster_bits; /* log2(cluster_size) */
	unsigned int cluster_size;   /* cluster size */
	unsigned char fats,fat_bits; /* number of FATs, FAT bits (12 or 16) */
	unsigned short fat_start;
	unsigned long fat_length;    /* FAT start & length (sec.) */
	unsigned long dir_start;
	unsigned short dir_entries;  /* root dir start & entries */
	unsigned long data_start;    /* first data sector */
	unsigned long max_cluster;   /* maximum cluster number */
	unsigned long root_cluster;  /* first cluster of the root directory */
	unsigned long fsinfo_sector; /* sector number of FAT32 fsinfo */
	struct semaphore fat_lock;
	unsigned int prev_free;      /* previously allocated cluster number */
	unsigned int free_clusters;  /* -1 if undefined */
	struct fat_mount_options options;
	struct nls_table *nls_disk;  /* Codepage used on disk */
	struct nls_table *nls_io;    /* Charset used for input and display */
	void *dir_ops;		     /* Opaque; default directory operations */
	int dir_per_block;	     /* dir entries per block */
	int dir_per_block_bits;	     /* log2(dir_per_block) */

	int fatent_shift;
	struct fatent_operations *fatent_ops;
	struct fat_posix_ops posix_ops;

	spinlock_t inode_hash_lock;
	struct hlist_head inode_hashtable[FAT_HASH_SIZE];

	/* dirent lookup hints */
	spinlock_t lkup_hint_lock;
	struct list_head lkup_hint_head;
	struct fat_lookup_hint lkup_hint[FAT_LKUP_HINT_MAX];
	unsigned lkup_hint_freemap:FAT_LKUP_HINT_MAX;

	/* GC fields */
	struct completion gc_com;
	u8 *gc_bitmap;
	int gc_quit;

#ifdef FAT_SOFTUPDATE
        /* Soft Update */
        struct list_head d_ud;
        struct list_head d_ud_head;
        struct list_head d_depend_private_head;
#endif
};

#define FAT_CACHE_VALID	0	/* special case for valid cache */

/*
 * MS-DOS file system inode data in memory
 */
struct msdos_inode_info {
	spinlock_t cache_lru_lock;
	struct list_head cache_lru;
	int nr_caches;
	/* for avoiding the race between fat_free() and fat_get_cluster() */
	unsigned int cache_valid_id;

	loff_t mmu_private;
	int i_start;		/* first cluster or 0 */
	int i_logstart;		/* logical first cluster */
	int i_attrs;		/* unused attribute bits */
	loff_t i_pos;		/* on-disk position of directory entry or 0 */
	struct hlist_node i_fat_hash;	/* hash by i_location */

	/* for Fast Sync Operation*/
	int i_new; /*1: new inode && not SYNCed yet*/
	int i_last_dclus; /*not SYNCed last cluster or 0*/
	int i_new_dclus; /*new cluster to be linked at i_last*/
	
#ifdef FAT_SOFTUPDATE
	/* for SoftUpdate Operation*/
	spinlock_t softupdate_ud_lock;
        struct fat_softupdate_ud *i_ud; /* dirent UD list */
        struct fat_softupdate_ud *i_ud_size; /* dirent UD list for size */
	struct list_head i_depend_private; /* private list for related this inode */
#endif /* FAT_SOFTUPDATE */

	struct inode vfs_inode;
};

static inline struct msdos_sb_info *MSDOS_SB(struct super_block *sb)
{
	return sb->s_fs_info;
}

static inline struct msdos_inode_info *MSDOS_I(struct inode *inode)
{
	return container_of(inode, struct msdos_inode_info, vfs_inode);
}

/* Return the FAT attribute byte for this inode */
static inline u8 fat_attr(struct inode *inode)
{
	return ((inode->i_mode & S_IWUGO) ? ATTR_NONE : ATTR_RO) |
		(S_ISDIR(inode->i_mode) ? ATTR_DIR : ATTR_NONE) |
		MSDOS_I(inode)->i_attrs;
}

static inline sector_t fat_clus_to_blknr(struct msdos_sb_info *sbi, int clus)
{
	return ((sector_t)clus - FAT_START_ENT) * sbi->sec_per_clus
		+ sbi->data_start;
}

static inline void fat16_towchar(wchar_t *dst, const __u8 *src, size_t len)
{
#ifdef __BIG_ENDIAN
	while (len--) {
		*dst++ = src[0] | (src[1] << 8);
		src += 2;
	}
#else
	memcpy(dst, src, len * 2);
#endif
}

static inline void fatwchar_to16(__u8 *dst, const wchar_t *src, size_t len)
{
#ifdef __BIG_ENDIAN
	while (len--) {
		dst[0] = *src & 0x00FF;
		dst[1] = (*src & 0xFF00) >> 8;
		dst += 2;
		src++;
	}
#else
	memcpy(dst, src, len * 2);
#endif
}

static inline loff_t fat_make_i_pos(struct super_block *sb,
				    struct buffer_head *bh,
				    struct msdos_dir_entry *de)
{
	return ((loff_t)bh->b_blocknr << MSDOS_SB(sb)->dir_per_block_bits)
		| (de - (struct msdos_dir_entry *)bh->b_data);
}

/* fat/cache.c */
extern void fat_cache_inval_inode(struct inode *inode);
extern int fat_get_cluster(struct inode *inode, int cluster,
			   int *fclus, int *dclus);
extern int fat_bmap(struct inode *inode, sector_t sector, sector_t *phys);

/* fat/dir.c */
extern struct file_operations fat_dir_operations;
extern int fat_search_long(struct inode *inode, const unsigned char *name,
			   int name_len, struct fat_slot_info *sinfo);
extern int fat_dir_empty(struct inode *dir);
extern int fat_subdirs(struct inode *dir);
extern int fat_scan(struct inode *dir, const unsigned char *name,
		    struct fat_slot_info *sinfo);
extern int fat_get_short_entry(struct inode *dir, loff_t *pos,
			       struct buffer_head **bh,
			       struct msdos_dir_entry **de);
extern int fat_get_dotdot_entry(struct inode *dir, struct buffer_head **bh,
				struct msdos_dir_entry **de, loff_t *i_pos);
extern int fat_alloc_new_dir(struct inode *dir, struct timespec *ts);
extern int fat_add_entries(struct inode *dir, void *slots, int nr_slots,
			   struct fat_slot_info *sinfo);
extern int fat_remove_entries(struct inode *dir, struct fat_slot_info *sinfo);

/* fat/fatent.c */
struct fat_entry {
	int entry;
	union {
		u8 *ent12_p[2];
		__le16 *ent16_p;
		__le32 *ent32_p;
	} u;
	int nr_bhs;
	struct buffer_head *bhs[2];
};

static inline void fatent_init(struct fat_entry *fatent)
{
	fatent->nr_bhs = 0;
	fatent->entry = 0;
	fatent->u.ent32_p = NULL;
	fatent->bhs[0] = fatent->bhs[1] = NULL;
}

static inline void fatent_set_entry(struct fat_entry *fatent, int entry)
{
	fatent->entry = entry;
	fatent->u.ent32_p = NULL;
}

static inline void fatent_brelse(struct fat_entry *fatent)
{
	int i;
	fatent->u.ent32_p = NULL;
	for (i = 0; i < fatent->nr_bhs; i++)
		brelse(fatent->bhs[i]);
	fatent->nr_bhs = 0;
	fatent->bhs[0] = fatent->bhs[1] = NULL;
}

extern void fat_ent_access_init(struct super_block *sb);
extern int fat_ent_read(struct inode *inode, struct fat_entry *fatent,
			int entry);
extern int fat_ent_write(struct inode *inode, struct fat_entry *fatent,
			 int new, int wait);
extern int fat_alloc_clusters(struct inode *inode, int *cluster,
			      int nr_cluster);
extern int fat_free_clusters(struct inode *inode, int cluster);
extern int fat_count_free_clusters(struct super_block *sb);

/* fat/file.c */
extern int fat_generic_ioctl(struct inode *inode, struct file *filp,
			     unsigned int cmd, unsigned long arg);
extern struct file_operations fat_file_operations;
extern struct file_operations fat_file_operations_nosync;
extern struct inode_operations fat_file_inode_operations;
extern struct inode_operations fat_symlink_inode_operations;
extern int fat_notify_change(struct dentry * dentry, struct iattr * attr);
extern void fat_truncate(struct inode *inode);

/* fat/inode.c */
extern void fat_attach(struct inode *inode, loff_t i_pos);
extern void fat_detach(struct inode *inode);
extern struct inode *fat_iget(struct super_block *sb, loff_t i_pos);
extern struct inode *fat_build_inode(struct super_block *sb,
			struct msdos_dir_entry *de, loff_t i_pos);
extern int fat_sync_inode(struct inode *inode);
extern int fat_fill_super(struct super_block *sb, void *data, int silent,
			struct inode_operations *fs_dir_inode_ops, int isvfat);
extern void fat_kill_super(struct super_block *sb);

/* fat/misc.c */
extern void fat_fs_panic(struct super_block *s, const char *fmt, ...);
extern void fat_clusters_flush(struct super_block *sb);
extern int fat_chain_add(struct inode *inode, int new_dclus, int nr_cluster);
extern time_t date_dos2unix(unsigned short time, unsigned short date);
extern void fat_date_unix2dos(time_t unix_date, __le16 *time, __le16 *date);
extern int fat_sync_bhs(struct buffer_head **bhs, int nr_bhs);
extern int fat_flush_bhs(struct buffer_head **bhs, int nr_bhs);

/* fat/gc.c */
#ifdef CONFIG_FAT_GC
extern void fat_start_gc(struct super_block *);
extern void fat_stop_gc(struct super_block *);
extern void fat_gc_mark_cluster_valid(struct msdos_sb_info *, int);
extern int fat_gc_mark_valid_entries(struct inode *);
#else
#define fat_start_gc(sb)
#define fat_stop_gc(sb)
#define fat_gc_mark_cluster_valid(sbi, clu)
#define fat_gc_mark_valid_entries(inode)
#endif

static inline int fat_syncdir(struct inode *dir)
{
	struct super_block *sb = dir->i_sb;
	struct msdos_sb_info *sbi = MSDOS_SB(sb);
	int ret = 0;
	
	if (!IS_DIRSYNC(dir) && sbi->options.batch_sync)
		ret = sync_blockdev(sb->s_bdev);

	return ret;
}

/*
 * vfat posix attr option "posix_attr" stuffs
 *
 * FEATURES
 *	
 * Following attributes/modes are supported in posix attributea mapping
 * in VFAT.
 *
 *   - FileType
 * 	This supports following special files and it's attributes;
 * 		symbolic link,	block device node,
 * 		char device node, fifo,	socket
 * 	Regular files/dirs also may have POSIX attributes.
 *   - DeviceFile
 * 	Major and minor number would be held at ctime
 * 	and both values are limited  to 255.
 *   - Owner's User ID/Group ID 
 * 	This can be used to distinguish root and others,
 * 	because this has just one bit width. 
 * 	Value of UID/GID for non-root user will be taken from uid/gid 
 * 	option on mounting. If nothing is specified, system uses 
 *	(u16)-1 as last resort. That means change-uid may affect on gid.
 *   - Permission for Group/Other (rwx)
 * 	Those modes will be kept in ctime_cs.
 * 	Also permission modes for "others" will be
 * 	same as "group", due to lack of fields.
 *	That means set-group-mode may affect on other-mode.
 *	On the other hand, set-other-mode has no affect to group-mode.
 *	
 *   - Permission for Owner (rwx)
 * 	These modes will be mapped to FAT attributes.
 * 	Just same as mapping under VFAT.
 *   - Others
 * 	no sticky, setgid nor setuid bits are not supported.
 * 
 *ALGORITHM FOR MAPPING DECISION
 *   - Regular file/dir
 * 	To distinguish regular files/dirs, look if this fat dir 
 * 	entry doesn't have ATTR_SYS, first. If it doesn't have 
 * 	ATTR_SYS, then check if TYPE field (MSB 3bits) in ctime_cs 
 * 	is equal to 7. If so, this regular file/dir is created and/or 
 * 	modified under VFAT with "posix_attr". And posix attribute 
 * 	mapping can be take place. Otherwise, conventional VFAT 
 * 	attribute mapping is used.
 *
 *   - Special file
 * 	To distinguish special files, look if this fat dir entry 
 * 	has ATTR_SYS, first. Also we need to check it not to have 
 *	ATT_EXT.
 *	If it has ATTR_SYS, then check 1st. LSB bit in ctime_cs, 
 *	referred as "special file flag".
 * 	If set,  this file is created under VFAT with "posix_attr". 
 * 	Look up TYPE field to decide special file type.
 *
 * 	This special file detection method has some flaw to make
 * 	potential confusion. E.g. some system file created under
 *	dos/win may be treated as special file.  However in most case,
 *	user don't create system file under dos/win.
 *	To reduce possiblity of this confusion, system makes
 *	sure special files except symlink have size ZERO.
 *	For symlink, system checks it's size not to exceed page size 
 *	and PATH_MAX.
 *
 *FAT DIR ENTRY FIELDS
 *
 *   - ctime_cs
 * 	    8bit byte
 * 	7 6 5 4 3 2 1 0
 * 	|===| | | | | |
 * 	TYPE  | | | | +- special file flag (valid if ATTR_SYS)
 * 	      | | | +--- User/Group ID(root or others)
 * 	      | | +----- !group X
 * 	      | +------- !group W
 * 	      +--------- !group R
 * 
 * 	  special file flag
 * 		Indicate this entry has posix attribute mapping.
 * 		This field is valid for fat dir entry, which 
 * 		have ATTR_SYS.
 * 
 * 	  special file TYPE
 * 		val	type on VFS(val)	Description
 * 		------------------------------------------------
 * 		0 	(place folder for backward compat)
 * 		1 	DT_LNK (10)		symbolic link
 * 		2	DT_BLK (6)		block dev
 * 		3	DT_CHR (4)		char dev
 * 		4	DT_FIFO (1)		fifo
 * 		5	DT_SOCK	(12)		socket
 * 	
 * 		7*)	(reserved for DT_REG/DT_DIR) 
 * 
 * 		*)Value 7 is reserved for regular file/dir (DT_REG/DT_DIR).
 * 		Normally ctime_cs would have 0-199 value to stand for 
 * 		up to 2sec. The value for DT_REG/DT_DIR is selected 
 * 		to be over this range to distinguish if file was created
 * 		under POSIX_ATTR or not.
 *
 *   - attr
 * 	FAT attribute	(val)		mapped attribute
 * 	------------------------------------------------
 * 	ATTR_RO		0x01 		!owner W
 * 	ATTR_HIDDEN	0x02		!owner R
 * 	ATTR_SYS	0x04
 * 	ATTR_VOLUME	0x08
 * 	ATTR_DIR	0x10		DIR
 * 	ATTR_ARCH	0x20		!owner X
 *
 *   - ctime
 * 		16bit word
 * 	f e d c b a 9 8 7 6 5 4 3 2 1 0
 * 	|=============| |-------------|
 * 	  major		  minor
 * 
 */

#define VFAT_POSIX_ATTR_VALID_MODE	(S_IFMT|S_IRWXU|S_IRWXG|S_IRWXO)

#define VFAT_CS_FMSK	0xe0
#define VFAT_CS_FSFT	5
#define VFAT_CS_FREG	0xe0

#define VFAT_CS_SPCF	0x01
#define VFAT_CS_UID	0x02
#define VFAT_CS_NXGRP	0x04
#define VFAT_CS_NWGRP	0x08
#define VFAT_CS_NRGRP	0x10

/* regular file/dir flag */
static inline int get_pxattr_regf(struct msdos_dir_entry *de)
{
	return (de->ctime_cs & VFAT_CS_FMSK) == VFAT_CS_FREG;
}
static inline void set_pxattr_regf(struct msdos_dir_entry *de, int val)
{
	val = val ? VFAT_CS_FMSK : 0;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_FMSK)));
}

/* file type */
static inline int get_pxattr_ftype(struct msdos_dir_entry *de)
{
	return ((de->ctime_cs & VFAT_CS_FMSK) >> VFAT_CS_FSFT);
}
static inline void set_pxattr_ftype(struct msdos_dir_entry *de, int val)
{
	val = (val  << VFAT_CS_FSFT) & VFAT_CS_FMSK;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_FMSK)));
}

/* special file flag */
static inline int get_pxattr_specf(struct msdos_dir_entry *de)
{
	return de->ctime_cs & VFAT_CS_SPCF;
}
static inline void set_pxattr_specf(struct msdos_dir_entry *de, int val)
{
	val = val ? VFAT_CS_SPCF : 0;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_SPCF)));
}

/* user r */
static inline int get_pxattr_ur(struct msdos_dir_entry *de)
{
	return !(de->attr & ATTR_HIDDEN);
}
static inline void set_pxattr_ur(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : ATTR_HIDDEN;
	de->attr = (val | (de->attr & ~ATTR_HIDDEN));
}

/* user w */
static inline int get_pxattr_uw(struct msdos_dir_entry *de)
{
	return !(de->attr & ATTR_RO);
}
static inline void set_pxattr_uw(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : ATTR_RO;
	de->attr = (val | (de->attr & ~ATTR_RO));
}

/* user x */
static inline int get_pxattr_ux(struct msdos_dir_entry *de)
{
	return !(de->attr & ATTR_ARCH);
}
static inline void set_pxattr_ux(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : ATTR_ARCH;
	de->attr = (val | (de->attr & ~ATTR_ARCH));
}

/* group r */
static inline int get_pxattr_gr(struct msdos_dir_entry *de)
{
	return !(de->ctime_cs & VFAT_CS_NRGRP);
}
static inline void set_pxattr_gr(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : VFAT_CS_NRGRP;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_NRGRP)));
}

/* group w */
static inline int get_pxattr_gw(struct msdos_dir_entry *de)
{
	return !(de->ctime_cs & VFAT_CS_NWGRP);
}
static inline void set_pxattr_gw(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : VFAT_CS_NWGRP;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_NWGRP)));
}

/* group x */
static inline int get_pxattr_gx(struct msdos_dir_entry *de)
{
	return !(de->ctime_cs & VFAT_CS_NXGRP);
}
static inline void set_pxattr_gx(struct msdos_dir_entry *de, int val)
{
	val = val ? 0 : VFAT_CS_NXGRP;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_NXGRP)));
}

/* user id */
static inline int get_pxattr_uid(struct msdos_dir_entry *de)
{
	return (de->ctime_cs & VFAT_CS_UID) != 0;
}
static inline void set_pxattr_uid(struct msdos_dir_entry *de, int val)
{
	val = val ? VFAT_CS_UID : 0;
	de->ctime_cs = (val | (de->ctime_cs & (~VFAT_CS_UID)));
}

/* driver major number */
static inline int get_pxattr_major(struct msdos_dir_entry *de)
{
	return ((le16_to_cpu(de->ctime) & 0xff00) >> 8);
}
static inline void set_pxattr_major(struct msdos_dir_entry *de, int val)
{
	val = (val & 0xff) << 8;
	de->ctime = cpu_to_le16((val | (le16_to_cpu(de->ctime) & 0x00ff)));
}

/* driver minor number */
static inline int get_pxattr_minor(struct msdos_dir_entry *de)
{
	return le16_to_cpu(de->ctime) & 0xff;
}
static inline void set_pxattr_minor(struct msdos_dir_entry *de, int val)
{
	val &= 0xff;
	de->ctime = cpu_to_le16(val | (le16_to_cpu(de->ctime) & 0xff00));
}

#ifdef FAT_DEBUG_PRINT
#define DPRINT printk
#else  /* FAT_DEBUG_PRINT */
#define DPRINT(format, args...) 
#endif /* FAT_DEBUG_PRINT */
#ifdef FAT_DEBUG_TRACE
#define DTRACE printk
#else  /* FAT_DEBUG_TRACE */
#define DTRACE(format, args...) 
#endif /* FAT_DEBUG_TRACE */

#ifdef FAT_SOFTUPDATE
#define FAT_SU_UD_DISC_CLSLINK       (0x100)
#define FAT_SU_UD_DISC_DIRENT        (0x101)
#define FAT_SU_UD_DISC_DIRENT_SIZE   (0x102)
/* Soft Update definision */
struct fat_softupdate_ud {
        struct list_head d_sb;
        struct list_head d_target;
        struct list_head d_depend;
        int flag_in_replace;
        int flag_ud_remove;
        int ud_disc;
        int entry_target;
        int entry_depend;
	loff_t i_pos_and_mask;
        loff_t size;
        struct fat_entry fat_ent_ud;
        struct inode *inode;
};
struct fat_softupdate_ud_head {
        struct super_block *sb;
        struct buffer_head *p_bh;
        struct list_head    d_sb_head;
        struct list_head    d_deppriv_head;
        struct list_head    d_target_head;
        struct list_head    d_depend_head;
};
struct fat_bh_private_data {
        struct fat_softupdate_ud_head ud;
};
struct fat_softupdate_depend_private {
        struct list_head    l_sb;
        struct list_head    l_inode;
        struct list_head    l_private;
        struct inode        *inode;
};
int fat_softupdate_test_free_private(struct buffer_head *bh);
int fat_softupdate_del_list_free_mem(struct buffer_head *bh, struct fat_softupdate_ud *p_clear_ud);
int fat_softupdate_rm_dependency(struct buffer_head *target_bhs, int rm_entry_target, int rm_entry_dep);
int fat_softupdate_test_rm_dependency(struct buffer_head *bh, int rm_entry_dep);
int fat_softupdate_rm_all_dependency(struct buffer_head *bh);
int fat_softupdate_add_dependency(struct inode *inode,
				  struct buffer_head *bh_target, int entry_target, 
				  struct buffer_head *bh_depend, int entry_depend);
int fat_softupdate_make_entry_UD(struct inode *inode, struct buffer_head *depend_bh, int depend_entry);
int fat_softupdate_list_entry_dependency(struct buffer_head *dirent_bh, struct inode *inode, struct msdos_dir_entry *raw_entry, int *wait);
int fat_softupdate_replace_ud_link(struct buffer_head *bhp);
void fat_end_buffer_write_sync(struct buffer_head *bh, int uptodate);
int fat_softupdate_make_dependency(struct inode *inode, struct fat_entry *target_ent, struct fat_entry *depend_ent);
/* procfs.c */
void fat_proc_unregister_global( const char *name );
int fat_proc_info_global_init( void );
int fat_UDList_in_proc( char *buffer, char **start, off_t offset,
			int count, int *eof, void *data );

#endif /* FAT_SOFTUPDATE */

#endif /* __KERNEL__ */

#endif
