/*
 * kernel/power/main.c - PM subsystem core functionality.
 *
 * Copyright (c) 2003 Patrick Mochel
 * Copyright (c) 2003 Open Source Development Lab
 *
 * Copyright 2005,2006 Sony Corporation
 * 
 * This file is released under the GPLv2
 *
 */

#include <linux/suspend.h>
#include <linux/kobject.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/pm.h>
#ifdef CONFIG_DEFERRED_RESUME
#include <linux/kthread.h>
#endif
#include <linux/device.h>
#include <linux/reboot.h>

#ifdef	CONFIG_SAFE_SUSPEND
extern int suspend_remount(void);
extern int resume_remount(void);
#endif

#include "power.h"

DECLARE_MUTEX(pm_sem);
atomic_t on_suspend = ATOMIC_INIT(0);

struct pm_ops * pm_ops = NULL;
suspend_disk_method_t pm_disk_mode = PM_DISK_SHUTDOWN;

/**
 *	pm_set_ops - Set the global power method table. 
 *	@ops:	Pointer to ops structure.
 */

void pm_set_ops(struct pm_ops * ops)
{
	down(&pm_sem);
	pm_ops = ops;
	up(&pm_sem);
}


/**
 *	suspend_prepare - Do prep work before entering low-power state.
 *	@state:		State we're entering.
 *
 *	This is common code that is called for each state that we're 
 *	entering. Allocate a console, stop all processes, then make sure
 *	the platform can enter the requested state.
 */

static int suspend_prepare(suspend_state_t state)
{
	int error = 0;

	if (!pm_ops || !pm_ops->enter)
		return -EPERM;

	pm_prepare_console();

	if (freeze_processes()) {
		error = -EAGAIN;
		goto Thaw;
	}

#ifdef	CONFIG_SAFE_SUSPEND
	(void) suspend_remount();
#endif
	if (pm_ops->prepare) {
		if ((error = pm_ops->prepare(state)))
			goto Thaw;
	}

	pr_debug("suspend devices\n");
	if ((error = device_suspend(PMSG_SUSPEND)))
		goto Finish;
	return 0;
 Finish:
	if (pm_ops->finish)
		pm_ops->finish(state);
 Thaw:
#ifdef	CONFIG_SAFE_SUSPEND
	(void) resume_remount();
#endif
	thaw_processes();
	pm_restore_console();
	return error;
}


static int suspend_enter(suspend_state_t state)
{
	int error = 0;
	unsigned long flags;

	local_irq_save(flags);

	if ((error = device_power_down(PMSG_SUSPEND)))
		goto Done;
	error = pm_ops->enter(state);
	device_power_up();
 Done:
	local_irq_restore(flags);
	return error;
}


/**
 *	suspend_finish - Do final work before exiting suspend sequence.
 *	@state:		State we're coming out of.
 *
 *	Call platform code to clean up, restart processes, and free the 
 *	console that we've allocated. This is not called for suspend-to-disk.
 */

static void suspend_finish(suspend_state_t state)
{
	device_resume();
	if (pm_ops && pm_ops->finish)
		pm_ops->finish(state);
#ifdef	CONFIG_SAFE_SUSPEND
	(void) resume_remount();
#endif
	thaw_processes();
	pm_restore_console();
}




char * pm_states[] = {
	[PM_SUSPEND_STANDBY]	= "standby",
	[PM_SUSPEND_MEM]	= "mem",
	[PM_SUSPEND_DISK]	= "disk",
	NULL,
};

#if defined(CONFIG_FAIL_SAFE_RESUME)
#define FAIL_SAFE_DEFAULT FAIL_SAFE_RESUME
#elif defined(CONFIG_FAIL_SAFE_REBOOT)
#define FAIL_SAFE_DEFAULT FAIL_SAFE_REBOOT
#else
#define FAIL_SAFE_DEFAULT FAIL_SAFE_HALT
#endif

enum {
	FAIL_SAFE_RESUME = 1,
	FAIL_SAFE_REBOOT,
	FAIL_SAFE_HALT,
	FAIL_SAFE_NUM
};
static unsigned fail_safe = FAIL_SAFE_DEFAULT;

#ifdef CONFIG_DEFERRED_RESUME

#include <linux/completion.h>

DECLARE_COMPLETION(resume_thread_wakeup);

struct task_struct *resume_thread_tsk = 0;
int enable_deferred_resume = 0;
u32 resume_state;


/*
 * We will replace pm_trace() with better instrumentation 
 * method, like LTT,  later.
 */

static int __tm_printk(const char *fmt, ...)
{
	char buf[512];
        va_list args;
        int i;
	unsigned long long tt;
	unsigned long us, ms;

	tt = sched_clock();
	ms = tt >> 10;  /* convert to usec */
	us = ms % 1000;
	ms = ms / 1000;

	i = sprintf(buf, "%6.6lu.%3.3lums:", ms,us);
        va_start(args, fmt);
        i = vsnprintf(buf+i, sizeof(buf)-i, fmt, args);
        va_end(args);
	printk(buf);

        return i;
}


void pm_trace(char *msg)
{
	static int last_cnt = 1;

	int c = pm_sem.count.counter;
	if (c == 0 || last_cnt != c)
		__tm_printk("[%3.3d]:pm_sem:%1.1d:<%s>:%s\n",
			current->pid,
			c,
			current->comm,
			msg);
	last_cnt = c;
}

#ifdef CONFIG_DEFERRED_RESUME_TEST
void pm_trace_force(char *msg)
{
	int c = pm_sem.count.counter;
	__tm_printk("[%3.3d]:pm_sem:%1.1d:<%s>:%s\n",
		    current->pid,
		    c,
		    current->comm,
		    msg);
	
}
#endif /* CONFIG_DEFERRED_RESUME_TEST */
#else
static inline void pm_trace(char *msg)
{
	return;
}
#endif /* CONFIG_DEFERRED_RESUME */

/**
 *	enter_state - Do common work of entering low-power state.
 *	@state:		pm_state structure for state we're entering.
 *
 *	Make sure we're the only ones trying to enter a sleep state. Fail
 *	if someone has beat us to it, since we don't want anything weird to
 *	happen when we wake up.
 *	Then, do the setup for suspend, enter the state, and cleaup (after
 *	we've woken up).
 */

static int enter_state(suspend_state_t state)
{
	int error;

	if (down_trylock(&pm_sem))
		return -EBUSY;

	atomic_inc(&on_suspend);
	smp_wmb();
#ifdef CONFIG_SOFTWARE_SUSPEND		
	if (state == PM_SUSPEND_DISK) {
		pr_debug("PM: Preparing system for suspend to disk\n");
		error = pm_suspend_disk();
		atomic_dec(&on_suspend);
		smp_wmb();
		goto Unlock;
	}
#endif

	/* Suspend is hard to get right on SMP. */
	if (num_online_cpus() != 1) {
		error = -EPERM;
		goto Unlock;
	}

	pr_debug("PM: Preparing system for suspend\n");
	if ((error = suspend_prepare(state))) {
		atomic_dec(&on_suspend);
		smp_wmb();
		goto Unlock;
	}

	pr_debug("PM: Entering state.\n");
	error = suspend_enter(state);
	atomic_dec(&on_suspend);
	smp_wmb();

#ifdef CONFIG_DEFERRED_RESUME
	pm_trace("Back");

	if (enable_deferred_resume && resume_thread_tsk) {
		if (!error) {
			extern void  thaw_process_group(int pgid);

			device_resume();

			/* wake up related threads */
			thaw_process_group(process_group(current));
			pm_trace("woke up minimal threads");

			/* wake up resume thread */
			resume_state=state;
			smp_wmb();
			complete(&resume_thread_wakeup);

			pr_debug("PM: Return without finishing up.\n");
			pm_trace("Return without finishing up");

			return 0;
		}
	}
#endif /* CONFIG_DEFERRED_RESUME */

	pr_debug("PM: Finishing up.\n");
	suspend_finish(state);

	pm_trace("Return with finishing up");

 Unlock:
	up(&pm_sem);

	if(error){
		switch(fail_safe){
		case FAIL_SAFE_REBOOT:
			machine_restart(NULL);
			break;
		case FAIL_SAFE_HALT:
			machine_halt();
			break;
		default:
			break;
		}
	}

	fail_safe = FAIL_SAFE_DEFAULT;
	return error;
}

/*
 * This is main interface to the outside world. It needs to be
 * called from process context.
 */
int software_suspend(void)
{
	return enter_state(PM_SUSPEND_DISK);
}


#ifdef CONFIG_DEFERRED_RESUME


static int resume_thread(void *state_p)
{
	struct k_sigaction sa;
	sigset_t blocked;
	u32 st;

	/* resume-thread is not freezable */
	current->flags |= PF_NOFREEZE;

	/* set scheduling params. */
	set_user_nice(current, 10);

	/* Block and flush all signals */
	sigfillset(&blocked);
	sigprocmask(SIG_BLOCK, &blocked, NULL);
	flush_signals(current);

	/* SIG_IGN makes children autoreap: see do_notify_parent(). */
	sa.sa.sa_handler = SIG_IGN;
	sa.sa.sa_flags = 0;
	siginitset(&sa.sa.sa_mask, sigmask(SIGCHLD));
	do_sigaction(SIGCHLD, &sa, (struct k_sigaction *)0);

	pm_trace("init");

	set_current_state(TASK_INTERRUPTIBLE);
	while (!kthread_should_stop()) {

		__set_current_state(TASK_RUNNING);

		pm_trace("waiting completion");

		wait_for_completion (&resume_thread_wakeup); /* wait until resume */
		st = *(u32 *) state_p;

		pm_trace("begin deferred finishing up");

#if defined(CONFIG_PRESERVE_SWSUSP_IMAGE)
		if (st == PM_SUSPEND_DISK) {
			pr_debug("PM(deferred, DISK): Finishing up.\n");
			finish_in_resume();
		} else {
			pr_debug("PM(deferred): Finishing up.\n");
			suspend_finish(st);
		}
#else
		pr_debug("PM(deferred): Finishing up.\n");
		suspend_finish(st);
#endif

		pm_trace("end deferred finishing up");

		up(&pm_sem);

		set_current_state(TASK_INTERRUPTIBLE);
	}
	__set_current_state(TASK_RUNNING);

	pm_trace("end");

	resume_thread_tsk = 0;
	enable_deferred_resume = 0;
	mb();
	return 0;
}

int __init resume_thread_init(void)
{
	struct task_struct *p = 0;
	p = kthread_create(resume_thread, &resume_state, "kresume");
	resume_thread_tsk = (IS_ERR(p)) ? NULL: p;
	wake_up_process(resume_thread_tsk);
	return 0;
}

__initcall(resume_thread_init);

#endif /* CONFIG_DEFERRED_RESUME */

/**
 *	pm_suspend - Externally visible function for suspending system.
 *	@state:		Enumarted value of state to enter.
 *
 *	Determine whether or not value is within range, get state 
 *	structure, and enter (above).
 */

int pm_suspend(suspend_state_t state)
{
	if (state > PM_SUSPEND_ON && state < PM_SUSPEND_MAX)
		return enter_state(state);
	return -EINVAL;
}



decl_subsys(power,NULL,NULL);


/**
 *	state - control system power state.
 *
 *	show() returns what states are supported, which is hard-coded to
 *	'standby' (Power-On Suspend), 'mem' (Suspend-to-RAM), and
 *	'disk' (Suspend-to-Disk).
 *
 *	store() accepts one of those strings, translates it into the 
 *	proper enumerated value, and initiates a suspend transition.
 */

static ssize_t state_show(struct subsystem * subsys, char * buf)
{
	int i;
	char * s = buf;

	for (i = 0; i < PM_SUSPEND_MAX; i++) {
		if (pm_states[i])
			s += sprintf(s,"%s ",pm_states[i]);
	}
	s += sprintf(s,"\n");
	return (s - buf);
}

static ssize_t state_store(struct subsystem * subsys, const char * buf, size_t n)
{
	suspend_state_t state = PM_SUSPEND_STANDBY;
	char ** s;
	char *p;
	int error;
	int len;

	p = memchr(buf, '\n', n);
	len = p ? p - buf : n;

	for (s = &pm_states[state]; state < PM_SUSPEND_MAX; s++, state++) {
		if (*s && !strncmp(buf, *s, len))
			break;
	}
	if (*s)
		error = enter_state(state);
	else
		error = -EINVAL;
	return error ? error : n;
}

power_attr(state);

#ifdef CONFIG_FAST_CLEAN_SHUTDOWN

#include <linux/reboot.h>
#include <linux/fs.h>

extern void tcp_shutdown_all(void); 

void fast_clean_shutdown(int fake_level)
{
	atomic_inc(&on_suspend);
	(void)freeze_processes();
	atomic_dec(&on_suspend);
	(void)suspend_remount();
	tcp_shutdown_all();

	if (fake_level) {
		atomic_inc(&on_suspend);
		thaw_processes();
		atomic_dec(&on_suspend);
		return;
	}

	device_shutdown(); /* try to flush cache on device side */
	machine_power_off();

	panic("fast_shutdown: can't power off\n");

	/* NOTREACHED */

}

static ssize_t fast_clean_shutdown_show(struct subsystem * subsys, char * buf)
{
	return sprintf(buf,"enabled\n");
}

static ssize_t fast_clean_shutdown_store(struct subsystem * subsys, const char * buf,
				  size_t n)
{

	int fake_level;
	char c;

	if (n>2) 
		return -EINVAL;
	c = buf[1];
	if (c!='\0' && c!='\n')
		return -EINVAL;

	fake_level = (buf[0] == '1' || buf[0] == 'y')  ? 0 : 1;
	fast_clean_shutdown(fake_level);

	return n;
}

power_attr(fast_clean_shutdown);


#endif /* CONFIG_FAST_CLEAN_SHUTDOWN */

#ifdef CONFIG_DEFERRED_RESUME

char * deferred_resume_states[] = {
	"disable",
	"enable",
	NULL
};


static ssize_t deferred_resume_show(struct subsystem * subsys, char * buf)
{
	char * s = buf;

	s += sprintf(s, deferred_resume_states[enable_deferred_resume]);
	s += sprintf(s, "\n");
	return (s - buf);
}

static ssize_t deferred_resume_store(struct subsystem * subsys, const char * buf, size_t n)
{
	char ** s;
	char *p;
	int len;
	int i;

	p = memchr(buf, '\n', n);
	len = p ? p - buf : n;

	i = 0;
	s = &deferred_resume_states[0]; 
	while (*s) {
		if (!strncmp(buf, *s, len))
			break;
		s++; i++;
	}

	if (!*s) {
		return -EINVAL;
	}

	if (resume_thread_tsk)
		enable_deferred_resume = i;
	return n;
}

power_attr(deferred_resume);

#endif /* CONFIG_DEFERRED_RESUME */


static ssize_t on_failure_show(struct subsystem * subsys, char * buf)
{
	static char *fst[3] = {"comeback","reboot","shutdown"};
	char *src = buf;
	int i;
	
	for(i=FAIL_SAFE_RESUME;i<FAIL_SAFE_NUM;i++)
		src += sprintf(src, "%s%s\n",fail_safe==i?"*":" ", fst[i-FAIL_SAFE_RESUME]);
	
	return (src - buf);
}

static ssize_t on_failure_store(struct subsystem * subsys, const char * buf, size_t n)
{
	char *endp;
	unsigned value;
	
	value = simple_strtoul(buf, &endp, 0);
	if(*endp || value<FAIL_SAFE_RESUME || value>=FAIL_SAFE_NUM)
		return -EINVAL;

	fail_safe = value;

	return n;
}

power_attr(on_failure);

static struct attribute * g[] = {
	&state_attr.attr,
#ifdef CONFIG_FAST_CLEAN_SHUTDOWN
	&fast_clean_shutdown_attr.attr,
#endif
#ifdef CONFIG_DEFERRED_RESUME 
	&deferred_resume_attr.attr,
#endif /* CONFIG_DEFERRED_RESUME */
	&on_failure_attr.attr,
	NULL,
};

static struct attribute_group attr_group = {
	.attrs = g,
};


static int __init pm_init(void)
{
	int error = subsystem_register(&power_subsys);
	if (!error)
		error = sysfs_create_group(&power_subsys.kset.kobj,&attr_group);
	return error;
}

core_initcall(pm_init);
