/*
 * drivers/power/process.c - Functions for starting/stopping processes on 
 *                           suspend transitions.
 *
 * Originally from swsusp.
 */


#undef DEBUG

#include <linux/smp_lock.h>
#include <linux/interrupt.h>
#include <linux/suspend.h>
#include <linux/module.h>

extern atomic_t on_suspend;
extern struct semaphore pm_sem;

/* 
 * Timeout for stopping processes
 */
#define TIMEOUT	(6 * HZ)


static inline int freezeable(struct task_struct * p)
{
	if ((p == current) || 
	    (p->flags & PF_NOFREEZE) ||
	    (p->exit_state == EXIT_ZOMBIE) ||
	    (p->exit_state == EXIT_DEAD) ||
	    (p->state == TASK_STOPPED) ||
	    (p->state == TASK_TRACED))
		return 0;
	return 1;
}

/* Refrigerator is place where frozen processes are stored :-). */
void refrigerator(unsigned long flag)
{
	/* Hmm, should we be allowed to suspend when there are realtime
	   processes around? */
	long save;
	save = current->state;
	current->state = TASK_UNINTERRUPTIBLE;

	/* Check refrigerator() on resume or after resume */
	if (!atomic_read(&on_suspend)) {
#ifdef DEBUG
		pr_debug("Strange: [%d] %s try to enter refrigerator again\n", 
			current->pid, current->comm);
		dump_stack();
#endif
		current->flags &= ~PF_FREEZE;
		mb();
		current->state = save;
		return;
	}

#ifdef CONFIG_DEFERRED_RESUME
	pr_debug("[%d] %s entered refrigerator\n", current->pid, current->comm);
#else
	pr_debug("%s entered refrigerator\n", current->comm);
#endif
	printk("=");
	current->flags &= ~PF_FREEZE;

	spin_lock_irq(&current->sighand->siglock);
	recalc_sigpending(); /* We sent fake signal, clean it up */
	spin_unlock_irq(&current->sighand->siglock);

	current->flags |= PF_FROZEN;
	while (current->flags & PF_FROZEN)
		schedule();
#ifdef CONFIG_DEFERRED_RESUME
	pr_debug("[%d] %s left refrigerator\n", current->pid, current->comm);
#else
	pr_debug("%s left refrigerator\n", current->comm);
#endif
	current->state = save;
}

/* 0 = success, else # of processes that we failed to stop */
static int freeze_processes_sub(int stop_kernel)
{
	int todo;
	unsigned long start_time;
	struct task_struct *g, *p;
	
	start_time = jiffies;
	do {
		todo = 0;
		read_lock(&tasklist_lock);
		do_each_thread(g, p) {
			unsigned long flags;
			if (!freezeable(p))
				continue;
			if (stop_kernel) {
				if (p->mm)
					continue;
			} else {
				if (!p->mm)
					continue;
			}
			if ((p->flags & PF_FROZEN) ||
			    (p->state == TASK_TRACED) ||
			    (p->state == TASK_STOPPED))
				continue;

			/* FIXME: smp problem here: we may not access other process' flags
			   without locking */
			p->flags |= PF_FREEZE;
			spin_lock_irqsave(&p->sighand->siglock, flags);
			signal_wake_up(p, 0);
			spin_unlock_irqrestore(&p->sighand->siglock, flags);
			todo++;
		} while_each_thread(g, p);
		read_unlock(&tasklist_lock);
		yield();			/* Yield is okay here */
		if (time_after(jiffies, start_time + TIMEOUT)) {
			printk( "\n" );
			printk(KERN_ERR " stopping tasks failed (%d tasks remaining)\n", todo );
			return todo;
		}
	} while(todo);
	return 0;
}

/* 0 = success, else # of processes that we failed to stop */
int freeze_processes(void)
{
	int ans;
	printk( "Stopping tasks: " );
	/* stop user threads first */
	if ((ans = freeze_processes_sub(0))) 
		return ans;
	if ((ans = freeze_processes_sub(1)))
		return ans;
	printk( "|\n" );
	BUG_ON(in_atomic());
	return 0;
}

void thaw_processes(void)
{
	struct task_struct *g, *p;

	printk( "Restarting tasks..." );
	/* preemption disabled to prevent user threads to run before
	 * kernel threads are resumed */
	preempt_disable();
	local_irq_disable();
	read_lock(&tasklist_lock);
	do_each_thread(g, p) {
		if (!freezeable(p))
			continue;
		if (p->flags & PF_FROZEN) {
			p->flags &= ~PF_FROZEN;
			wake_up_process(p);
			pr_debug("thaw: [%d] %s\n", p->pid, p->comm);
		} else
#ifdef CONFIG_DEFERRED_RESUME
			printk(KERN_INFO " [%d] %s not stopped\n", p->pid, p->comm );
#else
			printk(KERN_INFO " Strange, %s not stopped\n", p->comm );
#endif
	} while_each_thread(g, p);

	read_unlock(&tasklist_lock);
	local_irq_enable();
	preempt_enable();
	schedule();
	printk( " done\n" );
}

EXPORT_SYMBOL(refrigerator);

#ifdef CONFIG_DEFERRED_RESUME

/* Thaw processes blong to PGID process group.  */

void thaw_process_group(int pgid)
{
	struct task_struct *p;

	read_lock(&tasklist_lock);
	do_each_task_pid(pgid, PIDTYPE_PGID, p) {
		if (!freezeable(p))
			continue;
		if (p->flags & PF_FROZEN) {
			p->flags &= ~PF_FROZEN;
			wake_up_process(p);
			pr_debug("thaw pgid(%d): [%d] %s\n", pgid, p->pid, p->comm);
		} else {
			printk(KERN_INFO " Thaw pgid(%d): Strange, [%d] %s not stopped\n", 
				pgid, p->pid, p->comm);
		}
	} while_each_task_pid(pgid, PIDTYPE_PGID, p);
	read_unlock(&tasklist_lock);
}

#endif /* CONFIG_DEFERRED_RESUME */
