/*
 * include/linux/lz77_defalte.c
 * 
 * lz77 deflate routine
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 */
//#define DEBUG
//#define CHECK
#ifdef __KERNEL__
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/lz77.h>
EXPORT_SYMBOL(lz77_deflate);
#else
#include <stdio.h>
#include <string.h>
#include "lz77.h"
typedef unsigned char u8;
#endif

static inline int lz77_search_match(unsigned char *s1, unsigned char *s2, unsigned char *s_e)
{
	int len=0;
	while(s1<s_e){
		if(*s1++ == *s2++)
			len++;
		else
			break;
	}
	return len;
}

static inline void lz77_init_index(u8 *src, int len, void *workspace)
{
	short *hash = WORKSPACE_HASH(workspace);
	short *index = WORKSPACE_INDEX(workspace);
	int i;

	memset(workspace, 0xFF, LZ77_WORKSPACE_SIZE(len));

	for(i=0;i<=(len-3);i++){
		short v = HASH_VALUE(src[i],src[i+1], src[i+2]);
		short *h = hash+v;

		#ifdef CHECK
		if(v>=(1<<HASH_BITS)||v<0){
			lz77_warning("Warning: hash size too big!\n");
		}
		#endif
		
		index[i] = *h;
		*h = i;
	}

	#ifdef CHECK
	for(i=0;i<len;i++){
		if(index[i]>=len||index[i]<-1){
			lz77_warning("Warning: index size too big!\n");
		}
	}
	#endif
}

static inline void lz77_exit_index(void *workspace)
{
	return ;
}

static inline int lz77_search_index(u8 *ss, u8 *st, u8 *se, int *ml, void *workspace)
{
	int max_len = MIN_LEN-1;//at least 3 byte matching for not to lose
	int match_s = -1;
	short *index = WORKSPACE_INDEX(workspace);
	short off = st-ss;
	short i = index[off];
	
	se = se>(st+MAX_LEN)?(st+MAX_LEN):se;
	while(i>=0 && i>(off-WIN_LEN)){
		int m_l;
		m_l = lz77_search_match(st, ss+i, se);
		if(m_l > max_len){
			max_len = m_l;
			match_s = i;
			if(max_len>=OPT_LEN){
				break;
			}
		}
		
		i = index[i];
	}

	if(match_s<0)
		return -1;
	
	*ml = max_len;
	return (off-match_s);
}

/*
input:
  src       source data buffer
  len       length of source data 
  dst       destination buffer
  dst_len   length of destination buffer
            it has to be at least 4 byte longer
            than source length for uncompressable case
  workspace workspace buffer
  wl        length of workspace
output:
  mode(lz77/raw)
  section 0:
     type_bit * 8
     lit/codeword *8
        lit: literal ==> 1 byte
        codeword: size_bit*4 + addr_bit*12 ==> 2 byte
  section n
  end mark
*/
int lz77_deflate(unsigned char *src, int len, unsigned char *dst, int dst_len, void *workspace, int wl)
{
	int max_len, bd;
	unsigned char *d = dst;
	unsigned char *de = dst+dst_len-1;
	unsigned char *s = src;
	unsigned char *se = src+len;
	unsigned char *p;
	int i = 0;
		
	if(!src || !dst || dst_len<(len+4) || wl<LZ77_WORKSPACE_SIZE(len)){
		return -1;
	}

	lz77_init_index(src, len, workspace);
	
	*d++ = LZ77_COMPRESSED;
	p = d++;
	
	while(1){
		u8 type=0;
		
		while(i<8 && s<se && d<de){
			bd = lz77_search_index(src, s, src+len, &max_len, workspace);

			if(bd < 0){
				*d++ = *s++;
			}
			else{
				int size = size_to_index(max_len);
				type |= (1<<i);
				*d++ = ((size&0x0F)<<4) | ((bd&0xF00)>>8);
				*d++ = bd&0xFF;
				s += index_to_size(size);
			}

			i++;
		}

		*p = type;
		if(i==8){
			i = 0;
			p = d++;
			continue;
		}

		break;
	}

	if((d-dst)<len){
		*p |= (1<<i);
		*d++ = 0;
		*d++ = 0;
	}
		
	if((d-dst)>=len){//uncompressable
		int l=dst_len-4;
		
		lz77_printf("uncompressble -> fall back to raw data: %d %d->%d!\n",
		       dst_len, len, d-dst);
		
		l=l>len?len:l;
		d = dst;
		
		*d++ = LZ77_RAW;
		*d++ = 0;
		*d++ = l&0xFF;
		*d++ = l>>8;
		
		memcpy(d, src, l);
		d += l;
	}

	lz77_printf("delfated: %d %d->%d\n", dst_len, len, d-dst);

	lz77_exit_index(workspace);
	return (d-dst);
}
