/*--------------------------------------------------------------------------*
 * drivers/usb/mcd/usb_mcd_plat_uds.c
 * 
 * USB controller driver
 * 
 * Copyright 2005,2006 Sony Corporation
 * 
 * This code is based on plat_uds.c,
 * Copyright 2005 Mentor Graphics Corporation
 *
 * This file is part of the Inventra Controller Driver for Linux.
 *
 * The Inventra Controller Driver for Linux is free software; you
 * can redistribute it and/or modify it under the terms of the GNU
 * General Public License version 2 as published by the Free Software
 * Foundation.
 *
 * The Inventra Controller Driver for Linux is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with The Inventra Controller Driver for Linux ; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 * ANY DOWNLOAD, USE, REPRODUCTION, MODIFICATION OR DISTRIBUTION
 * OF THIS DRIVER INDICATES YOUR COMPLETE AND UNCONDITIONAL ACCEPTANCE
 * OF THOSE TERMS.THIS DRIVER IS PROVIDED "AS IS" AND MENTOR GRAPHICS
 * MAKES NO WARRANTIES, EXPRESS OR IMPLIED, RELATED TO THIS DRIVER.
 * MENTOR GRAPHICS SPECIFICALLY DISCLAIMS ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY; FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT.  MENTOR GRAPHICS DOES NOT PROVIDE SUPPORT
 * SERVICES OR UPDATES FOR THIS DRIVER, EVEN IF YOU ARE A MENTOR
 * GRAPHICS SUPPORT CUSTOMER.
 *---------------------------------------------------------------------------*/

/**
 * Introduction.
 * The ICD works like the other Linux HCDs/Gadgets: it is threadless,
 * so it does everything either in response to an interrupt, 
 * or during a call from an upper layer.
 * It implements a virtual root hub, so as to make uniform use
 * of the Linux hub driver.
 * 
 * The Linux (host-side) USB core has no concept of binding (the authors
 * apparently missed the point of the pipe discussion in the USB spec).
 * Instead, class drivers simply submit URBs, and an HCD may reject
 * or defer them if sufficient resources are not available.
 * This means class drivers have no way to know if their requirements 
 * can possibly be fulfilled, and may be blocked indefinitely by others,
 * without the end-user knowing why.
 * Therefore, whether things will work depends on the order of URB submissions
 * (which is dictated by the order of device insertion and/or driver loading
 * and thread scheduling).
 * 
 * The URB encodes pipe information in an integer, 
 * requiring table searches (hurting performance).
 * 
 * For the HDRC, local endpoint 0 is the only choice for control traffic,
 * so it is reprogrammed as needed, and locked during transfers.
 * Bulk transfers are queued to the available local endpoint with 
 * the smallest possible FIFO in the given direction
 * that will accomodate the transactions.
 * 
 * A typical response to the completion of a periodic URB is immediate
 * submission of another one, so the HCD does not assume it can reprogram
 * a local periodic-targetted endpoint for another purpose.
 * Instead, submission of a periodic URB is taken as a permanent situation,
 * so that endpoint is untouched.
 * One could imagine reprogramming periodic endpoints for other uses
 * between their polling intervals, effectively interleaving traffic on them.
 * Unfortunately, this assumes no device would ever NAK periodic tokens.
 * This is because the core no notification to software when an attempted
 * periodic transaction is NAKed (its NAKlimit feature is only for 
 * control/bulk).
 */

/*
 * Optional macros:
 * 
 * MUSB_DEBUG		0 => absolutely no diagnostics
 * 			1 => minimal diagnostics (basic operational states)
 * 			2 => 1 + detailed debugging of interface with USB core
 * 			3 => 2 + internal debugging (e.g. every register write)
 * 			4 => 3 + shared-IRQ-related checking
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/usb.h>

#include <asm/uaccess.h>
#include <linux/dmapool.h>

/*==================================================================*/
/*	Defined Macro													*/
/*==================================================================*/
#define TEST_BULKERR			/* for DEBUG */
#define	MUSB_DMA_BUS2VIRT 		/* for Bus Address */

/*==============================*/
/*	Macro in DMA			 	*/
/*==============================*/
// DMA_HOST_CODE			/* DMA code for host become valid */
// DMA_HOST_CODE_TX_IRQ	/* to enable DMA code in Tx IRQ */
// DMA_HOST_CODE_RX_IRQ	/* to enable DMA code in Rx IRQ */
// DMA_HOST_DPB_SET		/* to set a mode whether single or double buffer for fifo */
// HOST_NULL_FLAG		/* to enable null_flag to do Null-transfer*/
// FIFO_FLASH				/* to enable fifo flush */
// HOST_CLR_REQPKT		/* for clear ReqPkt in using AutoSet */
// OVER_DATA_ERR			/* to detect Over data */
// DMA_HOST_CODE_PROG		/* to enable DMA code in ProgramEnd() */

//#define CONSISTENT_SYNC_NEW	/* to flush cache for transfer_dma */
//#define DMA_HOST_UNLINK_CODE	/* to cancel ** incomplete!! ***/
#define MCD_PM_NON_CORRUPT		/* for deleting Message  */

#ifdef CONFIG_USB_DEBUG
    #define DEBUG 
#else
    #undef DEBUG
#endif

#include <linux/usb.h>
#ifndef MUSB_LINUX_MV21
#include "hcd.h"
#define HAS_USB_TT_MULTI
#endif

#define  MUSB_DYNFIFO_USED

#include "usb_mcd_defs.h"

#ifdef MUSB_GADGET
MGC_LinuxCd* MGC_pFirstDriver = NULL;
#endif

#include "usb_mcd_fifo.h"



#ifdef DMA_HOST_CODE

#include "usb_mcd_dma.h"
#endif
#if defined(CONSISTENT_SYNC)||defined(CONSISTENT_SYNC_NEW)
	#include<linux/dma-mapping.h>
#endif

#ifdef MUSB_V26
#define RETURN_IRQ_HANDLED return(IRQ_HANDLED)
#define RETURN_IRQ_NONE return(IRQ_NONE)
#endif

#ifdef MUSB_V24
typedef void irqreturn_t;
#define RETURN_IRQ_HANDLED return
#define RETURN_IRQ_NONE return
#endif

/****************************** CONSTANTS ********************************/

#define DRIVER_VERSION "1.04"
#define DRIVER_AUTHOR "Sony Corporation"
#define DRIVER_DESC "Inventra Controller Driver"

/** how much to "scale" response timeouts */
#define MUSB_MAX_RETRIES         4

/** time (millseconds) to wait before a restart */
#define MUSB_RESTART_TIME        5000

/** how many babbles to allow before giving up */
#define MUSB_MAX_BABBLE_COUNT    10


/******************************* FORWARDS ********************************/

/* Linux USBD glue */
#ifdef MUSB_V26
#if defined(CONFIG_USB_MCD_HOST)
STATIC int MGC_LinuxSubmitUrb(struct urb* pUrb, int iMemFlags);

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,5) ) 
STATIC int MGC_LinuxUnlinkUrb(struct urb* pUrb);
#else
STATIC int MGC_LinuxUnlinkUrb(struct urb* pUrb, int status);
#endif
STATIC void MGC_LinuxDisable(struct usb_device* , struct usb_host_endpoint *);
#endif /* CONFIG_USB_MCD_HOST */
#endif

#ifdef MUSB_V24
STATIC int MGC_LinuxSubmitUrb(struct urb* pUrb);
STATIC int MGC_LinuxUnlinkUrb(struct urb* pUrb);
#endif

#ifdef MCD_USE_ALLOCDEV
STATIC int MGC_LinuxAllocDevice(struct usb_device* pDevice);
#endif
#ifdef MCD_USE_FREEDEV
STATIC int MGC_LinuxFreeDevice(struct usb_device* pDevice);
#endif
#if defined(CONFIG_USB_MCD_HOST)
STATIC int MGC_LinuxGetFrame(struct usb_device* pDevice);

/* virtual hub glue */
STATIC void MGC_LinuxSetPortPower(void* pPrivateData, uint8_t bPortIndex,
				  uint8_t bPower);
STATIC void MGC_LinuxSetPortEnable(void* pPrivateData, uint8_t bPortIndex,
				   uint8_t bEnable);
STATIC void MGC_LinuxSetPortSuspend(void* pPrivateData, uint8_t bPortIndex,
				    uint8_t bSuspend);
STATIC void MGC_LinuxSetPortReset(void* pPrivateData, uint8_t bPortIndex,
				  uint8_t bReset);

/* driver functions */
#endif
STATIC MGC_LinuxCd* MGC_LinuxInitController(void* pDevice, uint16_t wType, 
	int nIpIrq, int nDmaIrq, void* pRegs, const char* pName);
#ifdef MCD_USE_SHUTDOWN
STATIC void MGC_LinuxShutdownController(MGC_LinuxCd* cd);
#endif
STATIC irqreturn_t MGC_LinuxIsr(int irq, void *__hci, struct pt_regs *r);
#if defined(CONFIG_USB_MCD_HOST)
/* HCD helpers */
STATIC uint8_t MGC_FindFirst1(unsigned int nValue);
STATIC void MGC_LinuxSetTimer(MGC_LinuxCd* pThis, 
	void (*pfFunc)(unsigned long),
	unsigned long pParam, unsigned long millisecs);
STATIC void MGC_LinuxStartNextUrb(MGC_LinuxCd* pThis, uint8_t bEnd);
STATIC void MGC_LinuxStartUrb(MGC_LinuxCd* pThis, uint8_t bEnd);
STATIC int MGC_LinuxFindEnd(MGC_LinuxCd* pThis, struct urb* pUrb);
STATIC uint8_t MGC_LinuxPacketRx(MGC_LinuxCd* pThis, uint8_t bEnd,
	uint8_t bIsochError);

/* HDRC functions */
STATIC void MGC_HdrcResetOff(unsigned long param);
STATIC void MGC_HdrcDropResume(unsigned long pParam);
#ifdef MCD_USE_RESTART 
STATIC void MGC_HdrcRestart(unsigned long pParam);
#endif
STATIC void MGC_HdrcProgramEnd(MGC_LinuxCd* pThis, uint8_t bEnd, 
	struct urb* pUrb, unsigned int nOut,
	uint8_t* pBuffer, uint32_t dwLength);
STATIC void MGC_HdrcStopEnd(MGC_LinuxCd* pThis, uint8_t bEnd);
STATIC void MGC_HdrcStartTx(MGC_LinuxCd* pThis, uint8_t bEnd);
STATIC uint8_t MGC_HdrcServiceHostDefault(MGC_LinuxCd* pThis, 
	uint16_t wCount, struct urb* pUrb);
STATIC void MGC_HdrcServiceDefaultEnd(MGC_LinuxCd* pThis);
STATIC void MGC_HdrcServiceTxAvail(MGC_LinuxCd* pThis, uint8_t bEnd);
STATIC void MGC_HdrcServiceRxReady(MGC_LinuxCd* pThis, uint8_t bEnd);
#endif
STATIC void MGC_HdrcDisable(MGC_LinuxCd* pThis);
STATIC uint8_t MGC_HdrcInit(uint16_t wType, MGC_LinuxCd* pThis);

#if defined(CONFIG_USB_MCD_HOST)
STATIC void* MGC_LinuxBufferAlloc(struct usb_bus* pBus, size_t nSize,
	int iMemFlags, dma_addr_t* pDmaAddress);
STATIC void MGC_LinuxBufferFree(struct usb_bus* pBus, size_t nSize,
	void* address, dma_addr_t DmaAddress);
#endif
uint8_t usb_mcd_buffer_create(MGC_LinuxCd*);
void usb_mcd_buffer_destroy(MGC_LinuxCd*);

#ifdef DMA_HOST_CODE
void USB_MCD_HostDmaCallback(MGC_DmaChannel* pChannel,uint8_t bEnd, uint8_t bTransmit, uint8_t bProtocol, void* pData);
#endif
#if MUSB_DEBUG > 0
STATIC void MGC_HdrcDumpRegs(MGC_LinuxCd* pThis, uint8_t bEnd);
#endif

#if defined(DMA_DEBUG_DATA)||defined(DMA_DEBUG_RX_DATA)
/* for DMA data check */
STATIC void usb_mcd_print_dma(uint8_t* pBuffer, uint32_t dwLength, uint32_t dwOffset);
#endif
#if defined(CONFIG_USB_MCD_HOST)
#define USB_MCD_MAXP(wMaxPacketSize) ((wMaxPacketSize) & 0x07ff)
STATIC inline uint16_t usb_mcd_maxpacket(struct usb_device *udev, int pipe, int is_out);
#endif
#ifdef MCD_PM_NON_CORRUPT
int usb_mcd_set_padnum(MGC_LinuxCd *);
#endif

/******************************* GLOBALS *********************************/

static const char	MGC_HcdName [] = "musb-hcd";
static unsigned int MGC_nIndex = 0;

#if defined(CONFIG_USB_MCD_HOST)
/* Linux USBD calls these */
static struct usb_operations MGC_LinuxOperations = 
{
	.get_frame_number =	MGC_LinuxGetFrame,
	.submit_urb =		MGC_LinuxSubmitUrb,
	.unlink_urb =		MGC_LinuxUnlinkUrb,
	.buffer_alloc =		MGC_LinuxBufferAlloc,
	.buffer_free =		MGC_LinuxBufferFree,
	.disable =			MGC_LinuxDisable,
	.hub_suspend =		NULL,
	.hub_resume	=		NULL
};
#endif
/*======================================================================*/
/* 	list used in setting dynamic fifos for Host 						*/
/*======================================================================*/
USB_MCD_DFIFO_LIST	usb_mcd_dfifo_setlist[USB_MCD_DFIFO_TOTAL_EP+1] /*include EP0 */
 = {
 	  /* ep_num | ep_dir(except EP0),	type,				fifo_size,				dpb(double buffer)	*/
	{ USB_MCD_DFIFO_EP0,				PIPE_CONTROL,		USB_MCD_DFIFO_CTRL_SZ,	USB_MCD_DFIFO_CTRL_DPB	},
	{ USB_MCD_DFIFO_EP1 | USB_DIR_IN,	PIPE_BULK,			USB_MCD_DFIFO_BULK_SZ,	USB_MCD_DFIFO_BULK_DPB	},
	{ USB_MCD_DFIFO_EP2 | USB_DIR_OUT,	PIPE_BULK,			USB_MCD_DFIFO_BULK_SZ,	USB_MCD_DFIFO_BULK_DPB	},
	{ USB_MCD_DFIFO_EP3 | USB_DIR_IN,	PIPE_INTERRUPT,		USB_MCD_DFIFO_INT_SZ,	USB_MCD_DFIFO_INT_DPB	},
	{ USB_MCD_DFIFO_EP4 | USB_DIR_OUT,	PIPE_INTERRUPT,		USB_MCD_DFIFO_INT_SZ,	USB_MCD_DFIFO_INT_DPB	},
	{ USB_MCD_DFIFO_EP5 | USB_DIR_IN,	PIPE_ISOCHRONOUS,	USB_MCD_DFIFO_ISO_SZ,	USB_MCD_DFIFO_ISO_DPB	},
	{ USB_MCD_DFIFO_EP6 | USB_DIR_OUT,	PIPE_ISOCHRONOUS,	USB_MCD_DFIFO_ISO_SZ,	USB_MCD_DFIFO_ISO_DPB	}
 };	

/**************************************************************************
 * Glue for virtual root hub
**************************************************************************/

#if defined(CONFIG_USB_MCD_HOST)
/* see virthub.h */
STATIC void MGC_LinuxSetPortPower(void* pPrivateData, uint8_t bPortIndex,
    uint8_t bPower)
{
    unsigned long flags;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pPrivateData;

    DBG(2, "<==\n");

    spin_lock_irqsave(&pThis->Lock, flags);

    if(bPower)
    {
        DBG(1, "Root port power on\n");
	MGC_HdrcStart(pThis);
    }
    else
    {
        DBG(1, "Root port power off\n");
	MGC_HdrcStop(pThis);
    }

    spin_unlock_irqrestore(&pThis->Lock, flags);
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/* see virthub.h */
STATIC void MGC_LinuxSetPortEnable(void* pPrivateData, uint8_t bPortIndex,
    uint8_t bEnable)
{    
    DBG(2, "<==\n");
    if (bEnable)
    {
	DBG(1, "Root port power enabled\n");
    }
    else
    {
    	DBG(1, "Root port power disabled\n");
    }
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/* see virthub.h */
STATIC void MGC_LinuxSetPortSuspend(void* pPrivateData, uint8_t bPortIndex,
	uint8_t bSuspend)
{
    uint8_t power;
    unsigned long flags;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pPrivateData;
    void* pBase = pThis->pRegs;

    DBG(2, "<==\n");
    
    spin_lock_irqsave(&pThis->Lock, flags);
    power = MGC_Read8(pBase, MGC_O_HDRC_POWER);

    if(bSuspend)
    {
        DBG(1, "Root port power suspended\n");
	MGC_Write8(pBase, MGC_O_HDRC_POWER, power | MGC_M_POWER_SUSPENDM);
    }
    else if(power & MGC_M_POWER_SUSPENDM)
    {
        DBG(1, "Root port power resumed\n");
	power &= ~(MGC_M_POWER_SUSPENDM | MGC_M_POWER_RESUME);
	MGC_Write8(pBase, MGC_O_HDRC_POWER, power | MGC_M_POWER_RESUME);
	WAIT_MS(20);
	MGC_Write8(pBase, MGC_O_HDRC_POWER, power);
    }
    
    spin_unlock_irqrestore(&pThis->Lock, flags);
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/* see virthub.h */
STATIC void MGC_LinuxSetPortReset(void* pPrivateData, uint8_t bPortIndex,
	uint8_t bReset)
{
    uint8_t power;
    unsigned long flags;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pPrivateData;
    void* pBase = pThis->pRegs;

    spin_lock_irqsave(&pThis->Lock, flags);
    power = MGC_Read8(pBase, MGC_O_HDRC_POWER) & 0xf0;

    DBG(2, "<==\n");
    
    if(bReset)
    {
        DBG(2, "Starting root port reset...\n");
	MGC_Write8(pBase, MGC_O_HDRC_POWER, power | MGC_M_POWER_RESET);
	MGC_LinuxSetTimer(pThis, MGC_HdrcResetOff, (unsigned long)pThis, 60);
    } else if(power & MGC_M_POWER_RESET) {
        DBG(2, "root port reset stopped\n");
	MGC_Write8(pBase, MGC_O_HDRC_POWER, power & ~MGC_M_POWER_RESET);
    }

    spin_unlock_irqrestore(&pThis->Lock, flags);
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/* Root speed need to be translated (addapted)
 */
STATIC uint8_t MGC_TranslateVirtualHubSpeed(uint8_t source) {
    uint8_t speed=2;
    
    switch ( source ) {
     	case 3: speed=0; break;
     	case 2: speed=1; break;
    }

    return speed;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Timer completion callback to turn off reset and get connection speed
 */
STATIC void MGC_HdrcResetOff(unsigned long param)
{
    uint8_t power;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)param;
    void* pBase = pThis->pRegs;

    DBG(2, "Stopping root port reset...\n");

    power = MGC_Read8(pBase, MGC_O_HDRC_POWER);
    MGC_Write8(pBase, MGC_O_HDRC_POWER, power & ~MGC_M_POWER_RESET);

    /* check for high-speed and set in root device if so */
    power = MGC_Read8(pBase, MGC_O_HDRC_POWER);
    if(power & MGC_M_POWER_HSMODE)
    {
	DBG(1, "high-speed device connected\n");
	pThis->bRootSpeed = 1;
    }

    MGC_VirtualHubPortResetDone(&(pThis->RootHub), 0, 
    	MGC_TranslateVirtualHubSpeed(pThis->bRootSpeed));
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Timer completion callback to finish resume handling started in ISR
 */
STATIC void MGC_HdrcDropResume(unsigned long pParam)
{
    uint8_t power;
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pParam;
    void* pBase = pThis->pRegs;
	
    DBG(2, "<==\n");

    power = MGC_Read8(pBase, MGC_O_HDRC_POWER);
    MGC_Write8(pBase, MGC_O_HDRC_POWER, power & ~MGC_M_POWER_RESUME);
    MGC_VirtualHubPortResumed(&(pThis->RootHub), 0);
}
#endif

#ifdef MCD_USE_RESTART
/**
 * Timer completion callback to request session again
 * (to avoid self-connecting)
 */
STATIC void MGC_HdrcRestart(unsigned long pParam)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pParam;

    MGC_HdrcStart(pThis);
}
#endif
/**************************************************************************
 * HDRC functions
**************************************************************************/
/**
 * Load an HDRC FIFO
 *
 * @param pBase base address of HDRC
 * @param bEnd local endpoint
 * @param wCount how many bytes to load
 * @param pSource data buffer
 */
void MGC_HdrcLoadFifo(const uint8_t* pBase, uint8_t bEnd, 
	uint16_t wCount, const uint8_t* pSource)
{
    uint16_t wIndex, wIndex32;
    uint16_t wCount32 = wCount >> 2;
    uint8_t bFifoOffset = MGC_FIFO_OFFSET(bEnd);

    DBG(1, "pBase=%p, bEnd=%d, wCount=0x%04x, pSrc=%p\n", \
	pBase, bEnd, wCount, pSource);

    /* doublewords when possible */
    for(wIndex = wIndex32 = 0; wIndex32 < wCount32; wIndex32++, wIndex += 4)
    {
        MGC_Write32(pBase, bFifoOffset, *((uint32_t*)&(pSource[wIndex])));
    }
    for(; wIndex < wCount; wIndex++)
    {
        MGC_Write8(pBase, bFifoOffset, pSource[wIndex]);
    }
#ifdef DEBUG_TX_FIFODATA
	usb_mcd_print_dma(pSource, wCount, 0);
#endif    
}
/**
 * Unload an HDRC FIFO
 *
 * @param pBase base address of HDRC
 * @param bEnd local endpoint
 * @param wCount how many bytes to unload
 * @param pDest data buffer
 */
void MGC_HdrcUnloadFifo(const uint8_t* pBase, uint8_t bEnd, 
	uint16_t wCount, uint8_t* pDest)
{
    uint16_t wIndex=0, wIndex32;
    uint16_t wCount32 = wCount >> 2;
    uint8_t bFifoOffset = MGC_FIFO_OFFSET(bEnd);

    DBG(2, "pBase=%p, bEnd=%d, wCount=0x%04x, pDest=%p\n", pBase, bEnd, \
	wCount, pDest);

    /* doublewords when possible */
    for(wIndex = wIndex32 = 0; wIndex32 < wCount32; wIndex32++, wIndex += 4)
    {
	    *((uint32_t*)&(pDest[wIndex])) = MGC_Read32(pBase, bFifoOffset);
    }
    
    while(wIndex < wCount)
    {
	    pDest[wIndex++]=MGC_Read8(pBase, bFifoOffset);
    }
#ifdef DEBUG_RX_FIFODATA
	usb_mcd_print_dma(pDest, wCount, 0);
#endif
}
#if defined(CONFIG_USB_MCD_HOST)
/*==================================================================*/
/**
 * Program an HDRC endpoint as per the given URB
 * @param pThis instance pointer
 * @param bEnd local endpoint
 * @param pURB URB pointer
 * @param nOut zero for Rx; non-zero for Tx
 * @param pBuffer buffer pointer
 * @param dwLength how many bytes to transmit or expect to receive
 */
/*==================================================================*/
STATIC void MGC_HdrcProgramEnd(MGC_LinuxCd* pThis, uint8_t bEnd, 
	struct urb* pUrb, unsigned int nOut,
	uint8_t* pBuffer, uint32_t dwLength)
{
    uint16_t wCsr, wLoadCount, wIntrTxE;
    struct usb_device* pParent;
#ifdef MUSB_LINUX_MV21
    int nChild;
    struct usb_device* pDevice;
#endif
#ifdef DMA_HOST_CODE
	uint32_t			dmaLength = 0;
    MGC_DmaController*	pDmaController	= NULL;
    MGC_DmaChannel* 	pDmaChannel		= NULL;
    uint8_t 			bAllocChannel;
    uint8_t 			bDmaOk = FALSE;	
	USB_MCD_DmaChStatus*	pDmaChStatus = pThis->aLocalEnd[bEnd].pChEpStatus;
#endif
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    unsigned int nPipe = pUrb->pipe;
    uint16_t wPacketSize = USB_MCD_MAXP(usb_mcd_maxpacket(pUrb->dev, nPipe, nOut));
    uint8_t bIsBulk = usb_pipebulk(nPipe);
    uint8_t bAddress = (uint8_t)usb_pipedevice(nPipe);
    uint8_t bRemoteEnd = (uint8_t)usb_pipeendpoint(nPipe);
    uint8_t bSpeed = (uint8_t)pUrb->dev->speed;
    uint8_t bInterval = (uint8_t)pUrb->interval;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
    uint8_t bStdType = 0;
    uint8_t bHubAddr = 0;
    uint8_t bHubPort = 0;
    uint8_t reg = 0;
    uint8_t bIsMulti = FALSE;
    uint8_t bDone = FALSE;
    unsigned long flags;
    
    DBG(2, "<== (bEnd=%d, pUrb=%p)\n", bEnd, pUrb);

    /* NOTE: there is always a parent due to the virtual root hub */
    /* parent hub address */
    pParent = pUrb->dev->parent;
    bHubAddr = (uint8_t)pParent->devnum;
    if(bHubAddr == pThis->RootHub.bAddress)
    {
        /* but not if parent is our virtual root hub */
	bHubAddr = 0;
    }

    /* parent port */
#ifdef MUSB_LINUX_MV21
    if((USB_SPEED_HIGH != bSpeed) && 
       (pParent->devnum != pThis->RootHub.bAddress))
    {
        /* walk up to first high-speed hub and remember our subtree's child */
	pDevice = pUrb->dev;
	while(pParent && (USB_SPEED_HIGH != pParent->speed) &&
	      (pParent->devnum != pThis->RootHub.bAddress))
	{
	    pDevice = pParent;
	    pParent = pParent->parent;
	}
	if(pParent && (pParent->devnum != pThis->RootHub.bAddress))
	{
	    /* correlate to port and check for multi-TT */
	    for(nChild = 0; nChild < pParent->maxchild; nChild++)
	    {
	        if(pParent->children[nChild] == pDevice)
		{
		    bHubPort = nChild + 1;
		    bIsMulti = (2 == pParent->descriptor.bDeviceProtocol);
		    break;
		}
	    }
	}
    }
#else
    /* if tt pointer, use its info */
    if(pUrb->dev->tt)
    {
	bHubPort = (uint8_t)pUrb->dev->ttport;
#ifdef HAS_USB_TT_MULTI
	bIsMulti = (uint8_t)pUrb->dev->tt->multi;
#endif
    }
#endif

    DBG(1, "end %d, device %d, parent %d, port %d, multi-tt: %d, speed:%d\n", \
	bEnd, pUrb->dev->devnum, bHubAddr, bHubPort, bIsMulti, pUrb->dev->speed ); 

    /* prepare endpoint registers according to flags */
    if(usb_pipeisoc(nPipe))
    {
#ifdef	DMA_HOST_CODE
    	pDmaChStatus->mode = DMA_HOST_MODE0;	/* DMA⡼ ⡼0 */
#endif
        bStdType = 1;
	if(pUrb->interval > 16)
	{
	    /* correct interval */
	    bInterval = MGC_FindFirst1(pUrb->interval);
	}
	if(bInterval < 1)
	{
	    bInterval = 1;
	}
    }
    else if(usb_pipeint(nPipe))
    {
#ifdef	DMA_HOST_CODE
    	pDmaChStatus->mode = DMA_HOST_MODE0;	/* DMA⡼ ⡼0 */
#endif
        bStdType = 3;
	if((USB_SPEED_HIGH == bSpeed) && (pUrb->interval > 255))
	{
	    /* correct interval for high-speed */
	    bInterval = MGC_FindFirst1(pUrb->interval);
	}
	if(bInterval < 1)
	{
	    bInterval = 1;
	}
    }
    else if(bIsBulk)
    {
#ifdef	DMA_HOST_CODE
#ifdef	DMA_BULK_TEST_MODE0
    	pDmaChStatus->mode = DMA_HOST_MODE0;	/*	DMA⡼ ⡼0 (ƥ) */
#else
    	pDmaChStatus->mode = DMA_HOST_MODE1;	/*	DMA⡼ ⡼ɣ(߷׻) */
#endif
#endif
        bStdType = 2;
	if(pUrb->interval > 16)
	{
	    /* correct interval */
	    bInterval = MGC_FindFirst1(pUrb->interval);
	}
    }
    else
    {
        nOut = 1;
	if(pUrb->interval > 16)
	{
	    /* correct interval */
	    bInterval = MGC_FindFirst1(pUrb->interval);
	}
    }
    reg = bStdType << 4;
    /* really NAKlimit in this case */
    if(bInterval < 2)
    {
	bInterval = 16;
    }

    reg |= (bRemoteEnd & 0xf);
    if(pThis->bIsMultipoint)
    {
	switch(bSpeed)
	{
	  case USB_SPEED_LOW:
	    reg |= 0xc0;
	    break;
	  case USB_SPEED_FULL:
	    reg |= 0x80;
	    break;
	  default:
	    reg |= 0x40;
	}
    }


    if(bIsBulk && pThis->bBulkSplit)
    {
    	wLoadCount = min((uint32_t)pEnd->wMaxPacketSizeTx, dwLength);
    }
    else 
    {
    	wLoadCount = min((uint32_t)wPacketSize, dwLength);
    }
	DBG(2,"PacketSize = %d wMaxPacketSizeTx = %d \n",dwLength, pEnd->wMaxPacketSizeTx);

#if  defined(DMA_HOST_CODE) && defined(DMA_HOST_CODE_PROG)
	DBG(1,"programEnd transfer_buffer_length=%d pDmaChStatus->dma_req_flag=%d \n"\
		,pUrb->transfer_buffer_length, pDmaChStatus->dma_req_flag);
	if(bEnd && (pUrb->transfer_buffer_length >0)){		/* ž¸ߤ롣*/
	/*==================================================================*/
	/* Mode1Ǥ RxžMAXP礭 */
		if(pDmaChStatus->dma_req_flag && (\
			((( usb_pipein(nPipe) && (pUrb->transfer_buffer_length > wPacketSize))\
			|| ( usb_pipeout(nPipe) && (pUrb->transfer_buffer_length >= wPacketSize)))\
				&& (pDmaChStatus->mode == DMA_HOST_MODE1))||\
					(pDmaChStatus->mode == DMA_HOST_MODE0))){	
			/* (MAXPʾ夫DMA⡼ɣˤޤDMA⡼ɣξ */

			/* candidate for DMA */
			bAllocChannel	= FALSE;
			pDmaController	= pThis->pDmaController;		/* get a operation's info  */
			pDmaChannel		= pEnd->pDmaChannel;			/* get a channel info to use in DMA*/
			if(pDmaController && !pDmaChannel){

				/* Allocate a channnel for DMA */
				pDmaChannel
                	= pEnd->pDmaChannel
                    	= pDmaController->pfDmaAllocateChannel(
                        	pDmaController, bEnd, (uint8_t)pEnd->bIsTx, bStdType, wPacketSize);
 
 	    		bAllocChannel = TRUE;
			}
			if(pDmaChannel){	/* if a channel is allocated */
				
				pDmaChStatus->dma_ch_flag = USB_MCD_FLAG_ON;	/* dma_ch_flag set */

				switch(pDmaChStatus->mode){
					case DMA_HOST_MODE0:
						
						/* get a size for DMA transfer */
						/* dwLength <-  pUrb->transfer_buffer_length */
						dmaLength =  (uint32_t)min(dwLength, (uint32_t)wPacketSize);
						pEnd->dwRequestSize = dmaLength;	/* žϿ */
						
						if(usb_pipein(nPipe)){
							pDmaChStatus->State	=	DMA_HOST_ST_CPU;	/* to CPU-mode	*/
							DBG(2,"Change to State CPU in DMAmode0 IN \n");
						}else if(usb_pipeout(nPipe)){
							pDmaChStatus->State	=	DMA_HOST_ST_DMA;	/* to DMA-mode	*/
							DBG(2,"Change to State DMA in DMAmode0 OUT \n");
						}else{
							printk("Error \n");
						}
						break;
					
					case DMA_HOST_MODE1:
						/* get a size for DMA transfer */
						/* dwLength <-  pUrb->transfer_buffer_length */
						
						if(usb_pipein(nPipe)){
							/* 1ѥåȾʤå */
							dmaLength =  (dwLength%wPacketSize)?((dwLength/wPacketSize)*wPacketSize):\
							((dwLength/wPacketSize)-1)*wPacketSize;
							dmaLength =  min(dmaLength, pDmaChannel->dwMaxLength); /* not used in case of mode1! */ 
						}else if(usb_pipeout(nPipe)){
							dmaLength =  min(dwLength, pDmaChannel->dwMaxLength);
						}
						pEnd->dwRequestSize =	dmaLength;			/* žϿ */
						pDmaChStatus->State	=	DMA_HOST_ST_DMA;	/* to DMA-mode */
						DBG(2,"Change to State DMA in DMAmode1 \n");
						break;
					case DMA_HOST_MODE_NON:
						printk("abnormal Error! \n");
						break;
				}
				/* CallBack */
				pDmaChannel->pfDmaCallback = USB_MCD_HostDmaCallback;
				/* CallBackΥǡpfDmaProgramChannelˤƥå */

				if(pDmaChStatus->State==DMA_HOST_ST_DMA){
				/* DMACž */
					bDmaOk = pDmaController->pfDmaProgramChannel(
                		pDmaChannel, (MGC_pfDmaCallback)(pDmaChannel->pfDmaCallback), (void*)pThis,
                			wPacketSize, pDmaChStatus->mode, __virt_to_bus(pBuffer), dmaLength);
	    			if(bDmaOk){
						pDmaChannel->dwActualLength = 0L;	/*  */
						wLoadCount = 0;						/* CPUžȤζ̽ѤΤDMAԤ0ˤ */
	    			}
	    			else if(bAllocChannel){
	    			/* DMA꼺Ԥ硢ĥͥƤϥͥ */
						pDmaController->pfDmaReleaseChannel(pDmaChannel);
						pEnd->pDmaChannel = NULL;
						pDmaChStatus->dma_ch_flag = USB_MCD_FLAG_OFF;		/* dma_ch_flag clear */
						bAllocChannel = FALSE;
	    			}
	    		}
			}else{	/* DMA channel not exist!	*/
				pDmaChStatus->State	=	DMA_HOST_ST_CPU;	/* set to CPU-mode */
				DBG(2,"could not get a DMAchannel and to State CPU \n");
			}
		}else{
		/* DMAtransfer can not do ! */
			pDmaChStatus->State	=	DMA_HOST_ST_CPU;	/* set to CPU-mode */
			DBG(2,"Change to State CPU \n");
		}
	}else if(bEnd){	/* transfer length is NULL */
	/*==================================================================*/
	/* NULLž */
		pDmaChStatus->State	=	DMA_HOST_ST_NULL;	/* to NULL-mode */
		DBG(2,"Change to State NULL \n");
		/* žλ */
		bDone = TRUE;
	}
#endif /* DMA_HOST_CODE end */

    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, bEnd);
    wCsr = MGC_M_TXCSR_MODE;

    /* transmit or receive? */
    if(nOut)
    {
        /* transmit */
		/* disable interrupt in case we flush */
		wIntrTxE = MGC_Read16(pBase, MGC_O_HDRC_INTRTXE);
		MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, wIntrTxE & ~(1 << bEnd));
        if(bEnd)
		{
	    	/* general endpoint */
	    	/* if not ready, flush and restore data toggle */
	    	if(!pEnd->bIsReady && pThis->bIsMultipoint)
	    	{
				pEnd->bIsReady = TRUE;
				/* twice in case of double packet buffering */
				MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, 
			       MGC_M_TXCSR_FLUSHFIFO | MGC_M_TXCSR_CLRDATATOG);
				MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, 
			       MGC_M_TXCSR_FLUSHFIFO | MGC_M_TXCSR_CLRDATATOG);
				/* data toggle */
				wCsr |= MGC_M_TXCSR_H_WR_DATATOGGLE;
				if(usb_gettoggle(pUrb->dev, pEnd->bEnd, 1))
				{
		    		wCsr |= MGC_M_TXCSR_H_DATATOGGLE;
				}
				else
				{
		    		wCsr &= ~MGC_M_TXCSR_H_DATATOGGLE;
				}
	    	}
		}
		else
		{
	    	/* endpoint 0: just flush */
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, bEnd,
			   MGC_M_CSR0_FLUSHFIFO);
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, bEnd,
			   MGC_M_CSR0_FLUSHFIFO);
		}

		if(pThis->bIsMultipoint)
		{
	    	/* target addr & hub addr/port */
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXFUNCADDR), 
		       bAddress);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBADDR), 
		       bIsMulti ? 0x80 | bHubAddr : bHubAddr);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBPORT), 
		       bHubPort);
	    	/* also, try Rx */
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXFUNCADDR), 
		       bAddress);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBADDR), 
		       bIsMulti ? 0x80 | bHubAddr : bHubAddr);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBPORT), 
		       bHubPort);
		}
		else
		{
	    	/* non-multipoint core */
	    	MGC_Write8(pBase, MGC_O_HDRC_FADDR, bAddress);
		}

		/* protocol/endpoint/interval/NAKlimit */
		if(bEnd)
		{
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_TXTYPE, bEnd, reg);
	    	if(bIsBulk && pThis->bBulkSplit)
	    	{
				MGC_WriteCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd, 
		    		wPacketSize | ((pEnd->wMaxPacketSizeTx / wPacketSize) - 1) << 11);
	    	}
	    	else
	    	{
				MGC_WriteCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd, wPacketSize);
	    	}
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_TXINTERVAL, bEnd, bInterval);
		}
		else
		{
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_NAKLIMIT0, 0, bInterval);
	    	if(pThis->bIsMultipoint)
	    	{
				MGC_WriteCsr8(pBase, MGC_O_HDRC_TYPE0, 0, reg & 0xc0);
	    	}
		}

#ifdef DMA_HOST_CODE
		if(pDmaChStatus->dma_req_flag && pDmaChStatus->dma_ch_flag){	/* DMA׵ͭ */
			switch(pDmaChStatus->mode){
				case DMA_HOST_MODE0:
					break;
				case DMA_HOST_MODE1:
					wCsr |= (MGC_M_TXCSR_AUTOSET | MGC_M_TXCSR_DMAENAB | MGC_M_TXCSR_DMAMODE);
					break;
				case DMA_HOST_MODE_NON:
					printk("abnormal Error! \n");
					break;
			}
		}
#endif /* DMA_HOST_CODE end */

		if(wLoadCount) {
	    	/* load FIFO */
	    	pEnd->dwRequestSize = wLoadCount;
	    	MGC_HdrcLoadFifo(pThis->pRegs, bEnd, wLoadCount, pBuffer);
		}
	
		/* re-enable interrupt */
		MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, wIntrTxE);

		/* write CSR */
		if(bEnd)
		{
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
		}
    }
    else
    {
        /* receive */

        /* if programmed for Tx, be sure it is ready for re-use */
        if(wCsr & MGC_M_TXCSR_MODE)
		{
	    	pEnd->bIsReady = FALSE;
	    	if(wCsr & MGC_M_TXCSR_FIFONOTEMPTY)
	    	{
	        	/* this shouldn't happen */
	        	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, 
			       MGC_M_TXCSR_FRCDATATOG);
	        	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, 
			       MGC_M_TXCSR_FRCDATATOG);
				ERR("switching end %d to Rx but Tx FIFO not empty\n", bEnd);
	    	}
	    	/* clear mode (and everything else) to enable Rx */
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, 0);
		}

		/* grab Rx residual if any */
		wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
		if(wCsr & MGC_M_RXCSR_RXPKTRDY)
		{
	    	DBG(1, "end %d Rx residual\n", bEnd); 
	    	bDone = MGC_LinuxPacketRx(pThis, bEnd, FALSE);
		}

		/* address */
		if(pThis->bIsMultipoint)
		{
	    	/* target addr & hub addr/port */
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXFUNCADDR), 
		       bAddress);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBADDR), 
		       bIsMulti ? 0x80 | bHubAddr : bHubAddr);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBPORT), 
		       bHubPort);
	    	/* also, try Tx */
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXFUNCADDR), 
		       bAddress);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBADDR), 
		       bIsMulti ? 0x80 | bHubAddr : bHubAddr);
	    	MGC_Write8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBPORT), 
		       bHubPort);
		}
		else
		{
	    	/* non-multipoint core */
	    	MGC_Write8(pBase, MGC_O_HDRC_FADDR, bAddress);
		}

		/* protocol/endpoint/interval/NAKlimit */
		if(bEnd)
		{
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_RXTYPE, bEnd, reg);
			if(bIsBulk && pThis->bBulkCombine)
	    	{
				MGC_WriteCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd,
		    		wPacketSize | ((pEnd->wMaxPacketSizeRx / wPacketSize) - 1) << 11);
	    	}
	    	else
	    	{
				MGC_WriteCsr16(pBase, MGC_O_HDRC_RXMAXP, bEnd, wPacketSize);
	    	}
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_RXINTERVAL, bEnd, bInterval);
		}
		else if(pThis->bIsMultipoint)
		{
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_TYPE0, 0, reg & 0xc0);
		}

		/* first time or re-program and shared FIFO, flush & clear toggle */
		if(!pEnd->bIsReady && pEnd->bIsSharedFifo)
		{
	    	/* twice in case of double packet buffering */
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, 
			   MGC_M_RXCSR_FLUSHFIFO | MGC_M_RXCSR_CLRDATATOG);
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, 
			   MGC_M_RXCSR_FLUSHFIFO | MGC_M_RXCSR_CLRDATATOG);
	    	pEnd->bIsReady = TRUE;
		}

		/* program data toggle if possibly switching use */
		if(!pEnd->bIsReady && pThis->bIsMultipoint)
		{
	    	wCsr = MGC_M_RXCSR_H_WR_DATATOGGLE;
	    	if(usb_gettoggle(pUrb->dev, pEnd->bEnd, 0))
	    	{
				wCsr |= MGC_M_RXCSR_H_DATATOGGLE;
	    	}
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
		}

		/* kick things off */
		if(bEnd && !bDone)
		{
			wCsr = 0;
#ifdef DMA_HOST_CODE
			if(pDmaChStatus->dma_req_flag && pDmaChStatus->dma_ch_flag){
				switch(pDmaChStatus->mode){
					case DMA_HOST_MODE0:
						break;
					case DMA_HOST_MODE1:
						wCsr |= (MGC_M_RXCSR_AUTOCLEAR | MGC_M_RXCSR_H_AUTOREQ | MGC_M_RXCSR_DMAENAB | MGC_M_RXCSR_DMAMODE);
						break;
					case DMA_HOST_MODE_NON:
						printk("abnormal Error! \n");
						break;
				}
			}
			if(pDmaChStatus->State == DMA_HOST_ST_DMA){
   	 			pDmaController->pfDmaEnableIntChannel(pDmaChannel);		/* DMAߵ */
   	 			pDmaChStatus->dma_trans_flag	=	USB_MCD_FLAG_ON;	/* DMAžե饰ON */
   	 			pDmaController->pfDmaStartChannel(pDmaChannel);			/* DMAž */
			}
#endif /* DMA_HOST_CODE */
#ifdef DMA_HOST_CODE
	    	wCsr |= MGC_M_RXCSR_H_REQPKT;
#else
	    	wCsr = MGC_M_RXCSR_H_REQPKT;
#endif
	    	if(usb_pipeint(nPipe))
	    	{
				wCsr |= MGC_M_RXCSR_DISNYET;
	    	}
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
		}
#ifdef DMA_HOST_CODE
		else{
			if(pDmaChStatus->State == DMA_HOST_ST_NULL){
				wCsr = MGC_M_RXCSR_H_REQPKT;	/* ReqPkt set */
				if(usb_pipeint(nPipe)){
					wCsr |= MGC_M_RXCSR_DISNYET;
	    		}
	    		MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
			}
		}
#endif /* DMA_HOST_CODE */	
    }/* receive end*/

    DBG(2, "done, unlocking pThis\n");
    spin_unlock_irqrestore(&pThis->Lock, flags);
}
#endif
/**
 * Try to stop traffic on the given local endpoint.
 */
STATIC void MGC_HdrcStopEnd(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t wCsr;
    unsigned long flags;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    const uint8_t reg=(bEnd)?MGC_O_HDRC_RXCSR:MGC_O_HDRC_CSR0;
    
    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, bEnd);
    wCsr = MGC_ReadCsr16(pBase, reg, bEnd);
    wCsr &= (bEnd)?~MGC_M_RXCSR_H_REQPKT:~MGC_M_CSR0_H_REQPKT;
    MGC_WriteCsr16(pBase, reg, bEnd, wCsr);
    spin_unlock_irqrestore(&pThis->Lock, flags);
}

/**
 * Flush the given local endpoint.
 */
static void
host_flush_ep(
    MGC_LinuxCd* pThis,
    uint8_t bEnd)
{
    uint16_t wCsr;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    
    MGC_SelectEnd(pBase, bEnd);

    if (bEnd) {
        /*  */
        wCsr = ( MGC_M_TXCSR_MODE |
                 MGC_M_TXCSR_CLRDATATOG |
                 MGC_M_TXCSR_FLUSHFIFO |
                 MGC_M_TXCSR_H_NAKTIMEOUT |
                 MGC_M_TXCSR_H_RXSTALL |
                 MGC_M_TXCSR_H_ERROR );
        MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, nEnd, wCsr);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, nEnd, wCsr);
        /*  */
        wCsr = ( MGC_M_RXCSR_CLRDATATOG |
                 MGC_M_RXCSR_FLUSHFIFO |
                 MGC_M_RXCSR_DATAERROR |
                 MGC_M_RXCSR_H_RXSTALL |
                 MGC_M_RXCSR_H_ERROR );
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, nEnd, wCsr);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, nEnd, wCsr);
    } else {
        /* EP0 */
        wCsr = ( MGC_M_CSR0_FLUSHFIFO |
                 MGC_M_CSR0_H_NAKTIMEOUT |
                 MGC_M_CSR0_H_RXSTALL);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsr);
    }
}
/**
 * Flush all endpoint.
 */
static void
host_flush_all_ep(
    MGC_LinuxCd* pThis)
{
    uint8_t bEnd;

    for (bEnd = 1; bEnd < pThis->bEndCount; bEnd++) {
        host_flush_ep(pThis, bEnd);
    }
}
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Start transmit.
 * Caller is responsible for locking shared resources.
 *
 * @param pThis instance pointer
 * @param bEnd local endpoint
 */
STATIC void MGC_HdrcStartTx(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t wCsr;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    DBG(2, "<==\n");

    /* NOTE: no locks here; caller should lock */
    MGC_SelectEnd(pBase, bEnd);
    
    if(bEnd)
    {
        wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
	wCsr |= MGC_M_TXCSR_TXPKTRDY;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
    }
    else
    {
	wCsr = MGC_M_CSR0_H_SETUPPKT | MGC_M_CSR0_TXPKTRDY;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, wCsr);
    }
    
}
#endif

/**
 * Interrupt Service Routine to record USB "global" interrupts.
 * Since these do not happen often and signify things of 
 * paramount importance, it seems OK to check them individually.
 *
 * @param pThis instance pointer
 * @param reg IntrUSB register contents
 */
void MGC_HdrcServiceUsb(MGC_LinuxCd* pThis, uint8_t reg)
{
    uint8_t bEnd;
    uint16_t wFrame;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    uint8_t devctl = MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);
#if defined(CONFIG_USB_MCD_HOST)||(MUSB_DEBUG > 0)
    uint8_t power = MGC_Read8(pBase, MGC_O_HDRC_POWER);
#endif
#ifdef MUSB_OTG
    union usb_otg_event otg_event;
#endif

    DBG(2, "<== Power=%02x, DevCtl=%02x, reg=0x%x\n", power, devctl, reg);
    DBG(1, "INTR" "%x\n", reg);

    if(reg & MGC_M_INTR_RESUME)
    {
	DBG(3, "RESUME\n");
	if(devctl & MGC_M_DEVCTL_HM)
	{
#if defined(CONFIG_USB_MCD_HOST)
	    power &= ~MGC_M_POWER_SUSPENDM;
	    MGC_Write8(pBase, MGC_O_HDRC_POWER, power | MGC_M_POWER_RESUME);
	    MGC_LinuxSetTimer(pThis, MGC_HdrcDropResume, 
		(unsigned long)pThis, 50); //20);
#endif
	}
#ifdef MUSB_GADGET
	else if( MUSB_IS_DEV(pThis) )
	{
	    MGC_GadgetResume(pThis);
	}
#endif
    }

#ifdef MUSB_OTG
    if(reg & MGC_M_INTR_SESSREQ)
    {
        DBG(-1, "SESSION_REQUEST\n");
        otg_event.type = USB_OTG_EVENT_TYPE_SRP;
        musb_otg_control_signal_interrupt(pThis->otg_control, &otg_event);
    }
#endif

    if(reg & MGC_M_INTR_VBUSERROR)
    {
	DBG(3, "VBUS_ERROR\n");
        ERR("Stopping host due to Vbus error\n");
	
	MGC_HdrcStop(pThis);
	MUSB_ERR_MODE(pThis, MUSB_ERR_VBUS);
#if defined(CONFIG_USB_MCD_HOST)	
	MGC_VirtualHubPortDisconnected(&(pThis->RootHub), 0);
	pThis->pRootDevice = NULL;
#endif
#ifdef MUSB_OTG
        musb_otg_control_notify_vbuserror(pThis->otg_control);
#endif
    }

    if(reg & MGC_M_INTR_SUSPEND)
    {
	DBG(3, "SUSPEND\n");
        if( MUSB_IS_HST(pThis) ) {
	    WARN("suspend detected\n");
	}
#ifdef MUSB_GADGET
	else if( MUSB_IS_DEV(pThis) )
	{
            /* enable_rchostޡ */
            if (devctl & MGC_M_DEVCTL_HR) {
                musb_otg_control_start_rchost(pThis->otg_control);
                host_flush_all_ep(pThis);
            }
	    MGC_GadgetSuspend(pThis);
	}
#endif
    }

    if(reg & MGC_M_INTR_CONNECT)
    {
	uint8_t bSpeed = 1;
    	uint8_t bHubSpeed = 2;

	DBG(3, "CONNECT\n");
#if defined(MUSB_OTG) || !defined(MUSB_GADGET)
	MUSB_HST_MODE(pThis);		
#ifdef MUSB_OTG
        otg_event.type = USB_OTG_EVENT_TYPE_CON;
        otg_event.con.value = 1;
        musb_otg_control_signal_interrupt(pThis->otg_control, &otg_event);
#endif

	if(devctl & MGC_M_DEVCTL_LSDEV)
	{
	    bSpeed = 3;
	    bHubSpeed = 0;
	} else if(devctl & MGC_M_DEVCTL_FSDEV)
	{
	    /* NOTE: full-speed is "speculative" until reset */
	    bSpeed = 2;
	    bHubSpeed = 1;
	}
	
	pThis->bRootSpeed = bSpeed;
	if(pThis->bIsMultipoint)
	{
	    /* set speed for EP0 */
	    MGC_SelectEnd(pBase, 0);
	    MGC_WriteCsr8(pBase, MGC_O_HDRC_TYPE0, 0, 
			  (bSpeed << 6));
	}
#if defined(CONFIG_USB_MCD_HOST)	
	MGC_VirtualHubPortConnected(&(pThis->RootHub), 0, 
		bHubSpeed);
#endif
#else
	ERR("GADGET mode, OTG support was not compiled");		
#endif		
    }

    if(reg & MGC_M_INTR_DISCONNECT)
    {
	DBG(3, "DISCONNECT()\n");	
	
#ifdef MUSB_OTG
        otg_event.type = USB_OTG_EVENT_TYPE_CON;
        otg_event.con.value = 0;
        musb_otg_control_signal_interrupt(pThis->otg_control, &otg_event);
#endif

	if( MUSB_IS_HST(pThis) )
	{
#if defined(CONFIG_USB_MCD_HOST)	
	
            DBG(1, "disconnecting VirtualHub\n");
	    MGC_VirtualHubPortDisconnected(&(pThis->RootHub), 0);
	    pThis->pRootDevice = NULL;

            host_flush_all_ep(pThis);

#ifdef MUSB_CONFIG_PROC_FS
	    if(pThis->pfDisconnectListener) {
		pThis->pfDisconnectListener(pThis->pDisconnectListenerParam);
	    }
#endif
#endif /* CONFIG_USB_MCD_HOST */
	}
    }
    
    if(reg & MGC_M_INTR_RESET)
    {
	DBG(3, "BUS RESET\n");

	if(devctl & MGC_M_DEVCTL_HM)
	{
	    /* we are babbling */
	    ERR("host transaction is babble\n");
	}
        else {
#ifdef MUSB_GADGET
            if (MUSB_IS_DEV(pThis)) {
                MGC_GadgetReset( pThis );
            }
#endif
	}
    }
    
    if(reg & MGC_M_INTR_SOF) {
	DBG(3, "START_OF_FRAME\n");

        /* start any periodic Tx transfers waiting for current frame */
        wFrame = MGC_Read16(pBase, MGC_O_HDRC_FRAME);
	for(bEnd = 1; 
	    (bEnd < pThis->bEndCount) && (pThis->wEndMask >= (1 << bEnd)); 
	    bEnd++)
	{
	    if(pThis->aLocalEnd[bEnd].dwWaitFrame >= wFrame)
	    {
	        pThis->aLocalEnd[bEnd].dwWaitFrame = 0;
#if defined(CONFIG_USB_MCD_HOST)
		MGC_HdrcStartTx(pThis, bEnd);
#endif
	    }
	}
    }
}
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Service the default endpoint (ep0) as host.
 * 
 * @param pThis this
 * @param wCount current byte count in FIFO
 * @param pUrb URB pointer for EP0
 * @return TRUE if more packets are required for this transaction
 */
STATIC uint8_t MGC_HdrcServiceHostDefault(MGC_LinuxCd* pThis, 
					   uint16_t wCount, 
					   struct urb* pUrb)
{
    uint8_t bMore = FALSE;
    uint8_t* pFifoDest = NULL;
    uint16_t wFifoCount = 0;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[0]);
    MUSB_DeviceRequest* pRequest = (MUSB_DeviceRequest*)pUrb->setup_packet;

    DBG(2, "<== (wCount=%04x, pUrb=%lx, bStage=%02x)\n",  \
    	wCount, (unsigned long)pUrb, pThis->bEnd0Stage); 

    if(MGC_END0_IN == pThis->bEnd0Stage)
    {
        /* we are receiving from peripheral */
#ifdef MUSB_DMA_BUS2VIRT
		if (!(pUrb->transfer_buffer)) {
        	pFifoDest = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
    	}else{
		pFifoDest = (uint8_t*)(pUrb->transfer_buffer);
		}
		pFifoDest += pUrb->actual_length;
 
    	if (!pFifoDest ) { // abort the transfer
        	ERR("no buffer was given, aborting\n");
       		return TRUE;
    	}

#else
        pFifoDest = pUrb->transfer_buffer + pUrb->actual_length;
#endif
		wFifoCount = min(wCount, ((uint16_t)(pUrb->transfer_buffer_length - pUrb->actual_length)));

    	DBG(3, "Receiving %d bytes in &%p[%d] (pUrb->actual_length=%u)\n", \
	    wFifoCount, pUrb->transfer_buffer, (unsigned int)pUrb->actual_length, \
	    pUrb->actual_length );

		DBG(3,"pUrb->transfer_dma=%p \n",(void *)pUrb->transfer_dma);
		MGC_HdrcUnloadFifo(pBase, 0, wFifoCount, pFifoDest);

#ifdef MUSB_CONFIG_PROC_FS
		pEnd->dwTotalRxBytes += wFifoCount;
		pEnd->dwTotalRxPackets++;
#endif
		pUrb->actual_length += wFifoCount;
		if((pUrb->actual_length < pUrb->transfer_buffer_length) &&
		   (wCount == pEnd->wPacketSize))
		{
		    bMore = TRUE;
		}
    }
    else
    {
        /* we are sending to peripheral */
        if((MGC_END0_START == pThis->bEnd0Stage) && 
	   (pRequest->bmRequestType & USB_DIR_IN))
	{
	    DBG(3, "just did setup, switching to IN\n");

	    /* this means we just did setup; switch to IN */
	    pThis->bEnd0Stage = MGC_END0_IN;
	    bMore = TRUE;
	    
#ifdef MUSB_CONFIG_PROC_FS
	    pEnd->dwTotalTxBytes += 8;
	    pEnd->dwTotalTxPackets++;
#endif
	}
	else if(pRequest->wLength && (MGC_END0_START == pThis->bEnd0Stage))
	{

	    pThis->bEnd0Stage = MGC_END0_OUT;

#ifdef MUSB_DMA_BUS2VIRT
        if (!(pUrb->transfer_buffer)) {
        	pFifoDest = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
       	}else{
        	pFifoDest = (uint8_t*)(pUrb->transfer_buffer);
        }
        pFifoDest += pUrb->actual_length;

       	if (!pFifoDest ) { // abort the transfer
        	ERR("no buffer was given, aborting\n");
            return TRUE;
        }
#else
	    pFifoDest = (uint8_t*)(pUrb->transfer_buffer + pUrb->actual_length);
#endif
	    wFifoCount = min(pEnd->wPacketSize, 
		((uint16_t)(pUrb->transfer_buffer_length - pUrb->actual_length)));
		
    	    DBG(3, "Sending %d bytes to %p\n", wFifoCount, pFifoDest); 

	    MGC_HdrcLoadFifo(pBase, 0, wFifoCount, pFifoDest);

#ifdef MUSB_CONFIG_PROC_FS
	    pEnd->dwTotalTxBytes += wFifoCount;
	    pEnd->dwTotalTxPackets++;
#endif
	    pEnd->dwRequestSize = wFifoCount;
	    pUrb->actual_length += wFifoCount;
	    if(pUrb->actual_length < pUrb->transfer_buffer_length)
	    {
	        bMore = TRUE;
	    }
	}
    }

    return bMore;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Handle default endpoint interrupt as host.
 * @param pThis this
 */
STATIC void MGC_HdrcServiceDefaultEnd(MGC_LinuxCd* pThis)
{
    uint16_t wCsrVal, wCount;
    uint8_t bVal;
    struct urb* pUrb;
    int status = USB_ST_NOERROR;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[0]);
    uint8_t bOutVal = 0;
    uint8_t bComplete = FALSE;
    uint8_t bError = FALSE;
    unsigned long flags;

    DBG(2, "<==\n");

    spin_lock(&pEnd->Lock);
    pUrb = pEnd->pUrb;    
    spin_unlock(&pEnd->Lock);

    spin_lock_irqsave(&pThis->Lock, flags);    
    MGC_SelectEnd(pBase, 0);
    wCsrVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0);
    wCount = MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0);
    bVal = (uint8_t)wCsrVal;

    /* check URB */
#ifdef MUSB_PARANOID
    if( pUrb && (pUrb->hcpriv != pEnd)) { \
        ERR("corrupt URB %lx!\n", (unsigned long)pUrb); 
	pUrb = NULL; 
    } 
#endif

    DBG(2, "CSR0=%04x, wCount=%04x\n", wCsrVal, wCount); 

    /* if we just did status stage, we are done */
    if(MGC_END0_STATUS == pThis->bEnd0Stage)
    {
	bComplete = TRUE;
    }

    /* prepare status */
    if((MGC_END0_START == pThis->bEnd0Stage) && !wCount &&
       (wCsrVal & MGC_M_CSR0_RXPKTRDY))
    {
	DBG(3, "missed data\n"); 

        /* just started and got Rx with no data, so probably missed data */
        status = USB_ST_SHORT_PACKET;
	bError = TRUE;
	//bComplete = TRUE;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, MGC_M_CSR0_FLUSHFIFO);
	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, MGC_M_CSR0_FLUSHFIFO);
    }
    
    if(bVal & MGC_M_CSR0_H_RXSTALL) {
	DBG(3, "stall\n"); 

        status = USB_ST_STALL;
	bError = TRUE;
    } else if(bVal & MGC_M_CSR0_H_ERROR) {
	DEBUG_CODE(3, ERR("no response (error)\n"); \
		MGC_HdrcDumpRegs(pThis, 0); );

        status = USB_ST_NORESPONSE;
	bError = TRUE;
    } else if(bVal & MGC_M_CSR0_H_NAKTIMEOUT) {
	DBG(3, "NAK timeout pEnd->bRetries=%d\n", pEnd->bRetries); 

        if( ++pEnd->bRetries < MUSB_MAX_RETRIES)
	{
	    /* cover it up if retries not exhausted */
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, 0);
	}
	else
	{
	    DEBUG_CODE(1, WARN("no response (NAK timeout)\n"); \
		MGC_HdrcDumpRegs(pThis, 0); );

	    status = USB_ST_NORESPONSE;
	    bError = TRUE;
	}
    }

    if(USB_ST_NORESPONSE == status)
    {
	DBG(1, "aborting\n"); 

	/* use the proper sequence to abort the transfer */
	if(bVal & MGC_M_CSR0_H_REQPKT)
	{
	    bVal &= ~MGC_M_CSR0_H_REQPKT;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bVal);
	    bVal &= ~MGC_M_CSR0_H_NAKTIMEOUT;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bVal);
	}
	else
	{
	    bVal |= MGC_M_CSR0_FLUSHFIFO;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bVal);
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bVal);
	    bVal &= ~MGC_M_CSR0_H_NAKTIMEOUT;
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bVal);
	}
	MGC_WriteCsr8(pBase, MGC_O_HDRC_NAKLIMIT0, 0, 0);
    }

    if(bError)
    {
	DBG(1, "handling error\n");

        /* clear it */
	MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, 0);
#ifdef MUSB_CONFIG_PROC_FS
        switch(pThis->bEnd0Stage)
	{
	case MGC_END0_START:
	case MGC_END0_OUT:
	    pEnd->dwErrorTxPackets++;
	    break;
	case MGC_END0_IN:
	    pEnd->dwErrorRxPackets++;
	    break;
	}
#endif

    }

    if(!pUrb)
    {
	/* stop endpoint since we have no place for its data */
	DBG(1, "no URB for end 0\n");

        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, MGC_M_CSR0_FLUSHFIFO);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, MGC_M_CSR0_FLUSHFIFO);
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, 0);

	/* start next URB that might be queued for it */
	spin_lock(&pEnd->Lock);
	MGC_LinuxStartNextUrb(pThis, 0);
	spin_unlock(&pEnd->Lock);
	spin_unlock_irqrestore(&pThis->Lock, flags);
        return;
    }
    
    if(!bComplete && !bError)
    {
        /* call common logic and prepare response */
        if( MGC_HdrcServiceHostDefault(pThis, wCount, pUrb) )
	{
	    /* more packets required */
	    bOutVal = (MGC_END0_IN == pThis->bEnd0Stage) ? 
	   	MGC_M_CSR0_H_REQPKT : MGC_M_CSR0_TXPKTRDY;

	   DBG(3, "Need more bytes bOutVal=%04x\n", bOutVal); 
	} else {
	    /* data transfer complete; perform status phase */
	    bOutVal = MGC_M_CSR0_H_STATUSPKT | 
	      (usb_pipeout(pUrb->pipe) ? MGC_M_CSR0_H_REQPKT : 
	       MGC_M_CSR0_TXPKTRDY);
	    /* flag status stage */
	    pThis->bEnd0Stage = MGC_END0_STATUS;

	   DBG(3, "Data transfer complete, status phase bOutVal=%04x\n", \
	   	bOutVal); 

	}
    }

    /* write CSR0 if needed */
    if(bOutVal)
    {
        MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, bOutVal);
    }
    

    /* call completion handler if done */
    if(bComplete || bError)
    {
        pUrb->status = status;

	DBG(2, "completing cntrl URB %p, status=%d, len=%x\n", \
	    pUrb, pUrb->status, pUrb->actual_length); 

	spin_unlock_irqrestore(&pThis->Lock, flags);
	if (pUrb->complete)
	{
	    pUrb->hcpriv=NULL;
	    COMPLETE_URB(pUrb, NULL);
	}

        /* start next URB if any */
	spin_lock(&pEnd->Lock);
        MGC_LinuxStartNextUrb(pThis, 0);
	spin_unlock(&pEnd->Lock);
    } else {
	spin_unlock_irqrestore(&pThis->Lock, flags);
    }

    DBG(3, "done\n");
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Service a Tx-Available interrupt for the given endpoint
 * @param pThis instance pointer
 * @param bEnd local endpoint
 */
STATIC void MGC_HdrcServiceTxAvail(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t	wVal;
    uint16_t	wLength;
    uint8_t		*pBuffer;
    struct urb	*pUrb;
    int			nPipe;
    uint8_t		bDone	= FALSE;
    MGC_LinuxLocalEnd	*pEnd = &(pThis->aLocalEnd[bEnd]);
    uint32_t	status	= USB_ST_NOERROR;
    uint8_t		*pBase	= (uint8_t*)pThis->pRegs;
    unsigned long		flags;

#ifdef HOST_NULL_FLAG
	USB_MCD_HostInfo	*HostInfo = &(pEnd->HostInfo);
#endif
#ifdef DMA_HOST_CODE
	uint16_t	wCsr	= 0;
	uint16_t	rest_sz	= 0;
    unsigned long	flags_dma;
    uint8_t		err_stat = 0;
    uint8_t		stat	= 0;
    USB_MCD_DmaChStatus *pDmaChStatus	=	pEnd->pChEpStatus;
	MGC_DmaController	*pDmaController	=	pThis->pDmaController;
    MGC_DmaChannel		*pDmaChannel	=	pEnd->pDmaChannel;

#endif
	struct usb_host_endpoint *ep;
	unsigned int epnum;

    DBG(2, "<==\n");

    spin_lock(&pEnd->Lock);
    pUrb = pEnd->pUrb;

    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, bEnd);
    wVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);

#if MUSB_DEBUG > 0
    /* check URB */
    if(pUrb && (pUrb->hcpriv != pEnd))
    {
        ERR("end %d has corrupt URB %lx!\n", bEnd, (unsigned long)pUrb);
	pUrb = NULL;
    }
#endif

    nPipe = pUrb ? pUrb->pipe : 0;
    DBG(2, "end %d TxCSR=%04x\n", bEnd, wVal); 

    /* check for errors */
    if(wVal & MGC_M_TXCSR_H_RXSTALL)
    {
		/* stall; record URB status */
        status = USB_ST_STALL;
		/* also record with USB core */
		if(pUrb) {
			epnum = usb_pipeendpoint(nPipe)& USB_ENDPOINT_NUMBER_MASK;
			if (usb_endpoint_out(usb_pipeendpoint(nPipe))) {
				ep = pUrb->dev->ep_out[epnum];
			} else {
				ep = pUrb->dev->ep_in[epnum];
			}
	    	USB_HALT_ENDPOINT(pUrb->dev, ep, 0);
	    	/* disableMGC_LinuxDisable()˴ؿ϶ */
		}
    }
    else if(wVal & MGC_M_TXCSR_H_ERROR)
    {
        status = USB_ST_NORESPONSE;
	/* do the proper sequence to abort the transfer */
	wVal &= ~MGC_M_TXCSR_FIFONOTEMPTY;
	wVal |= MGC_M_TXCSR_FLUSHFIFO;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wVal);
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wVal);
    }
    else if(wVal & MGC_M_TXCSR_H_NAKTIMEOUT)
    {
        if(++pEnd->bRetries < MUSB_MAX_RETRIES)
	{
	    /* cover it up if retries not exhausted */
	    MGC_WriteCsr16(pBase, MGC_O_HDRC_CSR0, 0, MGC_M_TXCSR_TXPKTRDY);
	    spin_unlock(&pEnd->Lock);
	    spin_unlock_irqrestore(&pThis->Lock, flags);
	    return;
	}
	status = USB_ST_NORESPONSE;
	/* do the proper sequence to abort the transfer */
	wVal &= ~MGC_M_TXCSR_FIFONOTEMPTY;
	wVal |= MGC_M_TXCSR_FLUSHFIFO;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wVal);
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wVal);
	MGC_WriteCsr8(pBase, MGC_O_HDRC_TXINTERVAL, bEnd, 0);
    }
    if(status)
    {
        /* reset error bits */
        wVal &= ~(MGC_M_TXCSR_H_ERROR | MGC_M_TXCSR_H_RXSTALL |
		  MGC_M_TXCSR_H_NAKTIMEOUT);
	wVal |= MGC_M_TXCSR_FRCDATATOG;
	MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wVal);
	bDone = TRUE;
#ifdef MUSB_CONFIG_PROC_FS
	pEnd->dwErrorTxPackets++;
#endif
    }
    else if(pUrb)
    {
#ifdef MUSB_CONFIG_PROC_FS
      pEnd->dwTotalTxBytes += pEnd->dwRequestSize;
      pEnd->dwTotalTxPackets++;	
#endif
#if defined(DMA_HOST_CODE) && defined(DMA_HOST_CODE_TX_IRQ)
/*=====================================================================================*/
		/* žȽ */
    	if(usb_pipeisoc(nPipe))
    	{
    	/*=============================================================================*/
	    /* isoch case ISOž */
	    	DBG(1,"***   ISO DMA transfer wolud be   *** \n");
			if(((pDmaChStatus)->dma_req_flag)&&((pDmaChStatus)->dma_ch_flag)){
			/* DMA׵ե饰ON  DMA̵ͥͭե饰ON ξ DMAž */
				/* DMA ISO TX */
				if((pDmaChStatus)->dma_trans_flag){	
				/* ϥ⡼1 ISOϥ⡼0ʤΤǸϤʤ */
    				spin_lock_irqsave(&pThis->Lock, flags_dma);		/* CPU߶ػ */

					/* Stop Channnel *//* Ϥ䤯ߤ뤿 */
   	 				stat = pDmaController->pfDmaStopChannel(pDmaChannel);
					/* Set TxCSR *//* ᤯ꤹ뤿 */
					/* DMAReqEnab <- 0 ,AutoSet <- 0 */
        			wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					wCsr &= ~(MGC_M_TXCSR_AUTOSET|MGC_M_TXCSR_DMAENAB);
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					/* DMAReqMode <- 0 */
        			wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					wCsr &= ~MGC_M_TXCSR_DMAMODE;
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);

					/* DMA߶ػ */
					stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
   	 				(pDmaChStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;	/* DMAžե饰OFF */

   	 				/* update size of tx data in case of Mode1 */
   	 				stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
   	 				if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 					printk("DMA mode1 Error BUS_ABORT \n"); 
   	 				}
   	 				if((pDmaChStatus)->mode==DMA_HOST_MODE0){
   	 				/* in case of ISO transfer in DMA mode0   */
   	 				
						/* update actual & status */
						pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length = pDmaChannel->dwActualLength;
						/* numbers of Tx Packet increments */
       					++pEnd->dwIsoPacket;
   	 				}else{
   	 					/* update buffer offset */
   	 					pEnd->dwOffset += pDmaChannel->dwActualLength;
   	 				}
   	 				spin_unlock_irqrestore(&pThis->Lock, flags_dma);	/* CPU */
				}else{
					/* update buffer offset */
   	 				pEnd->dwOffset += pEnd->dwRequestSize;
   	 			
   	 				if((pDmaChStatus)->mode==DMA_HOST_MODE0){
   	 				/* in case of ISO transfer in DMA mode0 */

						/* update actual & status */
						pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length = pEnd->dwRequestSize;
						/* numbers of Tx Packet increments */
       					++pEnd->dwIsoPacket;
       				}
       			}
       			/* žλȽ */
       			bDone = (pEnd->dwIsoPacket >= pUrb->number_of_packets);
       			
       			if(bDone){
       				(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
       			}else{
       				if((pDmaChStatus)->mode==DMA_HOST_MODE0){
       					(pDmaChStatus)->State = DMA_HOST_ST_DMA;	/* to DMA state */
       					DBG(1,"Change State DMA in DMAmode0 \n");
					}else{
						(pDmaChStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
						DBG(1,"Change State CPU in DMAmode0");
					}
       			}
       			/* DMA֤ʤ */
       			if((pDmaChStatus)->State == DMA_HOST_ST_DMA){
					/* ɥ쥹Ѵ */
					if((pDmaChStatus)->dma_req_flag){
					/* get reference buffer address */
                	pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma))
                		+ pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
					}else{
#ifdef MUSB_DMA_BUS2VIRT
            			if (!(pUrb->transfer_buffer)) {
                			pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            			}else{
                			pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            			}
            			pBuffer +=  pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
            			if (!pBuffer ) { // abort the transfer
                			ERR("no buffer was given, aborting\n");
                			return;
            			}
#else
	        			pBuffer = pUrb->transfer_buffer +
		  					pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
#endif
					}
					wLength = pUrb->iso_frame_desc[pEnd->dwIsoPacket].length;
					/* set size of a next packet */
					pEnd->dwRequestSize = wLength;
					
					/* Set Callback */
					pDmaChannel->pfDmaCallback = USB_MCD_HostDmaCallback;

       				/* set DMA register parameter  */
       				stat = pDmaController->pfDmaProgramChannel(
        					pDmaChannel,
        					(MGC_pfDmaCallback)(pDmaChannel->pfDmaCallback),
        					(void*)pThis,
        					(uint16_t)(pEnd->wPacketSize),
        					(pDmaChStatus)->mode,
        					__virt_to_bus(pBuffer),
        					wLength);
    				if (!stat) {
   	 					if((pDmaChStatus)->dma_ch_flag){
   	 						/* release DMA channel */
   	 						pDmaController->pfDmaReleaseChannel(pDmaChannel);
							pEnd->pDmaChannel = NULL;
   	 						/* DMA channel flag set off */
   	 						(pDmaChStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
						}
   	 					wLength = 0;
						(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
						DBG(1,"Change State IDLE in DMAmode0\n");
						err_stat = 1;
						bDone = TRUE;
    				}
					/* initialize  DMA transfer size */
					pDmaChannel->dwActualLength = 0;
				}else{
				/* CPU⡼ɤξ */
#ifdef MUSB_DMA_BUS2VIRT
            		if (!(pUrb->transfer_buffer)) {
                		pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            		}else{
                		pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            		}
            		pBuffer +=  pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
            		if (!pBuffer ) { // abort the transfer
                		ERR("no buffer was given, aborting\n");
                		return;
            		}
#else
	        		pBuffer = pUrb->transfer_buffer +
		  				pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
#endif
					wLength = pUrb->iso_frame_desc[pEnd->dwIsoPacket].length;
					/* load next packet */
	        		MGC_HdrcLoadFifo(pBase, bEnd, wLength, pBuffer);
					pEnd->dwRequestSize = wLength;
				}
			}else{
			/*=================================================================================*/
			/* CPUž DMAžκǸCPUžˤʤ*/
				DBG(1," ***   ISO CPU transfer    *** \n");
			
				/* update buffer offset */
   	 			pEnd->dwOffset += pEnd->dwRequestSize;
				/* ISO TX */
       			/* žλȽ */
       			bDone = (++pEnd->dwIsoPacket >= pUrb->number_of_packets);
       			if(bDone){
       				(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
       				DBG(1,"CPU bDone Change State IDLE \n");
       			}else{
					(pDmaChStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
					DBG(1,"bDone Change State CPU \n");
       			}				
				if((pDmaChStatus)->dma_req_flag){
					/* get reference buffer address */
                	pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma))
                		+ pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
				}else{
#ifdef MUSB_DMA_BUS2VIRT
            		if (!(pUrb->transfer_buffer)) {
                		pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            		}else{
                		pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            		}
            		pBuffer +=  pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
            		if (!pBuffer ) { // abort the transfer
                		ERR("no buffer was given, aborting\n");
                		return;
            		}
#else
	        		pBuffer = pUrb->transfer_buffer +
		  				pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
#endif
				}
				wLength = pUrb->iso_frame_desc[pEnd->dwIsoPacket].length;
				/* load next packet */
	        	MGC_HdrcLoadFifo(pBase, bEnd, wLength, pBuffer);
				pEnd->dwRequestSize = wLength;
			}
    	}
    	else
    	{
    	/*=================================================================================*/
    	/* Non-iso case */
    		DBG(2,"***   NonISO DMA transfer wolud be    *** \n");
		
			if(((pDmaChStatus)->dma_req_flag)&&((pDmaChStatus)->dma_ch_flag)){
			/* DMA׵ե饰ON  DMA̵ͥͭե饰ON ξ DMAž */
				/* DMA Non-ISO TX */
				if((pDmaChStatus)->dma_trans_flag){
				/* DMAžե饰ONξ *//* ⡼1 */
    				spin_lock_irqsave(&pThis->Lock, flags_dma);		/* CPU߶ػ */

					/* Set TxCSR */
					/* Stop Channnel */
   	 				stat = pDmaController->pfDmaStopChannel(pDmaChannel);
					/* DMAReqEnab <- 0 ,AutoSet <- 0 */
        			wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					wCsr &= ~(MGC_M_TXCSR_AUTOSET|MGC_M_TXCSR_DMAENAB);
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					/* DMAReqMode <- 0 */
        			wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					wCsr &= ~MGC_M_TXCSR_DMAMODE;
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					/* DMA߶ػ */
					stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
   	 				(pDmaChStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;	/* DMAžե饰OFF */

   	 				/* update size of tx data in case of Mode1 */
   	 				stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
   	 				if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 					printk("DMA mode1 Error BUS_ABORT \n"); 
   	 				}
					/* update buffer offset */
   	 				pEnd->dwOffset += pDmaChannel->dwActualLength;
   	 					
   	 				if((pDmaChStatus)->dma_ch_flag && (pDmaChStatus)->mode==DMA_HOST_MODE1){
   	 				/* DMA channel exist and DMA mode1 */
   	 			
   	 					/* DMA channel release */
   	 					pDmaController->pfDmaReleaseChannel(pDmaChannel);
						pEnd->pDmaChannel = NULL;
					
						/* DMA channel flag set off */
   	 					(pDmaChStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
					}
					spin_unlock_irqrestore(&pThis->Lock, flags_dma);	/* CPU */
				}else{
				/* in case of CPU transfer in DMA mode1 */
					/* Update Buffer offset */
					pEnd->dwOffset += pEnd->dwRequestSize;
					DBG(2,"NonISO Mode0 pEnd->dwOffset=%d \n", pEnd->dwOffset);
				}
				/* ɥ쥹Ѵ */
				/* get reference buffer address */
                pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma)) + pEnd->dwOffset;
				/* get the rest of data size to transfer */
				rest_sz = pUrb->transfer_buffer_length - pEnd->dwOffset;
				if(pUrb->transfer_buffer_length < pEnd->dwOffset){
					printk("*** rest_sz calcu Error value:[%d] !! ***",rest_sz);
				}
				DBG(2,"NonISO Mode0 rest_sz=%d  \n",rest_sz);
				
				if(rest_sz > (uint16_t)NULL){
					switch((pDmaChStatus)->mode){
						/* DMA Mode0 */
						case DMA_HOST_MODE0:
							(pDmaChStatus)->State = DMA_HOST_ST_DMA;	/* to DMA state */
							DBG(2,"NonISO data in DMAmode0 to State DMA! \n");
							/* get transfer size (by max MAXP) */
	   		 				wLength = min(pEnd->wPacketSize, (uint16_t)(rest_sz));
	   		 				DBG(2,"NonISO Mode0 wLength=%d  \n",wLength);
	        				/* Set Tx size */
							pEnd->dwRequestSize = wLength;

    						/* DMAХå򥳡뤷ȤNULLƤΤǺ */
							/* CallBack */
							pDmaChannel->pfDmaCallback = USB_MCD_HostDmaCallback;

       						/* set DMA register parameter  */
       						stat = pDmaController->pfDmaProgramChannel(
        						pDmaChannel,
        						(MGC_pfDmaCallback)(pDmaChannel->pfDmaCallback),
        						(void*)pThis,
        						(uint16_t)(pEnd->wPacketSize),
        						(pDmaChStatus)->mode,
        						__virt_to_bus(pBuffer),
        						wLength);
    						if (!stat) {
   	 							if((pDmaChStatus)->dma_ch_flag){
   	 								/* release DMA channel */
   	 								pDmaController->pfDmaReleaseChannel(pDmaChannel);
									pEnd->pDmaChannel = NULL;
   	 								/* DMA channel flag set off */
   	 								(pDmaChStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
								}
   	 							wLength = 0;
								(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
								DBG(2,"NonISO  Change State IDLE! in DMAmode0 \n");
								err_stat = 1;
								bDone = TRUE;
    						}
							/* initialize  DMA transfer size */
							pDmaChannel->dwActualLength = 0;
							break;
						/* DMA Mode1 */
						case DMA_HOST_MODE1:
						
							(pDmaChStatus)->State = DMA_HOST_ST_CPU;	/* CPU֤ */
							DBG(2," Change State CPU! in NonISO Mode1\n");
							/* get transfer size (by max MAXP) */
	   		 				wLength = min(pEnd->wPacketSize, (uint16_t)(rest_sz));
	   		 				/* FIFO Load */
	        				MGC_HdrcLoadFifo(pBase, bEnd, wLength, pBuffer);
	        				/* Set Tx size */
							pEnd->dwRequestSize = wLength;
							break;
						case DMA_HOST_MODE_NON:
							printk("abnormal Error! \n");
							break;
					}
				}else if(rest_sz == (uint16_t)NULL){
				/* next tx size is NULL  */
					if((pDmaChStatus)->State == DMA_HOST_ST_NULL){
						(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* IDLE֤ */
						DBG(2,"Change State NULL to IDLE! in NonISO DMA\n");
							/* transfer end */
							bDone = TRUE;
					}else{
						if(HostInfo->null_flag&&(pEnd->dwRequestSize%pEnd->wPacketSize)){
						/* NULL transfer to be needed */
							(pDmaChStatus)->State = DMA_HOST_ST_NULL;	/* NULL֤ */
							DBG(2,"Change State NULL! in NonISO DMA \n");
						}else{
							(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* IDLE֤ */
							DBG(2,"Change State IDLE! in NonISO DMA \n");
							/* transfer end */
							bDone = TRUE;
						}
					}
				}
			}else{
			/*=================================================================================*/
			/* Non-ISO TX by CPU */
			DBG(2," ***   NonISO CPU transfer   *** \n");
			
			/* update buffer offset */
   	 		pEnd->dwOffset += pEnd->dwRequestSize;
			/* CPUž DMAžκǸCPUžˤʤ */
				if((pDmaChStatus)->dma_req_flag){
					pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma)) + pEnd->dwOffset;
				}else{
#ifdef MUSB_DMA_BUS2VIRT
		            if (!(pUrb->transfer_buffer)) {
                		pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            		}else{
                		pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            		}
            		pBuffer +=  pEnd->dwOffset;
            		if (!pBuffer ) { // abort the transfer
                	ERR("no buffer was given, aborting\n");
					return;
            		}
#else
	    			pBuffer = pUrb->transfer_buffer + pEnd->dwOffset;
#endif
				}
				/* get the rest of data size to transfer */
				rest_sz = pUrb->transfer_buffer_length - pEnd->dwOffset;
				if(pUrb->transfer_buffer_length < pEnd->dwOffset){
					printk("*** rest_sz calcu Error value:[%d] !! ***",rest_sz);
				}				
				if(rest_sz > (uint16_t)NULL){
					DBG(2," CPU data > NULL \n");
					/* to CPU state */
					(pDmaChStatus)->State = DMA_HOST_ST_CPU;
					DBG(2,"Change State CPU in NonISO CPU\n");
					/* get transfer size (by max MAXP) */
	   		 		wLength = min(pEnd->wPacketSize, (uint16_t)(rest_sz));
	   		 		/* FIFO Load */
	        		MGC_HdrcLoadFifo(pBase, bEnd, wLength, pBuffer);
	        		/* Set Tx size */
					pEnd->dwRequestSize = wLength;
					
				}else if(rest_sz == (uint16_t)NULL){
				/* next tx size is NULL  */
					DBG(2,"CPU rest data == NULL \n");
					if((pDmaChStatus)->State == DMA_HOST_ST_NULL){
					
						(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
						DBG(2,"Change State IDLE in NonISO CPU \n");
						/* transfer end */
						bDone = TRUE;
					}else{
						if((HostInfo->null_flag)&&(pEnd->dwRequestSize==pEnd->wPacketSize)){
						/* NULL transfer to be needed */
							(pDmaChStatus)->State = DMA_HOST_ST_NULL;	/* NULL֤ */
							DBG(2,"Change State NULL in NonISO CPU \n");
						}else{
							(pDmaChStatus)->State = DMA_HOST_ST_IDLE;	/* IDLE֤ */
							DBG(2,"Change State IDLE in NonISO CPU \n");
							/* transfer end */
							bDone = TRUE;
						}
					}
				}else{
					printk("Error!  \n");
				}
			}
    	}
	}else{
	/* URBʤξ */
        
        /* no URB; proceed to next */
		MGC_LinuxStartNextUrb(pThis, bEnd);
		spin_unlock(&pEnd->Lock);
		spin_unlock_irqrestore(&pThis->Lock, flags);
		return;
	}
#endif // DMA_HOST_CODE

    if(bDone)
    {
#ifdef DMA_HOST_CODE
		if((pDmaChStatus)->dma_req_flag){
			(pDmaChStatus)->dma_req_flag	= USB_MCD_FLAG_OFF;
		}
#endif /* DMA_HOST_CODE */
    	/* set status */
    	pUrb->status = status;
		pUrb->actual_length = pEnd->dwOffset;
		DBG(2,"pUrb->status=%d  pUrb->actual_length=%d \n",pUrb->status, pUrb->actual_length);
		
    	/* call completion handler */
		if(usb_pipebulk(nPipe))
		{
	    	/* we re-use bulk, so re-programming required */
	    	pEnd->bIsReady = FALSE;
	    	/* release claim if borrowed */
	    	if((bEnd != pThis->bBulkTxEnd) &&
	       	(pThis->bBulkTxEnd != pThis->bBulkRxEnd))
	    	{
			pEnd->bIsClaimed = FALSE;
	    	}
	   		/* save data toggle */
	    	usb_settoggle(pUrb->dev, pEnd->bEnd, 1,
			  	(wVal & MGC_M_TXCSR_H_DATATOGGLE) ? 1 : 0);
		}

		DBG(1, "completing Tx URB %lx, status=%d, len=%x\n", \
	    	(unsigned long)pUrb, pUrb->status, pUrb->actual_length); 

		spin_unlock(&pEnd->Lock);
		spin_unlock_irqrestore(&pThis->Lock, flags);
		if(pUrb->complete)
		{
	    	pUrb->hcpriv=NULL;
	    	COMPLETE_URB(pUrb, NULL);
	    	DBG(2,"*** bDone Completed!! ***\n");
		}

		/* proceed to next */
		spin_lock(&pEnd->Lock);
		MGC_LinuxStartNextUrb(pThis, bEnd);
		spin_unlock(&pEnd->Lock);
    }
    else
    {
#ifdef DMA_HOST_CODE
	//=============================================================================
		if((pDmaChStatus)->State != DMA_HOST_ST_DMA){
		/* DMAְʳ DMA⡼0 DMA߻TxPktRdyåȤ롣*/

	    	DBG(2,"*** SET TXPKTRDY!! to be continued ***\n");
		   
    		/* start next transaction */
    		MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, MGC_M_TXCSR_TXPKTRDY);
			spin_unlock(&pEnd->Lock);
			spin_unlock_irqrestore(&pThis->Lock, flags);
	}
/* TX_RESTART_DMA */
	else{
		    /* DMAȤProgramEnd()ǹԤ*/

			pDmaChannel = pEnd->pDmaChannel;	/* ProgramEndǼͥ */
			
    		/* DMAߵ */
    		stat = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
    		if (!stat) {
        		printk("Cannot enable int dma channel %d\n", stat);
    		}
    		DBG(2," Tx EnableIntChannel! pDmaChannel=%p \n",pDmaChannel);
			/* DMAžե饰ON */
			(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_ON;
					
			/* DMAͥư */
			stat = pDmaController->pfDmaStartChannel(pDmaChannel);
    		if (!stat) {
        		printk("Cannot start dma channel %d\n", stat);
    		}
    		DBG(2,"Tx StartChannel! \n");
	}
/* TX_RESTART_DMA */
#endif /* DMA_HOST_CODE */
    } 
    
    DBG(3, "done\n");
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Service an Rx-Ready interrupt for the given endpoint
 * @param pThis instance pointer
 * @param bEnd local endpoint
 */
STATIC void MGC_HdrcServiceRxReady(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t wVal;
    struct urb* pUrb;
    int nPipe;
    uint8_t bIsochError = FALSE;
    uint8_t bDone = FALSE;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
    uint32_t status = USB_ST_NOERROR;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    unsigned long flags;
#ifdef DMA_HOST_CODE
    USB_MCD_DmaChStatus *pDmaChStatus   =   pEnd->pChEpStatus;
#endif
	struct usb_host_endpoint *ep;
	unsigned int epnum;

    DBG(2, "<==\n");

    spin_lock_irqsave(&pThis->Lock, flags);
    MGC_SelectEnd(pBase, bEnd);
    wVal = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);

    spin_lock(&pEnd->Lock);
    pUrb = pEnd->pUrb;
    nPipe = pUrb ? pUrb->pipe : 0;
    spin_unlock(&pEnd->Lock);

#if MUSB_DEBUG > 0
    /* check URB */
    if (pUrb && (pUrb->hcpriv != pEnd) )
    {
        printk(KERN_INFO "%s: end %d has corrupt URB %lx (hcpriv=%lx)!\n", 
	       __FUNCTION__, bEnd, (unsigned long)pUrb, 
	       (unsigned long)pUrb->hcpriv);
		pUrb = NULL;
    }
#endif


    DBG(2, "end %d RxCSR=%04x\n", bEnd, wVal); 

    /* check for errors */
    if(wVal & MGC_M_RXCSR_H_RXSTALL)
    {
		DBG(1, "end %d STALL\n", bEnd); 

		/* stall; record URB status */
        status = USB_ST_STALL;
		/* also record with USB core */
		if(pUrb)
		{
			epnum = usb_pipeendpoint(nPipe)& USB_ENDPOINT_NUMBER_MASK;
			if (usb_endpoint_out(usb_pipeendpoint(nPipe))) {
				ep = pUrb->dev->ep_out[epnum];
			} else {
				ep = pUrb->dev->ep_in[epnum];
			}
	    	USB_HALT_ENDPOINT(pUrb->dev, ep, 0);
	    	/* disableMGC_LinuxDisable()˴ؿ϶ */
		}
    }
    else if(wVal & MGC_M_RXCSR_H_ERROR)
    {

		DBG(1, "end %d Rx error\n", bEnd); 
		DEBUG_CODE(1, MGC_HdrcDumpRegs(pThis, bEnd); );

        status = USB_ST_NORESPONSE;
		/* do the proper sequence to abort the transfer */
		wVal &= ~MGC_M_RXCSR_H_REQPKT;
		MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
		MGC_WriteCsr8(pBase, MGC_O_HDRC_RXINTERVAL, bEnd, 0);
    }
    else if(wVal & MGC_M_RXCSR_DATAERROR)
    {
		DBG(1, "end %d Rx data error\n", bEnd); 
		DEBUG_CODE(1, MGC_HdrcDumpRegs(pThis, bEnd); );

		if(PIPE_BULK == pEnd->bTrafficType)
		{
	    	if(++pEnd->bRetries < MUSB_MAX_RETRIES)
	    	{
	        	/* cover it up if retries not exhausted */
	        	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd,
			       MGC_M_RXCSR_H_REQPKT);
			spin_unlock_irqrestore(&pThis->Lock, flags);
			return;
	    	}
	    	status = USB_ST_NORESPONSE;
	    	/* do the proper sequence to abort the transfer */
	    	wVal &= ~MGC_M_RXCSR_H_REQPKT;
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
	    	MGC_WriteCsr8(pBase, MGC_O_HDRC_RXINTERVAL, bEnd, 0);
		}
		else if(PIPE_ISOCHRONOUS == pEnd->bTrafficType)
		{
	    	bIsochError = TRUE;
	    	DBG(1, "end %d Isochronous error\n", bEnd); 
		}
    }
    
    /* if no errors, be sure a packet is ready for unloading */
    if(!(wVal & MGC_M_RXCSR_RXPKTRDY) && !status)
    {
        status = USB_ST_INTERNALERROR;
		DBG(1, "Rx interrupt with no errors or packet!\n");

		/* do the proper sequence to abort the transfer */
		wVal &= ~MGC_M_RXCSR_H_REQPKT;
		MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
    }

    if(status)
    {
        /* reset error bits */
		DBG(1, "end %d Rx error, status=%d\n", bEnd, status); 
		DEBUG_CODE(1, MGC_HdrcDumpRegs(pThis, bEnd); );

        wVal &= ~(MGC_M_RXCSR_H_ERROR | MGC_M_RXCSR_DATAERROR | 
		  MGC_M_RXCSR_H_RXSTALL | MGC_M_RXCSR_RXPKTRDY);
		MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
		bDone = TRUE;
#ifdef MUSB_CONFIG_PROC_FS
		pEnd->dwErrorRxPackets++;
#endif
    }
    else
    {
        if(pUrb)
		{
	    /* we are expecting traffic */
	   		bDone = MGC_LinuxPacketRx(pThis, bEnd, bIsochError);
    	}
    	else
    	{
	    /* stop endpoint since we have no place for its data */
	    	DBG(1, "no URB on end %d Rx!\n", bEnd); 

	    	wVal |= MGC_M_RXCSR_FLUSHFIFO;
	    	wVal &= ~MGC_M_RXCSR_H_REQPKT;
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
	    	wVal &= ~(MGC_M_RXCSR_FLUSHFIFO | MGC_M_RXCSR_RXPKTRDY);
	    	MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);

	    	/* start next URB that might be queued for it */
	    	spin_lock(&pEnd->Lock);
	    	spin_unlock_irqrestore(&pThis->Lock, flags);
	    	MGC_LinuxStartNextUrb(pThis, bEnd);
	    	spin_unlock(&pEnd->Lock);
	    	return;
    	}
    }

    if(bDone)
    {
    
		/* save data toggle if re-using */
#ifdef DMA_HOST_CODE
		if(pUrb){
			if(usb_pipebulk(nPipe))
			{
#endif
	    		/* we re-use bulk, so re-programming required */
	    		pEnd->bIsReady = FALSE;
	    		/* release claim if borrowed */
	    		if((bEnd != pThis->bBulkRxEnd) &&
	       		(pThis->bBulkTxEnd != pThis->bBulkRxEnd))
	    		{
					pEnd->bIsClaimed = FALSE;
	    		}
	    		/* save data toggle */
	    		usb_settoggle(pUrb->dev, pEnd->bEnd, 0,
					(wVal & MGC_M_RXCSR_H_DATATOGGLE) ? 1 : 0);
#ifdef DMA_HOST_CODE
			}
#endif			
		}
		spin_unlock_irqrestore(&pThis->Lock, flags);
	
		if(pUrb) {
	    	/* set status */
#ifdef DMA_HOST_CODE
			if(status != USB_ST_NOERROR){
				pUrb->status = status;
			}else{
				pUrb->status = USB_ST_NOERROR;
				/* if pUrb->status is not USB_ST_NOERROR, It uses status from PacketRx() */
			}
#endif
	    	DBG(1, "completing Rx URB %lx, end=%d, status=%d, len=%x\n", \
			(unsigned long)pUrb, bEnd, status, pUrb->actual_length); 

	    	/* call completion handler */
	    	if(pUrb->complete)
	    	{
				pUrb->hcpriv=NULL;
				COMPLETE_URB(pUrb, NULL);
	    	}
		}   
		spin_lock(&pEnd->Lock);
		MGC_LinuxStartNextUrb(pThis, bEnd);
		spin_unlock(&pEnd->Lock);
    } else {
#ifdef DMA_HOST_CODE
		if((pDmaChStatus)->State != DMA_HOST_ST_DMA){
	
        	/* continue by clearing RxPktRdy and setting ReqPkt */
			wVal &= ~MGC_M_RXCSR_RXPKTRDY;
    	    	wVal |= MGC_M_RXCSR_H_REQPKT;
			MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wVal);
		}
#endif
		spin_unlock_irqrestore(&pThis->Lock, flags);
    }
    
}
#endif
/**
* Program the HDRC to start (enable interrupts, etc.).
*/
void MGC_HdrcStart(MGC_LinuxCd* pThis)
{
    uint8_t bEnd=1;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    DBG(2, "<==\n");

    /* TODO: always set ISOUPDATE in POWER (periph mode) and leave it on! */

    /*  Set INT enable registers, enable interrupts */
#ifndef MUSB_POLL
    MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, pThis->wEndMask);
    MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, pThis->wEndMask & 0xfffe);
    MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE,
               MGC_M_INTR_SUSPEND | MGC_M_INTR_RESUME | MGC_M_INTR_RESET |
               MGC_M_INTR_BABBLE | MGC_M_INTR_CONNECT | MGC_M_INTR_DISCONNECT |
               MGC_M_INTR_SESSREQ | MGC_M_INTR_VBUSERROR )
#endif
    
    MGC_Write8(pBase, MGC_O_HDRC_TESTMODE, 0);

    /* enable high-speed/low-power and start session */
    MGC_Write8(pBase, MGC_O_HDRC_POWER,
#ifdef MUSB_OTG
	MGC_M_POWER_HSENAB);
#else
	MGC_M_POWER_SOFTCONN | MGC_M_POWER_HSENAB);
#endif

/* when MUSB_OTG is defined, customer code will start a session  
 * at the proper time. When GADGET is defined and OTG is not
 * defined well... that's a problem :)
 */
#ifndef MUSB_GADGET
    /* enable high-speed/low-power and start session & suspend IM host*/
    MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, MGC_M_DEVCTL_SESSION);
#else
#ifndef MUSB_OTG                
    {
    uint8_t state=MGC_Read8(pBase, MGC_O_HDRC_DEVCTL);	
    MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, state & ~MGC_M_DEVCTL_SESSION);
    }
#endif
#endif

    /* init the local ends */
#ifdef MUSB_CONFIG_PROC_FS
    pThis->aLocalEnd[0].dwTotalRxBytes = 0;
    pThis->aLocalEnd[0].dwTotalRxPackets = 0;
    pThis->aLocalEnd[0].dwErrorRxPackets = 0;
    pThis->aLocalEnd[0].dwTotalTxBytes = 0;
    pThis->aLocalEnd[0].dwTotalTxPackets = 0;
    pThis->aLocalEnd[0].dwErrorTxPackets = 0;
#endif
    for(bEnd=1; bEnd < pThis->bEndCount; bEnd++) {    
    	spin_lock( &pThis->aLocalEnd[bEnd].Lock );
#ifdef MUSB_CONFIG_PROC_FS
	pThis->aLocalEnd[bEnd].dwTotalRxBytes = 0;
	pThis->aLocalEnd[bEnd].dwTotalRxPackets = 0;
	pThis->aLocalEnd[bEnd].dwErrorRxPackets = 0;
	pThis->aLocalEnd[bEnd].dwTotalTxBytes = 0;
	pThis->aLocalEnd[bEnd].dwTotalTxPackets = 0;
	pThis->aLocalEnd[bEnd].dwErrorTxPackets = 0;
#endif
    	INIT_LIST_HEAD( &(pThis->aLocalEnd[bEnd].urb_list) );
	pThis->aLocalEnd[bEnd].bIsClaimed=FALSE;
    	spin_unlock( &pThis->aLocalEnd[bEnd].Lock );
    }
}
/**
 * Disable the HDRC (disable & flush interrupts).
 */
STATIC void MGC_HdrcDisable(MGC_LinuxCd* pThis)
{
    uint16_t temp;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    DBG(2, "<==\n");

    /* disable interrupts */
    MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, 0);
    MGC_Write16(pBase, MGC_O_HDRC_INTRTXE, 0);
    MGC_Write16(pBase, MGC_O_HDRC_INTRRXE, 0);

    /* off */
    MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, 0);
    MGC_Write8(pBase, MGC_O_HDRC_POWER, MGC_M_POWER_ENSUSPEND);

    /*  flush pending interrupts */
    temp = MGC_Read8(pBase, MGC_O_HDRC_INTRUSB);
    temp = MGC_Read16(pBase, MGC_O_HDRC_INTRTX);
    temp = MGC_Read16(pBase, MGC_O_HDRC_INTRRX);

    DBG(3, "Interrupts disabled\n");
}
/**
* Program the HDRC to stop (disable interrupts, etc.).
*/
void MGC_HdrcStop(MGC_LinuxCd* pThis)
{
    uint8_t bEnd;

    DBG(2, "<==\n");

    MGC_HdrcDisable(pThis);

    /* flush endpoints */
    for(bEnd = 0; bEnd < pThis->bEndCount; bEnd++)
    {
	MGC_HdrcStopEnd(pThis, bEnd);
    }
}

/**
* Discover HDRC configuration
*/
STATIC uint8_t MGC_HdrcInit(uint16_t wType, MGC_LinuxCd* pThis)
{
#ifdef MUSB_AHB_ID
    uint32_t dwData;
#endif
    MGC_LinuxLocalEnd* pEnd;
    uint8_t bEnd;
#if MUSB_DEBUG > 0
    uint16_t wRelease, wRelMajor, wRelMinor;
#endif /* MUSB_DEBUG > 0 */
#ifndef MCD_PM_CODE
    char aInfo[78];
    uint8_t	reg   = 0;	/* for reading FIFOSIZE register */
#endif
#if MUSB_DEBUG > 0
    char aRevision[32];
    char aDate[12];
    uint8_t bType = 0;
#endif /* MUSB_DEBUG > 0 */
    void* pBase = pThis->pRegs;
#ifndef MUSB_C_DYNFIFO_DEF
    /* how many of a given size/direction found: */
    uint8_t b2kTxEndCount = 0;
    uint8_t b2kRxEndCount = 0;
    uint8_t b1kTxEndCount = 0;
    uint8_t b1kRxEndCount = 0;
    /* the smallest 2k or 1k ends in Tx or Rx direction: */
    uint8_t b2kTxEnd = 0;
    uint8_t b2kRxEnd = 0;
    uint8_t b1kTxEnd = 0;
    uint8_t b1kRxEnd = 0;
    /* for tracking smallest: */
    uint16_t w2kTxSize = 0;
    uint16_t w1kTxSize = 0;
    uint16_t w2kRxSize = 0;
    uint16_t w1kRxSize = 0;
#endif	/* !MUSB_C_DYNFIFO_DEF */

    DBG(2, "<==\n");

#ifdef MUSB_OTG                
    MGC_Write8(pBase, MGC_O_HDRC_DEVCTL, 0);
#endif

    /* log core options */
    MGC_SelectEnd(pBase, 0);

#ifndef MCD_PM_CODE
    reg = MGC_ReadCsr8(pBase, MGC_O_HDRC_CONFIGDATA, 0);

    strcpy(aInfo,(reg & MGC_M_CONFIGDATA_UTMIDW)?"UTMI-16":"UTMI-8");
    if(reg & MGC_M_CONFIGDATA_DYNFIFO) {
        strcat(aInfo, ", dyn FIFOs");
    }
    if(reg & MGC_M_CONFIGDATA_MPRXE) {
    	pThis->bBulkCombine = TRUE;
        strcat(aInfo, ", bulk combine");
    }
    if(reg & MGC_M_CONFIGDATA_MPTXE) {
    	pThis->bBulkSplit = TRUE;
        strcat(aInfo, ", bulk split");
    }
    if(reg & MGC_M_CONFIGDATA_HBRXE) {
        strcat(aInfo, ", HB-ISO Rx");
    }
    if(reg & MGC_M_CONFIGDATA_HBTXE) {
        strcat(aInfo, ", HB-ISO Tx");
    }
    if(reg & MGC_M_CONFIGDATA_SOFTCONE) {
        strcat(aInfo, ", SoftConn");
    }
    
    printk(KERN_INFO "%s: ConfigData=%02x (%s)\n", __FUNCTION__, reg, aInfo);
#endif /* MCD_PM_CODE */
#if MUSB_DEBUG > 0
    /* log release info */
    wRelease = MGC_Read16(pBase, 0x6c);
#ifdef MUSB_AHB_ID
    dwData = MGC_Read32(pBase, 0x404);
    sprintf(aDate, "%04d-%02x-%02x", (dwData & 0xffff), 
	(dwData >> 16) & 0xff,
	(dwData >> 24) & 0xff);
    dwData = MGC_Read32(pBase, 0x408);
    printk("ID2=%lx\n", (long unsigned)dwData);
    dwData = MGC_Read32(pBase, 0x40c);
    printk("ID3=%lx\n", (long unsigned)dwData);
    bType = MGC_Read8(pBase, 0x400);
    pThis->bIsMultipoint=('M' == bType)
    	? TRUE : FALSE;
#else
    pThis->bIsMultipoint=(MUSB_CONTROLLER_MHDRC == wType)
    	? TRUE : FALSE;
#endif

    wRelMajor = (wRelease >> 10) & 0x1f;
    wRelMinor = wRelease & 0x3ff;
    snprintf(aRevision, 32, "%d.%d%s", wRelMajor, 
	    wRelMinor, (wRelease & 0x8000) ? "RC" : "");
    printk(KERN_INFO "%s: %cDRC version %s %s\n", __FUNCTION__, 
	   bType, aRevision, aDate);
#endif /* MUSB_DEBUG > 0 */

#ifndef MCD_PM_CODE
#ifdef MUSB_C_DYNFIFO_DEF
    if(!(reg & MGC_M_CONFIGDATA_DYNFIFO))
    {
        printk(KERN_ERR "%s: Built for Dynamic FIFOs but not detected in hardware; please rebuild\n",
	       __FUNCTION__);
	return FALSE;
    }
#else
    if(reg & MGC_M_CONFIGDATA_DYNFIFO)
    {
        printk(KERN_ERR "%s: Dynamic FIFOs detected but support not built; please rebuild\n",
	       __FUNCTION__);
	return FALSE;
    }
#endif
#endif /* MCD_PM_CODE */

    pThis->bBulkTxEnd = 0;
    pThis->bBulkRxEnd = 0;

	/* EP0 setting */
	musb_fifo_map(pThis, USB_MCD_DFIFO_EP0, USB_MCD_DFIFO_CTRL_SZ, USB_MCD_DFIFO_CTRL_DPB+1, TRUE/* Host */);

    /* discover endpoint configuration */
#ifndef MCD_PM_CODE
    pThis->bEndCount = 1;
#endif
    pThis->wEndMask = 1;
    pEnd = &(pThis->aLocalEnd[0]);
    spin_lock_init(&pEnd->Lock);
    INIT_LIST_HEAD(&(pEnd->urb_list));
    pEnd->wMaxPacketSizeTx = MGC_END0_FIFOSIZE;
    pEnd->wMaxPacketSizeRx = MGC_END0_FIFOSIZE;
    pEnd->bLocalEnd = 0;
#if MUSB_DEBUG > 0
    pEnd->dwPadFront = MGC_PAD_FRONT;
    pEnd->dwPadBack = MGC_PAD_BACK;
#endif

    for(bEnd = 1; bEnd < MUSB_C_NUM_EPS; bEnd++)
    {
        MGC_SelectEnd(pBase, bEnd);
		musb_fifo_map(pThis, \
				usb_mcd_dfifo_setlist[bEnd].ep,\
				usb_mcd_dfifo_setlist[bEnd].fifo_sz,\
				usb_mcd_dfifo_setlist[bEnd].dpb+1,\
				TRUE	/* Host */
		);
		pEnd = &(pThis->aLocalEnd[bEnd]);	
		if(usb_mcd_dfifo_setlist[bEnd].ep & USB_DIR_IN){
			pEnd->wMaxPacketSizeRx = \
				usb_mcd_dfifo_setlist[bEnd].fifo_sz * usb_mcd_dfifo_setlist[bEnd].dpb;
			pEnd->wMaxPacketSizeTx = 0; 
			if(usb_mcd_dfifo_setlist[bEnd].type == PIPE_BULK){
				pThis->bBulkRxEnd = bEnd;
			}
		}else{
			pEnd->wMaxPacketSizeTx = \
				usb_mcd_dfifo_setlist[bEnd].fifo_sz * usb_mcd_dfifo_setlist[bEnd].dpb;
			pEnd->wMaxPacketSizeRx = 0; 
			if(usb_mcd_dfifo_setlist[bEnd].type == PIPE_BULK){
				pThis->bBulkTxEnd = bEnd;
			}
		}
		DBG(2,"wMaxPacketSizeTx=%x wMaxPacketSizeRx=%x\n",pEnd->wMaxPacketSizeTx, pEnd->wMaxPacketSizeRx);
		pEnd->bIsSharedFifo = FALSE;
		spin_lock_init( &pEnd->Lock );
		INIT_LIST_HEAD(&(pEnd->urb_list));
		pEnd->bLocalEnd = bEnd;

#if MUSB_DEBUG > 0
		pEnd->dwPadFront = MGC_PAD_FRONT;
		pEnd->dwPadBack = MGC_PAD_BACK;
#endif
#ifndef MCD_PM_CODE
		pThis->bEndCount++;
#endif
		pThis->wEndMask |= (1 << bEnd);
    }

    if(pThis->bBulkTxEnd)
    {
	pEnd = &(pThis->aLocalEnd[pThis->bBulkTxEnd]);
	pEnd->bIsClaimed = TRUE;
    }
    if(pThis->bBulkRxEnd)
    {
	pEnd = &(pThis->aLocalEnd[pThis->bBulkRxEnd]);
	pEnd->bIsClaimed = TRUE;
    }

    return TRUE;
}

#if MUSB_DEBUG > 0
/**
 *
 */
STATIC uint8_t MGC_IsCorrupt(MGC_LinuxCd* pThis)
{
  uint8_t bEnd;
  MGC_LinuxLocalEnd* pEnd;
  uint8_t bResult = FALSE;

  if(MGC_PAD_FRONT != pThis->dwPadFront)
  {
    printk(KERN_INFO "musb: front pad corrupted (%x)\n", pThis->dwPadFront);
    bResult = TRUE;
  }
  if(MGC_PAD_BACK != pThis->dwPadBack)
  {
    printk(KERN_INFO "musb: back pad corrupted (%x)\n", pThis->dwPadBack);
    bResult = TRUE;
  }
  for(bEnd = 0; bEnd < pThis->bEndCount; bEnd++)
    {
      pEnd = &(pThis->aLocalEnd[bEnd]);
      if(MGC_PAD_FRONT != pEnd->dwPadFront)
	{
	  printk(KERN_INFO "musb: end %d front pad corrupted (%x)\n", 
		 bEnd, pEnd->dwPadFront);
	  bResult = TRUE;
	}
      if(MGC_PAD_BACK != pEnd->dwPadBack)
	{
	  printk(KERN_INFO "musb: end %d back pad corrupted (%x)\n", 
		 bEnd, pEnd->dwPadBack);
	  bResult = TRUE;
	}
    }
  return bResult;
}

/**
 * Dump core registers whose reads are non-destructive.
 */
STATIC void MGC_HdrcDumpRegs(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint8_t* pBase = (uint8_t*)pThis->pRegs;

    MGC_SelectEnd(pBase, bEnd);
    
    if(!bEnd)
    {
	printk(KERN_INFO " 0: CSR0=%04x, Count0=%02x, Type0=%02x, NAKlimit0=%02x\n",
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_CSR0, 0),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_COUNT0, 0),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_TYPE0, 0),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_NAKLIMIT0, 0));
    }
    else
    {
	printk(KERN_INFO "%2d: TxCSR=%04x, TxMaxP=%04x, TxType=%02x, TxInterval=%02x\n",
	       bEnd,
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd),
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_TXTYPE, bEnd),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_TXINTERVAL, bEnd));
	printk(KERN_INFO "    RxCSR=%04x, RxMaxP=%04x, RxType=%02x, RxInterval=%02x, RxCount=%04x\n",
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd),
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_TXMAXP, bEnd),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_TXTYPE, bEnd),
	       MGC_ReadCsr8(pBase, MGC_O_HDRC_TXINTERVAL, bEnd),
	       MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd));
    }
    
    if( pThis->bIsHost && pThis->bIsMultipoint)
    {
    printk(KERN_INFO "    TxAddr=%02x, TxHubAddr=%02x, TxHubPort=%02x\n",
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXFUNCADDR)),
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBADDR)),
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_TXHUBPORT)));
    printk(KERN_INFO "    RxAddr=%02x, RxHubAddr=%02x, RxHubPort=%02x\n",
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXFUNCADDR)),
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBADDR)),
	   MGC_Read8(pBase, MGC_BUSCTL_OFFSET(bEnd, MGC_O_HDRC_RXHUBPORT)));
    }
}
#endif

/**************************************************************************
 * Linux HCD functions
**************************************************************************/
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Find first (lowest-order) 1 bit
 * @param nValue value in which to search
 * @return bit position (0 could mean no bit; caller should check)
 */
STATIC uint8_t MGC_FindFirst1(unsigned int nValue)
{
    unsigned int nWork = nValue;
    uint8_t bResult;

    for(bResult = 0; bResult < 32; bResult++)
    {
	if(nWork & 1)
	{
	    return bResult;
	}
	nWork >>= 1;
    }
    return bResult;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Generic timer creation.
 * @param pThis instance pointer
 * @param pfFunc timer fire callback
 * @param pParam parameter for callback
 * @param millisecs how many milliseconds to set
 */
STATIC void MGC_LinuxSetTimer(MGC_LinuxCd* pThis, 
			      void (*pfFunc)(unsigned long),
			      unsigned long pParam, unsigned long millisecs)
{
    DBG(2, "<==\n");

    init_timer(&(pThis->Timer));
    pThis->Timer.function = pfFunc;
    pThis->Timer.data = (unsigned long)pParam;
    pThis->Timer.expires = jiffies + (HZ * millisecs) / 1000;
    add_timer (&(pThis->Timer));
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Find an endpoint for the given pipe
 *
 * @param pThis instance pointer
 * @param pURB URB pointer
 * @return suitable local endpoint
 * @return -1 if nothing appropriate
 */
STATIC int MGC_LinuxFindEnd(MGC_LinuxCd* pThis, struct urb* pUrb)
{
    MGC_LinuxLocalEnd* pEnd;
    uint8_t bDirOk, bTrafficOk, bSizeOk, bExact;
    int nEnd;
    int32_t dwDiff;
    uint16_t wBestDiff = 0xffff;
    uint16_t wBestExactDiff = 0xffff;
    int nBestEnd = -1;
    int nBestExactEnd = -1;
    unsigned int nPipe = pUrb->pipe;
    unsigned int nOut = usb_pipeout(nPipe);
    uint16_t wPacketSize = USB_MCD_MAXP(usb_mcd_maxpacket(pUrb->dev, nPipe, nOut));
    uint8_t bEnd = usb_pipeendpoint(nPipe);
    uint8_t bIsBulk = usb_pipebulk(nPipe);
    uint8_t bAddress = (uint8_t)usb_pipedevice(nPipe);
    
    DBG(2, "<==\n");

    /* control is always EP0, and can always be queued */
    if(usb_pipecontrol(pUrb->pipe))
    {
    	DBG(3, "is a control pipe use ep0\n");
        return 0;
    }

    /* use a reserved one for bulk if any */
    if(bIsBulk)
    {
	if(nOut && pThis->bBulkTxEnd)
	{
	    DBG(3, "use the bulk tx end\n");
	    return pThis->bBulkTxEnd;
	} else if(!nOut && pThis->bBulkRxEnd)
	{
	    DBG(3, "use the bulk rx end\n");
	    return pThis->bBulkRxEnd;
	}
    }

    /* scan, remembering exact match and best match (bulk only) */
    for(nEnd = 1; nEnd < pThis->bEndCount; nEnd++)
    {
        pEnd = &(pThis->aLocalEnd[nEnd]);
	/* consider only if direction is possible */
	bDirOk = (nOut && pEnd->wMaxPacketSizeTx) || 
	  (!nOut && pEnd->wMaxPacketSizeRx);
	/* consider only if size is possible (in the given direction) */
	bSizeOk = (nOut && (pEnd->wMaxPacketSizeTx >= wPacketSize)) ||
	  (!nOut && (pEnd->wMaxPacketSizeRx >= wPacketSize));
	/* consider only traffic type */
	bTrafficOk = (usb_pipetype(nPipe) == pEnd->bTrafficType);
	
	if(bDirOk && bSizeOk)
	{
	    /* convenient computations */
	    dwDiff = nOut ? (pEnd->wMaxPacketSizeTx - wPacketSize) : 
	      (pEnd->wMaxPacketSizeRx - wPacketSize);
	    bExact = bTrafficOk && (pEnd->bEnd == bEnd) &&
		(pEnd->bAddress == bAddress);

	    /* bulk: best size match not claimed (we only claim periodic) */
	    if(bIsBulk && !pEnd->bIsClaimed && (wBestDiff > dwDiff))
	    {
	        wBestDiff = (uint16_t)dwDiff;
		nBestEnd = nEnd;
		/* prefer end already in right direction (to avoid flush) */
		if((wBestExactDiff > dwDiff) && (nOut == (int)pEnd->bIsTx))
		{
		    wBestExactDiff = (uint16_t)dwDiff;
		    nBestExactEnd = nEnd;
		}
	    }
	    else if(!bIsBulk && (nEnd != pThis->bBulkTxEnd) &&
		(nEnd != pThis->bBulkRxEnd))
	    {
		/* periodic: exact match if present; otherwise best unclaimed */
		if(bExact)
		{
		    nBestExactEnd = nEnd;
		    break;
		}
		else if(!pEnd->bIsClaimed && (wBestDiff > dwDiff))
		{
	            wBestDiff = (uint16_t)dwDiff;
		    nBestEnd = nEnd;
		}
	    }
	}
	
    }

    DBG(2, "(out=%d, size=%d, proto=%d, addr=%d, end=%d, urb=%lx) = %d\n", \
	   nOut, wPacketSize, usb_pipetype(nPipe), \
	   bAddress, bEnd, (unsigned long)pUrb, \
	   (nBestExactEnd >= 0) ? nBestExactEnd : nBestEnd); 

    return (nBestExactEnd >= 0) ? nBestExactEnd : nBestEnd;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Receive a packet
 * @requires pThis->Lock locked
 * @return TRUE if URB is complete
 */
STATIC uint8_t MGC_LinuxPacketRx(MGC_LinuxCd* pThis, uint8_t bEnd,
				 uint8_t bIsochError)
{
    uint16_t wRxCount;
    uint16_t wLength	= 0;
    uint8_t* pBuffer	= NULL;
    uint16_t wCsr;
    uint8_t bDone 		= FALSE;
    uint8_t* pBase 		= (uint8_t*)pThis->pRegs;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
    struct urb* pUrb = pEnd->pUrb;
    int nPipe=0;
    void *buffer=NULL;
    
#ifdef HOST_NULL_FLAG
	USB_MCD_HostInfo *HostInfo = &(pEnd->HostInfo);
#endif

#ifdef DMA_HOST_CODE
    uint8_t	err_stat 	= 0;
    uint8_t stat		= 0;
	MGC_DmaController* pDmaController	= pThis->pDmaController;
	MGC_DmaChannel* pDmaChannel 		= pEnd->pDmaChannel;
	USB_MCD_DmaChStatus *pDmaChStatus   = pEnd->pChEpStatus;
#endif

    // MGC_SelectEnd(pBase, bEnd);
    wRxCount = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);
    DBG(2, "end %d RxCount=%04x\n", bEnd, wRxCount); 

#ifdef MUSB_PARANOID
    if ( !pUrb || ((pUrb->transfer_buffer_length - pEnd->dwOffset)<0) )  {
	ERR("ERROR during Rx: pUrb=%p, pUrb->transfer_buffer_length=%d pEnd->dwOffset=%d\n", \
	   pUrb, pUrb->transfer_buffer_length, pEnd->dwOffset );
	return TRUE;
    }    
#endif

    nPipe=pUrb->pipe;

    buffer=pUrb->transfer_buffer;
    DBG(2, "pUrb->transfer_flags=0x%x pUrb->transfer_buffer=%p\n",\
		pUrb->transfer_flags, pUrb->transfer_buffer);
    DBG(2, "pUrb->transfer_dma=%p \n",(void *)pUrb->transfer_dma);


#if	defined(DMA_HOST_CODE)
/*===========================================================================*/
	if((pDmaChStatus)->dma_req_flag){
	/* DMA request exist */
		buffer = (void*)(__bus_to_virt(pUrb->transfer_dma));
		DBG(2,"buffer=%p transfer_dma used in DMAmode \n",buffer);
	}else{
#ifdef	MUSB_DMA_BUS2VIRT 
    	if ( !(pUrb->transfer_buffer) ) {
			buffer = (void*)(__bus_to_virt(pUrb->transfer_dma));
			DBG(2,"buffer=%p transfer_dma used in not DMAmode \n",buffer);
		}
		else{
			DBG(2,"buffer=%p transfer_buffer used \n",buffer);
		}
#else
		buffer = (void*)(pUrb->transfer_buffer);
	    DBG(2,"buffer=%p \n",buffer);
#endif
	}

#endif
/*===========================================================================*/

    if ( !buffer ) { // abort the transfer
		ERR("Rx requested but no buffer was given, aborting\n");
		return TRUE;  
    }

#if defined(DMA_HOST_CODE) && defined(DMA_HOST_CODE_RX_IRQ)

	if( usb_pipeisoc(nPipe) )	/* ISOž	*/
   	{
   	/* ISO case */
   		/*===========================================================================*/
   		if(((pEnd->pChEpStatus)->dma_req_flag)&&((pEnd->pChEpStatus)->dma_ch_flag)){
		/* DMA transfer is requested and DMA channel exist */

			/* get buffer address to store rx data */
			pBuffer = buffer + pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;

			/*===========================================================================*/
			if(wRxCount == (uint16_t)NULL){
			/* NULLξ */
			
				/* received data size */
				wLength = (uint16_t)NULL;
				/* update actual & status */
				pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length = wLength;
				/* store error status */
				if(bIsochError)
				{
	    			pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_CRC;
	    			pUrb->error_count++;
				}
				else
				{
	    			pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_NOERROR;
				}
				/* see if we are done */ 
       			bDone = (++pEnd->dwIsoPacket >= pUrb->number_of_packets);

				if(!bDone){	/* ž̤λ */
					(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
					DBG(2,"Change State CPU at NULL data in ISO DMA \n");
				}else{		/* žλ */
					(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
					DBG(2,"Change State CPU at NULL data in ISO DMA \n");
					/* DMA channel exist */
					if((pEnd->pChEpStatus)->dma_ch_flag){
   	 					/* DMAͥ */
   	 					pDmaController->pfDmaReleaseChannel(pDmaChannel);
						DBG(2,"ReleaseChannel at NULL data in ISO DMA\n");
						pEnd->pDmaChannel = NULL;
   	 				
   	 					(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;	/* DMA̵ͥͭե饰OFF */
   	 				}
				}
				/* to be continued in common block; ³ν϶̽ǹԤ*/
			/*===========================================================================*/
			}else{
			/* NULLʳ */
				
				/* get a size to load data within MAXP as maximum */
				wLength = min((unsigned int)wRxCount,
		    		pUrb->iso_frame_desc[pEnd->dwIsoPacket].length);
				/* update size of data in EP */
				pEnd->dwOffset += wLength;
				/* update actual & status */
				pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length = wLength;
				
				/* store error status */
				if(bIsochError)
				{
	    			pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_CRC;
	    			pUrb->error_count++;
				}
				else
				{
	    			pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_NOERROR;
				}
				
				(pEnd->pChEpStatus)->State = DMA_HOST_ST_DMA;	/* to DMA state */
				DBG(2,"Change State DMA at non NULL data in ISO DMA \n");
				/* ++pEnd->dwIsoPacket DMAߤǹԤ(DMAmode0) */
				
				/* ³ν϶̽ǹԤ*/
			}
		/*===========================================================================*/
		}else{
		/* ISO CPUž */
			/* DMA channel exist */
			if((pEnd->pChEpStatus)->dma_ch_flag){
   	 			/* Release DMA channel */
   	 			pDmaController->pfDmaReleaseChannel(pDmaChannel);
				DBG(2,"ReleaseChannel data in ISO CPU\n");
				pEnd->pDmaChannel = NULL;
   	 			(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;	/* DMA̵ͥͭե饰OFF */
   	 		}
			
			/* get buffer address to store rx data */
			pBuffer = buffer + pUrb->iso_frame_desc[pEnd->dwIsoPacket].offset;
			/* get a size to load data */
			wLength = min((unsigned int)wRxCount,
		    	pUrb->iso_frame_desc[pEnd->dwIsoPacket].length);
			/* update size of data in EP */
			pEnd->dwOffset += wLength;
			/* update actual & status */
			pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length = wLength;
				
			/* store error status */
			if(bIsochError)
			{
	    		pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_CRC;
	    		pUrb->error_count++;
			}
			else
			{
	    		pUrb->iso_frame_desc[pEnd->dwIsoPacket].status = USB_ST_NOERROR;
			}
			
			/* see if we are done */
       		bDone = (++pEnd->dwIsoPacket >= pUrb->number_of_packets);
       			
			if(bDone){	/* žλ */
				(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
				DBG(2,"Change State IDLE in ISO CPU \n");
			}else{		/* ž̤λ */
				(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
				DBG(2,"Change State CPU in ISO CPU \n");
			}
    		/* ³ν϶̽ǹԤ*/
    	}
    } else {
    /*===========================================================================*/
    /* NON ISO  */
    	
    	if(((pEnd->pChEpStatus)->dma_req_flag)&&((pEnd->pChEpStatus)->dma_ch_flag)){
		 /*===========================================================================*/
   		 /* NonISO DMA */
    	/* DMA */
    		if(((pEnd->pChEpStatus)->dma_trans_flag)&&((pEnd->pChEpStatus)->State==DMA_HOST_ST_DMA)){
   	 			/* DMA */
   	 			stat = pDmaController->pfDmaStopChannel(pDmaChannel);
				DBG(2,"Release Channel in Non ISO DMA \n");
				/* DMA߶ػ */
				stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
				DBG(2,"Disable Channel in Non ISO DMA \n");
    			/* get length of data  transfered in DMA */
   	 			stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
				DBG(2,"GetChannelStatus in Non ISO DMA \n");
   	 			if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 				printk("DMA mode1 Error BUS_ABORT \n"); 
   	 			}
				
				/* update length received */
		   	 	pEnd->dwOffset +=  pDmaChannel->dwActualLength;

   	 			(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;	/* DMAžե饰OFF */
   	 	
   	 		}
   	 		/*===========================================================================*/
    		if(wRxCount == pEnd->wPacketSize){	/* ǧ */
			/* MAXP Packet */

    			pBuffer = buffer + pEnd->dwOffset;
				/* get length */
				wLength = min((unsigned int)wRxCount, 
		   			pUrb->transfer_buffer_length - pEnd->dwOffset);
   	 			
   	 			if(wLength < wRxCount){
   	 			/* Сǡ */
   	 				bDone		= TRUE;	/* žλ */
   	 				err_stat	= 1;	/* ̽FIFOեåԤ*/
   	 				(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
					DBG(2,"Change State IDLE at MAXP in Non ISO IDLE \n");
  	 			}else{
   	 			/* Сǡʤ */
   	 				switch((pEnd->pChEpStatus)->mode){
   	 					case DMA_HOST_MODE0:
   	 						pEnd->dwRequestSize = wRxCount;   	 					
   	 				   	 	/* žλȽDMA߽ˤƹԤ */
   	 						(pEnd->pChEpStatus)->State = DMA_HOST_ST_DMA;	/* to DMA state */
							DBG(2,"Change State DMA at MAXP in NonISO \n");
   	 						break;
   	 					
   	 					case DMA_HOST_MODE1:
							/* update length */
							pEnd->dwOffset += wLength;

							/* see if we are done */
							bDone = (pEnd->dwOffset >= pUrb->transfer_buffer_length) ||\
	   							(wRxCount < pEnd->wPacketSize);
   	 						
   	 						if(bDone){
   	 							if((HostInfo->null_flag)){
   	 							
   	 								(pEnd->pChEpStatus)->State = DMA_HOST_ST_NULL;	/* to NULL state */
									DBG(2,"Change State NULL at MAXP in Non ISO IDLE \n");
   	 								/* transfer end */
   	 								bDone = FALSE;
   	 							}else{
   	 								(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
									DBG(2,"Change State IDLE at MAXP in Non ISO IDLE \n");
   	 							}
   	 						}else{
   	 							(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;		/* to CPU state */
								DBG(2,"Change State CPU at MAXP in Non ISO IDLE \n");

   	 						}
   	 						break;
						case DMA_HOST_MODE_NON:
							printk("abnormal Error! \n");
							break;
   	 				}
   	 			}
   	 			/* ³ν϶̽ǹԤ*/
   	 		
   	 		/*===========================================================================*/
    		}else if(wRxCount > (uint16_t)NULL && wRxCount < pEnd->wPacketSize){
    		/* SHORT Packet */
    			
    			pBuffer = buffer + pEnd->dwOffset;
				/* get length */
				wLength = min((unsigned int)wRxCount, 
		   			pUrb->transfer_buffer_length - pEnd->dwOffset);
				/* see if we are done */

   	 			if(wLength < wRxCount){
   	 			/* Сǡ */
					/* update length */
					pEnd->dwOffset += wLength;
   	 				err_stat = 1;	/* ̽FIFOեåԤ */
   	 				bDone = TRUE;
   	 				(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
   	 			}else{
   	 			/* Сǡʤ */
   	 				switch((pEnd->pChEpStatus)->mode){
   	 					case DMA_HOST_MODE0:
   	 						pEnd->dwRequestSize = wRxCount;
   	 						(pEnd->pChEpStatus)->State = DMA_HOST_ST_DMA;	/* to DMA state */
							DBG(2,"Change State DMA at Short in NonISO \n");
							break;
   	 					case DMA_HOST_MODE1:
							/* update length */
							pEnd->dwOffset += wLength;
   	 						bDone = TRUE;
   	 						(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
							DBG(2,"Change State IDLE at Short in NonISO \n");
   	 						break;
						case DMA_HOST_MODE_NON:
							printk("abnormal Error! \n");
							break;
   	 				}
   	 			}
   	 			/* ³ν϶̽ǹԤ*/
    		
    		/*===========================================================================*/
    		}else if(wRxCount == (uint16_t)NULL){
    		/* NULL Packet */
    			/* get length */
				wLength = (uint16_t)NULL;
				/* transfer end */
   	 			bDone = TRUE;
   	 			
   	 			(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
				DBG(2,"Change State IDLE at NULL in Non ISO IDLE \n");
   	 				
    		}else{
    		/* ʤ */
    		}
    		/*===========================================================================*/
    		if(((pEnd->pChEpStatus)->dma_ch_flag)&&((pEnd->pChEpStatus)->State!=DMA_HOST_ST_DMA)){
   	 			
   	 			/* release DMA channel */
   	 			pDmaController->pfDmaReleaseChannel(pDmaChannel);
				DBG(2,"Release Channel in Non ISO IDLE \n");
				pEnd->pDmaChannel = NULL;
   	 				
   	 			/* clear channel flag */
   	 			(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
   	 		}
    	}else{
    	/*===========================================================================*/
    	/* NON ISO by CPU */
    		pBuffer = buffer + pEnd->dwOffset;
			/* get length */
			wLength = min((unsigned int)wRxCount, 
		   	pUrb->transfer_buffer_length - pEnd->dwOffset);
			/* update length */
			pEnd->dwOffset += wLength;
			/* see if we are done */
			bDone = (pEnd->dwOffset >= pUrb->transfer_buffer_length) ||\
	   		(wRxCount < pEnd->wPacketSize);

#ifdef OVER_DATA_ERR	   		
   	 		if(wLength < wRxCount){
   	 		/* Сǡ */
   	 		bDone		= TRUE;
   	 		err_stat	= 1;	/* ̽FIFOեåԤ */
   	 		(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
			DBG(2,"Change State IDLE at OVERDATA in NonISO CPU \n");
   	 		}else{
#endif
   	 		/* Сǡ*/
   	 			if(bDone){
   	 			if((wRxCount == pEnd->wPacketSize) && pEnd->HostInfo.null_flag){
   	 					(pEnd->pChEpStatus)->State = DMA_HOST_ST_NULL;	/* to NULL state */
   	 					bDone = FALSE;
						DBG(2,"Change NULL State RxCount=%d \n",wRxCount);
   	 				}else{
   	 					(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
						DBG(2,"Change IDLE State RxCount=%d \n",wRxCount);
   	 				}
   	 			}else{
   	 				(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;		/* to CPU state */
					DBG(2,"Change CPU State RxCount=%d \n",wRxCount);
   	 			}
#ifdef OVER_DATA_ERR	   		
   	 		}
#endif
   	 		/* ³ν϶̽ǹԤ*/
    	}
    }
    /*===========================================================================*/

	if((pEnd->pChEpStatus)->State == DMA_HOST_ST_DMA){
	/* DMA state case*/
    	/* DMAХå򥳡뤷ȤNULLƤΤǺ */
		/* CallBack */
		pDmaChannel->pfDmaCallback = USB_MCD_HostDmaCallback;
 	
    	/* DMA setting */
       	stat = pDmaController->pfDmaProgramChannel(
        	pDmaChannel,
        	(MGC_pfDmaCallback)(pDmaChannel->pfDmaCallback),
        	(void*)pThis,
        	(uint16_t)(pEnd->wPacketSize),
        	(pEnd->pChEpStatus)->mode,
        	__virt_to_bus(pBuffer),
        	wLength);
       	if (!stat) {
   	 		if((pEnd->pChEpStatus)->dma_ch_flag){
   	 			/* release DMA channel */
   	 			pDmaController->pfDmaReleaseChannel(pDmaChannel);
				DBG(2,"Release Channel by Program error in DMA State\n");
				pEnd->pDmaChannel = NULL;
   	 			(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;	/* DMA channel flag OFF */
   	 		}
			(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;		/* to IDLE state */
			DBG(2,"Change State IDLE by Program error in DMA \n"); 
			/* žλ */
        	bDone =	TRUE;
    	}
		/* set actual length */
		pDmaChannel->dwActualLength = 0;	/* valid in Mode0 */
		/* DMA interrupt enable */
   	 	pDmaController->pfDmaEnableIntChannel(pDmaChannel);
   	 	/* DMA transfer flag set on */
   	 	pDmaChStatus->dma_trans_flag	=	USB_MCD_FLAG_ON;
   	 	/* Start transferring  by DMA  */
   	 	pDmaController->pfDmaStartChannel(pDmaChannel);
		DBG(2,"StartChannel in DMA \n");

#ifdef MUSB_CONFIG_PROC_FS
    	pEnd->dwTotalRxBytes += wLength;
    	pEnd->dwTotalRxPackets += (wLength/(pEnd->wPacketSize)); 
#endif

	}else{
	/* non DMA state case */
	
		/* FIFO unload */
		MGC_HdrcUnloadFifo(pBase, bEnd, wLength, pBuffer);

#ifdef MUSB_CONFIG_PROC_FS
    	pEnd->dwTotalRxBytes += wLength;
    	pEnd->dwTotalRxPackets++;			
#endif	

		if(bDone){/* if transfer ended */
			/* FIFO flash */
			if(err_stat == TRUE){
			/* Over data received */
				pUrb->status = USB_ST_DATAOVERRUN;
			}
			wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
#ifdef HOST_CLR_REQPKT
			wCsr &= ~(MGC_M_RXCSR_RXPKTRDY|MGC_M_RXCSR_H_REQPKT);
#else
			wCsr &= ~(MGC_M_RXCSR_RXPKTRDY);
#endif			
#ifdef FIFO_FLASH
			wCsr |= MGC_M_RXCSR_FLUSHFIFO;
#endif
			/* Clear RxPktRdy, FIFOեå */
			MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
			if(HostInfo->db_buff){	/* in case of Double buffer */
				wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
#ifdef HOST_CLR_REQPKT
				wCsr &= ~(MGC_M_RXCSR_RXPKTRDY|MGC_M_RXCSR_H_REQPKT);
#else
				wCsr &= ~(MGC_M_RXCSR_RXPKTRDY);
#endif			
#ifdef FIFO_FLASH
				wCsr |= MGC_M_RXCSR_FLUSHFIFO;
#endif
				/* Clear RxPktRdy, FIFOեå */
				MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
			}
			/* FIFO status free */
	    	pEnd->bIsReady = TRUE;
			/* Update transfer size */
			pUrb->actual_length = pEnd->dwOffset;
		}
	}
	DBG(2,"PacketRxend  \n");
	DEBUG_CODE(3, if ( bDone ) { printk(KERN_INFO \
    "%s: completing %d-packet isoch URB; len=%x, errors=%d\n", \
	__FUNCTION__, pUrb->number_of_packets, pUrb->actual_length, \
	pUrb->error_count); } );

	return bDone;

#endif /* DMA_HOST_CODE */
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Start the next URB on an endpoint. Wants the _endpoint_ to be locked.
 * It might call MGC_LinuxStartUrb pThis needs to be locked as well.
 *
 * @param pThis instance pointer
 * @param pUrb pointer to just-d URB
 * @param bEnd local endpoint
 */
STATIC void MGC_LinuxStartNextUrb(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
    struct urb* pUrb = pEnd->pUrb;

    DBG(2, "<==\n");

    /* interrupt IN URBs are automatically re-submitted unless unlinked */
    if(bEnd && pUrb && usb_pipeint(pUrb->pipe) && usb_pipein(pUrb->pipe) &&
       pEnd->bIsClaimed)
    {
	pUrb->hcpriv = pEnd;
	MGC_LinuxStartUrb(pThis, bEnd);
	return;
    }

    /* check for linked URB */
    if(!list_empty(&pEnd->urb_list))
    {
        if(pEnd->pUrb) {
	    usb_put_urb(pEnd->pUrb);
	}
	
        pUrb = list_entry(pEnd->urb_list.next, struct urb, urb_list);
	list_del(&pUrb->urb_list);
	pEnd->pUrb = pUrb;

	DBG(1, "dequeued URB %lx\n", (unsigned long)pEnd->pUrb); 

    }
    else
    {
	pEnd->pUrb = NULL;
	DBG(1, "end %d idle\n", bEnd); 
    }

    if(pEnd->pUrb)
    {
	MGC_LinuxStartUrb(pThis, bEnd);
    }
}
#endif

#if defined(CONFIG_USB_MCD_HOST)
/**
 * Start the current URB on an endpoint; wants ep to be 
 * locked and pThis to be locked as well; end must be claimed from 
 * the caller
 */
STATIC void MGC_LinuxStartUrb(MGC_LinuxCd* pThis, uint8_t bEnd)
{
    uint16_t wFrame;
    uint32_t dwLength;
    void* pBuffer;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
    struct urb* pUrb = pEnd->pUrb;
    unsigned int nPipe = pUrb->pipe;
    unsigned int nOut = usb_pipeout(nPipe);
    uint16_t wPacketSize = USB_MCD_MAXP(usb_mcd_maxpacket(pUrb->dev, nPipe, nOut));
    uint8_t bAddress = (uint8_t)usb_pipedevice(nPipe);
    uint8_t bRemoteEnd = (uint8_t)usb_pipeendpoint(nPipe);

#ifdef DMA_HOST_CODE
	uint8_t bSizeVal;
    uint8_t stat;
	MGC_DmaController* pDmaController = pThis->pDmaController;
	MGC_DmaChannel* pDmaChannel;
	USB_MCD_DmaChStatus *pDmaChStatus = pThis->aLocalEnd[bEnd].pChEpStatus;
#endif   
    DBG(2, "<==\n");

    /* if no root device, assume this must be it */
    if(!pThis->pRootDevice)
    {
	pThis->pRootDevice = pUrb->dev;
	switch(pThis->bRootSpeed)
	{
	  case 1:
	    pThis->pRootDevice->speed = USB_SPEED_HIGH;
	    break;
	  case 2:
	    pThis->pRootDevice->speed = USB_SPEED_FULL;
	    break;
	  case 3:
	    pThis->pRootDevice->speed = USB_SPEED_LOW;
	    break;
	}
    }

    /* indicate in progress */
    pUrb->status = -EINPROGRESS;
    pUrb->actual_length = 0;
    pUrb->error_count = 0;
    
    DBG(1, "(%p): nOut=%d, wPacketSize=%d, bAddr=%d, bEnd=%d \n", pUrb, nOut, wPacketSize, bAddress, bRemoteEnd); 
    DBG(1,"pUrb->transfer_buffer=%p \n",pUrb->transfer_buffer);

    DBG(1, "pUrb->transfer_dma=%p \n",(void *)pUrb->transfer_dma);

    /* remember software state */
    pEnd->dwOffset = 0;
    pEnd->dwRequestSize = 0;
    pEnd->dwIsoPacket = 0;
    pEnd->dwWaitFrame = 0;
    pEnd->bRetries = 0;
    pEnd->wPacketSize = wPacketSize;
    pEnd->bAddress = bAddress;
    pEnd->bEnd = bRemoteEnd;
    pEnd->bTrafficType = (uint8_t)usb_pipetype(nPipe);
    pEnd->bIsTx=(nOut)?TRUE:FALSE;

#ifdef DMA_HOST_CODE
	/* Initialize */
	pDmaChStatus->index 			=	0;
    pDmaChStatus->Ep				=	bEnd;					/*	EPֹ楻å */
	pDmaChStatus->State 			=	DMA_HOST_ST_IDLE;		/* State for transfer set */
	pDmaChStatus->mode				=	DMA_HOST_MODE_NON;
	pDmaChStatus->dma_trans_flag	=	0;
	pDmaChStatus->dma_req_flag		=	0;
	pDmaChStatus->dma_ch_flag		=	0;
	pDmaChStatus->dwStartAddress	=	0;
	
	pEnd->HostInfo.null_flag		=	0;
	pEnd->HostInfo.db_buff			=	TRUE;
#endif
	
#ifdef DMA_HOST_CODE
#ifdef DMA_HOST_DPB_SET /* Хåե⡼(single or double) */
		/* DPB */
   		MGC_SelectEnd(pBase, bEnd);
#endif
#ifdef DMA_REQ_SET_CHANGE
	if((pUrb->transfer_flags&URB_NO_TRANSFER_DMA_MAP)&&(pUrb->transfer_buffer==NULL)){
#else
	if(pUrb->transfer_flags&URB_NO_TRANSFER_DMA_MAP){
#endif /* DMA_REQ_SET_CHANGE */
	/* DMAž׵᤬ */
#ifdef DMA_HOST_DISABLE
		pDmaChStatus->dma_req_flag = USB_MCD_FLAG_OFF;		/* dma_req_flag set off */
#else
		pDmaChStatus->dma_req_flag = USB_MCD_FLAG_ON;		/* dma_req_flag set on */
#endif
#ifdef DMA_HOST_DPB_SET
		/* DPB */
    	/* EP0ʳξꤹ sizeϤꤵƤΤȤ롣*/
    	if (!bEnd || pEnd->bIsTx) { /* OUT */
       		bSizeVal = MGC_Read8(pBase, MGC_O_HDRC_TXFIFOSZ);
    		bSizeVal &= ~MGC_M_FIFOSZ_DPB;
        	MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, bSizeVal);
    	}
    	if (!bEnd || !pEnd->bIsTx) {	/* IN */
    		bSizeVal = MGC_Read8(pBase, MGC_O_HDRC_RXFIFOSZ);
    		bSizeVal &= ~MGC_M_FIFOSZ_DPB;
        	MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, bSizeVal);
    	}
   		
		pEnd->HostInfo.db_buff	=	FALSE;			/* single buffer mode */
#endif	/* DMA_HOST_DPB_SET end */
	}else{
		pDmaChStatus->dma_req_flag = USB_MCD_FLAG_OFF;		/* dma_req_flag set off */
		pDmaChStatus->State = DMA_HOST_ST_CPU;				/* to CPU state */

#ifdef DMA_HOST_DPB_SET
    	/* EP0ʳξꤹ sizeϤꤵƤΤȤ롣*/
    	if (!bEnd || pEnd->bIsTx) { /* OUT */
       		bSizeVal = MGC_Read8(pBase, MGC_O_HDRC_TXFIFOSZ);
    		bSizeVal |= usb_mcd_dfifo_setlist[bEnd].dpb<<4;
    		DBG(2,"EP[%d] dpb=%x",bEnd, usb_mcd_dfifo_setlist[bEnd].dpb<<4);

        	MGC_Write8(pBase, MGC_O_HDRC_TXFIFOSZ, bSizeVal);
    	}
    	if (!bEnd || !pEnd->bIsTx) {	/* IN */
    		bSizeVal = MGC_Read8(pBase, MGC_O_HDRC_RXFIFOSZ);
    		bSizeVal |= usb_mcd_dfifo_setlist[bEnd].dpb<<4;
    		DBG(2,"EP[%d] dpb=%x",bEnd, usb_mcd_dfifo_setlist[bEnd].dpb<<4);

        	MGC_Write8(pBase, MGC_O_HDRC_RXFIFOSZ, bSizeVal);
    	}
   		
		pEnd->HostInfo.db_buff	=	TRUE;			/* double buffer mode */
#endif /* DMA_HOST_DPB_SET */
	}
#endif
#ifdef DMA_HOST_CODE
#ifdef HOST_NULL_FLAG
	/* NULL flag set	*/
	if(pUrb->transfer_flags&URB_ZERO_PACKET){
		pEnd->HostInfo.null_flag = USB_MCD_FLAG_ON;
	}else{
#ifdef HOST_NULL_FLAG_FORCE
		pEnd->HostInfo.null_flag = USB_MCD_FLAG_ON;
#else
		pEnd->HostInfo.null_flag = USB_MCD_FLAG_OFF;
#endif
	}
#else
	pEnd->HostInfo.null_flag = USB_MCD_FLAG_OFF;	/* for test */
#endif 
#endif
    /* pEnd->bIsClaimed=(usb_pipeisoc(nPipe) || usb_pipeint(nPipe)) ?TRUE:FALSE; 
     * end must be claimed from my caller
    */
    
    if(usb_pipecontrol(nPipe))
    {
		/* control transfers always start with an OUT */
		nOut = 1;
		pThis->bEnd0Stage = MGC_END0_START;
    }

    /* gather right source of data */
    if(usb_pipeisoc(nPipe))
    {
#ifdef TEST_BULKERR
		DBG(2,"StartUrb PIPE ISO  \n");
#endif
#ifdef DMA_HOST_CODE
		/*== get address of a buffer =======================================*/
		if((pDmaChStatus)->dma_req_flag){
			pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
		}else{
#ifdef MUSB_DMA_BUS2VIRT
			if (!(pUrb->transfer_buffer)) {
				pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            }else{
            	pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            }
            if (!pBuffer ) { // abort the transfer
            	ERR("no buffer was given, aborting\n");
				return;
            }
#else /* MUSB_DMA_BUS2VIRT */
	    	pBuffer = pUrb->transfer_buffer;
#endif
		}
		pBuffer +=  pUrb->iso_frame_desc[0].offset;

#endif /* DMA_HOST_CODE */
		/*== get transfer length ==== =======================================*/
		dwLength = pUrb->iso_frame_desc[0].length;
		DBG(1, "trans Buffer[virt]=%p length=%d \n",pBuffer, dwLength);
    }
    else if(usb_pipecontrol(nPipe))
    {
#ifdef TEST_BULKERR
        DBG(2,"StartUrb PIPE Control  \n");
#endif
		pBuffer = pUrb->setup_packet;
		dwLength = 8;
    }
    else
    {
#ifdef TEST_BULKERR
        DBG(2,"StartUrb PIPE Bulk or Int  \n");
#endif

#ifdef DMA_HOST_CODE
		/*== get address of a buffer =======================================*/
		if((pDmaChStatus)->dma_req_flag){
			pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
		}else{
#ifdef MUSB_DMA_BUS2VIRT
			if (!(pUrb->transfer_buffer)) {
				pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma));
            }else{
            	pBuffer = (uint8_t*)(pUrb->transfer_buffer);
            }
            if (!pBuffer ) { // abort the transfer
            	ERR("no buffer was given, aborting\n");
				return;
            }
#else	/* MUSB_DMA_BUS2VIRT */
	    	pBuffer = pUrb->transfer_buffer;
#endif	/* MUSB_DMA_BUS2VIRT */
		}
#endif /* DMA_HOST_CODE */
		/*== get transfer length ==== =======================================*/
		dwLength = pUrb->transfer_buffer_length;
	 	DBG(1, "trans Buffer[virt]=%p length=%d \n",pBuffer, dwLength);
    }
    
    /* Configure endpoint */	
#ifdef TEST_BULKERR
    DBG(2,"StartUrb pBuffer=%p\n",pBuffer);
#endif
#ifdef CONSISTENT_SYNC_NEW
	if(nOut){
		consistent_sync(pBuffer, dwLength, DMA_TO_DEVICE);
	}else{
		consistent_sync(pBuffer, dwLength, DMA_FROM_DEVICE);	
	}
#endif

	/*== transfer setting ===========================================*/
    MGC_HdrcProgramEnd(pThis, bEnd, pUrb, nOut, pBuffer, dwLength);
    /*===============================================================*/

    /* if transmit, start if it is time */
    if(nOut)
    {
#ifdef DMA_HOST_CODE
		if(((pEnd->pChEpStatus)->State)!=DMA_HOST_ST_DMA){	/* DMA֤Ǥʤ */
#endif
			/* determine if the time is right for a periodic transfer */
			if( usb_pipeisoc(nPipe) || usb_pipeint(nPipe) )
			{
	    		DBG(3, "check whether there's still time for periodic Tx\n");

	    		pEnd->dwIsoPacket = 0;
	    		wFrame = MGC_Read16(pBase, MGC_O_HDRC_FRAME);
	    		if((pUrb->transfer_flags & USB_ISO_ASAP) || 
	       			(wFrame >= pUrb->start_frame))
	    		{
					pEnd->dwWaitFrame = 0;
					MGC_HdrcStartTx(pThis, bEnd);
	    		}
	    		else
	    		{
					pEnd->dwWaitFrame = pUrb->start_frame;
					/* enable SOF interrupt so we can count down */
					MGC_Write8(pBase, MGC_O_HDRC_INTRUSBE, 0xff);
	    		}
			}
			else
			{
	    		DBG(3, "starting Tx\n");
	    		MGC_HdrcStartTx(pThis, bEnd);
			}
#ifdef DMA_HOST_CODE
		}else{
		    /* DMAȤProgramEnd()ǹԤ*/

			pDmaChannel = pEnd->pDmaChannel;	/* ProgramEndǼͥ */
			
    		/* DMAߵ */
    		stat = pDmaController->pfDmaEnableIntChannel(pDmaChannel);
    		if (!stat) {
        		DBG(2,"Cannot enable int dma channel %d\n", stat);
    		}
    		DBG(2,"StartUrb: Tx EnableIntChannel! pDmaChannel=%p \n",pDmaChannel);
			/* DMAžե饰ON */
			(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_ON;
					
			/* DMAͥư */
			stat = pDmaController->pfDmaStartChannel(pDmaChannel);
    		if (!stat) {
        		DBG(2,"Cannot start dma channel %d\n", stat);
    		}
    		DBG(2,"StartUrb: Tx StartChannel! \n");
		}
#endif
    }
}
#endif

#if defined(CONFIG_USB_MCD_HOST)
/*
 * Submit an URB, either to the virtual root hut or to a real device;
 * it also checks the URB to make sure it's valid.
 * This is called by the Linux USB core. TSubmit Urb lock pThis
 * and the End to use, so make sure the caller releases its locks.
 *
 * also set the hcpriv member to the localEnd
 
 * @param pUrb URB pointer (urb = USB request block data structure)
 * @return status code (0 succes)
 */
#ifdef MUSB_V26
STATIC int MGC_LinuxSubmitUrb(struct urb* pUrb, int iMemFlags)
#endif
#ifdef MUSB_V24
STATIC int MGC_LinuxSubmitUrb(struct urb* pUrb)
#endif
{
    MGC_LinuxCd* pThis;
    int nEnd;
    int bustime;
    MGC_LinuxLocalEnd* pEnd;
#if MUSB_DEBUG > 0
    MUSB_DeviceRequest* pRequest;
#endif
#ifdef MUSB_V24
    struct urb* pNextUrb;
#endif

    uint8_t bCheckBandwidth = FALSE;
    unsigned int pipe = pUrb ? pUrb->pipe : 0;
    
    DBG(2, "<==\n");
    DBG(3, "MGC_SubmitUrb=%p \n",(void*)pUrb->transfer_dma);             
    DBG(3, "buflen=%d \n",pUrb->transfer_buffer_length);             

#ifdef MUSB_PARANOID
    if (!pUrb || !pUrb->dev || !pUrb->dev->bus || !pUrb->dev->bus->hcpriv)
    {
	DBG(1, "invalid URB\n"); 
	DEBUG_CODE(3, if ( !pUrb ) { printk(KERN_ERR "pUrb is null\n"); } \
		else if ( !pUrb->dev ) { printk(KERN_ERR "pUrb->dev is null\n"); } \
		else if ( !pUrb->dev->bus ) { \
			printk(KERN_ERR "pUrb->dev->bus is null\n"); } \
		else if ( !pUrb->dev->bus->hcpriv ) { \
			printk(KERN_ERR "null instance ptr\n"); \
			}; );
	return -EINVAL;
    }

    if ( USB_ENDPOINT_HALTED(pUrb->dev, usb_pipeendpoint (pipe), usb_pipeout (pipe)) )
    {
	DBG(1, "endpoint_halted\n");
	return -EPIPE;
    }
#endif
    
    pThis = (MGC_LinuxCd*)pUrb->dev->bus->hcpriv;
    
#ifdef MUSB_PARANOID
	if(MGC_IsCorrupt(pThis)) { 
		printk(KERN_INFO "corrupted stopping before submit\n"); 
		MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED); 
		return -ENOENT; 
	}
#endif

    if( MUSB_IS_ERR(pThis) ) {
	return -ENODEV;    
    }

    /* if it is a request to the virtual root hub, delegate */
    if( usb_pipedevice(pipe) == pThis->RootHub.bAddress)
    {
	DBG(3, "delegating to root hub\n");
	return MGC_VirtualHubSubmitUrb(&(pThis->RootHub), pUrb);
    }

    /* if we are not connected, error */
    if( !pThis->bIsHost )
    {
	DBG(3, "controller not connected\n");
	return -ENODEV;
    }

    /* if no root device, assume this must be it */
    if(!pThis->pRootDevice)
    {
	pThis->pRootDevice = pUrb->dev;
    }

    /* some drivers try to confuse us by linking periodic URBs BOTH ways */
    if(!pUrb->bandwidth && (usb_pipeisoc(pipe) || usb_pipeint(pipe)))
    {
        bCheckBandwidth = TRUE;
    }

    /* find appropriate local endpoint to do it */
    nEnd = MGC_LinuxFindEnd(pThis, pUrb);
    if(nEnd < 0)
    {
	pUrb->status = USB_ST_URB_REQUEST_ERROR;
	ERR("no resource for proto=%d, addr=%d, end=%d\n", \
	       usb_pipetype(pipe), usb_pipedevice(pipe), \
	       usb_pipeendpoint(pipe)); \
	return USB_ST_URB_REQUEST_ERROR;
    }

    DBG(1, "end %d claimed for proto=%d, addr=%d, end=%d\n", nEnd, \
	usb_pipetype(pipe), usb_pipedevice(pipe),usb_pipeendpoint(pipe)); 
    DBG(3, "MGC_SubmitUrb=%p \n",(void *)pUrb->transfer_dma);             
    
    /* check bandwidth if needed */
    if(bCheckBandwidth)
    {
	bustime = usb_check_bandwidth(pUrb->dev, pUrb);
	if(bustime < 0)
	{
	    return bustime;
	}
	
	usb_claim_bandwidth(pUrb->dev, pUrb, bustime, 
			    usb_pipeisoc(pipe) ? 1 : 0);

#ifdef MUSB_V24
	/* propagate through linked URBs */
	pNextUrb = pUrb->next;
	while(pNextUrb && (0 == pNextUrb->bandwidth))
	{
	    pNextUrb->bandwidth = bustime;
	    pNextUrb = pNextUrb->next;
	}
#endif

    }

    /* increment urb's reference count, we now control it. */
    pUrb = usb_get_urb(pUrb);

    /* queue or start the URB */
    DBG(2, "pUrb=%lx, end=%d, bufsize=%x\n", (unsigned long)pUrb, \
    	nEnd, pUrb->transfer_buffer_length); 
    DBG(3, "MGC_SubmitUrb=%p \n",(void *)pUrb->transfer_dma);             
	   
    DEBUG_CODE(2, if(!nEnd) { \
	pRequest = (MUSB_DeviceRequest*)pUrb->setup_packet; \
	printk(KERN_INFO "%s: bmRequestType=%02x, bRequest=%02x, wLength=%04x\n", \
	       __FUNCTION__, pRequest->bmRequestType, pRequest->bRequest, \
	       le16_to_cpu(pRequest->wLength)); \
    } );


    /* queue URBs */
    pEnd = &(pThis->aLocalEnd[nEnd]);
    spin_lock(&pEnd->Lock);
    pEnd->bIsClaimed=(usb_pipeisoc( pUrb->pipe ) || 
    	usb_pipeint( pUrb->pipe )) ?TRUE:FALSE;
    
    pUrb->hcpriv = pEnd; /* assign the URB to the endpoint */
    list_add_tail(&pUrb->urb_list, &pEnd->urb_list);    
    DBG(1, "queued URB %p for end %d\n", pUrb, nEnd);

    DBG(3, "MGC_SubmitUrb=%p \n",(void *)pUrb->transfer_dma);             
    DBG(1, "check pEnd->pUrb %p \n", pEnd->pUrb);
    if(!pEnd->pUrb) {
	MGC_LinuxStartNextUrb(pThis, (uint8_t)nEnd);
    }
    spin_unlock(&pEnd->Lock);

#ifdef MUSB_PARANOID
    if ( MGC_IsCorrupt(pThis) ) { 
	printk(KERN_INFO "%s:%d:stopping after submit\n",
		__FUNCTION__, __LINE__); 
	MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED); 
	return -ENOENT; 
    }
#endif

    DEBUG_CODE(5, dump_urb(pUrb); );

    return 0;    
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Cancel URB.
 * @param pUrb URB pointer
 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,5) ) 
STATIC int MGC_LinuxUnlinkUrb(struct urb* pUrb)
#else 
STATIC int MGC_LinuxUnlinkUrb(struct urb* pUrb, int status)
#endif
{
    int pipe;
    MGC_LinuxCd* pThis;
    MGC_LinuxLocalEnd* pEnd;
#ifdef DMA_HOST_UNLINK_CODE
    uint16_t wCsr;
	MGC_DmaController* pDmaController; 
    uint8_t* pBase;
    uint8_t skip_flag = FALSE;
    unsigned long flags;
    uint8_t stat = 0;
#endif
    DBG(2, "<== pUrb=%p\n", pUrb);
    
    if (!pUrb || !pUrb->hcpriv)
    {
	return -EINVAL;
    }

    if (!pUrb->dev || !pUrb->dev->bus)
    {
	return -ENODEV;
    }

    pThis = (MGC_LinuxCd*)pUrb->dev->bus->hcpriv;
    if(!pThis)
    {
	return -ENODEV;
    }
	
#ifdef MUSB_PARANOID
    if(MGC_IsCorrupt(pThis))
    {
	printk(KERN_INFO "stopping before unlink\n");
	MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED); 
	return -EINVAL;
    }
#endif

    /* if it is a request to the virtual root hub, delegate */
    if (usb_pipedevice (pUrb->pipe) == pThis->RootHub.bAddress)
    {
	return MGC_VirtualHubUnlinkUrb(&(pThis->RootHub), pUrb);
    }

    /* stop local end if valid */
    pEnd = (MGC_LinuxLocalEnd*)pUrb->hcpriv;
#ifdef MUSB_PARANOID
    if((pEnd < &(pThis->aLocalEnd[0])) ||
	(pEnd > &(pThis->aLocalEnd[MUSB_C_NUM_EPS-1])))
    {
        /* somehow, we got passed a dangling URB pointer */
        return -EINVAL;
    }
#endif

    pipe = pUrb->pipe;    

    spin_lock(&pEnd->Lock);
#ifdef V24
    /* delete from list if it was ever added to one */
    if(pUrb->urb_list.next)
    {
	list_del(&pUrb->urb_list);
    }
#endif

    if(usb_pipeisoc(pipe) || usb_pipeint(pipe))
    {
        /* periodic: shutdown end and any claim */
        MGC_HdrcStopEnd(pThis, pEnd->bLocalEnd);
        pEnd->bIsClaimed = FALSE;
	pEnd->pUrb = NULL;
    } else {

#ifdef DMA_HOST_UNLINK_CODE
	/* Get Controller */
	pDmaController = pThis->pDmaController;
    /* Get Base Address */
    pBase = (uint8_t*)pThis->pRegs;

    /* CPU߶ػ */
	spin_lock_irqsave(&pThis->Lock, flags);
    	/* Check State */
    	switch ((pEnd->pChEpStatus)->State){
    		case DMA_HOST_ST_IDLE:
    		case DMA_HOST_ST_CPU:
    		case DMA_HOST_ST_NULL:
				if(usb_pipein(pipe)){	/* IN */
					
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
					DBG(2,"Read RXCSR=%x \n",wCsr);
					/* RxPktRdyꥢ */
					wCsr &= ~MGC_M_RXCSR_RXPKTRDY;
#ifdef HOST_CLR_REQPKT
					wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_H_REQPKT|MGC_M_RXCSR_AUTOCLEAR|MGC_M_RXCSR_DMAENAB);
#else
					wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_AUTOCLEAR|MGC_M_RXCSR_DMAENAB);
#endif					
					MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);

					/* clear DMAMODE */
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
					wCsr &= ~MGC_M_RXCSR_DMAMODE;
					/* FIFOեå */
					wCsr |= MGC_M_RXCSR_FLUSHFIFO;
					/* Clear RxPktRdy, FIFOեå */
					MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
					
					if((pEnd->HostInfo).db_buff){	/* in case of Double buffer */
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						wCsr |= MGC_M_RXCSR_FLUSHFIFO;
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
					}
				}else if(usb_pipeout(pipe)){ /* OUT*/
					/* TxPktRdyꥢ */
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					/* clear DMAENAB, AUTOSET */
					wCsr &= ~(MGC_M_TXCSR_AUTOSET |MGC_M_TXCSR_DMAENAB|MGC_M_TXCSR_TXPKTRDY);
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					/* clear DMAMODE*/
					wCsr &= ~MGC_M_TXCSR_DMAMODE;
					/* FIFOեå */
					wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
					wCsr |= MGC_M_TXCSR_FLUSHFIFO;
					
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					
					if((pEnd->HostInfo).db_buff){	/* in case of Double buffer */
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
						wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
						wCsr |= MGC_M_TXCSR_FLUSHFIFO;
						/* Clear FIFOեå */
						MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					}
				}
				if((pEnd->pChEpStatus)->dma_ch_flag){	/* DMAͥե饰ͭ */
					/* ͥͭξ */
					/* Release Channel */
					pDmaController->pfDmaReleaseChannel(pEnd->pDmaChannel);
    				DBG(2,"Release Channel \n");
					pEnd->pDmaChannel = NULL;
					(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
				}
				(pEnd->pChEpStatus)->dma_req_flag = USB_MCD_FLAG_OFF;
				(pEnd->pChEpStatus)->State = DMA_HOST_ST_CANCEL;
				DBG(2,"Change State CANCEL \n");
				
				/* FIFO status free */
	    		pEnd->bIsReady = TRUE;
				
    			break;
    		case DMA_HOST_ST_DMA:
				/* Stop Channel */
				stat = pDmaController->pfDmaStopChannel(pEnd->pDmaChannel);
    			DBG(2,"Stop Channel in Rx MODE0 \n");
				if (!stat) {
        			DBG(2,"Cannot stop dma channel %d\n", stat);
    			}
    			/* Disable Channel interrupt */
    			stat = pDmaController->pfDmaDisableIntChannel(pEnd->pDmaChannel);
    			if (!stat) {
        			DBG(2,"Cannot disable int dma channel %d\n", stat);
    			}
				(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;
						
				if(usb_pipein(pipe)){	/* IN */
					switch((pEnd->pChEpStatus)->mode){
						case DMA_HOST_MODE0:
							if((pEnd->pChEpStatus)->dma_trans_flag){	/* DMAžե饰ͭ */
								(pEnd->pChEpStatus)->State = DMA_HOST_ST_CANCEL;
								DBG(2,"Change State CANCEL \n");
								/* DMAߤǥ󥻥Ԥ */
								skip_flag = TRUE;
							}
							break;
					}
					if(!skip_flag){
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						DBG(2,"Read RXCSR=%x \n",wCsr);
						/* RxPktRdyꥢ */
						wCsr &= ~MGC_M_RXCSR_RXPKTRDY;
#ifdef HOST_CLR_REQPKT
						wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_H_REQPKT|MGC_M_RXCSR_AUTOCLEAR|MGC_M_RXCSR_DMAENAB);
#else
						wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_AUTOCLEAR|MGC_M_RXCSR_DMAENAB);
#endif					
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);

						/* clear DMAMODE */
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						wCsr &= ~MGC_M_RXCSR_DMAMODE;
						/* FIFOեå */
						wCsr |= MGC_M_RXCSR_FLUSHFIFO;
						/* Clear RxPktRdy, FIFOեå */
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);

						if((pEnd->HostInfo).db_buff){	/* in case of Double buffer */
							wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
							wCsr |= MGC_M_RXCSR_FLUSHFIFO;
							MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
						}
					}
				}else if(usb_pipeout(pipe)){ /* OUT */
					/* TxPktRdyꥢ */
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					/* clear DMAENAB, AUTOSET */
					wCsr &= ~(MGC_M_TXCSR_AUTOSET |MGC_M_TXCSR_DMAENAB|MGC_M_TXCSR_TXPKTRDY);
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					/* clear DMAMODE*/
					wCsr &= ~MGC_M_TXCSR_DMAMODE;
					/* FIFOեå */
					wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
					wCsr |= MGC_M_TXCSR_FLUSHFIFO;
					
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					
					if((pEnd->HostInfo).db_buff){	/* in case of Double buffer */
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
						wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
						wCsr |= MGC_M_TXCSR_FLUSHFIFO;
						/* Clear FIFOեå */
						MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					}    				
    			}
				if(!skip_flag){
					if((pEnd->pChEpStatus)->dma_ch_flag){	/* DMAͥե饰ͭ */
						/* ͥͭξ */
						/* Release Channel */
						pDmaController->pfDmaReleaseChannel(pEnd->pDmaChannel);
    					DBG(2,"Release Channel \n");
						pEnd->pDmaChannel = NULL;
						(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
					}
					(pEnd->pChEpStatus)->dma_req_flag = USB_MCD_FLAG_OFF;
					/* to State CANCEL */
					(pEnd->pChEpStatus)->State = DMA_HOST_ST_CANCEL;
					DBG(2,"Change State CANCEL \n");

					/* FIFO status free */
	    			pEnd->bIsReady = TRUE;
				}
    			break;
    		case DMA_HOST_ST_CANCEL:
    			/* 	*/
    			break;
    	}
		/* CPUߵ */
		spin_unlock_irqrestore(&pThis->Lock, flags);
#endif
		/* start next URB that might be queued for this end */
		MGC_LinuxStartNextUrb(pThis, pEnd->bLocalEnd);
    }

    pUrb->hcpriv = NULL;    
    spin_unlock(&pEnd->Lock);

    /* release bandwidth if necessary */
    if(pUrb->bandwidth && (usb_pipeisoc(pipe) || usb_pipeint(pipe)))
    {
	usb_release_bandwidth(pUrb->dev, pUrb, 
		usb_pipeisoc(pipe) ? 1 : 0);
    }

    if (pUrb->complete && (pUrb->transfer_flags & USB_ASYNC_UNLINK))
    {
	pUrb->status = -ENOENT;
	pUrb->actual_length = 0;
	
 	DBG(1, "completing URB %p due to unlink\n", pUrb);		
	/* Structure gets somehow corrupted... */
	pUrb->hcpriv=NULL;
	COMPLETE_URB(pUrb, NULL); 	
 	DBG(1, "URB %p completed\n", pUrb);
    } else {
	pUrb->status = -ENOENT;
    }
    usb_put_urb(pUrb);

#ifdef MUSB_PARANOID
    if(MGC_IsCorrupt(pThis))
    {
	ERR("stopping after unlink\n");
	MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED); 
	return -EINVAL;
    }
#endif

    DBG(2, "==>\n");

    return 0;
}
#endif

#if defined(MUSB_V26) || defined(MUSB_GADGET) 
/*==============================================================================*/
/* DMA-Coherent Buffers															*/
/*==============================================================================*/
/* FIXME tune these based on pool statistics ... */
static const size_t	pool_max [HCD_BUFFER_POOLS] = {
	/* platforms without dma-friendly caches might need to
	 * prevent cacheline sharing...
	 */
	32,
	128,
	512,
	PAGE_SIZE / 2
	/* bigger --> allocate pages */
};

/*==============================================================================*/
/* Allocate memory for a buffer that might use DMA.								*/
/*																				*/
/* retrun addr:			pointer of a virtual address							*/
/* @param pThis 		MGC_LinuxCd* to hold areas of dma_pool					*/
/* @param size:			the length of memory to get								*/
/* @param mem_flags:	the same Flags for memory area as kmalloc's flags		*/
/* @param dma:			pointer of a bus address								*/
/*							NULL when DMAble memeory is not requested.			*/
/*==============================================================================*/
void* MGC_AllocBufferMemory(MGC_LinuxCd* pThis, size_t size, int mem_flags, dma_addr_t* dma) {

	int 			i;
    void* addr = NULL;

	DBG(1,"size =%d \n",size);
	/* some USB hosts just use PIO */
#if defined(CONFIG_USB_MCD_HOST)
	if (!(pThis->pBus->controller->dma_mask)) {
#else
	if (!(pThis->controller->dma_mask)) {
#endif
		/* in case without DMAC */
		*dma = ~(dma_addr_t) 0;
		KMALLOC(addr, size, mem_flags);
		DBG(1, "malloced pAddr=%p, pDma=%p\n", addr, (void *)*dma); 
		return addr;
	}

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (size <= pool_max [i]){
			addr = dma_pool_alloc (pThis->pool [i], mem_flags, dma);
			DBG(1,"dma_pool_alloc pool[%d] pAddr=%p pDma=%p mem_flags=%d \n",\
				i, addr, dma, mem_flags);
			return addr;
		}
	}
#if defined(CONFIG_USB_MCD_HOST)
	addr = dma_alloc_coherent ((struct device*)pThis->pBus->controller, size, dma, 0);
#else
	addr = dma_alloc_coherent ((struct device*)pThis->controller, size, dma, 0);
#endif	
	DBG(1, "dma_alloc_coherent pAddr=%p, pDma=%p mem_flags=0 <-Force \n", addr, (void *)*dma); 
	return addr;
	
}

/*==============================================================================*/
/* Free memory previously allocated with AllocBufferMemory. 					*/
/* @param pThis 		MGC_LinuxCd* to hold areas of dma_pool					*/
/* @param size:			the length of memory to free							*/
/* @param addr:			virtual address for tha bus address(dma)				*/
/* @param dma:			the pointer of a bus address							*/
/*==============================================================================*/
void MGC_FreeBufferMemory(MGC_LinuxCd* pThis, size_t size, void *addr, dma_addr_t dma) {
	int 			i;

    DBG(1, "<== freeing size=%d, pAddr=%p, pDma=%p\n", size, addr, (void *)dma);
	if (!addr){
		DBG(1, "address NULL!\n");
		return;
	}
#if defined(CONFIG_USB_MCD_HOST)
	if (!(pThis->pBus->controller->dma_mask)) {
#else
	if (!(pThis->controller->dma_mask)) {
#endif
		KFREE(addr);
		DBG(1, "free address \n");
		return;
	}

	for (i = 0; i < HCD_BUFFER_POOLS; i++) {
		if (size <= pool_max [i]) {
			dma_pool_free (pThis->pool [i], addr, dma);
			DBG(1,"dma_pool_free pool[%d] pAddr=%p pDma=%p \n", i, addr, (void *)dma);
			return;
		}
	}
#if defined(CONFIG_USB_MCD_HOST)
	dma_free_coherent ((struct device*)pThis->pBus->controller, size, addr, dma);
#else
	dma_free_coherent ((struct device*)pThis->controller, size, addr, dma);
#endif
	DBG(1,"dma_free_coherent \n");
}
#endif

#if defined(MUSB_V26)
#if defined(CONFIG_USB_MCD_HOST)
/*==============================================================================*/
/* Called to Allocate memory using MGC_AllocBufferMemory. 						*/
/* @param	pBus		struct usb_bus* 										*/
/* @param	nSize		the length of memory to get								*/
/* @param	iMemFlags	the same Flags for memory area as kmalloc's flags		*/
/* @param	pDmaAddress pointer of a bus address								*/
/*							NULL when DMAble memeory is not requested.			*/
/*==============================================================================*/
STATIC void* MGC_LinuxBufferAlloc(struct usb_bus* pBus, size_t nSize,
	int iMemFlags, dma_addr_t* pDmaAddress)
{
    /* for now, just kmalloc it */
    MGC_LinuxCd* pThis=(MGC_LinuxCd*)pBus->hcpriv;

#ifdef MUSB_PARANOID
    if ( !pThis ) {
    	ERR("cannot find the controller, cannot allocate the memory\n");
	return 0;
    }
#endif

    DBG(2, "<== allocating memory on bus (%s), %d, pDmaAddress=%p\n", 
    	pBus->bus_name, pBus->busnum, (void *)*pDmaAddress );
    return MGC_AllocBufferMemory(pThis, nSize, iMemFlags, pDmaAddress);
}

/*==============================================================================*/
/* Called to free memory using MGC_LinuxBufferFree. 							*/
/* @param pBus			struct usb_bus* 										*/
/* @param size:			the length of memory to free							*/
/* @param addr:			virtual address for tha bus address(dma)				*/
/* @param dma:			the pointer of a bus address							*/
/*==============================================================================*/
STATIC void MGC_LinuxBufferFree(struct usb_bus* pBus, size_t nSize,
				void* address, dma_addr_t dma)
{
    MGC_LinuxCd* pThis=(MGC_LinuxCd*)pBus->hcpriv;

#ifdef MUSB_PARANOID
    if ( !pThis ) {
	KFREE(address);
    	ERR("cannot find the controller, cannot free the memory properly\n");
	return;
    }
#endif

    MGC_FreeBufferMemory(pThis, nSize, address, dma);
}
#endif /* defined(CONFIG_USB_MCD_HOST) */
#endif

#ifdef MUSB_V26
#if defined(CONFIG_USB_MCD_HOST)
/* disable an endpoint */
STATIC void MGC_LinuxDisable(struct usb_device* pDevice, struct usb_host_endpoint *ep)
{

}
#endif
#endif

#ifdef MCD_USE_ALLOCDEV 
/**
 * Private per-device allocation
 * @param pDevice Linux USBD device pointer
 * @return status code
 */
STATIC int MGC_LinuxAllocDevice(struct usb_device *pDevice)
{
    
    DBG(2, "<==\n");
    return 0;
    
}
#endif
#ifdef MCD_USE_FREEDEV
/**
 * Private per-device cleanup
 * @param pDevice Linux USBD device pointer
 * @return 0 (success)
 */
STATIC int MGC_LinuxFreeDevice (struct usb_device * pDevice)
{
    DBG(2, "<==\n");
    return 0;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/**
 * Get the current frame number
 * @param usb_dev pointer to USB device
 * @return frame number
 */
STATIC int MGC_LinuxGetFrame(struct usb_device* pDevice)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pDevice->bus->hcpriv;
    uint8_t* pBase = (uint8_t*)pThis->pRegs;
    
    DBG(2, "<==\n");

    return (int)MGC_Read16(pBase, MGC_O_HDRC_FRAME);
}
#endif
#ifdef CONFIG_USB_MCD_PCI
/**************************************************************************
 * ISR for PCI system
**************************************************************************/

/**
 * MGC_LinuxPciIsr PCI system ISR
 * USB IRQ and DMA IRQ are same irq number.
 * This routine call USB IRQ and DMA IRQ.
 * @param irq interrupt line associated with the controller 
 * @param __cd data structure for the pci IRQ data
 * @param r holds the snapshot of the processor's context before 
 *             the processor entered interrupt code. (not used here) 
 */
STATIC irqreturn_t
MGC_LinuxPciIsr(int irq, void *__cd, struct pt_regs *r)
{
    irqreturn_t ip_ret;
    irqreturn_t dma_ret;
    
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)__cd;

    ip_ret = MGC_LinuxIsr(irq, (void*)pThis, r);

#ifdef DMA_HOST_CODE
    if (pThis->pDmaController->pfDmaControllerIsr) {
        dma_ret
            = (pThis->pDmaController->pfDmaControllerIsr)(
                irq, (void*)pThis->pDmaController, r);

        /* ACK PCI interrupt if applicable */
        *(volatile uint32_t*)(pThis->pIack) = 3;
    }
    else {
        dma_ret = IRQ_NONE;
    }
#else
    dma_ret = IRQ_NONE;
#endif /* DMA_HOST_CODE */
    DBG(5, "IRQ_NONE %d %d\n",
        (ip_ret == IRQ_NONE), (dma_ret == IRQ_NONE));

    return (((ip_ret == IRQ_NONE) && (dma_ret == IRQ_NONE))
            ? IRQ_NONE
            : IRQ_HANDLED);
}
#endif
/**************************************************************************
 * Linux driver hooks, including PCI
**************************************************************************/

/**
 * ISR
 * @param irq interrupt line associated with the controller 
 * @param hci data structure for the host controller
 * @param r holds the snapshot of the processor's context before 
 *             the processor entered interrupt code. (not used here) 
 */
STATIC irqreturn_t MGC_LinuxIsr(int irq, void *__hci, struct pt_regs *r)
{
    uint8_t bShift;
    uint32_t nSource;
    uint32_t reg;
    uint8_t bIntrUsbValue;
    uint16_t wIntrTxValue, wIntrRxValue;
    uint8_t bIndexSave;
#if MUSB_DEBUG > 0
#if defined(CONFIG_USB_MCD_HOST)    	
    uint16_t wIntrTxCheck, wIntrRxCheck;
#endif
#endif
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)__hci;
    const void* pBase = pThis->pRegs;

    /* read registers (this clear the interrupt) */
    bIntrUsbValue = MGC_Read8(pBase, MGC_O_HDRC_INTRUSB);
    wIntrTxValue = MGC_Read16(pBase, MGC_O_HDRC_INTRTX);
    wIntrRxValue = MGC_Read16(pBase, MGC_O_HDRC_INTRRX);

    /* index쥸¸ - Ǹ */
    bIndexSave = MGC_Read8(pBase, MGC_O_HDRC_INDEX);

    /* determine if we were the cause; return if not */
    if(pThis->nIsPci)
    {
	nSource = *(volatile uint32_t*)(pThis->pIack);	
    	DEBUG_CODE(10,  if ( !nSource) { \
		printk(KERN_INFO "%s: PCI IRQ [mode=%s] nSource=%d\n", \
    		__FUNCTION__, MUSB_MODE(pThis), nSource); } );
    } else {
	nSource = bIntrUsbValue | wIntrTxValue | wIntrRxValue;
    	DEBUG_CODE(10,  if ( !nSource) { \
		printk(KERN_INFO "%s: IRQ [mode=%s] nSource=%d\n", \
    		__FUNCTION__, MUSB_MODE(pThis), nSource); } );
    }
    
#ifndef MUSB_POLL
    if (!nSource)
    {
	RETURN_IRQ_NONE;
    }
#endif

    DBG(4, "[%ld]: IRQ RECEIVED [devmode=%s] IntrUSB=%02x, IntrTx=%04x, IntrRx=%04x\n", \
	jiffies, MUSB_MODE(pThis), bIntrUsbValue, wIntrTxValue, wIntrRxValue); 
	
#ifdef MUSB_PARANOID
    /* corruption check */
    if(MGC_IsCorrupt(pThis))
    {
	printk(KERN_INFO "stopping before ISR, the controller structure is corrupted\n");
	MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED); 

#ifndef MUSB_POLL    
	/* ACK PCI interrupt if applicable */
	if(pThis->nIsPci)
	{
	    *(volatile uint32_t*)(pThis->pIack) = 3;
	}
#endif
	RETURN_IRQ_HANDLED;
    }
#endif

    /* ignore requests when in error */
    if( MUSB_IS_ERR(pThis) ) {
	RETURN_IRQ_HANDLED;    
    }
    
   /* the core can interrupt us for multiple reasons, I.E. more than an
    * interrupt line can be asserted; service the globa interrupt first.
    * Global interrups are used to signal connect/disconnect/vbuserr etc. */
    if(bIntrUsbValue)
    {
    	MGC_HdrcServiceUsb(pThis, bIntrUsbValue);
    }
    	    
    /* each bit of wIntrTxValue represent an endpoint, EP0 gets a 
     * "SPECIAL" treatment
     */	
    reg = wIntrTxValue;
    /* EP0 */
    if(reg & 1 )
    {
        if( MUSB_IS_HST(pThis) )
	{
#if defined(CONFIG_USB_MCD_HOST)
#ifdef MUSB_CONFIG_PROC_FS
	    if(pThis->pfDefaultEndHandler)
	    {
		pThis->pfDefaultEndHandler(pThis->pDefaultEndHandlerParam);
	    } else
#endif
	    	MGC_HdrcServiceDefaultEnd(pThis);
#endif
	}
#ifdef MUSB_GADGET
	else if( MUSB_IS_DEV(pThis) )
	{
	    	MGC_HdrcServiceDeviceDefaultEnd(pThis);
	}	
#endif
	else { 
	    ERR("pending interrupt!!!\n");
	}	

    }
    		
    /* TX on endpoints 1-15 */
    bShift = 1;
    reg >>= 1;
    while(reg)
    {
        if(reg & 1)
	{
	    if( MUSB_IS_HST(pThis)  )
	    {
#if defined(CONFIG_USB_MCD_HOST)
		MGC_HdrcServiceTxAvail(pThis, bShift);
#endif
	    }
#ifdef MUSB_GADGET
	    else if( MUSB_IS_DEV(pThis) )
	    {
		MGC_HdrcServiceDeviceTxAvail(pThis, bShift);
	    }
#endif
	    else { 
		ERR("pending interrupt!!! \n");
	    }
	}
	reg >>= 1;
	bShift++;
    }
#if defined(CONFIG_USB_MCD_HOST)    	
    DEBUG_CODE(10, wIntrTxCheck = MGC_Read16(pBase, MGC_O_HDRC_INTRTX); \
	if(wIntrTxCheck && (wIntrTxCheck == wIntrTxValue)) { \
	    printk(KERN_ERR 
		"%s:%d: Unhandled TX interrupt, wIntrTx=%04x wIntrTxCheck=%04x; DRC \
		stopped\n", __FUNCTION__, __LINE__, wIntrTxValue, wIntrTxCheck); \
	    for(bShift = 0; bShift < pThis->bEndCount; bShift++) { \
		MGC_HdrcDumpRegs(pThis, bShift); \
	    } \
	    MGC_HdrcStop(pThis); \
	    MUSB_ERR_MODE(pThis, MUSB_ERR_IRQ); \
	    MGC_VirtualHubPortDisconnected(&(pThis->RootHub), 0); \
	    pThis->pRootDevice = NULL; \
	} );
#endif

    /* RX on endpoints 1-15 */
    reg = wIntrRxValue;
    bShift = 1;
    reg >>= 1;
    while(reg)
    {
        if(reg & 1)
	{
	    if( MUSB_IS_HST(pThis)  )
	    {
#if defined(CONFIG_USB_MCD_HOST)	    
	        MGC_HdrcServiceRxReady(pThis, bShift);
#endif
	    }
#ifdef MUSB_GADGET
	    else if( MUSB_IS_DEV(pThis) )
	    {
		MGC_HdrcServiceDeviceRxReady(pThis, bShift);
	    }
#endif
	    else { 
		ERR("pending interrupt!!! \n");
	    }
	}
	reg >>= 1;
	bShift++;
    }
#if defined(CONFIG_USB_MCD_HOST)    	
    DEBUG_CODE(10,  wIntrRxCheck = MGC_Read16(pBase, MGC_O_HDRC_INTRRX); \
	if(wIntrRxCheck && (wIntrRxCheck == wIntrRxValue)) { \
		printk(KERN_ERR "%s:%d: Unhandled RX interrupt, IntrRx=%04x; IntrRxCheck=%04x DRC stopped\n", \
		       __FUNCTION__, __LINE__, wIntrRxValue, wIntrRxCheck); \
		for(bShift = 0; bShift < pThis->bEndCount; bShift++) { \
		    MGC_HdrcDumpRegs(pThis, bShift); \
		} \
		MGC_HdrcStop(pThis); \
	    	MUSB_ERR_MODE(pThis, MUSB_ERR_IRQ); \
		MGC_VirtualHubPortDisconnected(&(pThis->RootHub), 0); \
		pThis->pRootDevice = NULL; \
	    });
#endif
#if MUSB_DEBUG > 0
    if(MGC_IsCorrupt(pThis))
    {
	printk(KERN_INFO "stopping after ISR\n");
	MUSB_ERR_MODE(pThis, MUSB_ERR_CORRUPTED);
    }
#endif

#ifndef MUSB_POLL    
    /* ACK PCI interrupt if applicable */
    if(pThis->nIsPci) {
	*(volatile uint32_t*)(pThis->pIack) = 3;
    }
#endif
    /* index쥸 */
    MGC_SelectEnd(pBase, bIndexSave)

    RETURN_IRQ_HANDLED;
}

void MGC_OtgUpdate(MGC_LinuxCd* pThis, uint8_t bVbusError, uint8_t bConnect)
{
}

#ifdef MUSB_OTG

/* ؿmusb_bus_check_vbus
 * Vbusξ֤
 */

enum usb_otg_vbus_stat
musb_bus_check_vbus(
    void* regs)
{
    enum usb_otg_vbus_stat vbus_stat;
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    switch ((devctl & MGC_M_DEVCTL_VBUS) >> MGC_S_DEVCTL_VBUS) {
      case 0x1:
        vbus_stat = USB_OTG_VBUS_STAT_LOW;
        break;
      case 0x2:
        vbus_stat = USB_OTG_VBUS_STAT_SESS;
        break;
      case 0x3:
        vbus_stat = USB_OTG_VBUS_STAT_VALID;
        break;
      default:                  /* 0x0 */
        vbus_stat = USB_OTG_VBUS_STAT_OFF;
        break;
    }

    return vbus_stat;
}
 

/*
 * ؿmusb_bus_enable_role_change
 * b-deviceHostReq򥻥åȤRoleChangeĤ롣
 */

int
musb_bus_enable_role_change(
    void* regs)
{
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    devctl |= MGC_M_DEVCTL_HR;
    MGC_Write8(regs, MGC_O_HDRC_DEVCTL, devctl);

    return 0;
}

/*
 * ؿmusb_bus_disable_role_change
 * b-deviceHostReq򥻥åȤRoleChangeԵĤȤ롣
 */

int
musb_bus_disable_role_change(
    void* regs)
{
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    devctl &= ~MGC_M_DEVCTL_HR;
    MGC_Write8(regs, MGC_O_HDRC_DEVCTL, devctl);

    return 0;
}

/*
 * ؿmusb_bus_suspend
 * RoleChangeȼSuspendԤ
 */

int
musb_bus_suspend(
    void* regs)
{
    uint8_t power = MGC_Read8(regs, MGC_O_HDRC_POWER);

    power |= MGC_M_POWER_SUSPENDM;
    MGC_Write8(regs, MGC_O_HDRC_POWER, power);

    return 0;
}

/*
 * ؿmusb_bus_start_session
 * a-deviceSessionӥåȤ򥻥åȤVbus򳫻Ϥ롣
 */

int
musb_bus_start_session(
    void* regs)
{
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    host_flush_all_ep(MGC_pFirstDriver);
    MGC_HdrcStart(MGC_pFirstDriver);

    devctl |= MGC_M_DEVCTL_SESSION;
    MGC_Write8(regs, MGC_O_HDRC_DEVCTL, devctl);

    MUSB_HST_MODE(MGC_pFirstDriver);

    return 0;
}

/*
 * ؿmusb_bus_end_session
 * a-deviceSessionӥåȤ򥯥ꥢVbusߤ롣
 */

int
musb_bus_end_session(
    void *regs)
{
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

#ifdef MCD_PM_CODE
	/* Stop Hdrc */
	MGC_HdrcStop(MGC_pFirstDriver);
#endif	
#if defined(CONFIG_USB_MCD_HOST)
    MGC_VirtualHubPortDisconnected(&(MGC_pFirstDriver->RootHub), 0);
#endif
    MGC_pFirstDriver->pRootDevice = NULL;
    devctl &= ~MGC_M_DEVCTL_SESSION;
    MGC_Write8(regs, MGC_O_HDRC_DEVCTL, devctl);

    MUSB_OTG_MODE(MGC_pFirstDriver);

    return 0;
}

/*
 * ؿmusb_bus_notify_vbus_change
 * b-devicevbusѲΤ
 */

int
musb_bus_notify_vbus_change(
    void* regs,
    uint8_t old_vbus,
    uint8_t new_vbus)
{
    uint8_t power = MGC_Read8(regs, MGC_O_HDRC_POWER);

    if (old_vbus == USB_OTG_VBUS_STAT_SESS
        && new_vbus == USB_OTG_VBUS_STAT_LOW
        && power & MGC_M_POWER_SOFTCONN
        && MUSB_IS_DEV(MGC_pFirstDriver)) {
        MGC_GadgetDisconnect(MGC_pFirstDriver);
    }

    return 0;
}

/*
 * ؿmusb_bus_request_session
 * idleSessionӥåȤ򥻥åȤSRPȯԤ롣
 */

int
musb_bus_request_session(
    void* regs)
{
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    devctl |= MGC_M_DEVCTL_SESSION;
    MGC_Write8(regs, MGC_O_HDRC_DEVCTL, devctl);

    return 0;
}

/*
 * ؿmusb_bus_set_speed
 * HSEnable򥻥åȤޤϲ롣
 */

int
musb_bus_set_speed(
    void* regs,
    unsigned int speed)
{
    uint8_t power = MGC_Read8(regs, MGC_O_HDRC_POWER);

    if (speed == USB_OTG_HS_ENABLE) {
        power |= MGC_M_POWER_HSENAB;
    }
    else {
        power &= ~MGC_M_POWER_HSENAB;
    }
    MGC_Write8(regs, MGC_O_HDRC_POWER, power);

    return 0;
}

/*
 * ؿmusb_bus_get_speed
 * Υԡɾ֤롣
 */

extern enum usb_otg_control_speed
musb_bus_get_speed(
    void* regs)
{
    enum usb_otg_control_speed speed;
    uint8_t power = MGC_Read8(regs, MGC_O_HDRC_POWER);
    uint8_t devctl = MGC_Read8(regs, MGC_O_HDRC_DEVCTL);

    if (power & MGC_M_POWER_HSMODE) {
        speed = USB_OTG_SPEED_HS;
    }
    else if (devctl & MGC_M_DEVCTL_HM) {
        /* HostˤLowSpeed⥵ݡ */
        if (devctl & MGC_M_DEVCTL_FSDEV) {
            speed = USB_OTG_SPEED_FS;
        }
        else {
            speed = USB_OTG_SPEED_LS;
        }
    }
    else {
        /* GadgetˤHSФʤɬFS */
        speed = USB_OTG_SPEED_FS;
    }

    return speed;
}

/*
 * ؿmusb_bus_set_test_mode
 * ƥȥ⡼ꤹ
 */

int
musb_bus_set_test_mode(
    void* regs,
    enum usb_otg_control_test_mode test_mode)
{
    uint8_t power;
    uint8_t testreg = MGC_Read8(regs, MGC_O_HDRC_TESTMODE);

    switch (test_mode) {
      case USB_OTG_TEST_MODE_NORMAL:
        testreg = 0;
        break;
      case USB_OTG_TEST_MODE_G_FORCE_PACKET:
        power = MGC_Read8(regs, MGC_O_HDRC_POWER);
        power &= ~MGC_M_POWER_ENSUSPEND;
        MGC_Write8(regs, MGC_O_HDRC_POWER, power);
        mdelay(50);             /* PHYƵư, ChirpԤ */
      case USB_OTG_TEST_MODE_G_PACKET:
        MGC_SelectEnd(regs, 0 /* EP0 */);
        MGC_HdrcLoadFifo(regs, 0 /* EP0 */, sizeof(MGC_aTestPacket), MGC_aTestPacket);
        MGC_WriteCsr16(regs, MGC_O_HDRC_CSR0, 0 /* EP0 */, MGC_M_CSR0_TXPKTRDY);
        testreg |= MGC_M_TEST_PACKET;
        break;
      case USB_OTG_TEST_MODE_G_K:
        testreg |= MGC_M_TEST_K;
        break;
      case USB_OTG_TEST_MODE_G_J:
        testreg |= MGC_M_TEST_J;
        break;
      case USB_OTG_TEST_MODE_G_SE0_NAK:
        testreg |= MGC_M_TEST_SE0_NAK;
        break;
      case USB_OTG_TEST_MODE_H_AS:
        goto no_set;
      case USB_OTG_TEST_MODE_H_LS:
        testreg |= MGC_M_TEST_FORCE_HOST;
        testreg &= ~(MGC_M_TEST_FORCE_HS | MGC_M_TEST_FORCE_FS);
        break;
      case USB_OTG_TEST_MODE_H_FS:
        testreg |= MGC_M_TEST_FORCE_HOST | MGC_M_TEST_FORCE_FS;
        testreg &= ~(MGC_M_TEST_FORCE_HS);
        break;
      case USB_OTG_TEST_MODE_H_HS:
        testreg |= MGC_M_TEST_FORCE_HOST | MGC_M_TEST_FORCE_FS;
        testreg &= ~(MGC_M_TEST_FORCE_FS);
        break;
      default:
        return -EINVAL;
    }

    MGC_Write8(regs, MGC_O_HDRC_TESTMODE, testreg);

  no_set:
    return 0;
}

/*
 * ؿmusb_bus_get_test_mode
 * ƥȥ⡼ɼ
 */

enum usb_otg_control_test_mode
musb_bus_get_test_mode(
    void *regs)
{
    enum usb_otg_control_test_mode test_mode;
    uint8_t testreg = MGC_Read8(regs, MGC_O_HDRC_TESTMODE);

    if (testreg == MGC_M_TEST_PACKET) {
        test_mode = USB_OTG_TEST_MODE_G_PACKET;
    }
    else if (testreg == MGC_M_TEST_K) {
        test_mode = USB_OTG_TEST_MODE_G_K;
    }
    else if (testreg == MGC_M_TEST_J) {
        test_mode = USB_OTG_TEST_MODE_G_J;
    }
    else if (testreg == MGC_M_TEST_SE0_NAK) {
        test_mode = USB_OTG_TEST_MODE_G_SE0_NAK;
    }
    else if (testreg == (MGC_M_TEST_FORCE_HS | MGC_M_TEST_FORCE_HOST)) {
        test_mode = USB_OTG_TEST_MODE_H_HS;
    }
    else if (testreg == (MGC_M_TEST_FORCE_FS | MGC_M_TEST_FORCE_HOST)) {
        test_mode = USB_OTG_TEST_MODE_H_FS;
    }
    else if (testreg == MGC_M_TEST_FORCE_HOST) {
        test_mode = USB_OTG_TEST_MODE_H_LS;
    }
    else if (testreg) {
        /* error */
        test_mode = 0;
    }
    else {
        test_mode = USB_OTG_TEST_MODE_NORMAL;
    }

    return test_mode;
}

#endif /* MUSB_OTG */

/**
 * Perform generic per-controller initialization.
 *
 * @param pDevice is pci_dev or platform_dev
 * @param nIpIrq IRQ (interpretation is system-dependent)
 * @param nDmaIrq IRQ (interpretation is system-dependent)
 * @param pRegs pointer to controller registers, 
 * assumed already mapped into kernel space
 * @param pName name for bus
 */
STATIC MGC_LinuxCd*
MGC_LinuxInitController(
    void* pDevice,
    uint16_t wType, 
    int nIpIrq,
    int nDmaIrq,
    void* pRegs,
    const char* pName)
{
    uint8_t bOk;
    MGC_LinuxCd* pThis;
#if (MUSB_DEBUG > 0)
    char buf[8], *bufp = buf;
#endif
#if defined(DMA_HOST_CODE) 
	uint8_t	i;
#endif    
#ifdef MCD_PM_CODE
    char aInfo[78];
    uint8_t	reg;	/* for reading FIFOSIZE register */
#endif
#if (MUSB_DEBUG > 0)
#ifndef __sparc__
    snprintf(buf, 8, "%d&%d", nIpIrq, nDmaIrq);
#else
    bufp = __irq_itoa(nIpIrq);
#endif
    DBG(1, KERN_INFO "%s: MUSBxDRC at 0x%lx, IRQ %s\n",
	   __FUNCTION__, (unsigned long)pRegs, bufp);
#endif /* (MUSB_DEBUG > 0) */
    
    /* allocate */
    KMALLOC(pThis, sizeof(MGC_LinuxCd), GFP_KERNEL);    
    if(!pThis) {
        printk(KERN_ERR "%s: kmalloc driver instance data failed\n",
	       __FUNCTION__);
	return NULL;
    }
    memset (pThis, 0, sizeof(MGC_LinuxCd));
#if MUSB_DEBUG > 0
    pThis->dwPadFront = MGC_PAD_FRONT;
    pThis->dwPadBack = MGC_PAD_BACK;
#endif
    strcpy(pThis->aName, pName);
    pThis->pRegs = pRegs;
    spin_lock_init(&pThis->Lock);
    
    DBG(1, KERN_INFO "%s: Driver instance data at 0x%lx\n", 
	   __FUNCTION__, (unsigned long)pThis);

#ifdef DMA_HOST_CODE
    pThis->pDmaController = MGC_HdrcDmaControllerFactory
        .pfNewDmaController((uint8_t*)pRegs, 4 /* ch count */, nDmaIrq);
#ifndef MCD_PM_CODE
    if(pThis->pDmaController)
    {
    	DBG(3,  "DMA Mode enabled\n");
        pThis->pDmaController->pfDmaStartController(
            pThis->pDmaController);
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 0;
    }
#endif /* MCD_PM_CODE */
#else    
 #ifdef MUSB_DMA
    pThis->pDmaController = MGC_HdrcDmaControllerFactory
        .pfNewDmaController((uint8_t*)pRegs, 4 /* ch count */, nDmaIrq);
 #ifndef MCD_PM_CODE
    if(pThis->pDmaController)
    {
    	DBG(3,  "DMA Mode enabled\n");
        pThis->pDmaController->pfDmaStartController(
            pThis->pDmaController);
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 0;
    }
 #endif /* MCD_PM_CODE */
 #endif
#endif /* DMA_HOST_CODE end */

#ifdef DMA_HOST_CODE
	for(i=0; i<MUSB_C_NUM_EPS;i++){
		KMALLOC((pThis->aLocalEnd[i]).pChEpStatus, sizeof(USB_MCD_DmaChStatus), GFP_KERNEL);
    	if(!((pThis->aLocalEnd[i]).pChEpStatus)) {
        	printk(KERN_ERR "%s: kmalloc failed pChEpStatus \n",__FUNCTION__);
			return NULL;
		}
		memset ((pThis->aLocalEnd[i]).pChEpStatus, 0, sizeof(USB_MCD_DmaChStatus));
	}
#endif
    /* be sure interrupts are disabled before connecting ISR */
    MGC_HdrcDisable(pThis);
    MGC_SelectEnd(pThis->pRegs, 0);

#ifdef MCD_PM_CODE
    reg = MGC_ReadCsr8(pThis->pRegs, MGC_O_HDRC_CONFIGDATA, 0);

    strcpy(aInfo,(reg & MGC_M_CONFIGDATA_UTMIDW)?"UTMI-16":"UTMI-8");
    if(reg & MGC_M_CONFIGDATA_DYNFIFO) {
        strcat(aInfo, ", dyn FIFOs");
    }
    if(reg & MGC_M_CONFIGDATA_MPRXE) {
    	pThis->bBulkCombine = TRUE;
        strcat(aInfo, ", bulk combine");
    }
    if(reg & MGC_M_CONFIGDATA_MPTXE) {
    	pThis->bBulkSplit = TRUE;
        strcat(aInfo, ", bulk split");
    }
    if(reg & MGC_M_CONFIGDATA_HBRXE) {
        strcat(aInfo, ", HB-ISO Rx");
    }
    if(reg & MGC_M_CONFIGDATA_HBTXE) {
        strcat(aInfo, ", HB-ISO Tx");
    }
    if(reg & MGC_M_CONFIGDATA_SOFTCONE) {
        strcat(aInfo, ", SoftConn");
    }
    
    DBG(1, KERN_INFO "%s: ConfigData=%02x (%s)\n", __FUNCTION__, reg, aInfo);
#ifdef MUSB_C_DYNFIFO_DEF
    if(!(reg & MGC_M_CONFIGDATA_DYNFIFO))
    {
        printk(KERN_ERR "%s: Built for Dynamic FIFOs but not detected in hardware; please rebuild\n",
	       __FUNCTION__);
	return FALSE;
    }
#else
    if(reg & MGC_M_CONFIGDATA_DYNFIFO)
    {
        printk(KERN_ERR "%s: Dynamic FIFOs detected but support not built; please rebuild\n",
	       __FUNCTION__);
	return FALSE;
    }
#endif
#endif /* MCD_PM_CODE */

#ifndef MCD_PM_CODE
    /* discover configuration */
    bOk = MGC_HdrcInit(wType, pThis);
    if ( !bOk ) {
        KFREE(pThis);	
	return NULL;    
    }
#else
#ifdef MCD_PM_NON_CORRUPT
	usb_mcd_set_padnum(pThis);
#else
    pThis->bEndCount = MUSB_C_NUM_EPS;
#endif /* MCD_PM_NON_CORRUPT */
#endif
    
#ifdef CONFIG_USB_MCD_PCI
    /* Integrator system - PCI board */
    /* attach to the IRQ */
    if (request_irq(nIpIrq, MGC_LinuxPciIsr, SA_SHIRQ, pThis->aName, pThis)) {
        err("request_irq %d failed!", nIpIrq);
        KFREE(pThis);
        return NULL;
    }
#else
    /* Embedded system */
    /* attach to the IRQ */
    if (request_irq(nIpIrq, MGC_LinuxIsr, SA_INTERRUPT, pThis->aName, pThis))
    {
        err("request_irq %d failed!", nIpIrq);
        KFREE(pThis);	
        return NULL;
    }
#ifdef MCD_PM_CODE
	disable_irq(nIpIrq);
#endif
    if (request_irq(nDmaIrq,
                    pThis->pDmaController->pfDmaControllerIsr,
                    SA_INTERRUPT,
                    pThis->aName,
                    pThis->pDmaController))
    {
        err("request_irq %d failed!", nDmaIrq);
        KFREE(pThis);	
        return NULL;
    }
#ifdef MCD_PM_CODE
	disable_irq(nDmaIrq);
#endif
#endif
    pThis->nIrq = nIpIrq;

#if !defined(CONFIG_USB_MCD_HOST)
	pThis->controller = (struct device*)pDevice;
#endif
#if defined(CONFIG_USB_MCD_HOST) 
    /* allocate and register bus */
    pThis->pBus = usb_alloc_bus( &MGC_LinuxOperations );
    if(!pThis->pBus)
    {
        dbg ("usb_alloc_bus fail");
	free_irq (nIpIrq, pThis);
#ifndef CONFIG_USB_MCD_PCI
        free_irq(nDmaIrq, pThis->pDmaController);
#endif
        KFREE(pThis);	
	return NULL;
    }
        
#ifdef MUSB_HAS_BUSNAME
    pThis->pBus->bus_name = pThis->aName;
#endif
    pThis->pBus->hcpriv = (void *)pThis;

    printk(KERN_INFO "%s: New bus @%#lx\n", 
	   __FUNCTION__, (unsigned long)pThis->pBus);
    
    /* register the bus */
#ifdef MUSB_V26
    pThis->pBus->controller = (struct device*)pDevice;
#endif
    usb_register_bus(pThis->pBus);
    
    /* init virtual root hub */
    pThis->PortServices.pPrivateData = pThis;
    pThis->PortServices.pfSetPortPower = MGC_LinuxSetPortPower;
    pThis->PortServices.pfSetPortEnable = MGC_LinuxSetPortEnable;
    pThis->PortServices.pfSetPortSuspend = MGC_LinuxSetPortSuspend;
    pThis->PortServices.pfSetPortReset = MGC_LinuxSetPortReset;
    
    bOk = MGC_VirtualHubInit(&(pThis->RootHub), pThis->pBus, 1, 
			     &(pThis->PortServices));
    if(!bOk)
    {
        dbg ("Virtual Hub init failed");
#ifndef MUSB_V26
	usb_free_bus(pThis->pBus);
#endif
	free_irq (nIpIrq, pThis);
#ifndef CONFIG_USB_MCD_PCI
        free_irq(nDmaIrq, pThis->pDmaController);
#endif
        KFREE(pThis);	
	return NULL;
    }
    
    pThis->pBus->root_hub = pThis->RootHub.pDevice;
#endif /* CONFIG_USB_MCD_HOST */
#ifdef DMA_HOST_CODE_ALLOC
    bOk = usb_mcd_buffer_create(pThis);
	if (bOk != 0) {
		DBG(1,"pool alloc fail \n");
		return NULL;
	}
#endif
#ifdef MUSB_GADGET
    if( !MGC_pFirstDriver ) {
	MGC_pFirstDriver = pThis;
    }
#endif

#ifdef MUSB_OTG
    pThis->otg_control = musb_otg_control_map(pName, pRegs);

    DBG(2, "OTG map %s(%p)\n", pName, (char *)pRegs);
    DBG(2, "OTG pThis=%p, bEndCount=%d\n", pThis, pThis->bEndCount);
#endif

/**
 * Preset the behavior: controller can be compiled as gadget
 * (MUSB_GADGET) or to support the OTG protocol (MUSB_OTG that
 * implies MUSB_GADGET). 
 */
#ifndef MUSB_OTG
#ifdef MUSB_GADGET
    MUSB_DEV_MODE(pThis); 
#else
    MUSB_HST_MODE(pThis); 
#endif
#else
    MUSB_OTG_MODE(pThis); 
#endif

#ifdef MUSB_CONFIG_PROC_FS
    MGC_LinuxCreateProcFs(pThis->aName, pThis);
#endif

#ifdef MUSB_PROC_TESTMUSB
    MGC_LinuxCreateTestProcFs(pThis->aName, pThis);
#endif

    return pThis;
}

/* A couple of hooks to enable HSET */
#ifdef MUSB_CONFIG_PROC_FS
void MGC_HdrcSetDisconnectListener(MGC_LinuxCd* pCd,
				   MGC_pfDisconnectListener pfListener,
				   void* pParam)
{
    pCd->pfDisconnectListener = pfListener;
    pCd->pDisconnectListenerParam = pParam;
}

void MGC_HdrcSetDefaultEndHandler(MGC_LinuxCd* pCd,
				   MGC_pfDefaultEndHandler pfHandler,
				   void* pParam)
{
    pCd->pfDefaultEndHandler = pfHandler;
    pCd->pDefaultEndHandlerParam = pParam;
}

#endif

#ifdef CONFIG_USB_MCD_PCI
/**
 * PCI initialization
 */
STATIC int __devinit MGC_LinuxInitPciController(struct pci_dev *pDevice, 
						int nIrq, 
						unsigned int nIackAddr, 
						unsigned int nIackSize, 
						void* pIack, 
						unsigned int nRegsAddr, 
						unsigned int nRegsSize, 
						void* pRegs)
{
    char name[32];
    MGC_LinuxCd* pThis;
    struct device *dev=NULL;

#ifdef MUSB_V26
    dev=&pDevice->dev;
#endif

    snprintf(name, 32, "musbhdrc%d", MGC_nIndex++);
    DBG(2, "for %s\n", name); 

    pThis = MGC_LinuxInitController(dev, 0, nIrq, 0, pRegs, name);
    if (!pThis) {
      return -1;
    }
	
    pThis->nIsPci = TRUE;
    pThis->nIackAddr = nIackAddr;
    pThis->nIackSize = nIackSize;
    pThis->nRegsAddr = nRegsAddr;
    pThis->nRegsSize = nRegsSize;
    pThis->pIack = pIack;

    DBG(5, "Initialized PCI controller\n");
    DBG(5, " nIackAddr=%d\n", pThis->nIackAddr); 
    DBG(5, " nIackSize=%d\n", pThis->nIackSize); 
    DBG(5, " nRegsAddr=%d\n", pThis->nRegsAddr); 
    DBG(5, " nRegsSize=%d\n", pThis->nRegsSize); 
    DBG(5, " pIack=%p\n", pThis->pIack); 

    pci_set_drvdata(pDevice, pThis);
    
    MGC_VirtualHubStart(&(pThis->RootHub));
    
    return 0;
}
#endif

/*
 * Release resources acquired by driver
 */
STATIC void MGC_LinuxCdFree(MGC_LinuxCd* pThis)
{
#ifdef DMA_HOST_CODE
	uint8_t i;
#endif
    DBG(2, "<==\n");
    
#ifdef MUSB_OTG
    musb_otg_control_unmap(pThis->otg_control);
#endif

    MGC_HdrcStop(pThis);
    MUSB_ERR_MODE(pThis, MUSB_ERR_SHUTDOWN);
    
#ifdef MUSB_CONFIG_PROC_FS
    MGC_LinuxDeleteProcFs(pThis);
#endif

#ifdef MUSB_PROC_TESTMUSB
    MGC_LinuxDeleteTestProcFs(pThis->aName, pThis);
#endif

#ifdef DMA_HOST_CODE
    if(pThis->pDmaController)
    {
        pThis->pDmaController->pfDmaStopController(
            pThis->pDmaController);
        MGC_HdrcDmaControllerFactory.pfDestroyDmaController(
            pThis->pDmaController);
    }
#else
#ifdef MUSB_DMA
    if(pThis->pDmaController)
    {
        pThis->pDmaController->pfDmaStopController(
            pThis->pDmaController);
        MGC_HdrcDmaControllerFactory.pfDestroyDmaController(
            pThis->pDmaController);
    }
#endif
#endif

#if defined(CONFIG_USB_MCD_HOST)
    MGC_VirtualHubStop(&pThis->RootHub);
    MGC_VirtualHubDestroy(&pThis->RootHub);
    
    if (pThis->pBus->root_hub)
    {
        usb_disconnect (&(pThis->pBus->root_hub));
    }
	
    WAIT_MS(1);

    usb_deregister_bus (pThis->pBus);
#endif    
    if(pThis->nIsPci ) {
    	if ( pThis->pIack ) {
            iounmap(pThis->pIack);
	}
	if(pThis->pRegs) {
            iounmap(pThis->pRegs);
	}
	if(pThis->nIackSize) {
            release_mem_region(pThis->nIackAddr, pThis->nIackSize);
	}
	if(pThis->nRegsSize) {
            release_mem_region(pThis->nRegsAddr, pThis->nRegsSize);
	}
    }
    
    if(pThis->nIrq) {
        free_irq (pThis->nIrq, pThis);
#ifndef CONFIG_USB_MCD_PCI
        free_irq(pThis->pDmaController->nIrq, pThis->pDmaController);
#endif
    }

#ifndef MUSB_V26
#if defined(CONFIG_USB_MCD_HOST)
    usb_free_bus (pThis->pBus);
#endif
#endif

#ifdef DMA_HOST_CODE
	for(i=0; i<MUSB_C_NUM_EPS;i++){
		KFREE((pThis->aLocalEnd[i]).pChEpStatus);
	}
#endif
#ifdef DMA_HOST_CODE_ALLOC
	usb_mcd_buffer_destroy(pThis);
#endif
    KFREE(pThis);	
}

#ifdef CONFIG_USB_MCD_PCI
STATIC int __devinit
MGC_PciProbe(struct pci_dev *dev, const struct pci_device_id *id)
{
    int retval;
    void* pIack;
    void* pRegs;
    unsigned int nIackAddr, nIackSize, nRegsAddr, nRegsSize;
    
    DBG(2, "<==\n");

    if (pci_enable_device(dev) < 0)
        return -ENODEV;
		
    if (!dev->irq) {
        err("found device with no IRQ assigned. check BIOS settings!");
	return -ENODEV;
    }
    
    pci_set_master(dev);

    /* structured error handling */
    do {
        if (!(pci_resource_flags(dev,0) & IORESOURCE_MEM))
	    break;
	if (!(pci_resource_flags(dev,1) & IORESOURCE_MEM))
	    break;
	
	nIackAddr = pci_resource_start(dev, 0);
	nIackSize = pci_resource_len(dev, 0);
	nRegsAddr = pci_resource_start(dev, 1);
	nRegsSize = pci_resource_len(dev, 1);
	
	if (!request_mem_region (nIackAddr, nIackSize, "MDRC_IRQ"))
	{
	    dbg ("controller already in use");
	    return -EBUSY;
	}
	if (!request_mem_region (nRegsAddr, nRegsSize, "MDRC_REG"))
	{
	    dbg ("controller already in use");
	    return -EBUSY;
	}
	pIack = ioremap_nocache (nIackAddr, nIackSize);
	pRegs = ioremap_nocache (nRegsAddr, nRegsSize);
	if (!pIack || !pRegs)
	{
	    dbg ("error mapping memory");
	    retval = -EFAULT;
	    release_mem_region (nRegsAddr, nRegsSize);
	    release_mem_region (nIackAddr, nIackSize);
	    err ("init %s fail, %d", dev->slot_name, retval);
	    return retval;
	}
	
	retval = MGC_LinuxInitPciController(dev, dev->irq, 
		nIackAddr, nIackSize, pIack, nRegsAddr, 
		nRegsSize, pRegs);					    						    
	if(0 != retval)
	{
	    iounmap(pRegs);
	    iounmap(pIack);
	    release_mem_region (nRegsAddr, nRegsSize);
	    release_mem_region (nIackAddr, nIackSize);
	    err ("init %s fail, %d", dev->slot_name, retval);
	}

	return retval;
    } while(0);

    return -ENODEV;
}

STATIC void MGC_PciRemove(struct pci_dev *dev)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pci_get_drvdata(dev);

	
    DBG(2, "<==\n");
    MGC_LinuxCdFree(pThis);
}

#ifdef CONFIG_PM
STATIC int MGC_PciSuspend(struct pci_dev *dev, u32 state)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pci_get_drvdata(dev);

    DBG(2, "<==\n");

/*#ifdef MUSB_DMA */
#if defined(MUSB_DMA)||defined(DMA_HOST_CODE)
    if(pThis->pDmaController)
    {
        pThis->pDmaController->pfDmaStopController(
            pThis->pDmaController);
    }
#endif

    MGC_HdrcStop(pThis);
    return 0;
}

STATIC int MGC_PciResume(struct pci_dev *dev)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)pci_get_drvdata(dev);
	
    DBG(2, "<==\n");

/* #ifdef MUSB_DMA */
#if defined(MUSB_DMA)||defined(DMA_HOST_CODE)

    if(pThis->pDmaController)
    {
        pThis->pDmaController->pfDmaStartController(
            pThis->pDmaController);
    }
#endif

    MGC_HdrcStart(pThis);
    return 0;
}
#endif

/*-------------------------------------------------------------------------*/

/* PCI driver selection metadata; PCI hotplugging uses this */
static const struct pci_device_id __devinitdata MGC_aLinuxPciId[] = 
{
	{

	/* handle only the Mentor controller */

	.class = 		0,
	.class_mask = 	0,
	.driver_data =	0,

	.vendor =	0x14ab,
	.device =	0x10,
	.subvendor =	PCI_ANY_ID,
	.subdevice =	PCI_ANY_ID,

	}, 
	
	{ /* end: all zeroes */ }
	
};

MODULE_DEVICE_TABLE (pci, MGC_aLinuxPciId);

/* pci driver glue; this is a "new style" PCI driver module */
static struct pci_driver MGC_LinuxPciDriver = {
	.name =		(char *) MGC_HcdName,
	.id_table =	MGC_aLinuxPciId,

	.probe =	MGC_PciProbe,
	.remove =	MGC_PciRemove,

#ifdef	CONFIG_PM
	.suspend =	MGC_PciSuspend,
	.resume =	MGC_PciResume,
#endif
};

#endif	/* CONFIG_USB_MCD_PCI */

/*-------------------------------------------------------------------------*/

#if !defined(CONFIG_USB_MCD_PCI)
static int __devinit
platform_probe(
    struct device* dev)
{
    int retval = 0;
    char name[32];
    struct platform_device* pdev;
    struct resource* res;
    void* pRegs;
    unsigned int dwIpIrq;
    unsigned int dwDmaIrq;
    unsigned short wType;
    MGC_LinuxCd* pThis;

    DBG(2, "<==\n");

    pdev = to_platform_device(dev);

    /* 쥸ΰ */
    res = platform_get_resource(pdev, IORESOURCE_IO, 0);
    if (res) {
        pRegs= ioremap_nocache(res->start, res->end - res->start);
        DBG(2, "USB adr=%lx/%lx\n", res->start, res->end);
    }
    else {
        goto no_dev_err;
    }
    /* ֹ */
    dwIpIrq = platform_get_irq_byname(pdev, "USB_HS");
    if (!dwIpIrq) {
        goto no_dev_err;
    }
    dwDmaIrq = platform_get_irq_byname(pdev, "USB_DMAC");
    if (!dwDmaIrq) {
        goto no_dev_err;
    }
    /* ɥ饤̾ */
    snprintf(name, 32, "musbhdrc%d", MGC_nIndex++);
    /* USBסP2P*/
    wType = MUSB_CONTROLLER_HDRC;

    /* ȥ */
    pThis = MGC_LinuxInitController(dev, wType, dwIpIrq, dwDmaIrq, pRegs, name);
    if(pThis) {
        /* CDǡ */
        dev_set_drvdata(dev, pThis);
#if defined(CONFIG_USB_MCD_HOST)
        /* Hub */
        MGC_VirtualHubStart(&(pThis->RootHub));
#endif
    }
    else {
        iounmap(pRegs);
        err("%s: controller %s failed to initialize\n", __FILE__, name);
        goto no_dev_err;
    }

    return retval;

  no_dev_err:
    return -ENODEV;
}

static int
platform_remove(
    struct device* dev)
{
    MGC_LinuxCd* pThis = (MGC_LinuxCd*)dev_get_drvdata(dev);
    void* pRegs = pThis->pRegs;

    DBG(2, "<==\n");
    MGC_LinuxCdFree(pThis);
    iounmap(pRegs);

    return 0;
}

/*-------------------------------------------------------------------------*/

/* pci driver glue; this is a "new style" PCI driver module */
static struct device_driver mcd_driver = {
    .name =	"usb_mcd",
    .bus =	&platform_bus_type,
    .probe =	platform_probe,
    .remove =	platform_remove,
};

#endif	/* !defined(CONFIG_USB_MCD_PCI) */

/*-------------------------------------------------------------------------*/

/**
 * Required initialization for any module.
 */
int __init
MGC_ModuleInit(
    void) 
{
    int rc = -ENODEV;

    DBG(1, KERN_INFO
           "%s: Initializing MUSB Driver [npci=%d][gadget=%s][otg=%s]\n", 
           __FILE__,
#if !defined(CONFIG_USB_MCD_PCI)
           1,
#else
           0,
#endif
#ifdef MUSB_GADGET
           "yes"
#else
           "no"
#endif
           ,
#ifdef MUSB_OTG
           "yes"
#else
           "no"
#endif
           );
#if defined(CONFIG_USB_MCD_HOST)
    if ( !usb_disabled() ) {
#endif
#ifdef MUSB_OTG
        musb_otg_control_init();

        DBG(2, "OTG Support enabled\n");
#endif

#if defined(CONFIG_USB_MCD_PCI)
        rc=pci_module_init (&MGC_LinuxPciDriver);
        DBG(2, "PCI sysinit completed [%d]%d\n", rc, MGC_pFirstDriver->bEndCount);
#else
        /* NON PCI machines */
        rc = driver_register(&mcd_driver);
        DBG(2, "driver_register [%d]%d\n",
            rc, MGC_pFirstDriver ? MGC_pFirstDriver->bEndCount : 0);
#endif

#if defined(MUSB_GADGET) && defined(MUSB_TEST_GADGET)
        if ( rc==0 ) {
            MGC_gtest_init();
        }
#endif
#if defined(CONFIG_USB_MCD_HOST)
    } else {    
        printk(KERN_INFO "%s: USB Disabled , exiting\n", __FILE__);   
    }
#endif
    return rc;
}
/* ŪLinkξtarget_init.cƤӽФ */
#if !defined(MUSB_STATIC_LINK)
module_init (MGC_ModuleInit);
#endif


/**
 * Required cleanup for any module
 */
void __exit MGC_ModuleCleanup (void) 
{	
#ifdef CONFIG_USB_MCD_PCI
    pci_unregister_driver (&MGC_LinuxPciDriver);
#else
    driver_unregister(&mcd_driver);
#endif

#ifdef MUSB_OTG
    musb_otg_control_exit();

    DBG(2, "OTG Support enabled\n");
#endif
}
module_exit (MGC_ModuleCleanup);

#if defined(CONFIG_USB_MCD_HOST)
#ifdef DMA_HOST_CODE
/*==========================================================================*/
/*	void USB_MCD_HostDmaCallback()											*/
/*																			*/
/*	Dma Callback function for Host											*/
/*	 this function was called in DMA interrupt handler						*/
/*	 was set by pfDmaProgramChannel()										*/
/*==========================================================================*/
void USB_MCD_HostDmaCallback(
    MGC_DmaChannel* pChannel,
    uint8_t bEnd, 
    uint8_t bTransmit,
    uint8_t bProtocol,
    void* pData)
{
    uint8_t stat;
    int16_t wLength;
    uint16_t wCsr;
    uint16_t wRxCount;
	MGC_LinuxCd* pThis = (MGC_LinuxCd*)pData;
    const void* pBase = pThis->pRegs;
    struct urb* pUrb;
    int nPipe;

    uint8_t bDone = FALSE;
    MGC_LinuxLocalEnd* pEnd = &(pThis->aLocalEnd[bEnd]);
	uint16_t	rest_sz;
    uint8_t 	status = USB_ST_NOERROR;
	unsigned long flags;
    uint8_t	err_stat =	0;
    uint8_t* pBuffer =	NULL;
	
#ifdef HOST_NULL_FLAG
	USB_MCD_HostInfo *HostInfo = &(pEnd->HostInfo);
#endif
	MGC_DmaController* pDmaController = pThis->pDmaController;
    MGC_DmaChannel* pDmaChannel = pEnd->pDmaChannel;
	
	USB_MCD_DmaChStatus *pDmaChStatus = pEnd->pChEpStatus;
    
    spin_lock(&pEnd->Lock);
    pUrb = pEnd->pUrb;			
    nPipe = pUrb ? pUrb->pipe : 0;	
    spin_unlock(&pEnd->Lock);

    MGC_SelectEnd(pBase, bEnd);
    
    if(nPipe){
		/*==================================================================================*/
		/*  Receive 																		*/
		/*==================================================================================*/
	    if(usb_pipein(nPipe)){
	    	switch(pDmaChStatus->mode){
	    		case DMA_HOST_MODE0:
					/* DMA⡼0 */
					if((pEnd->pChEpStatus)->dma_trans_flag){
					
						/* Stop Channel */
						stat = pDmaController->pfDmaStopChannel(pDmaChannel);
    					DBG(2,"Stop Channel in Rx MODE0 \n");
						if (!stat) {
        					DBG(2,"Cannot stop dma channel %d\n", stat);
    					}
    					/* Disable Channel interrupt */
    					stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
    					if (!stat) {
        					DBG(2,"Cannot disable int dma channel %d\n", stat);
    					}
						(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;
					
						(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU; /* to CPU state */
    					DBG(2,"Change State CPU in Rx MODE0 \n");

   	 					/* update size of tx data in case of Mode0 */
   	 					stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
    					DBG(2,"GetChannelStatus in Rx MODE0 \n");
   	 					if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 						printk("Rx mode0 Error BUS_ABORT \n"); 
   	 					}
						if( pDmaChannel->dwActualLength != pEnd->dwRequestSize){
        					printk("Error size not equal! [dwActualLength]:%d\
        					 [dwRequestSize]:%d\n", pDmaChannel->dwActualLength ,pEnd->dwRequestSize);
						}
						/* update buffer offset */
   	 					pEnd->dwOffset += pDmaChannel->dwActualLength;
   	 					/* ISOξӤ˼ѥåȥꤹ */
   	 					
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						DBG(2,"Mode0 read RXCSR=%x \n",wCsr);
						if((pDmaChannel->dwActualLength < pEnd->wPacketSize)||!(wCsr&MGC_M_RXCSR_AUTOCLEAR)){
						/* RxPktRdyޥ˥奢 */
							wCsr &= ~MGC_M_RXCSR_RXPKTRDY;
						}
#ifdef HOST_CLR_REQPKT
						wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_H_REQPKT|MGC_M_RXCSR_AUTOCLEAR);
#else
						wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_AUTOCLEAR);
#endif					
						DBG(2,"Mode0 write RXCSR=%x \n",wCsr);
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
		       				
						if(usb_pipeisoc(nPipe)){
						/* ISO */
							/* set packet size a each frame for ISO */
							pUrb->iso_frame_desc[pEnd->dwIsoPacket].actual_length =  pDmaChannel->dwActualLength;
							
							bDone = (++pEnd->dwIsoPacket >= pUrb->number_of_packets);
							
							if(bDone){
							/* transfer end */
								if((pEnd->pChEpStatus)->dma_ch_flag){
								/* ͥͭξ */
									/* Release Channel */
									pDmaController->pfDmaReleaseChannel(pDmaChannel);
    								DBG(2,"Release Channel at transfer comp in Rx MODE0 \n");
									pEnd->pDmaChannel = NULL;
					  					
						  			(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
						  		}
						  		(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
    							DBG(2,"State IDLE at transfer comp in ISO Rx MODE0 \n");
								bDone = TRUE;
	    						status = USB_ST_NOERROR;
							}else{
								(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
    							DBG(2,"Change State CPU in ISO Rx MODE0 \n");
							}
						}else{
						/* Non-ISO */
					
							if(pEnd->wPacketSize ==  pDmaChannel->dwActualLength){
							/* Maxp size case */
								/* get size of rest data to transfer */
								rest_sz = pUrb->transfer_buffer_length - pEnd->dwOffset;
								if(pUrb->transfer_buffer_length < pEnd->dwOffset){
									printk("*** rest_sz calcu Error value:[%d] !! ***",rest_sz);
								}
								/* check rest size of the data */
								if(rest_sz > (uint16_t)NULL){	/* ĥͭ */
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
    								DBG(2,"Change State CPU in NonISO Rx MODE0 \n");
								}else if(rest_sz == (uint16_t)NULL){ /* NULLξ */
									if(HostInfo->null_flag){
									/* needed to transfer NULL data */
										(pEnd->pChEpStatus)->State = DMA_HOST_ST_NULL;	/* to NULL state */
    								DBG(2,"Change State NULL at NULL in NonISO Rx MODE0 \n");
									}else{
									/* transfer end*/
										if((pEnd->pChEpStatus)->dma_ch_flag){
										/* ͥͭξ */
											/* Release Cahnnel */
											pDmaController->pfDmaReleaseChannel(pDmaChannel);
											DBG(2,"Release Channel in NonISO Rx MODE0 \n");
											pEnd->pDmaChannel = NULL;
						  					(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
						  				}
						  				(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
    									DBG(2,"Change State IDLE at NULL in NonISO Rx MODE0 \n");

										status = USB_ST_NOERROR;
										bDone = TRUE;
									}
								}
							}else if(pEnd->wPacketSize > pDmaChannel->dwActualLength \
									&& pDmaChannel->dwActualLength>(uint16_t)NULL){
							/* Short packet size case */
								if((pEnd->pChEpStatus)->dma_ch_flag){
								/* ͥͭξ */
									/* Release Channel */
									pDmaController->pfDmaReleaseChannel(pDmaChannel);
									DBG(2,"Release Channel at Short n NonISO Rx MODE0 \n");
									pEnd->pDmaChannel = NULL;
					  
						  			(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
						  		}
						  		(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
    							DBG(2,"Change State IDLE at Short in NonISO Rx MODE0 \n");
	    						status = USB_ST_NOERROR;

							}else if(pDmaChannel->dwActualLength==(uint16_t)NULL){
					  			/* ꤨʤ */
					  			printk("Error!  \n");
							}else{
								printk("Condition Error! in Rx Mode0 \n");
								/* ʤ */
							}
						}
					}
	    			break;
	    			
	    		case DMA_HOST_MODE1:
					/* DMA⡼1 */
					if((pEnd->pChEpStatus)->dma_trans_flag){
					/* DMA channel exist */
						stat = pDmaController->pfDmaStopChannel(pDmaChannel);
    					DBG(2,"Stop Channel in Rx MODE1 \n");
    					if (!stat) {
        					DBG(2,"Cannot stop dma channel %d\n", stat);
    					}
    					stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
    					if (!stat) {
        					DBG(2,"Cannot disable int dma channel %d\n", stat);
    					}
						(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;
					
						/* get size of data transfered in DMA */
						stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
						DBG(2,"GetChannelStatus in Rx MODE1\n");
   	 					if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 						printk( "Rx mode1 Error BUS_ABORT \n"); 
   	 					}
		      			pEnd->dwOffset += pDmaChannel->dwActualLength;
						DBG(2,"Rx Length=%d in Rx MODE1 \n",pEnd->dwOffset);

						if((pEnd->pChEpStatus)->dma_ch_flag){
						/* ͥͭξ */
							/* Release Channel */
							pDmaController->pfDmaReleaseChannel(pDmaChannel);
							DBG(2,"Release Channel in Rx MODE1 \n");
							pEnd->pDmaChannel = NULL;
							(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
						}
						
						/*--- CPU߶ػ ---*/
    					spin_lock_irqsave(&pThis->Lock, flags);
						
						(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
						DBG(2,"Change State CPU in Rx MODE1\n");

						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						/* clear DMAENAB, AUTOREQ, AUTOCLEAR, (REQPKT)*/
						wCsr &= ~(MGC_M_RXCSR_H_AUTOREQ|MGC_M_RXCSR_H_REQPKT|MGC_M_RXCSR_AUTOCLEAR|MGC_M_RXCSR_DMAENAB);						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);

						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						/* clear DMAMODE */
						wCsr &= ~MGC_M_RXCSR_DMAMODE;
						MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
						
						/* Υѥåȼǧ */
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
						
						if(wCsr&MGC_M_RXCSR_RXPKTRDY){
						/* MGC_M_RXCSR_RXPKTRDY was detected !*/
						
							wRxCount = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCOUNT, bEnd);    
    						DBG(2, "end %d RxCount=%04x \n", bEnd, wRxCount); 
							
							wLength = min((unsigned int)wRxCount,\
		   							pUrb->transfer_buffer_length - pEnd->dwOffset);
		   					DBG(2, "wLength=%d \n", wLength); 

							/* CPUž DMAžκǸCPUžˤʤ */
							if((pDmaChStatus)->dma_req_flag){
								pBuffer = (uint8_t*)(__bus_to_virt(pUrb->transfer_dma)) + pEnd->dwOffset;
								DBG(2, "buffer=%p transfer_dma used in DMAmode1 in DMA Rx handler \n",pBuffer);
							}else{
#ifdef	MUSB_DMA_BUS2VIRT 
    						if ( !(pUrb->transfer_buffer) ) {
								pBuffer = (void*)__bus_to_virt(pUrb->transfer_dma) + pEnd->dwOffset;
								DBG(2, "buffer=%p transfer_dma used in not DMAmode \n",pBuffer);
							}else{
								pBuffer = (void*)pUrb->transfer_buffer + pEnd->dwOffset;
								DBG(2, "buffer=%p transfer_buffer used \n",pBuffer);
							}
#else
							pBuffer = (void*)__bus_to_virt(pUrb->transfer_buffer) + pEnd->dwOffset;
							
#endif
							}
#ifdef CONSISTENT_SYNC
							consistent_sync(pBuffer, pEnd->dwOffset, DMA_FROM_DEVICE);
#endif
							MGC_HdrcUnloadFifo(pBase, bEnd, wLength, pBuffer);

							/* update length */
							pEnd->dwOffset += wLength;
							
							if(wRxCount == pEnd->wPacketSize){
							/* MAXP packet received */
								if(pUrb->transfer_buffer_length < pEnd->dwOffset){
								/* Over data received */
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
									DBG(2,"Change State IDLE in Rx MODE1\n");
									err_stat = TRUE;
									bDone = TRUE;
									status = USB_ST_DATAOVERRUN;
									
								}else{
								
									/* get size of rest data */
									rest_sz = pUrb->transfer_buffer_length - pEnd->dwOffset;
									if(pUrb->transfer_buffer_length < pEnd->dwOffset){
										printk("*** rest_sz calcu Error value:[%d] !! ***",rest_sz);
									}									
									if(rest_sz>(uint16_t)NULL){
									/* ĥͭ */
										(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
										DBG(2,"Change State CPU in Rx MODE1\n");
									}else if(rest_sz==(uint16_t)NULL){
									/* ĥʤ */
										if(HostInfo->null_flag){
										/* needed to transfer NULL data */
											(pEnd->pChEpStatus)->State = DMA_HOST_ST_NULL;	/* to NULL state */
											DBG(2,"Change State NULL in Rx MODE1\n");
										}else{
										/* needed not to transfer NULL data */
											(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
											DBG(2,"Change State IDLE in Rx MODE1\n");
											bDone = TRUE;
											status = USB_ST_NOERROR;
										}
									}else{
										/* ʤ */
									}
								}
							}else if(wRxCount < pEnd->wPacketSize && wRxCount >(uint16_t)NULL){
							/* SHORT packet received */
								(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
								DBG(2,"Change State IDLE at Short data in Rx MODE1\n");
								bDone = TRUE;
								if(pUrb->transfer_buffer_length < pEnd->dwOffset){
								/* Over data received */
									err_stat = TRUE;
									status = USB_ST_DATAOVERRUN;
								}else{
									status = USB_ST_NOERROR;
								}
							}else if(wRxCount == (uint16_t)NULL){
								/* NULL packet received */
								(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
								DBG(2,"Change State IDLE at NULL data in Rx MODE1\n");
								bDone = TRUE;
								status = USB_ST_NOERROR;
							}else{
							/* ʤ */
							}
							/* Because RxPktRdy exist, clear it. */
							wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
#ifdef HOST_CLR_REQPKT
							wCsr &= ~(MGC_M_RXCSR_RXPKTRDY|MGC_M_RXCSR_H_REQPKT);
#else
							wCsr &= ~(MGC_M_RXCSR_RXPKTRDY);
#endif			
							/* Clear RxPktRdy*/
							MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
						}else{
							/* in case  not RxPktRdy exist */
							
							/* get size of rest data */
							rest_sz = pUrb->transfer_buffer_length - pEnd->dwOffset;
							if(pUrb->transfer_buffer_length < pEnd->dwOffset){
								printk("*** rest_sz calcu Error value:[%d] !! ***",rest_sz);
							}
							if(rest_sz>(uint16_t)NULL){
							/* ĥͭ */
							(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
							DBG(2,"Change State CPU at non NULL no RxPktRdy in Rx MODE1 \n");
										
							}else if(rest_sz==(uint16_t)NULL){
							/* ĥʤ */
								if(HostInfo->null_flag){
								/* needed to transfer NULL data */
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_NULL;	/* to NULL state */
									DBG(2,"Change State NULL at NULL no RxPktRdy in Rx MODE1 \n");

								}else{
								/* needed not to transfer NULL data */
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
									DBG(2,"Change State IDLE at NULL no RxPktRdy in Rx MODE1 \n");
									bDone = TRUE;
									status = USB_ST_NOERROR;
								}
							}else{
								/* ʤ */
							}
						}
						/*--- CPU ---*/
						spin_unlock_irqrestore(&pThis->Lock, flags);
					}
	    			break;
				case DMA_HOST_MODE_NON:
					printk("abnormal Error! \n");
					break;
	    	}
	    	if(bDone || (pEnd->pChEpStatus)->State == DMA_HOST_ST_IDLE){ /* if transfer ended */
			
				/* FIFO flash */
				if(status != USB_ST_NOERROR){
					pUrb->status = status;
				}else{
					pUrb->status = USB_ST_NOERROR;
					DBG(2,"using  pUrb->status = USB_ST_NOERROR \n");
					/* if pUrb->status is not USB_ST_NOERROR, It uses status from PacketRx() */
				}

				wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
				wCsr |= MGC_M_RXCSR_FLUSHFIFO;
				/* Clear RxPktRdy, FIFOեå */
				MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
				if(HostInfo->db_buff){	/* in case of Double buffer */
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
#ifdef HOST_CLR_REQPKT
					wCsr &= ~(MGC_M_RXCSR_RXPKTRDY|MGC_M_RXCSR_H_REQPKT);
#else
					wCsr &= ~(MGC_M_RXCSR_RXPKTRDY);
#endif			
					wCsr |= MGC_M_RXCSR_FLUSHFIFO;
					/* Clear RxPktRdy, FIFOեå */
					MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
				}
				
				/* FIFO status free */
	    		pEnd->bIsReady = TRUE;
				/* Update transfer size */
				pUrb->actual_length = pEnd->dwOffset;
				DBG(2,"pUrb->actual_length=%d \n",pUrb->actual_length);
#ifdef DMARX_BUF
				DBG(2,"Rx DMA data check \n");
				usb_mcd_print_dma(((uint8_t*)(__bus_to_virt(pUrb->transfer_dma))), pUrb->actual_length, 0);
#endif
				/* Complete Callback */
				if(pUrb->complete)
	    		{
					pUrb->hcpriv=NULL;
					COMPLETE_URB(pUrb, NULL);
					DBG(2, "complete! in DMA Rx handler \n");
				}else{
					DBG(2, "No complete func ! in DMA Rx handler \n");
				}
				spin_lock(&pEnd->Lock);
				MGC_LinuxStartNextUrb(pThis, bEnd);
				spin_unlock(&pEnd->Lock);
				return;
	    	}else{
	    	/* to be continued transferring */
	    		/* Set ReqPkt*/
				wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd);
				wCsr |=  MGC_M_RXCSR_H_REQPKT;
				MGC_WriteCsr16(pBase, MGC_O_HDRC_RXCSR, bEnd, wCsr);
				DBG(2,"*** SET REQPKT ***\n");
				return;
	    	}

		/*==================================================================================*/
		/*  Transmit 																		*/
		/*==================================================================================*/
	    }else if(usb_pipeout(nPipe)){
	    	switch(pDmaChStatus->mode){
	    		case DMA_HOST_MODE0:
					if((pEnd->pChEpStatus)->dma_trans_flag){
						/* DMAžե饰ON */
							/* DMAͥ */
							stat = pDmaController->pfDmaStopChannel(pDmaChannel);
    						if (!stat) {
        						DBG(2,"Cannot stop dma channel %d\n", stat);
    						}
    						/* DMAԵ */
    						stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
    						if (!stat) {
        						DBG(2,"Cannot disable int dma channel %d\n", stat);
    						}
							(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;
							/* get size of data transfered in DMA */
							stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
   	 						if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 							printk("Tx mode0 Error BUS_ABORT \n"); 
   	 						}

							if(pDmaChannel->dwActualLength != pEnd->dwRequestSize){
        						DBG(2,"ERROR size not equal! [dwActualLength]:%d \
        						 [dwRequestSize]:%d\n",pDmaChannel->dwActualLength ,pEnd->dwRequestSize);
							}

		      				DBG(2,"TX Mode0 pEnd->dwRequestSize=%d \n",pEnd->dwRequestSize);
							(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
						}
	    			break;
	    		case DMA_HOST_MODE1:
						/* CPU߶ػ */
    					spin_lock_irqsave(&pThis->Lock, flags);
    					
						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
						/* clear DMAENAB, AUTOSET */
						wCsr &= ~(MGC_M_TXCSR_AUTOSET |MGC_M_TXCSR_DMAENAB);
						MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);    					

						wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
						/* clear DMAMODE*/
						wCsr &= ~MGC_M_TXCSR_DMAMODE;
						MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
						
						if((pEnd->pChEpStatus)->dma_trans_flag){
						/* DMAžե饰ON */
						
							/* DMAͥ */
							stat = pDmaController->pfDmaStopChannel(pDmaChannel);
    						if (!stat) {
        						DBG(2,"Cannot stop dma channel %d\n", stat);
    						}
    						/* DMAԵ */
    						stat = pDmaController->pfDmaDisableIntChannel(pDmaChannel);
    						if (!stat) {
        						DBG(2,"Cannot disable int dma channel %d\n", stat);
    						}
							(pEnd->pChEpStatus)->dma_trans_flag = USB_MCD_FLAG_OFF;
							
							if((pEnd->pChEpStatus)->mode == DMA_HOST_MODE1){
							/* in case of DMA mode1 */
								
								/* get size of data transfered by DMA */
								stat = pDmaController->pfDmaGetChannelStatus(pDmaChannel);
   	 							if(stat == MGC_DMA_STATUS_BUS_ABORT){
   	 								printk("Tx mode1 Error BUS_ABORT \n"); 
   	 							}
								if(pDmaChannel->dwActualLength != pEnd->dwRequestSize){
        							printk("ERROR size not equal! [dwActualLength]:%d \
        							 [dwRequestSize]:%d\n",pDmaChannel->dwActualLength ,pEnd->dwRequestSize);
								}
								/* update buffer offset */
   	 							pEnd->dwOffset += pDmaChannel->dwActualLength;
   	 							/* mode1ǤDMAλ˥򹹿롣̩ˤϺǸΥѥåȤUSBХǤ
   	 							äƤʤ⤢롣CPUž߻˽λȽǤʤ롣*/
   	 							DBG(2,"Tx mode1 pEnd->dwOffset = %d \n",pEnd->dwOffset);
   	 						}
   	 						if((pEnd->pChEpStatus)->dma_ch_flag){
   	 						/* DMA channel exist and DMA mode1(implicit condition) */
   	 			
   	 							/* DMA channel release */
   	 							pDmaController->pfDmaReleaseChannel(pDmaChannel);
								pEnd->pDmaChannel = NULL;
								/* DMA channel flag set off */
   	 							(pEnd->pChEpStatus)->dma_ch_flag = USB_MCD_FLAG_OFF;
							}
							/* check TxCSR whether TxPktRdy was clear */
							wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
							/* clear DMAMODE*/
							if(wCsr&MGC_M_TXCSR_TXPKTRDY){
							/* TXPKTRDY exist */
								(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
								DBG(2,"Tx mode1 to CPU TxPktRdy exist! \n");
							}else{
								if(pEnd->dwRequestSize%pEnd->wPacketSize){
								/* SHORTѥåȤꤵƤ */
									wCsr =	MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
									/* set TxPktRdy */
									wCsr |= MGC_M_TXCSR_TXPKTRDY;
									MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
								
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_CPU;	/* to CPU state */
									DBG(2,"DmaHostCb Tx mode1 to CPU \n");
								}else{
								/* MAXPž줿AutoSetˤTxPktRdyåȤƤ롣*/
									(pEnd->pChEpStatus)->State = DMA_HOST_ST_IDLE;	/* to IDLE state */
									DBG(2,"DmaHostCb Tx mode1 to IDLE bDone! \n");
									bDone = TRUE;
									status = USB_ST_NOERROR;
								}
							}
						}
						/* CPU */
						spin_unlock_irqrestore(&pThis->Lock, flags);
	    			break;
				case DMA_HOST_MODE_NON:
					printk("abnormal Error! \n");
					break;

	    	}
	    	if(bDone || (pEnd->pChEpStatus)->State == DMA_HOST_ST_IDLE){ /* if transfer ended */
				/* FIFOեå */
				wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
				wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
				wCsr |= MGC_M_TXCSR_FLUSHFIFO;
				MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
				if(HostInfo->db_buff){	/* in case of Double buffer */
					wCsr = MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					wCsr &= ~MGC_M_TXCSR_FIFONOTEMPTY;
					wCsr |= MGC_M_TXCSR_FLUSHFIFO;
					/* Clear RxPktRdy, FIFOեå */
					MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
				}
				/* FIFO status free */
	    		pEnd->bIsReady = TRUE;
				
	    		/* žλ*/
				pUrb->actual_length = pEnd->dwOffset;
				DBG(2,"Tx mode1 pUrb->actual_length=%d \n",pUrb->actual_length);
				pUrb->status = status;
								
				/* Complete Callback */
				if(pUrb->complete)
	    		{
					pUrb->hcpriv=NULL;
					COMPLETE_URB(pUrb, NULL);
					DBG(2, "*** Complete! in DMA Tx handler ***\n");
				}else{
					DBG(2, "***No complete func ! in DMA Tx handler ***\n");
				}
				spin_lock(&pEnd->Lock);
				MGC_LinuxStartNextUrb(pThis, bEnd);
				spin_unlock(&pEnd->Lock);
	
				return;	
	    	}else{
	    		if(pDmaChStatus->mode == DMA_HOST_MODE0){
					wCsr =	MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
					if(pEnd->dwRequestSize< pEnd->wPacketSize ||!(wCsr&MGC_M_TXCSR_AUTOSET)){
					/* TxPktRdyå AutoSetʤ */
						wCsr =	MGC_ReadCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd);
						/* set TxPktRdy */
						wCsr |= MGC_M_TXCSR_TXPKTRDY;
						MGC_WriteCsr16(pBase, MGC_O_HDRC_TXCSR, bEnd, wCsr);
					}
	    		}
	    		return;
	    	}
	    }else{
			printk("Error \n");
		}
	}else{
		DBG(2,"Non Pipe !  \n");
		/* ʤ */
	}
}
#endif
#endif /* CONFIG_USB_MCD_HOST */
/*==========================================================================*/
/*	usb_mcd_buffer_create													*/
/*	  create pools to allocate a buffer for DMA 							*/
/*  @param pThis MGC_LinuxCd*	to hold areas of dma_pool					*/
/*==========================================================================*/
uint8_t usb_mcd_buffer_create(MGC_LinuxCd* pThis){
	char	name [16];
	int 	i, size;

	DBG(1,"<==\n");
	
	for (i = 0; i < HCD_BUFFER_POOLS; i++) { 
		if (!(size = pool_max [i]))
			continue;
		snprintf (name, sizeof name, "buffer-%d", size);
		pThis->pool [i] 
#if defined(CONFIG_USB_MCD_HOST)
			= dma_pool_create (name, (struct device*)pThis->pBus->controller,
				size, size, 0);
#else
			= dma_pool_create (name, (struct device*)pThis->controller,
				size, size, 0);
#endif
		DBG(1,"pThis->pool [%d] =%p size=%d \n",i, pThis->pool [i], size);
		if (!(pThis->pool [i])) {
			DBG(1,"not get a pool-area! call usb_mcd_buffer_destroy \n");
			usb_mcd_buffer_destroy(pThis);
			return -ENOMEM;
		}
	}
	return 0;
}
/*==========================================================================*/
/*	usb_mcd_buffer_destroy													*/
/*	  release pools to allocate a buffer for DMA 							*/
/*  @param pThis MGC_LinuxCd*	to hold areas of dma_pool					*/
/*==========================================================================*/
void usb_mcd_buffer_destroy(MGC_LinuxCd* pThis){
	int		i;

	DBG(1,"<==\n");
	
	for (i = 0; i < HCD_BUFFER_POOLS; i++) { 
		struct dma_pool		*pool = pThis->pool [i];
		if (pool) {
			dma_pool_destroy (pool);
			pThis->pool[i] = NULL;
		DBG(1,"pThis->pool [%d] =%p \n",i, pool);
		}
	}
}

#if defined(DMA_DEBUG_DATA)||defined(DMA_DEBUG_RX_DATA)
/*==========================================================================*/
/*	usb_mcd_print_dma														*/
/*	  print a data to transfer in DMA			 							*/
/*  @param uint8_t* pBuffer		a target buffer to print					*/
/*  @param uint32_t dwLength	length of data to print						*/
/*	@param uint32_t dwOffset	a start position in target buffer to print	*/
/*==========================================================================*/
STATIC void usb_mcd_print_dma(uint8_t* pBuffer, uint32_t dwLength, uint32_t dwOffset){
	uint32_t	i;
	uint8_t* 	pData = pBuffer + dwOffset;
	
	printk("Data check pBuffer=%p Length=%d Offset=%d \n", pBuffer, dwLength, dwOffset);
	
	/* print data */

	for(i=0; i <dwLength; i++){
		printk("%x ",((uint8_t*)(pData))[i]);
		if(((i+1)%16)==0){
			printk("\n");
		}
	}
	printk("\n");

	return;
}
#endif
#if defined(CONFIG_USB_MCD_HOST)
/*==========================================================================*/
/*	usb_mcd_maxpacket														*/
/*	Τusb.husb_maxpacket()											*/
/*	MaxPacketSize롣												*/
/*	EPξʤΤǷٹ𤬤Ǥ뤿˥δؿ򥳡뤹뤳Ȥˤ롣	*/
/*==========================================================================*/
STATIC inline uint16_t 
	usb_mcd_maxpacket(struct usb_device *udev, int pipe, int is_out)
{
	struct usb_host_endpoint	*ep;
	unsigned	epnum = usb_pipeendpoint(pipe);

	if (is_out) {
		ep = udev->ep_out[epnum];
	} else {
		ep = udev->ep_in[epnum];
	}
	if (!ep)
		return 0;

	return le16_to_cpu(ep->desc.wMaxPacketSize);
}
#endif
#ifdef MCD_PM_CODE
/*==========================================================================*/
/*	musb_bus_start_hdrc()													*/
/*	start_controlǥ뤵												*/
/*==========================================================================*/
int musb_bus_start_hdrc(void){

    unsigned short wType;
    uint8_t bOk;

#ifdef DMA_HOST_CODE
    if(MGC_pFirstDriver->pDmaController)
    {
    	DBG(3,  "DMA Mode enabled\n");
    	/* Enable Dmac interrupt */
        MGC_pFirstDriver->pDmaController->pfDmaStartController(\
        	MGC_pFirstDriver->pDmaController
        );
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 1;
    }
#else    
 #ifdef MUSB_DMA
    if(MGC_pFirstDriver->pDmaController)
    {
    	DBG(3,  "DMA Mode enabled\n");
    	/* Enable Dmac interrupt */
        MGC_pFirstDriver->pDmaController->pfDmaStartController( \
            MGC_pFirstDriver->pDmaController
        );
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 1;
    }
 #endif
#endif /* DMA_HOST_CODE end */

	/* Disable USB interrupt */
    MGC_HdrcDisable(MGC_pFirstDriver);

    /* discover configuration */
    wType = MUSB_CONTROLLER_HDRC;
    bOk = MGC_HdrcInit(wType, MGC_pFirstDriver);
    if ( !bOk ) {
		return 1;    
    }
    enable_irq(MGC_pFirstDriver->nIrq);					/* USB 	*/
    enable_irq(MGC_pFirstDriver->pDmaController->nIrq);	/* DMAC */

	return 0;
}
/*==========================================================================*/
/*	musb_bus_stop_hdrc()													*/
/*	stop_controlǥ뤵												*/
/*==========================================================================*/
int musb_bus_stop_hdrc(void){

#ifdef DMA_HOST_CODE
    if(MGC_pFirstDriver->pDmaController)
    {
    	DBG(3,  "DMA Mode disabled\n");
    	/* Enable Dmac interrupt */
        MGC_pFirstDriver->pDmaController->pfDmaStopController(\
        	MGC_pFirstDriver->pDmaController
        );
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 1;
    }
#else    
 #ifdef MUSB_DMA
    if(MGC_pFirstDriver->pDmaController)
    {
    	DBG(3,  "DMA Mode disabled\n");
    	/* Enable Dmac interrupt */
        MGC_pFirstDriver->pDmaController->pfDmaStopController( \
            MGC_pFirstDriver->pDmaController
        );
    }
    else {
        err("cannot dma controller is not allocated\n");
        return 1;
    }
 #endif
#endif /* DMA_HOST_CODE end */
	/* HdrcStop */
	MGC_HdrcStop(MGC_pFirstDriver);
	
    disable_irq(MGC_pFirstDriver->nIrq);					/* USB 	*/
    disable_irq(MGC_pFirstDriver->pDmaController->nIrq);	/* DMAC */

return 0;

}
/*==========================================================================*/
/*	musb_bus_start_hdrc_int()												*/
/*	called in start_gadget to enable HDRC-interrupt 						*/
/*==========================================================================*/
int musb_bus_start_hdrc_int(void){

    MGC_GadgetStart(MGC_pFirstDriver);
	
	return 0;
}
/*==========================================================================*/
/*	musb_bus_stop_hdrc_int()												*/
/*	called in start_gadget to disable HDRC-interrupt 						*/
/*==========================================================================*/
int musb_bus_stop_hdrc_int(void){

    MGC_GadgetStop(MGC_pFirstDriver);
	
	return 0;
}

#ifdef MCD_PM_NON_CORRUPT

/*==========================================================================*/
/*	usb_mcd_set_padnum()													*/
/*	called in InitController to delete print Corrupt						*/
/*==========================================================================*/
int usb_mcd_set_padnum(MGC_LinuxCd *pParam){

    MGC_LinuxLocalEnd* pEnd;
    uint8_t bEnd;
    
    pParam->wEndMask = 1;
    pEnd = &(pParam->aLocalEnd[0]);
    spin_lock_init(&pEnd->Lock);
    INIT_LIST_HEAD(&(pEnd->urb_list));
    pEnd->wMaxPacketSizeTx = MGC_END0_FIFOSIZE;
    pEnd->wMaxPacketSizeRx = MGC_END0_FIFOSIZE;
    pEnd->bLocalEnd = 0;
#if MUSB_DEBUG > 0
    pEnd->dwPadFront = MGC_PAD_FRONT;
    pEnd->dwPadBack = MGC_PAD_BACK;
#endif

    pParam->bEndCount = 1;
    for(bEnd = 1; bEnd < MUSB_C_NUM_EPS; bEnd++)
    {
		pEnd = &(pParam->aLocalEnd[bEnd]);
		spin_lock_init( &pEnd->Lock );
		INIT_LIST_HEAD(&(pEnd->urb_list));
		pEnd->bLocalEnd = bEnd;
#if MUSB_DEBUG > 0
		pEnd->dwPadFront = MGC_PAD_FRONT;
		pEnd->dwPadBack = MGC_PAD_BACK;
#endif
		pParam->bEndCount++;
		pParam->wEndMask |= (1 << bEnd);
    }
    return 0;
}
#endif

#endif /* MCD_PM_CODE */
/* this module is always GPL, the gadget might not... */
#define DRIVER_INFO DRIVER_DESC " " DRIVER_VERSION
MODULE_DESCRIPTION (DRIVER_INFO);
MODULE_AUTHOR (DRIVER_AUTHOR);
MODULE_LICENSE ("GPL");

